/*
 *  Copyright (C) 1999, 2000  Heikki Tauriainen <Heikki.Tauriainen@hut.fi>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "BuchiAutomaton.h"

namespace Graph
{

/******************************************************************************
 *
 * Function definitions for class BuchiAutomaton.
 *
 *****************************************************************************/

/* ========================================================================= */
BuchiAutomaton::BuchiAutomaton
  (size_type initial_number_of_states, size_type initstate,
   unsigned long int number_of_accept_sets) :
  initial_state(initstate), number_of_acceptance_sets(number_of_accept_sets)
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructor for class BuchiAutomaton. Initializes an
 *                automaton with a given number of states and a given
 *                initial state.
 *
 * Arguments:     initial_number_of_states  --  Initial size of the automaton
 *                                              (can be grown later).
 *                initstate                 --  Index of the automaton's
 *                                              initial state.
 *                number_of_accept_sets     --  Number of acceptance sets in
 *                                              the automaton.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  if (initstate >= initial_number_of_states
      && !(initial_number_of_states == 0 && initstate == 0))
    throw NodeIndexException();

  nodes.reserve(initial_number_of_states);
  for (size_type i = 0; i < initial_number_of_states; ++i)
    nodes.push_back(new BuchiState(number_of_accept_sets));
}

/* ========================================================================= */
BuchiAutomaton& BuchiAutomaton::operator=(const BuchiAutomaton& automaton)
/* ----------------------------------------------------------------------------
 *
 * Description:   Assignment operator for class BuchiAutomaton. Assigns the
 *                contents of a BuchiAutomaton to another one.
 *
 * Argument:      automaton  --  A constant reference to a BuchiAutomaton whose
 *                               contents are to be copied.
 *
 * Returns:       A reference to the BuchiAutomaton assigned to.
 *
 * ------------------------------------------------------------------------- */
{
  if (&automaton != this)
  {
    Graph<GraphEdgeContainer>::operator=(automaton);
    initial_state = automaton.initial_state;
    number_of_acceptance_sets = automaton.number_of_acceptance_sets;
  }

  return *this;
}

/* ========================================================================= */
void BuchiAutomaton::clear()
/* ----------------------------------------------------------------------------
 *
 * Description:   Makes the automaton empty.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Graph<GraphEdgeContainer>::clear();
  initial_state = 0;
  number_of_acceptance_sets = 0;
}

/* ========================================================================= */
BuchiAutomaton::size_type BuchiAutomaton::expand(size_type node_count)
/* ----------------------------------------------------------------------------
 *
 * Description:   Inserts a given number of states to a BuchiAutomaton.
 *
 * Argument:      node_count  --  Number of states to be inserted.
 *
 * Returns:       The index of the last inserted state.
 *
 * ------------------------------------------------------------------------- */
{
  nodes.reserve(nodes.size() + node_count);

  while (node_count > 0)
  {
    nodes.push_back(new BuchiState(number_of_acceptance_sets));
    node_count--;
  }

  return nodes.size() - 1;
}

/* ========================================================================= */
void BuchiAutomaton::regularize()
/* ----------------------------------------------------------------------------
 *
 * Description:   Converts a generalized Bchi automaton (i.e., an automaton
 *                with more than one set of accepting states) into an automaton
 *                with only one set of accepting states.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  if (empty())
    return;

  /*
   *  A vector of pointers to Nodes (actually to BuchiStates) for storing
   *  intermediate results.
   */

  vector<Graph<GraphEdgeContainer>::Node*,
         ALLOC(Graph<GraphEdgeContainer>::Node*) > reordered_nodes;

  /*
   *  If the automaton currently has more than one set of accepting states,
   *  first expand the automaton into one that contains a linearly greater
   *  number of states (in the number of acceptance sets) than the current
   *  automaton.
   *
   *  That is, if there are n acceptance sets, make n `copies' of the
   *  original automaton. This results in n `levels' of states, where each
   *  `level' corresponds to a particular acceptance set. Then redirect
   *  some transitions from level m to level  (m + 1) % n  such that if a
   *  state on level  m  belongs to acceptance set  m, connect the state
   *  to its target state on level  (m + 1) % n. Otherwise, connect the state
   *  to its target on level  m.
   *
   *  In addition, make each state belonging to acceptance set 0 accepting
   *  also in the expanded automaton.
   */

  if (number_of_acceptance_sets > 1)
  {
    unsigned long int expanded_size = size() * number_of_acceptance_sets;
    size_type current_state, target_state;

    reordered_nodes.reserve(expanded_size);
    for (size_type state = 0; state < expanded_size; state++)
      reordered_nodes.push_back(new BuchiState(1));

    for (size_type state = 0; state < size(); state++)
    {
      for (GraphEdgeContainer::const_iterator transition
	     = nodes[state]->edges().begin();
	   transition != nodes[state]->edges().end();
	   ++transition)
      {
	for (unsigned long int accept_set = 0;
	     accept_set < number_of_acceptance_sets;
	     accept_set++)
	{
	  current_state = state + accept_set * size();
	  target_state = (*transition)->targetNode() + accept_set * size();
	  if (operator[](state).acceptanceSets().test(accept_set))
	  {
	    if (accept_set == number_of_acceptance_sets - 1)
	      target_state = (*transition)->targetNode();
	    else
	      target_state += size();
	  }

	  static_cast<BuchiState*>(reordered_nodes[current_state])
	    ->connectTo(target_state,
		        static_cast<BuchiTransition*>(*transition)->guard());
	}
      }

      if (operator[](state).acceptanceSets().test(0))
	static_cast<BuchiState*>(reordered_nodes[state])
	  ->acceptanceSets().set(0);

      delete nodes[state];
    }

    nodes = reordered_nodes;
    reordered_nodes.clear();

    number_of_acceptance_sets = 1;
  }	

  /*
   *  Remove redundant states of the automaton by performing a depth-first
   *  search in the automaton starting from its initial state. During the
   *  search, all visited states are copied to a new automaton, which will
   *  consist of exactly all the reachable states of the automaton at
   *  the end of the search.
   */

  Bitset visited(nodes.size());
  stack<size_type, deque<size_type, ALLOC(size_type) > > state_stack;
  size_type current_state, target_state;

  /*
   *  Mapping from old state identifiers to new state identifiers.
   */

  vector<size_type, ALLOC(size_type) > state_number_mapping(nodes.size());

  /*
   *  Initialize the depth-first search stack with the initial state of the
   *  automaton.
   */

  state_stack.push(initial_state);
  visited.set(initial_state);
  state_number_mapping[initial_state] = 0;
  reordered_nodes.push_back(new BuchiState(1));

  if (number_of_acceptance_sets > 0 &&
      operator[](initial_state).acceptanceSets().test(0))
    static_cast<BuchiState*>(reordered_nodes.back())->acceptanceSets().set(0);

  while (!state_stack.empty())
  {
    /*
     *  Pop a state from the stack.
     */

    current_state = state_stack.top();
    state_stack.pop();

    for (GraphEdgeContainer::const_iterator transition = 
           nodes[current_state]->edges().begin();
         transition != nodes[current_state]->edges().end();
         ++transition)
    {
      /*
       *  Process the transitions of the current state. If there exists a
       *  transition to a state not yet included in the new automaton, create
       *  a new state and insert it to the new automaton. Push the target state
       *  also on the stack so that its successors will be eventually
       *  processed.
       */

      target_state = (*transition)->targetNode();

      if (!visited[target_state])
      {
        visited.set(target_state);
        state_stack.push(target_state);

        state_number_mapping[target_state] = reordered_nodes.size();
        reordered_nodes.push_back(new BuchiState(1));

	if (number_of_acceptance_sets > 0
	    && operator[](target_state).acceptanceSets().test(0))
	  static_cast<BuchiState*>(reordered_nodes.back())
	    ->acceptanceSets().set(0);
      }

      /*
       *  Adjust the transitions between the states of the old and the new
       *  automata.
       */

      static_cast<BuchiState*>
        (reordered_nodes[state_number_mapping[current_state]])
          ->connectTo(state_number_mapping[target_state],
                      static_cast<const BuchiTransition*>(*transition)
                        ->guard());
    }

    /*
     *  After processing a state of the old automaton, release the memory
     *  allocated to the state.
     */

    delete nodes[current_state];
  }

  /*
   *  Release the memory allocated to unreachable states.
   */

  for (size_type state = 0; state < nodes.size(); ++state)
    if (!visited[state])
      delete nodes[state];

  nodes = reordered_nodes;
  initial_state = 0;
}

/* ========================================================================= */
void BuchiAutomaton::read(istream& input_stream)
/* ----------------------------------------------------------------------------
 *
 * Description:   Reads an automaton description (which may represent a
 *                generalized Bchi automaton) from a stream and stores it
 *                into the automaton object, converting it to a regular
 *                Bchi automaton if necessary.
 *
 * Argument:      input_stream  --  A reference to an input stream.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  size_type number_of_states, current_state, neighbor_state;
  unsigned long int acceptance_set;
  Exceptional_istream einput_stream(&input_stream, ios::failbit | ios::badbit);

  clear();

  try 
  {
    /*
     *  Read in the number of states in the generalized Bchi automaton.
     */

    einput_stream >> number_of_states;

    /*
     *  If the automaton is empty, do nothing.
     */

    if (number_of_states == 0)
      return;

    einput_stream >> number_of_acceptance_sets;

    /*
     *  Allocate space for the regular Bchi automaton that will be constructed
     *  from the generalized Bchi automaton.
     */

    nodes.reserve(number_of_states);
    for (size_type i = 0; i < number_of_states; ++i)
      nodes.push_back(new BuchiState(number_of_acceptance_sets));

    /*
     *  The automaton state numbers will be mapped from input file identifiers
     *  to the interval [0...(number of states - 1)].
     */

    map<long int, size_type, less<long int>, ALLOC(size_type) >
      state_number_map;
    pair<long int, size_type> state_mapping(0, 0);

    pair<map<long int, size_type, less<long int>, ALLOC(size_type) >
             ::const_iterator,
         bool>
      state_finder;

    /*
     *  Also the acceptance set numbers will be mapped to the interval
     *  [0...(number of acceptance sets - 1)].
     */

    map<long int, unsigned long int, less<long int>, ALLOC(unsigned long int) >
      acceptance_set_map;
    pair<long int, unsigned long int> acceptance_set_mapping(0, 0);

    pair<map<long int, unsigned long int, less<long int>,
             ALLOC(unsigned long int) >::const_iterator,
         bool>
      acceptance_set_finder;

    /*
     *  A bitset is used to keep information about states which have been
     *  processed. (This is used to verify that each state is defined exactly
     *  once in the input.)
     */

    Bitset processed_states(number_of_states);

    bool initial_state_fixed = false;       /* will be set to true after    */
                                            /* processing the initial state */
                                            /* of the automaton             */

    int is_initial;

    LtlFormula* guard = 0;

    for (size_type i = 0; i < number_of_states; ++i)
    {
      /*
       *  Begin processing a new state by reading a state id from the stream.
       */

      einput_stream >> state_mapping.first;

      /*
       *  Try to insert a new state mapping into the state identifier map.
       *  If an insertion actually occurs (there is no element in the map with
       *  key equal to `state_mapping.first'), increment the state counter. In
       *  any case, set the value of `current_state' to the map value.
       *  (Note: It is an error to redefine a state that has already been
       *  processed.)
       */

      state_finder = state_number_map.insert(state_mapping);

      if (!state_finder.second)  /* no insertion occurred */
      {
        current_state = (state_finder.first)->second;
        if (processed_states.test(current_state))
          throw AutomatonParseException("State redefinition encountered.");
      }
      else
      {
        if (state_mapping.second >= number_of_states)
          throw AutomatonParseException("Number of different state "
                                        "identifiers does not match the size"
                                        " of the Bchi automaton.");

        current_state = state_mapping.second;
        state_mapping.second++;
      }

      /* 
       *  Check whether the current state is an initial state. (There must be
       *  exactly one initial state.)
       */

      einput_stream >> is_initial;

      if (is_initial != 0)
      {
        if (!initial_state_fixed)
        {
          initial_state = current_state;
          initial_state_fixed = true;
        }
        else
          throw AutomatonParseException("Multiple initial state definitions.");
      }

      /*
       *  Determine which acceptance sets the current state belongs to.
       *  The numbers of the acceptance sets are mapped to the proper
       *  interval, again by introducing mappings for acceptance set numbers
       *  whenever necessary.
       */

      operator[](current_state).acceptanceSets().clear();

      while (1)
      {
        einput_stream >> acceptance_set_mapping.first;

        if (acceptance_set_mapping.first == -1)
          break;

        acceptance_set_finder =
          acceptance_set_map.insert(acceptance_set_mapping);

        if (!acceptance_set_finder.second)
          acceptance_set = (acceptance_set_finder.first)->second;
        else
        {
          if (acceptance_set_mapping.second >= number_of_acceptance_sets)
            throw AutomatonParseException("Number of acceptance sets "
                                          "does not match automaton state "
                                          "definitions.");

          acceptance_set = acceptance_set_mapping.second;
          ++acceptance_set_mapping.second;
        }

	operator[](current_state).acceptanceSets().set(acceptance_set);
      }

      /*
       *  Process the transitions from the state to other states. Read a
       *  target state id and add a mapping for it in the translation table if
       *  necessary. Then, read the propositional formula guarding the
       *  transition and connect the current state to the target using the
       *  guard formula.
       */

      while (1)
      {
        einput_stream >> state_mapping.first;

        if (state_mapping.first == -1)
          break;

        state_finder = state_number_map.insert(state_mapping);
        if (!state_finder.second)
          neighbor_state = (state_finder.first)->second;
        else
        {
          if (state_mapping.second >= number_of_states)
            throw AutomatonParseException("Number of different state "
                                          "identifiers does not match the size"
                                          " of the Bchi automaton.");

          neighbor_state = state_mapping.second;
          state_mapping.second++;
        }

        try
        {
          FormulaString guard_formula;

          input_stream >> guard_formula;

          guard = formulaStringToLtlFormula(guard_formula);

          if (!guard->propositional())
          {
            delete guard;
            throw AutomatonParseException("Illegal operators in guard "
                                          "formula.");
          }

	  operator[](current_state).connectTo(neighbor_state, guard);
        }
        catch (FormulaString::ParseErrorException&)
        {
          throw AutomatonParseException("Error parsing guard formula.");
        }
      }

      processed_states.set(current_state);
    }

    if (!initial_state_fixed)
      throw AutomatonParseException("No initial state specified.");
    
    if (processed_states.count() != number_of_states)
      throw AutomatonParseException("Incomplete automaton definition.");
  }
  catch (IOException& e)
  {
    clear();
    if (input_stream.eof())
      throw AutomatonParseException("Unexpected end of input.");
    else
      throw;
  }
  catch (...)
  {
    clear();
    throw;
  }
}

/* ========================================================================= */
BuchiAutomaton* BuchiAutomaton::intersect
  (const BuchiAutomaton& a1, const BuchiAutomaton& a2,
   map<size_type, StateIdPair, less<size_type>, ALLOC(StateIdPair) >*
     intersection_state_mapping)
/* ----------------------------------------------------------------------------
 *
 * Description:   Computes the intersection of two Bchi automata and returns
 *                a pointer to the intersection of the two automata.
 *
 * Arguments:     a1, a2                      --  References to two constant
 *                                                Bchi automata.
 *                intersection_state_mapping  --  An (optional) pointer to a
 *                                                map which can be used to find
 *                                                out the state identifiers of
 *                                                the original automata to
 *                                                which a particular state in
 *                                                the intersection corresponds.
 *
 * Returns:       A newly allocated BuchiAutomaton representing the
 *                intersection of the two automata.
 *
 * ------------------------------------------------------------------------- */
{
  if (intersection_state_mapping != 0)
    intersection_state_mapping->clear();

  /*
   *  If either of the original automata is empty, the intersection is also
   *  empty.
   */

  if (a1.empty() || a2.empty())
    return new BuchiAutomaton(0, 0, 0);

  BuchiAutomaton* automaton;

  /*
   *  Determine the number of acceptance sets in the intersection.
   */

  bool a1_has_no_acceptance_sets = (a1.number_of_acceptance_sets == 0);
  bool a2_has_no_acceptance_sets = (a2.number_of_acceptance_sets == 0);

  unsigned long int number_of_intersection_acceptance_sets;

  if (a1_has_no_acceptance_sets && a2_has_no_acceptance_sets)
    number_of_intersection_acceptance_sets = 0;
  else
    number_of_intersection_acceptance_sets = a1.number_of_acceptance_sets
                                             + a2.number_of_acceptance_sets;

  automaton = new BuchiAutomaton(1, 0, number_of_intersection_acceptance_sets);

  /*
   *  A stack for processing pairs of states of the original automata.
   */

  stack<const StateIdPair*, deque<const StateIdPair*,
        ALLOC(const StateIdPair*) > >
    unprocessed_states;

  /*
   *  `state_mapping' maps pairs of states of the original automata to the
   *  states of the new automaton.
   */

  map<StateIdPair, size_type, less<StateIdPair>, ALLOC(size_type) >
    state_mapping;

  size_type first_free_id = 1;                /* First free identifier for a
					       * new state in the intersection
					       * automaton.
					       */

  const StateIdPair* state_pair;              /* Pointer to pair of two state
					       * identifiers of the original
					       * automata.
					       */

  size_type intersect_state;                  /* `Current' state in the
					       * intersection automaton.
					       */

  bool intersect_state_valid;                 /* True if the current state has
					       * been determined by using the
					       * mapping.
					       */

  Bitset* intersect_acceptance_sets;          /* Pointers for accessing the */
  const Bitset* acceptance_sets;              /* acceptance sets of the new
					       * and the original automata.
					       */

  const GraphEdgeContainer* transitions1;     /* Pointers for accessing the */
  const GraphEdgeContainer* transitions2;     /* transitions of the two
					       * original automata.
					       */

  const LtlFormula* guard1;                   /* Pointers for accessing the */
  const LtlFormula* guard2;                   /* guard formulas of the      */
  LtlFormula* new_guard;                      /* transitions of the
					       * automata.
					       */

  /*
   *  Insert the initial state into the state mapping.
   */

  state_mapping.insert(make_pair(make_pair(a1.initial_state, a2.initial_state),
				 0));
  unprocessed_states.push(&(state_mapping.begin()->first));

  /*
   *  Adjust the acceptance sets of the initial state of the intersection.
   */

  intersect_acceptance_sets = &((*automaton)[0].acceptanceSets());

  if (!a1_has_no_acceptance_sets)
  {
    acceptance_sets = &(a1[a1.initial_state].acceptanceSets());

    for (unsigned long int accept_set = 0;
	 accept_set < a1.number_of_acceptance_sets;
	 accept_set++)
    {
      if (acceptance_sets->test(accept_set))
	intersect_acceptance_sets->set(accept_set);
    }
  }

  if (!a2_has_no_acceptance_sets)
  {
    acceptance_sets = &(a2[a2.initial_state].acceptanceSets());

    for (unsigned long int accept_set = 0;
	 accept_set < a2.number_of_acceptance_sets;
	 accept_set++)
    {
      if (acceptance_sets->test(accept_set))
	intersect_acceptance_sets->set(a1.number_of_acceptance_sets
				       + accept_set);
    }
  }

  /*
   *  Pop pairs of states of the two original automata until all states have
   *  been processed.
   */

  while (!unprocessed_states.empty())
  {
    if (::user_break)
    {
      delete automaton;
      throw UserBreakException();
    }

    intersect_state_valid = false;
    state_pair = unprocessed_states.top();
    unprocessed_states.pop();

    /*
     *  Loop through the transitions of the two original automata. If the
     *  conjunction of the guard formulae of any two transitions is
     *  satisfiable, insert a new transition into the intersection automaton.
     *  Create new states in the intersection automaton if necessary.
     */

    transitions1 = &a1[state_pair->first].edges();
    transitions2 = &a2[state_pair->second].edges();

    for (GraphEdgeContainer::const_iterator tr1 = transitions1->begin();
	 tr1 != transitions1->end();
	 ++tr1)
    {
      guard1 = &(static_cast<const BuchiTransition*>(*tr1)->guard());

      for (GraphEdgeContainer::const_iterator tr2 = transitions2->begin();
	   tr2 != transitions2->end();
	   ++tr2)
      {
	guard2 = &(static_cast<const BuchiTransition*>(*tr2)->guard());

        new_guard = new And(*guard1, *guard2);

	if (new_guard->satisfiable())
	{
	  /*
	   *  Determine the `current' state of the intersection automaton.
	   */

	  if (!intersect_state_valid)
	  {
	    intersect_state = state_mapping[*state_pair];
	    intersect_state_valid = true;
	  }

	  /*
	   *  Test whether the state pair pointed to by the two transitions
	   *  is already in the intersection.
	   */

	  pair<map<StateIdPair, size_type, less<StateIdPair>,
                   ALLOC(size_type) >::iterator,
               bool>
	    check_state;

	  check_state
	    = state_mapping.insert(make_pair(make_pair((*tr1)->targetNode(),
						       (*tr2)->targetNode()),
					     first_free_id));

	  if (check_state.second)  /* insertion occurred? */
	  {
	    automaton->expand();

	    /*
	     *  Determine the acceptance sets to which the new state in the
	     *  intersection automaton belongs.
	     */

	    intersect_acceptance_sets
	      = &((*automaton)[first_free_id].acceptanceSets());

	    if (!a1_has_no_acceptance_sets)
	    {
	      acceptance_sets = &(a1[check_state.first->first.first].
				    acceptanceSets());

	      for (unsigned long int accept_set = 0;
		   accept_set < a1.number_of_acceptance_sets;
		   accept_set++)
	      {
		if (acceptance_sets->test(accept_set))
		  intersect_acceptance_sets->set(accept_set);
	      }
	    }

	    if (!a2_has_no_acceptance_sets)
	    {
	      acceptance_sets = &(a2[check_state.first->first.second].
				    acceptanceSets());

	      for (unsigned long int accept_set = 0;
		   accept_set < a2.number_of_acceptance_sets;
		   accept_set++)
	      {
		if (acceptance_sets->test(accept_set))
		  intersect_acceptance_sets->set(a1.number_of_acceptance_sets
						 + accept_set);
	      }
	    }

	    /*
	     *  Connect the `current' state of the intersection automaton to
	     *  the new state.
	     */

	    (*automaton)[intersect_state].connectTo(first_free_id, new_guard);
	    first_free_id++;
	    unprocessed_states.push(&(check_state.first->first));
	  }
	  else
	    (*automaton)[intersect_state].connectTo(check_state.first->second,
						    new_guard);

	  intersect_state_valid = false;
	}
	else
	  delete new_guard;
      }
    }
  }

  if (intersection_state_mapping != 0)
  {
    for (map<StateIdPair, size_type, less<StateIdPair>, ALLOC(size_type) >
	   ::const_iterator mapping = state_mapping.begin();
	 mapping != state_mapping.end();
	 ++mapping)
      intersection_state_mapping->insert(make_pair(mapping->second,
						   mapping->first));
  }

  return automaton;
}

/* ========================================================================= */
void BuchiAutomaton::print
  (ostream& stream, int indent, GraphOutputFormat fmt) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Writes information about a BuchiAutomaton to a stream.
 *
 * Arguments:     stream  --  A reference to an output stream.
 *                indent  --  Number of spaces to leave to the left of output.
 *                fmt     --  Determines the format of the output.
 * 
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);

  if (fmt == DOT)
    estream << string(indent, ' ') + "digraph G {\n";

  if (nodes.empty())
  {
    if (fmt == NORMAL)
      estream << string(indent, ' ') + "The Bchi automaton is empty.\n";
  }
  else
  {
    if (fmt == NORMAL)
    {
      pair<size_type, unsigned long int> statistics = stats();
      pair<size_type, unsigned long int> reachable_part_statistics
        = subgraphStats(initial_state);

      estream << string(indent, ' ') + "The Bchi automaton consists of\n"
                 + string(indent + 4, ' ')
              << statistics.first
              << " states and\n" + string(indent + 4, ' ')
              << statistics.second
              << " transitions.\n" + string(indent, ' ')
                 + "The automaton has "
	      << number_of_acceptance_sets
	      << " sets of accepting states.\n" + string(indent, ' ')
                 + "The reachable part of the automaton contains\n"
                 + string(indent + 4, ' ')
              << reachable_part_statistics.first
              << " states and\n" + string(indent + 4, ' ')
              << reachable_part_statistics.second
              << " transitions.\n" + string(indent, ' ') + "Initial state: "
              << initial_state
              << '\n';
    }
    else if (fmt == DOT)
    {
      estream << string(indent, ' ') + "  init [style=invis];\n"
                 + string(indent, ' ') + "  init->n"
	      << initial_state
	      << ";\n";
    }

    size_type s = nodes.size();
    for (size_type state = 0; state < s; ++state)
    {
      estream << string(indent, ' ');
      if (fmt == NORMAL)
      {
        estream << "State " << state << ":\n";
        operator[](state).print(stream, indent + 4);
      }
      else if (fmt == DOT)
      {
        GraphEdgeContainer::const_iterator transition;
	bool first_printed = false;

        estream << "  n"
		<< state
		<< string(" [shape=circle,label=\"")
                << state
                << "\\n{";

	for (unsigned long int accept_set = 0;
	     accept_set < number_of_acceptance_sets;
	     accept_set++)
	{
	  if (operator[](state).acceptanceSets().test(accept_set))
	  {
	    if (first_printed)
	      estream << ',';
	    else
	      first_printed = true;

	    estream << accept_set;
	  }
	}

	estream << "}\",fontsize=12];\n";

        for (transition = nodes[state]->edges().begin();
             transition != nodes[state]->edges().end();
             ++transition)
        {
          estream << string(indent + 2, ' ') + 'n' << state;
          (*transition)->print(stream, indent, fmt);
          estream << ";\n";
        }
      }
    }
  }

  if (fmt == DOT)
    estream << string(indent, ' ') + "}\n";

  estream.flush();
}



/******************************************************************************
 *
 * Function definitions for class BuchiAutomaton::BuchiTransition.
 *
 *****************************************************************************/

/* ========================================================================= */
BuchiAutomaton::BuchiTransition& BuchiAutomaton::BuchiTransition::operator=
  (const BuchiTransition& transition)
/* ----------------------------------------------------------------------------
 *
 * Description:   Assignment operator for class
 *                BuchiAutomaton::BuchiTransition. Assigns the value of a
 *                BuchiTransition to another one.
 *
 * Arguments:     transition  --  BuchiTransition to be assigned to "this"
 *                                transition.
 *
 * Returns:       A reference to the BuchiTransition assigned to.
 *
 * ------------------------------------------------------------------------- */
{
  if (&transition != this)
  {
    delete guard_formula;
    guard_formula = transition.guard_formula->clone();
    Edge::operator=(transition);
  }

  return *this;
}

/* ========================================================================= */
void BuchiAutomaton::BuchiTransition::print
  (ostream& stream, int indent, GraphOutputFormat fmt) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Writes information about a transition between two states of
 *                a Bchi automaton.
 *
 * Arguments:     stream  --  A reference to an output stream.
 *                indent  --  Number of spaces to leave to the left of output.
 *                fmt     --  Determines the format of output.
 * 
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);

  if (fmt == NORMAL)
  {
    estream << string(indent, ' ') + "Transition to state "
            << targetNode()
            << " [ guard: " << *guard_formula << " ]\n";
  }
  else if (fmt == DOT)
  {
    ostrstream formula_stream;
    formula_stream << *guard_formula << ends;
    formula_stream.freeze();
    char* formula = formula_stream.str();

    estream << " -> n" << targetNode() << " [label=\"";
    for (unsigned long int i = 0; formula[i] != '\0'; ++i)
    {
      if (formula[i] == '/')
      {
        estream << "&&";
        i++;
      }
      else if (formula[i] == '\\')
      {
        estream << "||";
        i++;
      }
      else
        estream << formula[i];
    }
    estream << "\",fontsize=10,fontname=\"Courier-Bold\"]";
    formula_stream.freeze(0);
  }

  estream.flush();
}



/******************************************************************************
 *
 * Function definitions for class BuchiAutomaton::BuchiState.
 *
 *****************************************************************************/

/* ========================================================================= */
void BuchiAutomaton::BuchiState::print
  (ostream& stream, int indent, GraphOutputFormat fmt) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Writes information about a state of a Bchi automaton.
 *
 * Arguments:     stream  --  A reference to an output stream.
 *                indent  --  Number of spaces to leave to the left of output.
 *                fmt     --  Determines the output format of the state.
 * 
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------ */
{
  if (fmt == DOT)
    return;

  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);

  estream << string(indent, ' ') + "Member of acceptance sets {";

  bool first_printed = false;

  for (unsigned long int accept_set = 0; accept_set < acceptance_sets.size();
       ++accept_set)
  {
    if (acceptance_sets[accept_set])
    {
      if (first_printed)
	estream << ", ";
      else
	first_printed = true;

      estream << accept_set;
    }
  }

  estream << "}\n";

  if (!edges().empty())
  {
    GraphEdgeContainer::const_iterator edge;

    for (edge = edges().begin(); edge != edges().end(); ++edge)
      (*edge)->print(stream, indent);

  } else
    estream << string(indent, ' ') + "No transitions to other states.\n";

  estream.flush();
}

}
