/*
 *  Copyright (C) 1999, 2000  Heikki Tauriainen <Heikki.Tauriainen@hut.fi>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef _BuchiAutomaton_h_
#define _BuchiAutomaton_h_

#include <deque>
#include <iostream>
#include <map>
#include <string>
#include <stack>
#include <set>
#include <utility>
#include <vector>
#include "Bitset.h"
#include "EdgeContainer.h"
#include "Exception.h"
#include "FormulaString.h"
#include "Graph.h"
#include "LtlFormula.h"

using namespace std;

namespace Graph
{

using LtlFormula::True;
using LtlFormula::And;
using LtlFormula::formulaStringToLtlFormula;
using LtlFormula::LtlFormula;
using FormulaString::FormulaString;

/******************************************************************************
 *
 * A class for representing Bchi automata with a single set of accepting
 * states.
 *
 *****************************************************************************/

class BuchiAutomaton : public Graph<GraphEdgeContainer>
{
private:

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

  class BuchiTransition;                            /* A class for representing
						     * the transitions between
						     * the states of the
						     * automaton.
						     */

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

  class BuchiState :                                /* A class for      */
    public Graph<GraphEdgeContainer>::Node          /* representing the */
  {                                                 /* states of the
                                                     * automaton.
						     */
  public:
    explicit BuchiState                             /* Constructor. */
      (unsigned long int num_of_acceptance_sets);

    /* default copy constructor */

    ~BuchiState();                                  /* Destructor. */

    /* default assignment operator */
  
    BuchiState* clone() const;                      /* Creates a copy of the
                                                     * state.
                                                     */

    /* `edges' inherited from Graph<EdgeContainer>::Node */
    
    void connectTo                                  /* Connects the state */
      (size_type state, const LtlFormula& guard);   /* to another state.  */ 
    void connectTo
      (size_type state, LtlFormula* guard);

    /* `disconnectFrom' inherited from Graph<EdgeContainer>::Node */

    /* `connected' inherited from Graph<EdgeContainer>::Node */

    Bitset& acceptanceSets();                       /* Tells the acceptance
						     * status of the state.
						     */
    
    void print                                      /* Writes information   */
      (ostream& stream = cout,                      /* about the state to a */
       int indent = 0,                              /* stream.              */
       GraphOutputFormat fmt = NORMAL)
      const;     

  private:
    Bitset acceptance_sets;                         /* Acceptance status of the
						     * state.
						     */
  };

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

public:
  explicit BuchiAutomaton                           /* Constructor. */
    (size_type initial_number_of_states = 0,
     size_type initstate = 0,
     unsigned long int number_of_accept_sets = 0);

  /* default copy constructor */

  ~BuchiAutomaton();                                /* Destructor. */

  BuchiAutomaton&                                   /* Assignment operator. */
    operator=(const BuchiAutomaton& automaton);

  BuchiState& operator[](size_type index) const;    /* Indexing operator. No
						     * range check is
						     * performed on the
						     * argument.
						     */

  BuchiState& node(size_type index) const;          /* Synonym for the indexing
                                                     * operator. This function
						     * also checks the range of
						     * the argument.
                                                     */

  /* `size' inherited from Graph<EdgeContainer> */

  /* `empty' inherited from Graph<EdgeContainer> */

  void clear();                                     /* Makes the automaton
						     * empty.
						     */

  size_type expand(size_type node_count = 1);       /* Inserts states to the
                                                     * automaton.
                                                     */

  void connect                                      /* Connects two states   */
    (size_type father, size_type child,             /* of the automaton with */
     const LtlFormula& guard = True());             /* a transition.         */

  /* `disconnect' inherited from Graph<EdgeContainer> */

  /* `connected' inherited from Graph<EdgeContainer> */

  /* `stats' inherited from Graph<EdgeContainer> */

  /* `subgraphStats' inherited from Graph<EdgeContainer> */

  size_type initialState() const;                   /* Get or set the       */
  size_type& initialState();                        /* initial state of the *
                                                     * automaton.
                                                     */

  unsigned long int numberOfAcceptanceSets() const; /* Returns the number of
						     * acceptance sets in the
						     * automaton.
						     */

  void regularize();                                /* Converts a generalized
                                                     * automaton to an
                                                     * automaton with one set
						     * of accepting states.
                                                     */

  void read(istream& input_stream);                 /* Reads the automaton
                                                     * from a stream.
                                                     */

  void print                                        /* Writes information  */
    (ostream& stream = cout,                        /* about the automaton */
     int indent = 0,                                /* to a stream in      */
     GraphOutputFormat fmt = NORMAL) const;         /* various formats
						     * (determined by the
						     * `fmt' argument).
						     */

  static BuchiAutomaton* intersect                  /* Computes the        */
    (const BuchiAutomaton& a1,                      /* intersection of two */
     const BuchiAutomaton& a2,                      /* Bchi automata.     */
     map<size_type, StateIdPair,
         less<size_type>, ALLOC(StateIdPair) >*
       intersection_state_mapping = 0);

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

  class AutomatonParseException;                    /* Class for reporting
						     * parse errors when
						     * reading an automaton
						     * description from a
						     * stream.
						     */

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

private:
  size_type initial_state;                          /* Identifier of the
                                                     * initial state of the
                                                     * automaton.
                                                     */

  unsigned long int number_of_acceptance_sets;      /* Number of acceptance
						     * sets in the automaton.
						     */
};



/******************************************************************************
 *
 * A class for representing the transitions of a Bchi automaton.
 *
 *****************************************************************************/

class BuchiAutomaton::BuchiTransition : public Graph<GraphEdgeContainer>::Edge
{
public:
  BuchiTransition                                   /* Constructor. */
    (size_type target, LtlFormula* formula);

  BuchiTransition                                   /* Copy constructor. */
    (const BuchiTransition& transition);

  ~BuchiTransition();                               /* Destructor. */

  bool operator==                                   /* Equality test between */
    (const BuchiTransition& transition);            /* two BuchiTransitions. */

  bool operator<                                    /* `Less than' relation */
    (const BuchiTransition& transition);            /* between two
						     * BuchiTransitions.
						     */

  BuchiTransition& operator=                        /* Assignment operator. */
    (const BuchiTransition& transition);

  BuchiTransition* clone() const;                   /* Creates a copy of the
                                                     * transition.
                                                     */

  /* `targetNode' inherited from Graph<EdgeContainer>::Edge */
    
  bool enabled                                      /* Tests whether the     */
    (const Bitset& truth_assignment) const;         /* transition is enabled
                                                     * in a given truth
                                                     * assignment for
                                                     * propositional
                                                     * variables.
                                                     */

  const LtlFormula& guard() const;                  /* Returns the
                                                     * propositional formula
                                                     * guarding the transition.
                                                     */
    
  void print                                        /* Writes information   */
    (ostream& stream = cout,                        /* about the transition */
     int indent = 0,                                /* to a stream in       */
     GraphOutputFormat fmt = NORMAL) const;         /* various formats
						     * (determined by the
						     * `fmt' argument).
						     */

private:
  const LtlFormula* guard_formula;                  /* The propositional
                                                     * formula guarding the
                                                     * transition.
                                                     */
};



/******************************************************************************
 *
 * A class for reporting parse errors when reading an automaton description
 * from a stream.
 *
 *****************************************************************************/

class BuchiAutomaton::AutomatonParseException : public Exception
{
public:
  AutomatonParseException                           /* Constructor. */
    (const string& msg = "Parse error.");

  /* default copy constructor */

  ~AutomatonParseException();                       /* Destructor. */

  AutomatonParseException&                          /* Assignment operator. */
    operator=(const AutomatonParseException& e);

  /* `what' inherited from class Exception */
};



/******************************************************************************
 *
 * Inline function definitions for class BuchiAutomaton.
 *
 *****************************************************************************/

/* ========================================================================= */
inline BuchiAutomaton::~BuchiAutomaton()
/* ----------------------------------------------------------------------------
 *
 * Description:   Destructor for class BuchiAutomaton.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline BuchiAutomaton::BuchiState& BuchiAutomaton::operator[]
  (size_type index) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Indexing operator for class BuchiAutomaton. This function can
 *                be used to refer to the individual states of the automaton.
 *                No range check will be performed on the argument.
 *
 * Argument:      index  --  Index of a state.
 *
 * Returns:       A reference to a state of the automaton.
 *
 * ------------------------------------------------------------------------- */
{
  return static_cast<BuchiState&>(*nodes[index]);
}

/* ========================================================================= */
inline BuchiAutomaton::BuchiState& BuchiAutomaton::node(size_type index) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Function for referring to a single state of a BuchiAutomaton.
 *                This function will perform a range check on the argument.
 *
 * Argument:      index  --  Index of a state.
 *
 * Returns:       A reference to a state of the automaton.
 *
 * ------------------------------------------------------------------------- */
{
  return static_cast<BuchiState&>(Graph<GraphEdgeContainer>::node(index));
}

/* ========================================================================= */
inline void BuchiAutomaton::connect
  (size_type father, size_type child, const LtlFormula& guard)
/* ----------------------------------------------------------------------------
 *
 * Description:   Connects two states of a BuchiAutomaton to each other, using
 *                a LtlFormula (which is actually a propositional formula) to
 *                guard the transition between the states. This function will
 *                check the range of the arguments.
 *
 * Arguments:     father  --  Source state.
 *                child   --  Target state.
 *                guard   --  A reference to a constant LtlFormula (a
 *                            propositional formula) guarding the transition.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  if (child >= nodes.size())
    throw NodeIndexException();

  node(father).connectTo(child, guard);
}

/* ========================================================================= */
inline BuchiAutomaton::size_type BuchiAutomaton::initialState() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Returns the initial state of the BuchiAutomaton by value.
 *
 * Arguments:     None.
 *
 * Returns:       Index of the initial state of the automaton.
 *
 * ------------------------------------------------------------------------- */
{
  return initial_state;
}

/* ========================================================================= */
inline BuchiAutomaton::size_type& BuchiAutomaton::initialState()
/* ----------------------------------------------------------------------------
 *
 * Description:   Returns the initial state of the BuchiAutomaton by reference.
 *                This function can therefore be used to change the initial
 *                state.
 *
 * Arguments:     None.
 *
 * Returns:       A reference to the value of the initial state.
 *
 * ------------------------------------------------------------------------- */
{
  return initial_state;
}

/* ========================================================================= */
inline unsigned long int BuchiAutomaton::numberOfAcceptanceSets() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Returns the number of acceptance sets in the automaton.
 *
 * Arguments:     None.
 *
 * Returns:       The number of acceptance sets in the automaton.
 *
 * ------------------------------------------------------------------------- */
{
  return number_of_acceptance_sets;
}

/* ========================================================================= */
inline istream& operator>>(istream& stream, BuchiAutomaton& automaton)
/* ----------------------------------------------------------------------------
 *
 * Description:   Defines an alternative method for reading an automaton from
 *                a stream by using the >> operator.
 *
 * Arguments:     stream     --  A reference to an input stream.
 *                automaton  --  A reference to the BuchiAutomaton.
 *
 * Returns:       A reference to the input stream.
 *
 * ------------------------------------------------------------------------- */
{
  automaton.read(stream);
  return stream;
}



/******************************************************************************
 *
 * Inline function definitions for class BuchiAutomaton::BuchiTransition.
 *
 *****************************************************************************/

/* ========================================================================= */
inline BuchiAutomaton::BuchiTransition::BuchiTransition
  (size_type target, LtlFormula* formula) :
  Edge(target), guard_formula(formula)
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructor for class BuchiAutomaton::BuchiTransition.
 *                Initializes a new transition to a BuchiState, guarded by a
 *                LtlFormula (which is actually a propositional formula).
 *                
 * Arguments:     target   --  Identifier of the target state of the automaton.
 *                formula  --  A pointer to a propositional formula guarding
 *                             the transition.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline BuchiAutomaton::BuchiTransition::BuchiTransition
  (const BuchiTransition& transition) :
  Edge(transition), guard_formula(transition.guard_formula->clone())
/* ----------------------------------------------------------------------------
 *
 * Description:   Copy constructor for class BuchiAutomaton::BuchiTransition.
 *                Creates a copy of a BuchiTransition object.
 *
 * Arguments:     transition  --  BuchiTransition to be copied.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline BuchiAutomaton::BuchiTransition::~BuchiTransition()
/* ----------------------------------------------------------------------------
 * 
 * Description:   Destructor for class BuchiAutomaton::BuchiTransition.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  delete guard_formula;
}

/* ========================================================================= */
inline bool BuchiAutomaton::BuchiTransition::operator==
  (const BuchiTransition& transition)
/* ----------------------------------------------------------------------------
 *
 * Description:   Equality relation for class BuchiTransition. Two
 *                transitions are `equal' if and only if their target nodes
 *                have the same identifier and if their guard formulae are
 *                syntactically identical.
 *
 * Argument:      transition  --  A reference to a constant BuchiTransition.
 *
 * Returns:       Truth value according to the relationship between the two
 *                transitions.
 *
 * ------------------------------------------------------------------------- */
{
  return (Edge::operator==(transition)
	  && string(*guard_formula) == string(*(transition.guard_formula)));
}

/* ========================================================================= */
inline bool BuchiAutomaton::BuchiTransition::operator<
  (const BuchiTransition& transition)
/* ----------------------------------------------------------------------------
 *
 * Description:   `Less than' relation for class BuchiTransition. A
 *                BuchiTransitions is `less than' another if and only if the
 *                identifier of its target node is less than that of the other
 *                or the target nodes agree but the guard formula of the
 *                first transition is lexicographically `less than' the
 *                other.
 *
 * Argument:      transition  --  A reference to a constant BuchiTransition.
 *
 * Returns:       Truth value according to the relationship between the two
 *                transitions.
 *
 * ------------------------------------------------------------------------- */
{
  return (Edge::operator<(transition)
	  || (Edge::operator==(transition)
	      && string(*guard_formula)
	           < string(*(transition.guard_formula))));
}

/* ========================================================================= */
inline BuchiAutomaton::BuchiTransition*
BuchiAutomaton::BuchiTransition::clone() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Creates an identical copy of a BuchiTransition.
 *
 * Arguments:     None.
 *
 * Returns:       A pointer to the copy.
 *
 * ------------------------------------------------------------------------- */
{
  return new BuchiTransition(*this);
}

/* ========================================================================= */
inline bool BuchiAutomaton::BuchiTransition::enabled
  (const Bitset& truth_assignment) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Determines whether the transition is enabled in a given
 *                truth assignment for propositional variables.
 *
 * Arguments:     truth_assignment  --  A reference to a constant Bitset
 *                                      representing an assignment of truth
 *                                      values to propositional variables.
 *
 * Returns:       A truth value telling whether the transition is enabled in
 *                the truth assignment.
 *
 * ------------------------------------------------------------------------- */
{ 
  return guard_formula->evaluate(truth_assignment);
}

/* ========================================================================= */
inline const LtlFormula& BuchiAutomaton::BuchiTransition::guard() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Returns the propositional formula guarding the transition
 *                (a LtlFormula object).
 *
 * Arguments:     None.
 *
 * Returns:       A reference to the constant propositional formula guarding
 *                the transition.
 *
 * ------------------------------------------------------------------------- */
{
  return *guard_formula;
}



/******************************************************************************
 *
 * Inline function definitions for class BuchiAutomaton::BuchiState.
 *
 *****************************************************************************/

/* ========================================================================= */
inline BuchiAutomaton::BuchiState::BuchiState
  (unsigned long int num_of_acceptance_sets) : 
  Node(), acceptance_sets(num_of_acceptance_sets)
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructor for class BuchiAutomaton::BuchiState. Initializes
 *                a new state for the automaton.
 *
 * Argument:      num_of_acceptance_sets  --  Number of acceptance sets in the
 *                                            automaton to which the state
 *                                            belongs.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline BuchiAutomaton::BuchiState::~BuchiState()
/* ----------------------------------------------------------------------------
 *
 * Description:   Destructor for class BuchiAutomaton::BuchiState.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline BuchiAutomaton::BuchiState* BuchiAutomaton::BuchiState::clone() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Creates an identical copy of a BuchiState.
 *
 * Arguments:     None.
 *
 * Returns:       A pointer to the copy.
 *
 * ------------------------------------------------------------------------- */
{
  return new BuchiState(*this);
}

/* ========================================================================= */
inline Bitset& BuchiAutomaton::BuchiState::acceptanceSets()
/* ----------------------------------------------------------------------------
 *
 * Description:   Returns the acceptance status of the BuchiState, i.e. the
 *                Bitset indicating the acceptance sets to which the state
 *                belongs.
 *
 * Arguments:     None.
 *
 * Returns:       A truth value according to the acceptance status of the
 *                state.
 *
 * ------------------------------------------------------------------------- */
{
  return acceptance_sets;
}

/* ========================================================================= */
inline void
BuchiAutomaton::BuchiState::connectTo
  (size_type state, const LtlFormula& guard = True())
/* ----------------------------------------------------------------------------
 *
 * Description:   Connects a BuchiState to another, using a given propositional
 *                formula (a LtlFormula object) to guard the transition
 *                between the states.
 *
 * Arguments:     state  --  Identifier of the target state.
 *                guard  --  A reference to a constant LtlFormula object
 *                           (a propositional formula) guarding the transition.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  outgoing_edges.insert(new BuchiTransition(state, guard.clone()));
}

/* ========================================================================= */
inline void
BuchiAutomaton::BuchiState::connectTo
  (size_type state, LtlFormula* guard = new True())
/* ----------------------------------------------------------------------------
 *
 * Description:   Connects a BuchiState to another, using a given propositional
 *                formula (a pointer to a LtlFormula object) to guard the
 *                transition between the states.
 *
 * Arguments:     state  --  Identifier of the target state.
 *                guard  --  A pointer to a LtlFormula object (a propositional
 *                           formula) guarding the transition.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  outgoing_edges.insert(new BuchiTransition(state, guard));
}



/******************************************************************************
 *
 * Inline function definitions for class
 * BuchiAutomaton::AutomatonParseException.
 *
 *****************************************************************************/

/* ========================================================================= */
inline BuchiAutomaton::AutomatonParseException::AutomatonParseException
  (const string& msg) :
  Exception(msg)
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructor for class
 *                BuchiAutomaton::AutomatonParseException. Initializes a new
 *                exception object with an error message.
 *
 * Argument:      msg  --  A reference to a constant string containing the
 *                         error message.
 *
 * Returns:       Nothing.
 * 
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline BuchiAutomaton::AutomatonParseException::~AutomatonParseException()
/* ----------------------------------------------------------------------------
 *
 * Description:   Destructor for class BuchiAutomaton::AutomatonParseException.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline BuchiAutomaton::AutomatonParseException&
BuchiAutomaton::AutomatonParseException::operator=
  (const AutomatonParseException& e)
/* ----------------------------------------------------------------------------
 *
 * Description:   Assignment operator for class
 *                BuchiAutomaton::AutomatonParseException. Assigns the contents
 *                of another object of the same type to the exception object.
 *
 * Argument:      e  --  A reference to a constant exception object of the
 *                       same type.
 *
 * Returns:       A reference to the exception object assigned to.
 *
 * ------------------------------------------------------------------------- */
{
  Exception::operator=(e);
  return *this;
}

}

#endif  /* _BuchiAutomaton_h_ */
