#include "Configuration.h"
#include "Configuration.tab.cc"

/******************************************************************************
 *
 * Function definitions for class Configuration.
 *
 *****************************************************************************/

/* ========================================================================= */
Configuration::Configuration()
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructor for class Configuration. Creates a new data
 *                structure for storing program configuration and initializes
 *                the configuration parameters to their default values.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  reset();
}

/* ========================================================================= */
Configuration::~Configuration()
/* ----------------------------------------------------------------------------
 *
 * Description:   Destructor for class Configuration.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  for (vector<AlgorithmInformation, ALLOC(AlgorithmInformation) >
	 ::const_iterator it = algorithms.begin();
       it != algorithms.end(); ++it)
  {
    delete it->name;
    delete it->path_to_program;
    delete it->extra_parameters;
  }
}

/* ========================================================================= */
void Configuration::read(int argc, char* argv[])
/* ----------------------------------------------------------------------------
 *
 * Description:   Reads the program configuration from a file.
 *
 * Arguments:     argc  --  Number of program command line parameters.
 *                argv  --  An array of C-style strings for the command line
 *                          parameters.
 *                These parameters are used for determining whether the
 *                configuration should be read from a different file than
 *                `config' (the default).
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  reset();

  /*
   *  Initialize the default operator priorities.
   */

  default_operator_priority = 10;
  symbol_priority.clear();
  symbol_priority[RandomFormulaParameters::ATOM] = 90;
  symbol_priority[RandomFormulaParameters::TRUE] = 5;
  symbol_priority[RandomFormulaParameters::FALSE] = 5;
  symbol_priority[RandomFormulaParameters::AND] = -1;
  symbol_priority[RandomFormulaParameters::OR] = -1;
  symbol_priority[RandomFormulaParameters::UNTIL] = -1;
  symbol_priority[RandomFormulaParameters::V] = -1;
  symbol_priority[RandomFormulaParameters::IMPLICATION] = -1;
  symbol_priority[RandomFormulaParameters::EQUIVALENCE] = -1;
  symbol_priority[RandomFormulaParameters::NOT] = -1;
  symbol_priority[RandomFormulaParameters::NEXT] = -1;
  symbol_priority[RandomFormulaParameters::FINALLY] = -1;
  symbol_priority[RandomFormulaParameters::GLOBALLY] = -1;

  /*
   *  Determine the name of the configuration file.
   */

  string filename = "config";

  for (int argument = argc - 1; argument > 0; argument--)
  {
    if (string(argv[argument]).substr(0, 13) == "--configfile=")
    {
      filename = string(argv[argument] + 13);
      break;
    }
    else if (string(argv[argument]).substr(0, 6) == "--help"
	     || string(argv[argument]).substr(0,3) == "-h")
    {
      showCommandLineHelp(argv[0]);
      exit(0);
    }
  }

  /*
   *  Read the configuration.
   */

  if ((yyin = fopen(filename.c_str(), "r")) == NULL)
    throw ConfigurationException(-1, "Error opening `"
                                     + string(filename) + "'");

  try
  {
    yyparse();
    fclose(yyin);
  }
  catch (ConfigurationException&)
  {
    fclose(yyin);
    throw;
  }
}

/* ========================================================================= */
void Configuration::readCommandLineParameters(int argc, char* argv[])
/* ----------------------------------------------------------------------------
 *
 * Description:   Parses the command line parameters passed to the program.
 *
 * Arguments:     argc  --  (Number of command line parameters) + 1.
 *                argv  --  Array of C-style strings storing the parameters.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  if (argc < 2)
    return;

  map<string, ParamInfo, less<string>, ALLOC(ParamInfo) >
    recognized_parameters;

  recognized_parameters["--configfile"] = ParamInfo(false, CCONFIGFILE, OTHER);
  recognized_parameters["--verbosity"] = ParamInfo(false, CVERBOSITY, OTHER);
  recognized_parameters["--interactive"]
    = ParamInfo(false, CINTERACTIVE, OTHER);
  recognized_parameters["--nopause"] = ParamInfo(true, CNOPAUSE, OTHER);
  recognized_parameters["--pauseonerror"]
    = ParamInfo(true, CPAUSEONERROR, OTHER);
  recognized_parameters["--pause"] = ParamInfo(true, CPAUSE, OTHER);
  recognized_parameters["--rounds"] = ParamInfo(false, CROUNDS, OTHER);
  recognized_parameters["--statespacechangeinterval"]
    = ParamInfo(false, CSTATESPACECHANGEINTERVAL, OTHER);
  recognized_parameters["--formulachangeinterval"]
    = ParamInfo(false, CFORMULACHANGEINTERVAL, OTHER);
  recognized_parameters["--randomseed"] = ParamInfo(false, CRANDOMSEED, OTHER);
  recognized_parameters["--synchronousproduct"]
    = ParamInfo(false, CSYNCHRONOUSPRODUCT, OTHER);
  recognized_parameters["--globalproduct"]
    = ParamInfo(true, CGLOBALPRODUCT, OTHER);
  recognized_parameters["--localproduct"]
    = ParamInfo(true, CLOCALPRODUCT, OTHER);
  recognized_parameters["--statespacesize"]
    = ParamInfo(false, CSTATESPACESIZE, OTHER);
  recognized_parameters["--statespacepropositions"]
    = ParamInfo(false, CSTATESPACEPROPOSITIONS, PROPOSITIONCOUNT);
  recognized_parameters["--edgeprobability"]
    = ParamInfo(false, CEDGEPROBABILITY, PROBABILITY);
  recognized_parameters["--truthprobability"]
    = ParamInfo(false, CTRUTHPROBABILITY, PROBABILITY);
  recognized_parameters["--statespacegeneratemode"]
    = ParamInfo(false, CSTATESPACEGENERATEMODE, OTHER);
  recognized_parameters["--randomconnectedgraph"]
    = ParamInfo(true, CRANDOMCONNECTEDGRAPH, OTHER);
  recognized_parameters["--randomgraph"]
    = ParamInfo(true, CRANDOMGRAPH, OTHER);
  recognized_parameters["--randompath"] = ParamInfo(true, CRANDOMPATH, OTHER);
  recognized_parameters["--enumeratedpath"]
    = ParamInfo(true, CENUMERATEDPATH, OTHER);
  recognized_parameters["--abbreviatedoperators"]
    = ParamInfo(true, CABBREVIATEDOPERATORS, OTHER);
  recognized_parameters["--noabbreviatedoperators"]
    = ParamInfo(true, CNOABBREVIATEDOPERATORS, OTHER);
  recognized_parameters["--formulaoutputmode"]
    = ParamInfo(false, CFORMULAOUTPUTMODE, OTHER);
  recognized_parameters["--outputnnf"] = ParamInfo(true, COUTPUTNNF, OTHER);
  recognized_parameters["--formulageneratemode"]
    = ParamInfo(false, CFORMULAGENERATEMODE, OTHER);
  recognized_parameters["--generatennf"]
    = ParamInfo(true, CGENERATENNF, OTHER);
  recognized_parameters["--formulapropositions"]
    = ParamInfo(false, CFORMULAPROPOSITIONS, PROPOSITIONCOUNT);
  recognized_parameters["--formulalength"]
    = ParamInfo(false, CFORMULALENGTH, OTHER);
  recognized_parameters["--propositionpriority"]
    = ParamInfo(false, CPROPOSITIONPRIORITY, PRIORITY);
  recognized_parameters["--truepriority"]
    = ParamInfo(false, CTRUEPRIORITY, PRIORITY);
  recognized_parameters["--falsepriority"]
    = ParamInfo(false, CFALSEPRIORITY, PRIORITY);
  recognized_parameters["--defaultoperatorpriority"]
    = ParamInfo(false, CDEFAULTOPERATORPRIORITY, PRIORITY);
  recognized_parameters["--andpriority"]
    = ParamInfo(false, CANDPRIORITY, PRIORITY);
  recognized_parameters["--orpriority"]
    = ParamInfo(false, CORPRIORITY, PRIORITY);
  recognized_parameters["--implicationpriority"]
    = ParamInfo(false, CIMPLICATIONPRIORITY, PRIORITY);
  recognized_parameters["--equivalencepriority"]
    = ParamInfo(false, CEQUIVALENCEPRIORITY, PRIORITY);
  recognized_parameters["--untilpriority"]
    = ParamInfo(false, CUNTILPRIORITY, PRIORITY);
  recognized_parameters["--vpriority"]
    = ParamInfo(false, CVPRIORITY, PRIORITY);
  recognized_parameters["--notpriority"]
    = ParamInfo(false, CNOTPRIORITY, PRIORITY);
  recognized_parameters["--nextpriority"]
    = ParamInfo(false, CNEXTPRIORITY, PRIORITY);
  recognized_parameters["--finallypriority"]
    = ParamInfo(false, CFINALLYPRIORITY, PRIORITY);
  recognized_parameters["--globallypriority"]
    = ParamInfo(false, CGLOBALLYPRIORITY, PRIORITY);
  recognized_parameters["--version"] = ParamInfo(true, CVERSION, OTHER);
  recognized_parameters["--help"]
    = recognized_parameters["-h"]
    = ParamInfo(true, CHELP, OTHER);
  recognized_parameters["--logfile"]
    = ParamInfo(false, CLOGFILE, OTHER);
  recognized_parameters["--formulafile"]
    = ParamInfo(false, CFORMULAFILE, OTHER);
  recognized_parameters["--enable"]
    = ParamInfo(false, CENABLE, OTHER);
  recognized_parameters["--disable"]
    = ParamInfo(false, CDISABLE, OTHER);

  string option, value;
  string::size_type index;
  map<string, ParamInfo>::const_iterator search_option;
  bool error;
  double d_value;
  long int i_value;

  for (int parameter = 1; parameter < argc; parameter++)
  {
    option = argv[parameter];

    error = (option.length() < 1 || option[0] != '-');

    if (!error)
    {
      index = option.find('=');

      value = option.substr(index + 1);
      option = option.substr(0, index);

      search_option = recognized_parameters.find(option);
      error = (search_option == recognized_parameters.end());
    }

    if (error)
      throw ConfigurationException(-1, "Unrecognized command line option (`"
				       + option + "').  Try --help for a "
                                                  "list of available "
                                                  "options.");

    if (search_option->second.standalone && index != string::npos)
      throw ConfigurationException(-1, "No value expected for option `"
				       + option + "'.");
    else if (!search_option->second.standalone
	     && (index == string::npos || value.empty()))
      throw ConfigurationException(-1, "Value missing for option `"
				       + option + "'.");

    switch (search_option->second.value_type)
    {
      case PRIORITY :
	{
	  i_value = parseCommandLineInteger(option, value);

	  checkIntegerRange(i_value, priority_range, false);

	  RandomFormulaParameters::Symbol symbol;

	  if (search_option->second.parameter_type == DEFAULTOPERATORPRIORITY)
	    default_operator_priority = i_value;
	  else
	  {
	    switch (search_option->second.parameter_type)
	    {
              case CPROPOSITIONPRIORITY :
		symbol = RandomFormulaParameters::ATOM;
		break;
              case CTRUEPRIORITY :
		symbol = RandomFormulaParameters::TRUE;
		break;
              case CFALSEPRIORITY :
		symbol = RandomFormulaParameters::FALSE;
		break;
              case CANDPRIORITY :
		symbol = RandomFormulaParameters::AND;
		break;
              case CORPRIORITY :
		symbol = RandomFormulaParameters::OR;
		break;
              case CUNTILPRIORITY :
		symbol = RandomFormulaParameters::UNTIL;
		break;
              case CVPRIORITY :
		symbol = RandomFormulaParameters::V;
		break;
              case CIMPLICATIONPRIORITY :
		symbol = RandomFormulaParameters::IMPLICATION;
		break;
              case CEQUIVALENCEPRIORITY :
		symbol = RandomFormulaParameters::EQUIVALENCE;
		break;
              case CNOTPRIORITY :
		symbol = RandomFormulaParameters::NOT;
		break;
              case CNEXTPRIORITY :
		symbol = RandomFormulaParameters::NEXT;
		break;
              case CFINALLYPRIORITY :
		symbol = RandomFormulaParameters::FINALLY;
		break;
              case CGLOBALLYPRIORITY :
		symbol = RandomFormulaParameters::GLOBALLY;
		break;
              default :
		break;
	    }
	    symbol_priority[symbol] = i_value;
	  }
	}
	break;
	
      case PROBABILITY :
	{
	  char* endptr;
	  d_value = strtod(value.c_str(), &endptr);

	  if (*endptr != '\0')
	    d_value = -1.0;

	  checkProbability(d_value);

	  if (search_option->second.parameter_type == CEDGEPROBABILITY)
	    statespace_options.edge_probability = d_value;
	  else
	    statespace_options.truth_probability = d_value;
	}
	break;

      case PROPOSITIONCOUNT :
	i_value = parseCommandLineInteger(option, value);

	checkIntegerRange(i_value, proposition_count_range, false);

	if (search_option->second.parameter_type == CSTATESPACEPROPOSITIONS)
	  statespace_options.atoms_per_state = i_value;
	else
	  formula_options.generation_parameters.number_of_available_variables
	    = i_value;

	break;

      default :
	switch (search_option->second.parameter_type)
	{
  	  case CVERBOSITY :
	    i_value = parseCommandLineInteger(option, value);
	    checkIntegerRange(i_value, verbosity_range, false);
	    global_options.verbosity = i_value;
	    break;

	  case CINTERACTIVE :
	    if (value == "always")
	      global_options.interactive = ALWAYS;
	    else if (value == "onerror")
	      global_options.interactive = ONERROR;
	    else if (value == "never")
	      global_options.interactive = NEVER;
	    else
	      error = true;
	    break;

  	  case CNOPAUSE :
	    global_options.interactive = NEVER;
	    break;

	  case CPAUSEONERROR :
	    global_options.interactive = ONERROR;
	    break;

	  case CPAUSE :
	    global_options.interactive = ALWAYS;
	    break;

 	  case CROUNDS :
	    i_value = parseCommandLineInteger(option, value);
	    checkIntegerRange(i_value, round_count_range, false);
	    global_options.number_of_rounds = i_value;
	    break;

  	  case CSTATESPACECHANGEINTERVAL :
	    i_value = parseCommandLineInteger(option, value);
	    checkIntegerRange(i_value, generation_range, false);
	    global_options.statespace_change_interval = i_value;
	    break;

	  case CFORMULACHANGEINTERVAL :
	    i_value = parseCommandLineInteger(option, value);
	    checkIntegerRange(i_value, generation_range, false);
	    global_options.formula_change_interval = i_value;
	    break;

	  case CRANDOMSEED :
	    i_value = parseCommandLineInteger(option, value);
	    SRAND(i_value);
	    break;

	  case CSYNCHRONOUSPRODUCT :
	    if (value == "local")
	      global_options.product_mode = LOCAL;
	    else if (value == "global")
	      global_options.product_mode = GLOBAL;
	    else
	      error = true;
	    break;

	  case CGLOBALPRODUCT :
	    global_options.product_mode = GLOBAL;
	    break;

	  case CLOCALPRODUCT :
	    global_options.product_mode = LOCAL;
	    break;

	  case CSTATESPACESIZE :
	    index = value.find("...");
	    if (index == string::npos)
	    {
	      i_value = parseCommandLineInteger(option, value);
	      if (i_value <= 0)
		throw ConfigurationException(-1, "Size of state space must be "
					         "positive.");
	      statespace_options.size
		= statespace_options.max_size
		= i_value;
	    }
	    else
	    {
	      long int min = parseCommandLineInteger(option + " (min)",
						     value.substr(0, index));
	      if (min <= 0)
		throw ConfigurationException(-1, "Minimum state space size "
					         "must be positive.");

	      statespace_options.size = min;

	      long int max = parseCommandLineInteger(option + " (max)",
						     value.substr(index + 3));

	      if (max < min)
		throw ConfigurationException(-1, "Minimum state space size "
					         "exceeds the maximum state "
					         "space size.");

	      statespace_options.max_size = max;
	    }
	    break;

	  case CSTATESPACEGENERATEMODE :
	    if (value == "randomconnectedgraph")
	      statespace_options.generate_mode = RANDOMCONNECTEDGRAPH;
	    else if (value == "randomgraph")
	      statespace_options.generate_mode = RANDOMGRAPH;
	    else if (value == "randompath")
	      statespace_options.generate_mode = RANDOMPATH;
	    else if (value == "enumeratedpath")
	      statespace_options.generate_mode = ENUMERATEDPATH;
	    else
	      error = true;
	    break;

	  case CRANDOMCONNECTEDGRAPH :
	    statespace_options.generate_mode = RANDOMCONNECTEDGRAPH;
	    break;
	    
	  case CRANDOMGRAPH :
	    statespace_options.generate_mode = RANDOMGRAPH;
	    break;

	  case CRANDOMPATH :
	    statespace_options.generate_mode = RANDOMPATH;
	    break;

	  case CENUMERATEDPATH :
	    statespace_options.generate_mode = ENUMERATEDPATH;
	    break;

	  case CABBREVIATEDOPERATORS :
	    formula_options.allow_abbreviated_operators = true;
	    break;

	  case CNOABBREVIATEDOPERATORS :
	    formula_options.allow_abbreviated_operators = false;
	    break;

	  case CFORMULAOUTPUTMODE :
	    if (value == "nnf")
	      formula_options.output_mode = NNF;
	    else if (value == "normal")
	      formula_options.output_mode = NORMAL;
	    else
	      error = true;
	    break;

	  case COUTPUTNNF : 
	    formula_options.output_mode = NNF;
	    break;

	  case CFORMULAGENERATEMODE :
	    if (value == "nnf")
	      formula_options.generate_mode = NNF;
	    else if (value == "normal")
	      formula_options.generate_mode = NORMAL;
	    else
	      error = true;
	    break;

	  case CGENERATENNF :
	    formula_options.generate_mode = NNF;
	    break;

	  case CFORMULALENGTH :
	    index = value.find("...");
	    if (index == string::npos)
	    {
	      i_value = parseCommandLineInteger(option, value);
	      if (i_value <= 0)
		throw ConfigurationException(-1, "Formula length must be "
					         "positive.");
	      formula_options.generation_parameters.length
		= formula_options.generation_parameters.max_length
		= i_value;
	    }
	    else
	    {
	      long int min = parseCommandLineInteger(option + " (min)",
						     value.substr(0, index));
	      if (min <= 0)
		throw ConfigurationException(-1, "Minimum formula length "
					         "must be positive.");

	      formula_options.generation_parameters.length = min;

	      long int max = parseCommandLineInteger(option + " (max)",
						     value.substr(index + 3));

	      if (max < min)
		throw ConfigurationException(-1, "Minimum formula length "
					         "exceeds the maximum formula "
					         "length.");

	      formula_options.generation_parameters.max_length = max;
	    }
	    break;

	  case CVERSION :
	    cout << "lbtt version 0.82  --  11 Aug 2000" << endl;
	    exit(0);
	    break;

	  case CLOGFILE :
	    global_options.transcript_filename = value;
	    break;

	  case CFORMULAFILE :
	    global_options.formula_input_filename = value;
	    break;

	  case CENABLE :
	  case CDISABLE :
	    i_value = parseCommandLineInteger(option, value);

	    if (i_value < 0)
	      throw ConfigurationException(-1, "Algorithm identifiers must be "
					       "nonnegative.");
	    else if (static_cast<vector<AlgorithmInformation>::size_type>
		       (i_value)
		     > algorithms.size() - 1)
	    {
	      ostrstream msg;
	      msg << "Invalid algorithm identifier (" << i_value << ")."
		  << ends;
	      msg.freeze();
	      string msgstring(msg.str());
	      msg.freeze(0);
	      throw ConfigurationException(-1, msgstring);
	    }

	    algorithms[i_value].enabled
	      = (search_option->second.parameter_type == CENABLE);

	    break;

	  default :
	    break;
	}
	break;
    }

    if (error)
      throw ConfigurationException(-1, "Unrecognized value (`" + value + "') "
				       + "for option `" + option + "'.");
  }
}

/* ========================================================================= */
void Configuration::validateAndInitialize()
/* ----------------------------------------------------------------------------
 *
 * Description:   Checks that the program configuration parameters (read from
 *                the configuration file or given in the command line) are
 *                valid. This function will also perform the final
 *                initializations for the configuration (in particular, the
 *                formula symbol priorities).
 *
 * Arguments:     None.
 *
 * Returns:       Nothing; throws a ConfigurationException if there are any
 *                errors.
 *
 * ------------------------------------------------------------------------- */
{
  /*
   *  Check that there is at least one algorithm available for use.
   */

  if (algorithms.empty())
    throw ConfigurationException
            (-1, "No algorithms defined in the configuration file.");

  /*
   *  The case where the number of available variables for propositional
   *  formulae is zero is equivalent to the case where propositional atoms are
   *  disallowed in the formulae altogether.
   */

  if (formula_options.generation_parameters.number_of_available_variables == 0)
    symbol_priority[RandomFormulaParameters::ATOM] = 0;

  /*
   *  Verify that at least one propositional symbol class (a Boolean constant
   *  or a propositional atom) has a nonzero priority.
   */

  if (symbol_priority[RandomFormulaParameters::ATOM] == 0
      && symbol_priority[RandomFormulaParameters::TRUE] == 0
      && symbol_priority[RandomFormulaParameters::FALSE] == 0)
    throw ConfigurationException(-1, "At least one propositional symbol "
                                     "must have nonzero priority.");

  /*
   *  If the operators ->, <->, <> and [] are disallowed, set their priorities 
   *  to zero.
   */

  if (!formula_options.allow_abbreviated_operators)
  {
    symbol_priority[RandomFormulaParameters::IMPLICATION] = 0;
    symbol_priority[RandomFormulaParameters::EQUIVALENCE] = 0;
    symbol_priority[RandomFormulaParameters::FINALLY] = 0;
    symbol_priority[RandomFormulaParameters::GLOBALLY] = 0;
  }

  /*
   *  Verify that at least one unary and one binary operator have a nonzero
   *  priority. Initialize also the priorities of the operators with the
   *  `default' priority.
   */

  bool unary_operator_allowed = false, binary_operator_allowed = false;

  for (map<int, int, less<int>, ALLOC(int) >::iterator
	 it = symbol_priority.begin();
       it != symbol_priority.end(); ++it)
  {
    if (it->second == -1)
      it->second = default_operator_priority;

    if (it->second > 0)
    {
      if (!unary_operator_allowed)
        unary_operator_allowed = 
          (it->first == RandomFormulaParameters::NOT 
           || it->first == RandomFormulaParameters::NEXT 
           || it->first == RandomFormulaParameters::FINALLY
           || it->first == RandomFormulaParameters::GLOBALLY);
      if (!binary_operator_allowed)
        binary_operator_allowed = 
          (it->first == RandomFormulaParameters::AND 
           || it->first == RandomFormulaParameters::OR 
           || it->first == RandomFormulaParameters::UNTIL
           || it->first == RandomFormulaParameters::V
           || it->first == RandomFormulaParameters::IMPLICATION
           || it->first == RandomFormulaParameters::EQUIVALENCE);
    }
  }

  if (!unary_operator_allowed || !binary_operator_allowed)
    throw ConfigurationException(-1, "At least one unary and one binary "
                                     "operator must have nonzero priority.");

  /*
   *  Collect the allowed symbols and operators with their priorities into
   *  their respective symbol lists from which the symbols are randomly
   *  selected when generating formulae.
   */

  for (map<int, int, less<int>, ALLOC(int) >::iterator
	 it = symbol_priority.begin();
       it != symbol_priority.end(); ++it)
  {
    if (it->second != 0)
    {
      switch (it->first)
      {
        case RandomFormulaParameters::ATOM :
        case RandomFormulaParameters::TRUE :
        case RandomFormulaParameters::FALSE :
          formula_options.generation_parameters.usePropositionalSymbol(*it);
          break;

        case RandomFormulaParameters::NOT :
          formula_options.generation_parameters.useShortFormulaOperator(*it);
          if (formula_options.generate_mode != NNF)
            formula_options.generation_parameters.useLongFormulaOperator(*it);
          break;

        case RandomFormulaParameters::NEXT :
          formula_options.generation_parameters.useShortFormulaOperator(*it);
          formula_options.generation_parameters.useLongFormulaOperator(*it);
          break;

        case RandomFormulaParameters::AND :
        case RandomFormulaParameters::OR :
        case RandomFormulaParameters::UNTIL :
        case RandomFormulaParameters::V :
          formula_options.generation_parameters.useLongFormulaOperator(*it);
          break;

        case RandomFormulaParameters::FINALLY :
        case RandomFormulaParameters::GLOBALLY :
          if (formula_options.allow_abbreviated_operators)
          {
            formula_options.generation_parameters.useShortFormulaOperator(*it);
            formula_options.generation_parameters.useLongFormulaOperator(*it);
          }
          break;

        case RandomFormulaParameters::IMPLICATION :
        case RandomFormulaParameters::EQUIVALENCE :
          if (formula_options.allow_abbreviated_operators)
            formula_options.generation_parameters.useLongFormulaOperator(*it);
          break;
      }
    }
  }
}

/* ========================================================================= */
void Configuration::print(ostream& stream, int indent) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Writes information about the program configuration into a
 *                stream.
 *
 * Argument:      stream  --  A reference to an output stream.
 *                indent  --  Number of spaces to leave to the left of output.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::badbit | ios::failbit);

  estream << string(indent, ' ') + "Program configuration:\n\n"
             + string(indent + 2, ' ')
          << configuration.global_options.number_of_rounds
          << string(" test round")
             + (configuration.global_options.number_of_rounds > 1 ? "s" : "")
             + ".\n" + string(indent + 2, ' ');

  if (configuration.global_options.interactive == ALWAYS)
    estream << "Pausing after every test round.\n";
  else if (configuration.global_options.interactive == ONERROR)
    estream << "Testing will be interrupted in case of an error.\n";
  else
    estream << "Running in non-interrupting mode.\n";

  estream << string(indent + 2, ' ')
             + "Computing "
             + (configuration.global_options.product_mode == GLOBAL
		? "global" : "restricted")
             + " synchronous products.\n";

  if (!global_options.transcript_filename.empty())
    estream << string(indent + 2, ' ') + "Writing error log to `"
               + global_options.transcript_filename + "'.\n";

  estream << '\n' + string(indent + 2, ' ') + "Algorithms:\n";

  vector<AlgorithmInformation, ALLOC(AlgorithmInformation) >::size_type
    algorithm_number = 0;
  
  for (vector<AlgorithmInformation, ALLOC(AlgorithmInformation) >
	 ::const_iterator a = algorithms.begin();
       a != algorithms.end();
       ++a)
  {
    estream << string(indent + 4, ' ')
            << algorithm_number
            << string(": `") + *(a->name) + '\'';

    if (!a->enabled)
      estream << " (disabled)";

    estream << '\n';

    algorithm_number++;
  }

  estream << endl
          << string(indent + 2, ' ') + "Random state spaces:\n"
             + string(indent + 4, ' ');

  switch (statespace_options.generate_mode)
  {
    case RANDOMGRAPH :
      estream << "Random graphs ";
      break;
    case RANDOMCONNECTEDGRAPH :
      estream << "Random connected graphs ";
      break;
    case RANDOMPATH :
      estream << "Random paths ";
      break;
    case ENUMERATEDPATH :
      estream << "Enumerated paths ";
      break;
    default :
      break;
  }

  estream << '('
	  << statespace_options.size;

  if (statespace_options.max_size != statespace_options.size
      && statespace_options.generate_mode != ENUMERATEDPATH)
    estream << "..." << statespace_options.max_size;

  estream << string(" state") + (statespace_options.max_size == 1 ? "" : "s")
             + ", "
          << statespace_options.atoms_per_state
          << string(" atomic proposition")
             + (statespace_options.atoms_per_state == 1 ? "" : "s")
             + ")\n" + string(indent + 4, ' ');

  if (global_options.statespace_change_interval == 0)
    estream << "Using a fixed state space.\n" + string(indent + 4, ' ');
  else
  {
    estream << "New state space will be generated after every ";
    if (global_options.statespace_change_interval > 1)
    {
      estream << global_options.statespace_change_interval;

      if (global_options.statespace_change_interval < 10
	  || global_options.statespace_change_interval > 20)
      {
	switch (global_options.statespace_change_interval % 10)
	{
	  case 1 : estream << "st"; break;
	  case 2 : estream << "nd"; break;
	  case 3 : estream << "rd"; break;
	  default : estream << "th"; break;
	}
      }
      else
	estream << "th";

      estream << ' ';
    }

    estream << "round.\n" + string(indent + 4, ' ');
  }

  if (statespace_options.generate_mode != ENUMERATEDPATH)
  {	
    if (statespace_options.generate_mode & GRAPH)
      estream << "Random edge probability: "
              << statespace_options.edge_probability
              << '\n' + string(indent + 4, ' ');

    estream << "Propositional truth probability: "
	    << statespace_options.truth_probability
	    << "\n";
  }

  estream << "\n" + string(indent + 2, ' ');

  if (global_options.formula_input_filename.empty())
  {
    estream << "Random LTL formulae:\n" + string(indent + 4, ' ')
	    << formula_options.generation_parameters.length;

    if (formula_options.generation_parameters.max_length
          != formula_options.generation_parameters.length)
      estream << "..." << formula_options.generation_parameters.max_length;

    estream << string(" symbol")
               + (formula_options.generation_parameters.max_length == 1
		  ? "" : "s")
               + ", "
	    << formula_options.generation_parameters.
                 number_of_available_variables
	    << string(" atomic proposition")
               + (formula_options.generation_parameters.
		    number_of_available_variables == 1 ? "" : "s");
  }
  else
    estream << "Reading LTL formulae from `"
               + global_options.formula_input_filename
               + "'.";

  estream << '\n' + string(indent + 4, ' ');

  if (global_options.formula_change_interval == 0)
    estream << "Using a fixed LTL formula.";
  else
  {
    estream << "New LTL formula will be "
	    << (global_options.formula_input_filename.empty()
		? "generate"
		: "rea")
	    << "d after every ";
    if (global_options.formula_change_interval > 1)
    {
      estream << global_options.formula_change_interval;

      if (global_options.formula_change_interval < 10
	  || global_options.formula_change_interval > 20)
      {
	switch (global_options.formula_change_interval % 10)
        {
	  case 1 : estream << "st"; break;
	  case 2 : estream << "nd"; break;
	  case 3 : estream << "rd"; break;
	  default : estream << "th"; break;
	}
      }
      else
	estream << "th";

      estream << ' ';
    }

    estream << "round.";
  }

  estream << '\n';

  if (global_options.formula_input_filename.empty()
      && formula_options.generate_mode == NNF)
    estream << string(indent + 4, ' ')
               + "Formulae will be generated into negation normal form.\n";
  else if (formula_options.output_mode == NNF)
    estream << string(indent + 4, ' ')
               + "Formulae will be converted into negation normal form.\n";

  if (global_options.formula_input_filename.empty())
  {
    estream << string(indent + 4, ' ') + "Operands in use (priority):\n"
               + string(indent + 6, ' ');

    bool first_printed = false;

    typedef pair<int, int> IntegerPair;

    for (vector<IntegerPair, ALLOC(IntegerPair) >::const_iterator op
	   = formula_options.generation_parameters.
	       propositional_symbol_priorities.begin();
	 op != formula_options.generation_parameters.
                 propositional_symbol_priorities.end();
	 ++op)
    {
      if (op->first == RandomFormulaParameters::NO_SYMBOL)
	continue;

      if (first_printed)
	estream << "; ";

      first_printed = true;

      switch (op->first)
      {
        case RandomFormulaParameters::ATOM :
	  estream << "propositions";
	  break;
        case RandomFormulaParameters::TRUE :
	  estream << "true";
	  break;
        case RandomFormulaParameters::FALSE :
	  estream << "false";
	  break;
        default :
	  break;
      }

      estream << " (" << op->second << ')';
    }

    estream << '\n'
	    << string(indent + 4, ' ') + "Operators in use (priority):\n"
               + string(indent + 6, ' ');

    int number_of_printed_operators = 0;

    for (map<int, int, less<int>, ALLOC(int) >::const_iterator op
	   = symbol_priority.begin();
	 op != symbol_priority.end();
	 ++op)
    {
      if (op->first == RandomFormulaParameters::ATOM
	  || op->first == RandomFormulaParameters::TRUE
	  || op->first == RandomFormulaParameters::FALSE
	  || op->second == 0)
	continue;

      if (number_of_printed_operators > 0)
      {
	estream << "; ";
	if (number_of_printed_operators % 6 == 0)
	  estream << '\n' << string(indent + 6, ' ');
      }

      number_of_printed_operators++;

      switch (op->first)
      {
        case RandomFormulaParameters::AND :
	  estream << "/\\";
	  break;
        case RandomFormulaParameters::OR :
	  estream << "\\/";
	  break;
        case RandomFormulaParameters::IMPLICATION :
	  estream << "->";
	  break;
        case RandomFormulaParameters::EQUIVALENCE :
	  estream << "<->";
	  break;
        case RandomFormulaParameters::UNTIL :
	  estream << 'U';
	  break;
        case RandomFormulaParameters::V :
	  estream << 'V';
	  break;
        case RandomFormulaParameters::NEXT :
	  estream << 'X';
	  break;
        case RandomFormulaParameters::NOT :
	  estream << '!';
	  break;
        case RandomFormulaParameters::FINALLY :
	  estream << "<>";
	  break;
        case RandomFormulaParameters::GLOBALLY :
	  estream << "[]";
	  break;
        default :
	  break;
      }

      estream << " (" << op->second;

      if (op->first == RandomFormulaParameters::NOT
	  && formula_options.generate_mode == NNF)
	estream << ", only preceding atoms";

      estream << ')';
    }

    estream << '\n';
  }

  estream << endl;
}

/* ========================================================================= */
void Configuration::showCommandLineHelp(const char* program_name)
/* ----------------------------------------------------------------------------
 *
 * Description:   Prints the list of command line options.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  cout << "Usage: " << program_name << " [OPTION]...\n\n"
          "General options:\n"
          "  --configfile=FILENAME       Read configuration from FILENAME\n"
          "  --disable=INTEGER           Disables testing of an algorithm\n"
          "  --enable=INTEGER            Enables testing of an algorithm\n"
          "  --formulafile=FILENAME      Read LTL formulae from FILENAME\n"
          "  -h, --help                  Show this help and exit\n"
          "  --interactive=MODE          Set the interactivity mode (`always',"
                                         " `onerror', \n"
          "                              `never')\n"
          "  --logfile=FILENAME          Write error log to FILENAME\n"
          "  --nopause                   Do not pause between test rounds "
                                         "(equivalent to\n"
          "                              --interactive=never)\n"
          "  --pauseonerror              Pause between test rounds in case of "
                                         "an error\n"
          "                              (equivalent to "
                                         "--interactive=onerror)\n"
          "  --pause                     Pause between every test round "
                                         "(equivalent to\n"
          "                              --interactive=always)\n"
          "  --rounds=NUMBER_OF_ROUNDS   Set number of test rounds (1-)\n"
          "  --randomseed=INTEGER        Set random seed\n"
          "  --verbosity=VERBOSITY       Set the verbosity of output (0-5)\n"
          "  --version                   Display program version and exit\n\n"
          "Synchronous product computation options:\n"
          "  --synchronousproduct=MODE   Set product computation mode "
                                         "(`global' or\n"
          "                              `local')\n"
          "  --globalproduct             Compute global synchronous products "
                                         "(equivalent\n"
          "                              to --synchronousproduct=global)\n"
          "  --localproduct              Compute restricted synchronous "
                                         "products\n"
          "                              (equivalent to "
                                         "--synchronousproduct=local)\n\n"
          "Reachability graph generation options:\n"
          "  --statespacechangeinterval=NUMBER_OF_ROUNDS\n"
          "                              Set state space generation interval "
                                         "in test\n"
          "                              rounds (0-)\n"
          "  --statespacesize=SIZE,\n"
          "  --statespacesize=MIN...MAX\n"
          "                              Set size of generated state spaces "
                                         "(1-)\n"
          "  --statespacepropositions=NUMBER_OF_PROPOSITIONS\n"
          "                              Set number of propositions per "
                                         "state (0-)\n"
          "  --edgeprobability=PROBABILITY\n"
          "                              Set random edge probability in state "
                                         "spaces\n"
          "                              (0.0--1.0)\n"
          "  --truthprobability=PROBABILITY\n"
          "                              Set truth probability of "
                                         "propositions (0.0--1.0)\n"
          "  --statespacegeneratemode=MODE\n"
          "                              Set state space generation mode\n"
          "                              (`randomconnectedgraph', "
                                         "`randomgraph',\n"
          "                              `randompath', `enumeratedpath')\n"
          "  --randomconnectedgraph      Generate connected graphs as state "
                                         "spaces\n"
          "                              (equivalent to\n"
          "                              --statespacegeneratemode="
                                         "randomconnectedgraph)\n"
          "  --randomgraph               Generate random graphs as state "
                                         "spaces\n"
          "                              (equivalent to\n"
          "                              --statespacegeneratemode="
                                         "randomgraph)\n"
          "  --randompath                Generate random paths as state "
                                         "spaces\n"
          "                              (equivalent to\n"
          "                              --statespacegeneratemode="
                                         "randompath)\n"
          "  --enumeratedpath            Enumerate all paths of a given size "
                                         "and a given\n"
          "                              number of propositions per state "
                                         "(equivalent to\n"
          "                              --statespacegeneratemode="
                                         "enumeratedpath)\n\n"
          "LTL formula generation options:\n"
          "  --formulachangeinterval=NUMBER_OF_ROUNDS\n"
          "                              Set formula generation interval in "
                                         "test rounds\n"
          "                              (0-)\n"
          "  --formulapropositions=NUMBER_OF_PROPOSITIONS\n"
          "                              Set maximum number of atomic "
                                         "propositions in\n"
          "                              generated formulae (0-)\n"
          "  --formulalength=NUMBER_OF_SYMBOLS,\n"
          "  --formulalength=MIN...MAX\n"
          "                              Set number of symbols in generated "
                                         "formulae (1-)\n"
          "  --formulageneratemode=MODE  Set formula generation mode "
                                         "(`normal', `nnf')\n"
          "  --generatennf               Generate formulae in negation normal "
	                                 "form\n"
          "                              (equivalent to "
                                         "--formulageneratemode=nnf)\n"
          "  --formulaoutputmode=MODE    Set formula output mode (`normal', "
                                         "`nnf')\n"
          "  --outputnnf                 Output formulae in negation normal "
                                         "form\n"
          "                              (equivalent to "
                                         "--formulaoutputmode=nnf)\n"
          "  --abbreviatedoperators,\n"
          "  --noabbreviatedoperators\n"    
          "                              Allow or disallow the operators ->, "
                                         "<->, <>, []\n"
          "                              in the generated formulae\n"
          "  --propositionpriority=PRIORITY\n"
          "                              Set priority for atomic "
                                         "propositions\n"
          "  --truepriority=PRIORITY     Set priority for the constant "
                                         "`true'\n"
          "  --falsepriority=PRIORITY    Set priority for the constant "
                                         "`false'\n"
          "  --defaultoperatorpriority=PRIORITY\n"
          "                              Set default priority for operators\n"
          "  --andpriority=PRIORITY      Set priority for the /\\ operator\n"
          "  --orpriority=PRIORITY       Set priority for the \\/ operator\n"
          "  --implicationpriority=PRIORITY\n"
          "                              Set priority for the -> operator\n"
          "  --equivalencepriority=PRIORITY\n"
          "                              Set priority for the <-> operator\n"
          "  --untilpriority=PRIORITY    Set priority for the Until operator\n"
          "  --vpriority=PRIORITY        Set priority for the Release operator"
                                         "\n"
          "  --notpriority=PRIORITY      Set priority for the negation "
                                         "operator\n"
          "  --nextpriority=PRIORITY     Set priority for the Next operator\n"
          "  --finallypriority=PRIORITY  Set priority for the <> operator\n"
          "  --globallypriority=PRIORITY Set priority for the [] operator\n";
}

/* ========================================================================= */
void Configuration::reset()
/* ----------------------------------------------------------------------------
 *
 * Description:   Resets the program configuration to the default
 *                configuration.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  global_options.verbosity = 5;
  global_options.interactive = ALWAYS;
  global_options.number_of_rounds = 10;
  global_options.statespace_change_interval = 10;
  global_options.formula_change_interval = 1;
  global_options.product_mode = GLOBAL;
  global_options.transcript_filename = "";
  global_options.formula_input_filename = "";

  statespace_options.size = 20;
  statespace_options.atoms_per_state = 5;
  statespace_options.edge_probability = 0.2;
  statespace_options.truth_probability = 0.5;
  statespace_options.generate_mode = RANDOMCONNECTEDGRAPH;

  formula_options.allow_abbreviated_operators = true;
  formula_options.output_mode = NORMAL;
  formula_options.generate_mode = NORMAL;

  formula_options.generation_parameters.reset();
}

/* ========================================================================= */
long int Configuration::parseCommandLineInteger
  (const string& option, const string& value) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Converts an integer (given as a string) into a long int. Used
 *                when processing command line parameters.
 *
 * Arguments:     option  --  A reference to a constant string giving a name of
 *                            a command line option.
 *                value   --  A reference to a string which is supposed to
 *                            contain an integer.
 *
 * Returns:       The value of the integer.
 *
 * ------------------------------------------------------------------------- */
{
  char* endptr;
  long int val = strtol(value.c_str(), &endptr, 10);

  if (*endptr != '\0' || value.empty())
    throw ConfigurationException(-1, "The value for `" + option + "' must "
				     "be a nonnegative integer.");

  if (val == LONG_MIN || val == LONG_MAX)
    throw ConfigurationException(-1, "The value for `" + option + "' is out "
				     "of range.");

  return val;
}

        

