/*
 *  Copyright (C) 1999, 2000  Heikki Tauriainen <Heikki.Tauriainen@hut.fi>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

%{
#include <cmath>
#include <climits>
#include <cstdlib>
#include <cstring>
#include <string>
%}

%option yylineno

%%

[ \t\n]*                    { /* Skip whitespace. */ }
"#"[^\n]*                   { /* Skip comments. */ }

"{"                         { return LBRACE; }
"}"                         { return RBRACE; }
"="                         { return EQUALS; }

algorithm                   { return ALGORITHM; }
name                        { return NAME; }
path                        { return PROGRAMPATH; }
parameters                  { return PARAMETERS; }
enabled                     { return ENABLED; }

globaloptions               { return GLOBALOPTIONS; }
verbosity                   { return VERBOSITY; }
interactive                 { return INTERACTIVE; }
rounds                      { return ROUNDS; }
statespacechangeinterval    { return STATESPACECHANGEINTERVAL; }
formulachangeinterval       { return FORMULACHANGEINTERVAL; }
randomseed                  { return RANDOMSEED; }
synchronousproduct          { return SYNCHRONOUSPRODUCT; }

statespaceoptions           { return STATESPACEOPTIONS; }
size                        { return SIZE; }
propositions                { return PROPOSITIONS; }
edgeprobability             { return EDGEPROBABILITY; }
truthprobability            { return TRUTHPROBABILITY; }

formulaoptions              { return FORMULAOPTIONS; }
abbreviatedoperators        { return ABBREVIATEDOPERATORS; }
outputmode                  { return OUTPUTMODE; }
generatemode                { return GENERATEMODE; }
length                      { return LENGTH; }
propositionpriority         { return PROPOSITIONPRIORITY; }
truepriority                { return TRUEPRIORITY; }
falsepriority               { return FALSEPRIORITY; }
defaultoperatorpriority     { return DEFAULTOPERATORPRIORITY; }
andpriority                 { return ANDPRIORITY; }
orpriority                  { return ORPRIORITY; }
implicationpriority         { return IMPLICATIONPRIORITY; }
equivalencepriority         { return EQUIVALENCEPRIORITY; }
untilpriority               { return UNTILPRIORITY; }
vpriority                   { return VPRIORITY; }
notpriority                 { return NOTPRIORITY; }
nextpriority                { return NEXTPRIORITY; }
finallypriority             { return FINALLYPRIORITY; }
globallypriority            { return GLOBALLYPRIORITY; }

true|yes                    {
                              yylval.truth_value = true;
                              return TRUTH_VALUE;
                            }

false|no                    {
                              yylval.truth_value = false;
                              return TRUTH_VALUE;
                            }

always                      {
                              yylval.interactivity_value =
                                Configuration::ALWAYS;
                              return INTERACTIVITY_VALUE;
                            }

never                       {
                              yylval.interactivity_value =
                                Configuration::NEVER;
                              return INTERACTIVITY_VALUE;
                            }

onerror                     {
                              yylval.interactivity_value =
                                Configuration::ONERROR;
                              return INTERACTIVITY_VALUE;
                            }

normal                      {
			      yylval.formula_mode_value = 
				Configuration::NORMAL;
			      return FORMULA_MODE_VALUE;
                            }

nnf                         {
                              yylval.formula_mode_value = Configuration::NNF;
			      return FORMULA_MODE_VALUE;
                            }

local                       {
                              yylval.product_type_value = Configuration::LOCAL;
			      return PRODUCT_TYPE_VALUE;
                            }

global                      {
                              yylval.product_type_value =
				Configuration::GLOBAL;
			      return PRODUCT_TYPE_VALUE;
                            }

randomgraph                 {
                              yylval.statespace_mode_value
				= Configuration::RANDOMGRAPH;
			      return STATESPACE_MODE_VALUE;
                            }

randomconnectedgraph        {
                              yylval.statespace_mode_value
				= Configuration::RANDOMCONNECTEDGRAPH;
			      return STATESPACE_MODE_VALUE;
                            }

randompath                  {
                              yylval.statespace_mode_value
				= Configuration::RANDOMPATH;
			      return STATESPACE_MODE_VALUE;
                            }

enumeratedpath              {
                              yylval.statespace_mode_value
				= Configuration::ENUMERATEDPATH;
			      return STATESPACE_MODE_VALUE;
                            }


"-"?[0-9]+"...""-"?[0-9]+   {
                              char *dot_ptr;
                              yylval.integer_interval.min
				= strtol(yytext, &dot_ptr, 10);
			      
			      if (yylval.integer_interval.min == LONG_MIN
				  || yylval.integer_interval.min == LONG_MAX)
                                throw Configuration::ConfigurationException
                                  (yylineno, "Integer out of range.");

			      dot_ptr += 3;
			      yylval.integer_interval.max
				= strtol(dot_ptr, 0, 10);

			      if (yylval.integer_interval.max == LONG_MIN
				  || yylval.integer_interval.max == LONG_MAX)
                                throw Configuration::ConfigurationException
                                  (yylineno, "Integer out of range.");

			      return INTEGER_INTERVAL;
                            }

"-"?[0-9]+                  {
                              yylval.integer = strtol(yytext, 0, 10);
                              if (yylval.integer == LONG_MIN
                                  || yylval.integer == LONG_MAX)
                                throw Configuration::ConfigurationException
                                  (yylineno, "Integer out of range.");
                              return INTEGER;
                            }

"-"?[0-9]*"."[0-9]+         {
                              if ((yylval.real = strtod(yytext, 0))
                                  == HUGE_VAL)
                                throw Configuration::ConfigurationException
                                  (yylineno, "Real number out of range.");
                              return REAL;
                            }

\"([^\n\"\\]*(\\[^\n])?)*\" {
                              unsigned long int len = strlen(yytext);
                              bool escape = false;
                              yylval.str = new string;
                              for (unsigned long int i = 1; i < len - 1; i++)
			      {
                                if (!escape && yytext[i] == '\\')
                                  escape = true;
                                else
				{
                                  escape = false;
                                  (*yylval.str) += yytext[i];
                                }
                              }
                              return STRING_CONSTANT;
                            }

.                           {
                              return UNKNOWN;
                            }


%%

int getLineNumber()
{
  return yylineno;
}
