/*
 *  Copyright (C) 1999, 2000  Heikki Tauriainen <Heikki.Tauriainen@hut.fi>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "FormulaString.h"

namespace FormulaString
{

/******************************************************************************
 *
 * Function definitions for class FormulaString.
 *
 *****************************************************************************/

/* ========================================================================= */
unsigned long int FormulaString::randomize
  (const RandomFormulaParameters& randomization_parameters)
/* ----------------------------------------------------------------------------
 *
 * Description:   Initializes a FormulaString with a random formula, whose
 *                generation is parameterized by the information in the
 *                function argument.
 *
 * Argument:      randomization_parameters  --  Information controlling the
 *                                              generation of the formula
 *                                              (length of the formula, maximum
 *                                              number of different variables 
 *                                              in the formula, priorities for
 *                                              different operators).
 *
 * Returns:       Length of the random formula.
 *
 * ------------------------------------------------------------------------- */
{
  unsigned long int length = randomization_parameters.length;

  if (randomization_parameters.max_length > length)
    length += LRAND(0, randomization_parameters.max_length - length + 1);

  formula.resize(0);

  typedef pair<int, int> SymbolPriority;

  vector<SymbolPriority, ALLOC(SymbolPriority) >::const_iterator
    symbol_priority;

  stack<unsigned long int> subformula_length_stack;
  subformula_length_stack.push(length);

  ostrstream formula_stream;

  long int x;
  int arity;

  while (!subformula_length_stack.empty())
  {
    /*
     *  Select a list of allowable symbols according to the target length.
     *  If the length is 1, only propositional variables and constants are
     *  allowed. If the length is 2, only unary operators are allowed.
     *  Otherwise all unary and binary operators are allowed.
     */

    switch (subformula_length_stack.top())
    {
      case 1 :
	x = LRAND(0, randomization_parameters.
		       propositional_symbol_priorities[0].second);
	symbol_priority = randomization_parameters.
	                    propositional_symbol_priorities.begin();
	break;
      case 2 :
        x = LRAND(0, randomization_parameters.
                       short_formula_operators[0].second);
	symbol_priority = randomization_parameters.
                            short_formula_operators.begin();
	break;
      default :
        x = LRAND(0, randomization_parameters.
                       long_formula_operators[0].second);
	symbol_priority = randomization_parameters.
	                    long_formula_operators.begin();
	break;
    }

    /*
     *  Using the selected list, choose a symbol from the list randomly,
     *  according to the priority distribution of the different symbols in
     *  the list. The list consists of pairs <symbol, priority>; however,
     *  the first element of the list does not denote any symbol: instead,
     *  it gives the sum of the priorities of the symbols in the list.
     */

    ++symbol_priority;
    while (x >= symbol_priority->second)
    {
      x -= symbol_priority->second;
      ++symbol_priority;
    }

    switch (symbol_priority->first)
    {
      /*
       *  Propositional atom: Append a random propositional variable to the end
       *                      of the formula.
       */

      case RandomFormulaParameters::ATOM :
      {
	formula_stream << 'p'
		       << LRAND(0, randomization_parameters.
	                             number_of_available_variables);

	arity = 0;
	break;
      }

      /*
       *  Propositional constants.
       */

      case RandomFormulaParameters::TRUE :
	formula_stream << LTL_TRUE;
	arity = 0;
	break;

      case RandomFormulaParameters::FALSE :
	formula_stream << LTL_FALSE;
	arity = 0;
	break;

      /*
       *  Unary operators: First append a random unary operator to the end of
       *                   the formula, then recursively generate a random
       *                   formula of length `length_of_formula' - 1.
       */

      case RandomFormulaParameters::NOT :
	formula_stream << LTL_NOT;
	arity = 1;
	break;

      case RandomFormulaParameters::NEXT :
	formula_stream << LTL_NEXT;
	arity = 1;
	break;

      case RandomFormulaParameters::FINALLY :
	formula_stream << LTL_FINALLY;
	arity = 1;
	break;

      case RandomFormulaParameters::GLOBALLY :
	formula_stream << LTL_GLOBALLY;
	arity = 1;
	break;

      /*
       *  Binary operators: First append a random binary operator to the end of
       *                    the formula, then prepare to generate two formulae
       *                    with their combined length `subformula_length' - 1.
       */

      case RandomFormulaParameters::AND :
	formula_stream << LTL_AND;
	arity = 2;
	break;

      case RandomFormulaParameters::OR :
	formula_stream << LTL_OR;
	arity = 2;
	break;

      case RandomFormulaParameters::UNTIL :
	formula_stream << LTL_UNTIL;
	arity = 2;
	break;

      case RandomFormulaParameters::V :
	formula_stream << LTL_V;
	arity = 2;
	break;

      case RandomFormulaParameters::IMPLICATION :
	formula_stream << LTL_IMPLICATION;
	arity = 2;
	break;

      case RandomFormulaParameters::EQUIVALENCE :
	formula_stream << LTL_EQUIVALENCE;
	arity = 2;
	break;
    }

    if (arity == 0)
    {
      subformula_length_stack.pop();
      if (!subformula_length_stack.empty())
	formula_stream << ' ';
    }
    else
    {
      formula_stream << ' ';
    
      if (arity == 1)
	subformula_length_stack.top()--;
      else if (arity == 2)
      {
	unsigned long int len = LRAND(1, subformula_length_stack.top() - 1);
	subformula_length_stack.top() -= len + 1;
	subformula_length_stack.push(len);
      }
    }
  }

  formula_stream << ends;

  formula_stream.freeze();
  formula = string(formula_stream.str());
  formula_stream.freeze(0);

  return length;
}

/* ========================================================================= */
void FormulaString::read(istream& input_stream)
/* ----------------------------------------------------------------------------
 *
 * Description:   Initializes a FormulaString by reading a formula from a
 *                stream.
 *
 * Argument:      input_stream  --  A reference to an input stream.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  formula = "";

  TraversalInfo<string*> info;
  info.data = &formula;

  try
  {
    FormulaString::template
      traverseStream<ReadAtom, ReadSymbol<LTL_TRUE>, ReadSymbol<LTL_FALSE>, 
                     ReadSymbol<LTL_NOT>, ReadSymbol<LTL_NEXT>,
                     ReadSymbol<LTL_FINALLY>, ReadSymbol<LTL_GLOBALLY>,
                     ReadSymbol<LTL_AND>, ReadSymbol<LTL_OR>,
                     ReadSymbol<LTL_IMPLICATION>, ReadSymbol<LTL_EQUIVALENCE>,
                     ReadSymbol<LTL_UNTIL>, ReadSymbol<LTL_V> >
      (PREORDER, input_stream, info);
  }
  catch (IOException&)
  {
    if (input_stream.eof() && !formula.empty())
      throw ParseErrorException("Error reading LTL formula: Unexpected end of "
				"input.");
    else
      throw;
  }

  formula.resize(formula.size() - 1);
}

/* ========================================================================= */
void FormulaString::read(const char* filename)
/* ----------------------------------------------------------------------------
 *
 * Description:   Initializes a FormulaString by reading a formula from a file.
 *
 * Argument:      filename  --  Name of the file.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  formula = "";

  ifstream formula_file;
  formula_file.open(filename, ios::in);

  if (!formula_file.good())
    throw FileOpenException(filename);

  try
  {
    read(formula_file);
    formula_file.close();
  }
  catch (...)
  {
    formula_file.close();
    throw;
  }
}

/*
 *  Strings to be used for instatiating the `traverse' template in function
 *  FormulaString::prettyPrint.
 */

char LTL_PRETTYPRINT_TRUE[]        = "true";
char LTL_PRETTYPRINT_FALSE[]       = "false";
char LTL_PRETTYPRINT_NOT[]         = "!";
char LTL_PRETTYPRINT_NEXT[]        = "X";
char LTL_PRETTYPRINT_FINALLY[]     = "<>";
char LTL_PRETTYPRINT_GLOBALLY[]    = "[]";
char LTL_PRETTYPRINT_AND[]         = "/\\";
char LTL_PRETTYPRINT_OR[]          = "\\/";
char LTL_PRETTYPRINT_IMPLICATION[] = "->";
char LTL_PRETTYPRINT_EQUIVALENCE[] = "<->";
char LTL_PRETTYPRINT_UNTIL[]       = "U";
char LTL_PRETTYPRINT_V[]           = "V";

/* ========================================================================= */
void FormulaString::prettyPrint(ostream& output_stream)
/* ----------------------------------------------------------------------------
 *
 * Description:   Writes a FormulaString in a parenthesized infix notation to a
 *                stream.
 *
 * Argument:      output_stream  --  Reference to the output stream.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  TraversalInfo<Exceptional_ostream*> info;
  Exceptional_ostream eoutput_stream(&output_stream,
                                     ios::failbit | ios::badbit);
  info.data = &eoutput_stream;

  traverse<FormatAtom,
           FormatConstant<LTL_PRETTYPRINT_TRUE>,
           FormatConstant<LTL_PRETTYPRINT_FALSE>,
           FormatUnary<LTL_PRETTYPRINT_NOT>,
           FormatUnary<LTL_PRETTYPRINT_NEXT>,
           FormatUnary<LTL_PRETTYPRINT_FINALLY>,
           FormatUnary<LTL_PRETTYPRINT_GLOBALLY>,
           FormatBinary<LTL_PRETTYPRINT_AND>,
           FormatBinary<LTL_PRETTYPRINT_OR>,
           FormatBinary<LTL_PRETTYPRINT_IMPLICATION>,
           FormatBinary<LTL_PRETTYPRINT_EQUIVALENCE>,
           FormatBinary<LTL_PRETTYPRINT_UNTIL>,
           FormatBinary<LTL_PRETTYPRINT_V> >
    (PREORDER | INORDER | POSTORDER, info);
}

/* ========================================================================= */
void FormulaString::negate()
/* ----------------------------------------------------------------------------
 *
 * Description:   Negates a FormulaString.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  string negated_formula = "! ";
  negated_formula += formula;
  formula = negated_formula;
}

/* ========================================================================= */
void FormulaString::toNegationNormalForm()
/* ----------------------------------------------------------------------------
 *
 * Description:   Converts the FormulaString into negation normal form, i.e.
 *                to a representation having no `abbreviated' operators
 *                (`finally', `globally', implication and equivalence) with
 *                all negations preceding propositional variables.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  TraversalInfo<NNFConvertData> info;
  info.data.negate_stack.push(false);
  info.data.result = new string;

  traverse<AtomToNNF,
           ConstantToNNF<true>,
           ConstantToNNF<false>,
           NegationToNNF,
           NextToNNF,
           AbbreviatedUnaryToNNF<LTL_UNTIL, LTL_V, false>,
           AbbreviatedUnaryToNNF<LTL_V, LTL_UNTIL, true>,
           BinaryToNNF<LTL_AND, LTL_OR, true, true>,
           BinaryToNNF<LTL_OR, LTL_AND, true, true>,
           BinaryToNNF<LTL_OR, LTL_AND, false, true>,
           EquivalenceToNNF,
           BinaryToNNF<LTL_UNTIL, LTL_V, true, true>,
           BinaryToNNF<LTL_V, LTL_UNTIL, true, true> >
    (PREORDER | INORDER | POSTORDER, info);

  formula = *info.data.result;
  delete info.data.result;
}

/* ========================================================================= */
FormulaString::FormulaSymbolType FormulaString::parseToken
  (const string& token)
/* ----------------------------------------------------------------------------
 *
 * Description:   Tells the type of a formula symbol.
 *    
 * Argument:      token  --  The token.
 *
 * Returns:       The type of the symbol (PROPOSITIONALSYMBOL, UNARYOPERATOR,
 *                BINARYOPERATOR or UNKNOWN).
 *
 * ------------------------------------------------------------------------- */
{
  if (token.empty()
      || (token.length() > 1 && token[0] != 'p')
      || (token.length() == 1 && token[0] == 'p'))
    return UNKNOWN;

  switch (token[0])
  {
    case LTL_PROPOSITION :
    {
      long int id;
      char* endptr;

      id = strtol(token.c_str() + 1, &endptr, 10);

      if (*endptr != '\0' || id < 0 || id == LONG_MIN || id == LONG_MAX)
	return UNKNOWN;

      return PROPOSITIONALSYMBOL;
    }

    case LTL_TRUE :
    case LTL_FALSE :
      return PROPOSITIONALSYMBOL;
        
    case LTL_NOT :
    case LTL_NEXT :
    case LTL_FINALLY :
    case LTL_GLOBALLY :
      return UNARYOPERATOR;
        
    case LTL_AND :
    case LTL_OR :
    case LTL_IMPLICATION :
    case LTL_EQUIVALENCE :
    case LTL_UNTIL :
    case LTL_V :
      return BINARYOPERATOR;
        
    default :
      return UNKNOWN;
  }
}

}
