/*
 *  Copyright (C) 1999, 2000  Heikki Tauriainen <Heikki.Tauriainen@hut.fi>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef _FormulaString_h_
#define _FormulaString_h_

#include <cstdlib>
#include <deque>
#include <fstream>
#include <stack>
#include <string>
#include <strstream>
#include <vector>
#include "Exception.h"
#include "MacroDefs.h"
#include "RandomFormulaParameters.h"

using namespace std;

namespace FormulaString
{

/******************************************************************************
 *
 * A class for LTL formulae represented as strings of a special form. The
 * string stores the formula in a prefix form using the following syntax
 * (successive tokens are always separated from each other by whitespace):
 *
 *    <FormulaString>   ::=   <proposition>
 *                          | <unary operator> <FormulaString>
 *                          | <binary operator> <FormulaString> <FormulaString>
 *
 *    <proposition>     ::=   'p'<nonnegative integer>
 *
 *    <unary operator>  ::=   '!'               (negation)
 *                          | 'X'               (`next')
 *                          | 'F'               (`finally')
 *                          | 'G'               (`globally')
 *
 *    <binary operator> ::=   '&'               (conjunction)
 *                          | '|'               (disjunction)
 *                          | 'i'               (implication)
 *                          | 'e'               (equivalence)
 *                          | 'U'               (`until')
 *                          | 'V'               (`V', the dual of `until')
 *
 *****************************************************************************/

class FormulaString
{
public:
  FormulaString();                                  /* Constructor. */

  /* default copy constructor */

  ~FormulaString();                                 /* Destructor. */

  /* default assignment operator */

  unsigned long int randomize                       /* Initializes the      */
    (const RandomFormulaParameters&                 /* object with a random */
       randomization_parameters);                   /* LTL formula.         */

  void read(istream& input_stream);                 /* Reads a formula from
                                                     * a stream.
						     */

  void read(const char* filename);                  /* Reads a formula from
                                                     * a file.
						     */

  void prettyPrint(ostream& output_stream = cout);  /* Prints the formula in
                                                     * parenthesized infix
                                                     * notation.
						     */

  void negate();                                    /* Negates the formula. */

  void toNegationNormalForm();                      /* Transforms the formula
                                                     * into negation normal
                                                     * form.
						     */

  static const char LTL_PROPOSITION = 'p';          /* Character constants */
  static const char LTL_TRUE        = 't';          /* for symbols of the  */
  static const char LTL_FALSE       = 'f';          /* FormulaString.      */
  static const char LTL_NOT         = '!';
  static const char LTL_NEXT        = 'X';
  static const char LTL_FINALLY     = 'F';
  static const char LTL_GLOBALLY    = 'G';
  static const char LTL_AND         = '&';
  static const char LTL_OR          = '|';
  static const char LTL_IMPLICATION = 'i';
  static const char LTL_EQUIVALENCE = 'e';
  static const char LTL_UNTIL       = 'U';
  static const char LTL_V           = 'V';

  enum FormulaSymbolType {PROPOSITIONALSYMBOL,      /* Constants for     */
                          UNARYOPERATOR,            /* different formula */
                          BINARYOPERATOR,           /* symbol types.     */
                          UNKNOWN};

  static FormulaSymbolType                          /* Determines the type  */
    parseToken(const string& token);                /* of a formula symbol. */

  friend ostream& operator<<                        /* for printing the     */
    (ostream& stream,                               /* formula using the << */
     const FormulaString& formula_string);          /* operator             */

  /*
   *  Declarations and definitions for constants and functions required when
   *  `traversing' a FormulaString. See the documentation for the template
   *  function `FormulaString::traverse' for information on how they can be
   *  used.
   */

  enum TraversalMode {PREORDER = 1, INORDER = 2,    /* Constants for the   */
                      POSTORDER = 4};               /* different traversal */
                                                    /* modes.              */

  template<class D>                                 /* Template structure    */
  struct TraversalInfo                              /* for maintaining the   */
  {                                                 /* state of the          */
    int operand;                                    /* traversal.            */
    D data;                                         /* `D' can be any data   */
  };                                                /* type.                 */

  template<class Atom, class True, class False,     /* Function for          */
           class Not, class Next, class Finally,    /* `traversing' a        */
           class Globally, class And, class Or,     /* FormulaString object. */
           class Implication, class Equivalence,
           class Until, class V, class InfoType>
  void traverse
    (int mode, TraversalInfo<InfoType>& info);

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

  class ParseErrorException : public Exception      /* A class for reporting */
                                                    /* errors when creating  */
                                                    /* or traversing a       */
  {                                                 /* FormulaString.        */
  public:
    ParseErrorException                             /* Constructor. */
      (const string& msg = "Parse error.");

    /* default copy constructor */

    ~ParseErrorException();                         /* Destructor. */

    ParseErrorException& operator=                  /* Assignment operator. */
      (const ParseErrorException& e);
  };

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

private:
  string formula;                                   /* The formula. */

  template<class Atom, class True, class False,     /* Function for          */
           class Not, class Next, class Finally,    /* `traversing' a        */
           class Globally, class And, class Or,     /* formula string,       */
           class Implication, class Equivalence,    /* reading token symbols */
           class Until, class V, class InfoType>    /* from a stream.        */
  static void traverseStream
    (int mode, istream& formula_stream,
     TraversalInfo<InfoType>& info);

  /*
   *  Definitions for classes defining function objects to be used when
   *  traversing a formula string in various contexts (e.g., when creating a
   *  FormulaString object by reading the tokens of the formula from a
   *  stream or translating a FormulaString to a different form).
   */

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   *
   *  Classes for defining function objects used when constructing a
   *  FormulaString object by reading tokens from a stream external to the
   *  object.
   */

  class ReadAtom
  {
  public:
    static void eval
      (FormulaString::TraversalInfo<string*>& info,  const string& atom);
  };

  template<char C>
  class ReadSymbol
  {
  public:
    static void eval(FormulaString::TraversalInfo<string*>& info);
  };

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   *
   *  Classes for function objects used when writing a FormulaString in a
   *  parenthesized infix notation.
   */

  class FormatAtom
  {
  public:
    static void eval
      (FormulaString::TraversalInfo<Exceptional_ostream*>& info,
       const string& atom);
  };

  template<char* C>
  class FormatConstant
  {
  public:
    static void eval(FormulaString::TraversalInfo<Exceptional_ostream*>& info);
  };

  template<char* C>
  class FormatUnary
  {
  public:
    static void eval(FormulaString::TraversalInfo<Exceptional_ostream*>& info);
  };

  template<char* C>
  class FormatBinary
  {
  public:
    static void eval(FormulaString::TraversalInfo<Exceptional_ostream*>& info);
  };

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   *
   *  Classes for function objects used when converting a FormulaString into
   *  negation normal form.
   */

  struct NNFConvertData
  {
    string* result;
    stack<bool, deque<bool, ALLOC(bool) > > negate_stack;
    stack<string*, deque<string*, ALLOC(string*) > > equivalence_operand_stack;
  };
 
  class AtomToNNF
  {
  public:
    static void eval
      (FormulaString::TraversalInfo<NNFConvertData>& info, const string& atom);
  };

  template<bool B>
  class ConstantToNNF
  {
  public:
    static void eval(FormulaString::TraversalInfo<NNFConvertData>& info);
  };

  class NegationToNNF
  {
  public:
    static void eval(FormulaString::TraversalInfo<NNFConvertData>& info);
  };

  class NextToNNF
  {
  public:
    static void eval(FormulaString::TraversalInfo<NNFConvertData>& info);
  };

  template<char OP1, char OP2, bool C>
  class AbbreviatedUnaryToNNF
  {
  public:
    static void eval(FormulaString::TraversalInfo<NNFConvertData>& info);
  };

  template<char OP1, char OP2, bool B1, bool B2>
  class BinaryToNNF
  {
  public:
    static void eval(FormulaString::TraversalInfo<NNFConvertData>& info);
  };

  class EquivalenceToNNF
  {
  public:
    static void eval(FormulaString::TraversalInfo<NNFConvertData>& info);
  };

  friend class FormulaString::EquivalenceToNNF;
};



/******************************************************************************
 *
 * Inline and template function definitions for class FormulaString.
 *
 *****************************************************************************/

/* ========================================================================= */
inline FormulaString::FormulaString()
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructor for class FormulaString. Initializes an empty
 *                FormulaString object.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline FormulaString::~FormulaString()
/* ----------------------------------------------------------------------------
 *
 * Description:   Destructor for class FormulaString.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
template<class Atom, class True, class False, class Not, class Next,
         class Finally, class Globally, class And, class Or, class Implication,
         class Equivalence, class Until, class V, class InfoType>
inline void FormulaString::traverse(int mode, TraversalInfo<InfoType>& info)
/* ----------------------------------------------------------------------------
 *
 * Description:   Template function for `traversing' a FormulaString. The
 *                function implements a general algorithm for traversing an
 *                imaginary tree representation of a FormulaString (the formula
 *                can be represented as a tree whose nodes correspond to the
 *                logical operators in the formula and whose leaves represent
 *                the propositional variables or constants in the formula).
 *
 *                The traversal algorithm performs a depth-first search in the 
 *                formula tree. The user may define what is to be done at each
 *                node of the tree by instatiating the template appropriately. 
 *                The user may also choose when this action is to be done (e.g.
 *                before or after visiting the successors of the node).
 *
 *                The template parameters define the actions to be performed
 *                when encountering a node of a specific type during the
 *                depth-first-search:
 *
 *                    `Atom'         defines the action to be performed in the
 *                                   tree leaves with a propositional variable
 *
 *                    `True',        define the actions to be performed in the
 *                    `False'        tree leaves with a propositional constant
 *
 *                    `Not',         define the actions to be performed at a
 *                    `Next',        tree node with the corresponding operator
 *                    `Finally',
 *                    `Globally',
 *                    `And',
 *                    `Or',
 *                    `Implication',
 *                    `Equivalence',
 *                    `Until'
 *                    `V'
 *
 *                To maintain information about the state of the traversal, an 
 *                object of type TraversalInfo<InfoType> is used. `InfoType' is
 *                any data type chosen by the user. It might e.g. be a stack
 *                for storing partial results gathered during the tree
 *                traversal or anything that best suits the user's needs,
 *                depending on the context.
 * 
 *                The TraversalInfo object is given to the template function
 *                as an argument when requesting a traversal of the
 *                FormulaString. This object is then always passed to the
 *                action functions in order for them to be able to access the
 *                state of the traversal.
 *
 *                The TraversalInfo structure contains two attributes for
 *                accessing the state of the traversal:
 *
 *                   operand  --  An integer telling the specific state of the
 *                                traversal in some node. May be either 0, 1 or
 *                                2. The values have the following meaning:
 *                                  0:  No successors of the node have been
 *                                      visited yet in the depth-first search.
 *                                  1:  One successor has been visited.
 *                                  2:  Two successors have been visited.
 *                                  (no other values are possible since each
 *                                  node can have at most two successors)
 *
 *                   data     --  User-definable data type containing all the
 *                                additional information required (e.g. it
 *                                can be the stack containing the partial
 *                                results; each action the manipulates this
 *                                stack, leaving the final result on the top
 *                                of the stack at the end of the traversal).
 *
 *                To implement the actions, each of the template parameter
 *                classes (excluding `InfoType') must define a public static
 *                member function `eval' that can be called with one
 *                TraversalInfo<InfoType>& argument, and in the case of
 *                `Atom', an additional constant string argument (the name of
 *                the atom).
 *
 *                There are a few examples on the use of this function in the
 *                files FormulaString.* .
 *
 * Arguments:     mode  --  Determines when the action functions are to be
 *                          called. This may be a combination of three
 *                          values:
 *                             FormulaString::PREORDER :
 *                                 The action is performed before visiting
 *                                 the successors of the node.
 *                             FormulaString::INORDER :
 *                                 The action is performed after visiting
 *                                 the nodes in the left subtree starting
 *                                 from the node but before the nodes in the
 *                                 right subtree. Note: For unary operators,
 *                                 INORDER has the same meaning as PREORDER.
 *                             FormulaString::POSTORDER :
 *                                 The action is performed after visiting
 *                                 all successors of the node.
 *                          Any combination of these values is also allowed;
 *                          thus, a value `PREORDER | POSTORDER' will cause
 *                          the action to be performed before and after
 *                          processing the successors of the node. (When
 *                          using the combinations, `info.operand' can assist
 *                          in determining how many times the action has been
 *                          called for the node.)
 *                info  --  A reference to a TraversalInfo<InfoType> structure
 *                          for maintaining the status of the traversal.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  istrstream stream(formula.c_str());
  FormulaString::template
    traverseStream<Atom, True, False, Not, Next, Finally, Globally, And, Or,
                   Implication, Equivalence, Until, V>
    (mode, stream, info);
}

/* ========================================================================= */
template<class Atom, class True, class False, class Not, class Next,
         class Finally, class Globally, class And, class Or, class Implication,
         class Equivalence, class Until, class V, class InfoType>
void FormulaString::traverseStream
  (int mode, istream& formula_stream, TraversalInfo<InfoType>& info)
/* ----------------------------------------------------------------------------
 *
 * Description:   The implementation of the traversal algorithm. This function
 *                reads the tokens from an input stream and then performs the
 *                actions according to the structure of the formula.
 *
 * Arguments:     mode            --  Traversal mode.
 *                formula_stream  --  Stream from where formula tokens are to
 *                                    be read.
 *                info            --  A data structure maintaining the state
 *                                    of the traversal.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  typedef pair<char, int> OperatorStackElement;

  string token;
  stack<OperatorStackElement, deque<OperatorStackElement,
                                    ALLOC(OperatorStackElement) >  >
    operator_stack;
  bool finished = false;
  FormulaSymbolType symbol;

  Exceptional_istream estream(&formula_stream, ios::failbit | ios::badbit);  

  try
  {
    while (!finished)
    {
      estream >> token;
        
      symbol = FormulaString::parseToken(token);
      
      switch (symbol)
      {
        case PROPOSITIONALSYMBOL :
        {
          /*
           *  Perform the action associated with a propositional variable or a
           *  constant.
           */

          switch (token[0])
          {
            case LTL_PROPOSITION :  Atom::eval(info, token); break;
            case LTL_TRUE :         True::eval(info); break;
            case LTL_FALSE :       False::eval(info); break;
          }
        
          /*
           *  Remove previous operators from the stack, performing inorder or
           *  postorder actions if necessary, until the stack becomes empty or
           *  a binary operator with an unprocessed successor is found in the
           *  stack.
           */

          bool continue_stack_unwinding = true;
          char op;

          while (!operator_stack.empty() && continue_stack_unwinding)
          {
            op = operator_stack.top().first;
            info.operand = operator_stack.top().second + 1;
            operator_stack.pop();
          
            symbol = FormulaString::parseToken(string(1, op));
          
            if (((symbol == UNARYOPERATOR) && (mode & POSTORDER))
                || ((symbol == BINARYOPERATOR)
                    && (((mode & INORDER) && info.operand == 1)
                        || ((mode & POSTORDER) && info.operand == 2))))
            {
              switch (op)
              {
                case LTL_NOT :                 Not::eval(info); break;
                case LTL_NEXT :               Next::eval(info); break;
                case LTL_FINALLY :         Finally::eval(info); break;
                case LTL_GLOBALLY :       Globally::eval(info); break;
                case LTL_AND :                 And::eval(info); break;
                case LTL_OR :                   Or::eval(info); break;
                case LTL_IMPLICATION : Implication::eval(info); break;
                case LTL_EQUIVALENCE : Equivalence::eval(info); break;
                case LTL_UNTIL :             Until::eval(info); break;
                case LTL_V :                     V::eval(info); break;
              }
            }
          
            if (symbol == BINARYOPERATOR && info.operand == 1)
            {
              operator_stack.push(make_pair(op, info.operand));
              continue_stack_unwinding = false;
            }
          }

          /*
           *  If the operator stack is empty, the traversal is complete.
           */
        
          if (operator_stack.empty())
            finished = true;
        }
      
        break;
      
      case UNARYOPERATOR :
      case BINARYOPERATOR :
        /*
         *  If necessary, perform a preorder or an inorder action of the
         *  operator. Then, push the operator on the operator stack.
         */

        if (mode & PREORDER 
            || (symbol == UNARYOPERATOR && (mode & INORDER)))
        {
          info.operand = 0;
          switch (token[0])
          {
            case LTL_NOT :                 Not::eval(info); break;
            case LTL_NEXT :               Next::eval(info); break;
            case LTL_FINALLY :         Finally::eval(info); break;
            case LTL_GLOBALLY :       Globally::eval(info); break;
            case LTL_AND :                 And::eval(info); break;
            case LTL_OR :                   Or::eval(info); break;
            case LTL_IMPLICATION : Implication::eval(info); break;
            case LTL_EQUIVALENCE : Equivalence::eval(info); break;
            case LTL_UNTIL :             Until::eval(info); break;
            case LTL_V :                     V::eval(info); break;
          }
        }
      
        operator_stack.push(make_pair(token[0], 0));
        break;
      
      default :
        throw ParseErrorException("Error reading LTL formula: Unrecognized "
                                  "token: `" + token + "'.");
        break;
      }
    }
  }
  catch (IOException&)
  {
    if (formula_stream.eof())
      throw IOException("Error reading LTL formula: Unexpected end of input.");

    throw IOException("Error reading LTL formula: I/O error.");
  }
}

/* ========================================================================= */
inline ostream& operator<<
  (ostream& stream, const FormulaString::FormulaString& formula_string)
/* ----------------------------------------------------------------------------
 *
 * Description:   Writes a FormulaString to a stream in prefix notation. For
 *                parenthesized infix output, use FormulaString::prettyPrint.
 *
 * Arguments:     stream          --  A reference to an output stream.
 *                formula_string  --  A reference to the constant FormulaString
 *                                    to be written to the stream.
 *
 * Returns:       A reference to the output stream.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);
  estream << formula_string.formula;
  return stream;
}

/* ========================================================================= */
inline istream& operator>>
  (istream& stream, FormulaString::FormulaString& formula_string)
/* ----------------------------------------------------------------------------
 *
 * Description:   Reads a FormulaString from a stream.
 *
 * Arguments:     stream          --  A reference to an input stream.
 *                formula_string  --  A reference to a FormulaString.
 * 
 * Returns:       A reference to the input stream.
 * 
 * ------------------------------------------------------------------------- */
{
  formula_string.read(stream);
  return stream;
}



/******************************************************************************
 *
 * Inline function definitions for class FormulaString::ParseErrorException.
 *
 *****************************************************************************/

/* ========================================================================= */
inline FormulaString::ParseErrorException::ParseErrorException
  (const string& msg) :
  Exception(msg)
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructor for class FormulaString::ParseErrorException.
 *                Initializes an exception object with an error message.
 *
 * Argument:      msg  --  Error message.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline FormulaString::ParseErrorException::~ParseErrorException()
/* ----------------------------------------------------------------------------
 *
 * Description:   Destructor for class FormulaString::ParseErrorException.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline FormulaString::ParseErrorException&
FormulaString::ParseErrorException::operator=(const ParseErrorException& e)
/* ----------------------------------------------------------------------------
 *
 * Description:   Assignment operator for class
 *                FormulaString::ParseErrorException.
 *
 * Argument:      e  --  A reference to a constant exception object of the
 *                       same type.
 *
 * Returns:       A reference to the exception object assigned to.
 *
 * ------------------------------------------------------------------------- */
{
  Exception::operator=(e);
  return *this;
}



/******************************************************************************
 *
 * Function definitions for template class FormulaString::ReadSymbol.
 *
 *****************************************************************************/

/* ========================================================================= */
template<char C>
inline void FormulaString::ReadSymbol::eval
  (FormulaString::TraversalInfo<string*>& info)
/* ----------------------------------------------------------------------------
 * 
 * Description:   Appends a symbol (a propositional constant or an operator,
 *                given as a template parameter) read from some stream to the
 *                string under construction. Used by
 *                FormulaString::read(istream&) when traversing a formula in
 *                an input stream.
 *
 * Arguments:     info  --  A reference to the TraversalInfo structure
 *                          containing the partially constructed result.
 *
 * Returns:       Nothing.
 * 
 * ------------------------------------------------------------------------- */
{
  *(info.data) += C;
  *(info.data) += ' ';
}



/******************************************************************************
 *
 * Inline function definitions for class FormulaString::ReadAtom.
 *
 *****************************************************************************/

/* ========================================================================= */
inline void FormulaString::ReadAtom::eval
  (FormulaString::TraversalInfo<string*>& info, const string& atom)
/* ----------------------------------------------------------------------------
 *
 * Description:   Appends a propositional atom read from some stream to the
 *                string being constructed. Used by
 *                FormulaString::read(istream&) when traversing a formula in 
 *                an input stream.
 *
 * Arguments:     info  --  A reference to the TraversalInfo structure
 *                          containing the partially constructed result.
 *                atom  --  Propositional atom read from a stream.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  *(info.data) += string(atom) + ' ';
}



/******************************************************************************
 *
 * Function definitions for template class FormulaString::FormatConstant.
 *
 *****************************************************************************/

/* ========================================================================= */
template<char* C>
inline void FormulaString::FormatConstant::eval
  (FormulaString::TraversalInfo<Exceptional_ostream*>& info)
/* ----------------------------------------------------------------------------
 *
 * Description:   Writes a propositional constant (given as a template
 *                parameter) to a stream. Used by FormulaString::prettyPrint()
 *                when traversing a FormulaString to translate it into
 *                parenthesized infix notation.
 *
 * Argument:      info  --  A reference to a TraversalInfo structure containing
 *                          a pointer to an output stream.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  *(info.data) << C;
}



/******************************************************************************
 *
 * Function definitions for template class FormulaString::FormatUnary.
 *
 *****************************************************************************/

/* ========================================================================= */
template<char* C>
inline void FormulaString::FormatUnary::eval
  (FormulaString::TraversalInfo<Exceptional_ostream*>& info)
/* ----------------------------------------------------------------------------
 *
 * Description:   Writes an unary operator (given as a template parameter) to a
 *                stream. Used by FormulaString::prettyPrint() when traversing
 *                a FormulaString to translate it into parenthesized infix
 *                notation.
 *
 * Argument:      info  --  A reference to a TraversalInfo structure containing
 *                          a pointer to an output stream.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  if (info.operand == 0)
    *(info.data) << C << ' ';
}



/******************************************************************************
 *
 * Function definitions for template class FormulaString::FormatBinary.
 *
 *****************************************************************************/

/* ========================================================================= */
template<char* C>
inline void FormulaString::FormatBinary::eval
  (FormulaString::TraversalInfo<Exceptional_ostream*>& info)
/* ----------------------------------------------------------------------------
 *
 * Description:   Translates a binary subformula (the binary operator is given
 *                as a template parameter) of a FormulaString object into
 *                parenthesized infix notation. Used by
 *                FormulaString::prettyPrint() when traversing a FormulaString.
 *
 * Argument:      info  --  A reference to a TraversalInfo structure containing
 *                          a pointer to an output stream.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  switch (info.operand)
  {
    case 0 : *(info.data) << '('; break;
    case 1 : *(info.data) << ' ' << C << ' '; break;
    default : *(info.data) << ')'; break;
  }
}



/******************************************************************************
 *
 * Inline function definitions for class FormulaString::FormatAtom.
 *
 *****************************************************************************/

/* ========================================================================= */
inline void FormulaString::FormatAtom::eval
  (FormulaString::TraversalInfo<Exceptional_ostream*>& info,
   const string& atom)
/* ----------------------------------------------------------------------------
 *
 * Description:   Writes a propositional variable to a stream. Used by
 *                FormulaString::prettyPrint() when traversing a formula to
 *                translate it into parenthesized infix notation.
 *
 * Arguments:     info  --  A reference to a TraversalInfo structure containing
 *                          a pointer to an output stream.
 *                atom  --  A propositional variable in string representation.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  *(info.data) << atom;
}



/******************************************************************************
 *
 * Inline function definitions for class FormulaString::AtomToNNF.
 *
 *****************************************************************************/

/* ========================================================================= */
inline void FormulaString::AtomToNNF::eval
  (FormulaString::TraversalInfo<NNFConvertData>& info, const string& atom)
/* ----------------------------------------------------------------------------
 *
 * Description:   Converts a propositional atom to negation normal form.
 *                Used by FormulaString::toNegationNormalForm() when traversing
 *                a FormulaString to perform the conversion.
 *
 *                This function implements the conversions
 *                    p  ==>  p  and
 *                   !p  ==> !p  .
 *
 * Arguments:     info  --  A reference to a TraversalInfo object containing
 *                          the information whether the propositional
 *                          variable needs to be negated, in addition with
 *                          the partially constructed result.
 *                atom  --  A propositional variable in string representation.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  if (info.data.negate_stack.top())
  {
    *(info.data.result) += LTL_NOT;
    *(info.data.result) += ' ';
  }
  
  *(info.data.result) += atom;
  
  info.data.negate_stack.pop();
  if (!info.data.negate_stack.empty())
    *(info.data.result) += ' ';
}



/******************************************************************************
 *
 * Function definitions for template class FormulaString::ConstantToNNF.
 *
 *****************************************************************************/

/* ========================================================================= */
template<bool B>
inline void FormulaString::ConstantToNNF::eval
  (FormulaString::TraversalInfo<NNFConvertData>& info)
/* ----------------------------------------------------------------------------
 *
 * Description:   Converts a propositional constant to negation normal form.
 *                Used by FormulaString::toNegationNormalForm() when traversing
 *                a FormulaString to perform the conversion.
 *
 *                This function implements the conversions
 *                   true  ==>  true        false  ==>  false
 *                  !true  ==> false       !false  ==>  true
 *
 * Argument:      info  --  A reference to a TraversalInfo object containing
 *                          the information whether the propositional
 *                          constant needs to be negated, in addition with
 *                          the partially constructed result.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  *(info.data.result) += ((info.data.negate_stack.top() == B) ? LTL_FALSE :
                                                                LTL_TRUE);
  info.data.negate_stack.pop();
  if (!info.data.negate_stack.empty())
    *(info.data.result) += ' ';
}



/******************************************************************************
 *
 * Inline function definitions for class FormulaString::NegationToNNF.
 *
 *****************************************************************************/

/* ========================================================================= */
inline void FormulaString::NegationToNNF::eval
  (FormulaString::TraversalInfo<NNFConvertData>& info)
/* ----------------------------------------------------------------------------
 *
 * Description:   Converts a negated propositional formula to negation normal
 *                form. Used by FormulaString::toNegationNormalForm() when
 *                traversing a FormulaString to perform the conversion.
 *
 *                This function implements the conversions
 *                    !p   ==>  NNF(p)   and
 *                   !!p   ==>  p        .
 *
 * Argument:      info  --  A reference to a TraversalInfo object containing
 *                          the information whether a negation needs to be
 *                          pushed over the negation operator (thus cancelling
 *                          the negation), in addition with the partially
 *                          constructed result.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  if (info.operand == 0)
    info.data.negate_stack.top() = !info.data.negate_stack.top();
}


 
/******************************************************************************
 *
 * Inline function definitions for class FormulaString::NextToNNF.
 *
 *****************************************************************************/

/* ========================================================================= */
inline void FormulaString::NextToNNF::eval
  (FormulaString::TraversalInfo<NNFConvertData>& info)
/* ----------------------------------------------------------------------------
 *
 * Description:   Converts a propositional formula with a Next operator to
 *                negation normal form. Used by 
 *                FormulaString::toNegationNormalForm() when traversing a
 *                FormulaString to perform the conversion.
 *
 *                This function implements the conversions
 *                    Xp  ==>  Xp   and
 *                   !Xp  ==>  X!p  .
 *
 * Argument:      info  --  A reference to a TraversalInfo object containing
 *                          the information whether a negation needs to be
 *                          pushed over the Next operator, in addition with the
 *                          partially constructed result.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  if (info.operand == 0)
  {
    *(info.data.result) += LTL_NEXT;
    *(info.data.result) += ' ';
  }
}



/******************************************************************************
 *
 * Function definitions for template class
 * FormulaString::AbbreviatedUnaryToNNF.
 *
 *****************************************************************************/

/* ========================================================================= */
template<char OP1, char OP2, bool C>
inline void FormulaString::AbbreviatedUnaryToNNF::eval
  (FormulaString::TraversalInfo<NNFConvertData>& info)
/* ----------------------------------------------------------------------------
 *
 * Description:   Converts a propositional formula with either a Finally or a
 *                Globally operator to negation normal form, expanding the
 *                operator to an equivalent representation. Used by 
 *                FormulaString::toNegationNormalForm() when traversing a
 *                FormulaString to perform the conversion.
 *
 *                This function implements the conversions
 *                    Fp  ==>  true U p        Gp  ==>  false V p
 *                   !Fp  ==>  false V !p     !Gp  ==>  true U !p
 *                instantiated by the template parameters.
 *
 * Argument:      info  --  A reference to a TraversalInfo object containing
 *                          the information whether a negation needs to be
 *                          pushed over the operator, in addition with the
 *                          partially constructed result. The presence of a
 *                          negation affects the way the operator is expanded.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  if (info.operand == 0)
  {
    bool negate = info.data.negate_stack.top();
    *(info.data.result) += (negate ? OP2 : OP1);
    *(info.data.result) += ' ';
    *(info.data.result) += (negate == C) ? LTL_TRUE : LTL_FALSE;
    *(info.data.result) += ' ';
  }
}


 
/******************************************************************************
 *
 * Function definitions for template class FormulaString::BinaryToNNF.
 *
 *****************************************************************************/

/* ========================================================================= */
template<char OP1, char OP2, bool B1, bool B2>
inline void FormulaString::BinaryToNNF::eval
  (FormulaString::TraversalInfo<NNFConvertData>& info)
/* ----------------------------------------------------------------------------
 *
 * Description:   Converts a propositional formula with a binary operator
 *                (other than an equivalence operator) to negation normal
 *                form. Used by 
 *                FormulaString::toNegationNormalForm() when traversing a
 *                FormulaString to perform the conversion.
 *
 *                This function implements the conversions
 *                  p1 /\ p2   ==>   p1 /\ p2        p1 \/ p2   ==>   p1 \/ p2
 *                !(p1 /\ p2)  ==>  !p1 \/ !p2     !(p1 \/ p2)  ==>  !p1 /\ !p2
 *
 *                  p1 -> p2   ==>  !p1 \/ p2      !(p1 -> p2)  ==>   p1 \/ !p2
 *
 *                   p1 U p2   ==>   p1 U p2          p1 V p2   ==>   p1 V p2
 *                 !(p1 U p2)  ==>  !p1 V !p2       !(p1 V p2)  ==>  !p1 U !p2
 *                instantiated by the template parameters.
 *
 * Argument:      info  --  A reference to a TraversalInfo object containing
 *                          the information whether a negation needs to be
 *                          pushed over the operator, in addition with the
 *                          partially constructed result. The presence of a
 *                          negation affects the way of rewriting the operator.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  if (info.operand == 0)
  {
    bool negate = info.data.negate_stack.top();
    
    info.data.negate_stack.top() = (negate ? B2 : !B2);
    info.data.negate_stack.push(negate ? B1 : !B1);
    
    *(info.data.result) += (negate ? OP2 : OP1);
    *(info.data.result) += ' ';
  }
}



/******************************************************************************
 *
 * Function definitions for template class FormulaString::EquivalenceToNNF.
 *
 *****************************************************************************/

/* ========================================================================= */
inline void FormulaString::EquivalenceToNNF::eval
  (FormulaString::TraversalInfo<NNFConvertData>& info)
/* ----------------------------------------------------------------------------
 *
 * Description:   Converts a propositional formula with an equivalence operator
 *                to negation normal form. Used by 
 *                FormulaString::toNegationNormalForm() when traversing a
 *                FormulaString to perform the conversion.
 *
 *                This function implements the conversions
 *                    p1 <-> p2   ==>  (p1 \/ !p2) /\ (!p1 \/ p2)
 *                  !(p1 <-> p2)  ==>  (p1 /\ !p2) \/ (!p1 \/ p2) 
 *
 * Argument:      info  --  A reference to a TraversalInfo object containing
 *                          the information whether a negation needs to be
 *                          pushed over the operator, in addition with the
 *                          partially constructed result. The presence of a
 *                          negation affects the way of rewriting the operator.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  switch (info.operand)
  {
    case 0 :
      {
        bool negate = info.data.negate_stack.top();
        *(info.data.result) += (negate ? LTL_OR : LTL_AND);
        *(info.data.result) += ' ';
        *(info.data.result) += (negate ? LTL_AND : LTL_OR);
        *(info.data.result) += ' ';
        
        info.data.equivalence_operand_stack.push(info.data.result);
        info.data.result = new string;
        
        info.data.negate_stack.push(false);
      }
      break;
        
    case 1 :
      info.data.equivalence_operand_stack.push(info.data.result);
      info.data.result = new string;
      
      info.data.negate_stack.push(false);
      break;
        
    case 2 :
      {
        /*
         *  The handling of equivalences must be done in a special way.
         *  After both of the (unnegated) operands of the equivalence operator
         *  have been processed, the operand formulae are negated and converted
         *  to negation normal form independently by traversing both
         *  subformulae separately (the recursive calls to
         *  FormulaString::traverseStream). Then, the result is constructed
         *  from the four computed subformulae.
         */

        string* left_operand = info.data.equivalence_operand_stack.top();
        info.data.equivalence_operand_stack.pop();
        string* right_operand = info.data.result;

        /*
         *  Compute the negation of the left operand of the equivalence
         *  operator in negation normal form.
         */
        
        istrstream left_operand_stream(left_operand->c_str());
          
        info.data.result = new string;
        info.data.negate_stack.push(true);

        FormulaString::template
          traverseStream<AtomToNNF, FormulaString::ConstantToNNF<true>,
                         FormulaString::ConstantToNNF<false>, NegationToNNF,
                         NextToNNF,
                         FormulaString::AbbreviatedUnaryToNNF
                           <LTL_UNTIL, LTL_V, false>,
                         FormulaString::AbbreviatedUnaryToNNF
                           <LTL_V, LTL_UNTIL, true>,
                         FormulaString::BinaryToNNF<LTL_AND, LTL_OR, true,
                                                    true>,
                         FormulaString::BinaryToNNF<LTL_OR, LTL_AND, true,
                                                    true>,
                         FormulaString::BinaryToNNF<LTL_OR, LTL_AND, false,
                                                    true>,
                         EquivalenceToNNF,
                         FormulaString::BinaryToNNF<LTL_UNTIL, LTL_V, true, 
                                                    true>,
                         FormulaString::BinaryToNNF<LTL_V, LTL_UNTIL, true,
                                                    true>
                        >
          (PREORDER | INORDER | POSTORDER, left_operand_stream, info);

        string* neg_left_operand = info.data.result;
           
        /*
         *  Compute the negation of the right operand of the equivalence
         *  operator in negation normal form.
         */

        istrstream right_operand_stream(right_operand->c_str());
          
        info.data.result = new string;
        info.data.negate_stack.push(true);

        FormulaString::template
          traverseStream<AtomToNNF, FormulaString::ConstantToNNF<true>,
                         FormulaString::ConstantToNNF<false>, NegationToNNF,
                         NextToNNF,
                         FormulaString::AbbreviatedUnaryToNNF
                           <LTL_UNTIL, LTL_V, false>,
                         FormulaString::AbbreviatedUnaryToNNF
                           <LTL_V, LTL_UNTIL, true>,
                         FormulaString::BinaryToNNF<LTL_AND, LTL_OR, true,
                                                    true>,
                         FormulaString::BinaryToNNF<LTL_OR, LTL_AND, true,
                                                    true>,
                         FormulaString::BinaryToNNF<LTL_OR, LTL_AND, false,
                                                    true>,
                         EquivalenceToNNF,
                         FormulaString::BinaryToNNF<LTL_UNTIL, LTL_V, true,
                                                    true>,
                         FormulaString::BinaryToNNF<LTL_V, LTL_UNTIL, true,
                                                    true>
                        >
          (PREORDER | INORDER | POSTORDER, right_operand_stream, info);
        
        string* neg_right_operand = info.data.result;

        /*
         *  Construct the result.
         */

        info.operand = 2;

        info.data.result = info.data.equivalence_operand_stack.top();
        info.data.equivalence_operand_stack.pop();
      
        *(info.data.result) += *left_operand + *neg_right_operand
                               + (info.data.negate_stack.top() ? LTL_AND :
                                                                 LTL_OR)
                               + ' ' + *neg_left_operand + *right_operand;
            
        info.data.negate_stack.pop();
        if (info.data.negate_stack.empty())
          info.data.result->resize(info.data.result->size() - 1);
        
        delete left_operand;
        delete neg_left_operand;
        delete right_operand;
        delete neg_right_operand;
      } 
      break;
  }
}

}

#endif  /* _FormulaString_h_ */



