/*
 *  Copyright (C) 1999, 2000  Heikki Tauriainen <Heikki.Tauriainen@hut.fi>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "LtlFormula.h"

namespace LtlFormula
{

/******************************************************************************
 *
 * Function definitions for class LtlFormula.
 *
 *****************************************************************************/

/* ========================================================================= */
LtlFormula& LtlFormula::operator=(const LtlFormula& f)
/* ----------------------------------------------------------------------------
 *
 * Description:   Assignment operator for class LtlFormula.
 *
 * Argument:      f  --  A reference to a constant LtlFormula that is assigned
 *                       to this LtlFormula.
 *
 * Returns:       A reference to the LtlFormula whose value was changed.
 *
 * ------------------------------------------------------------------------- */
{
  if (&f != this)
  {
    info_flags = f.info_flags;
  }
 
  return *this;
}

/* ========================================================================= */
bool LtlFormula::evaluatePath
  (const StateSpace& reachability_graph,
   const vector<StateSpace::size_type, ALLOC(StateSpace::size_type) >&
     path_states,
   StateSpace::size_type loop_state,
   Bitset* result_set) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Evaluates the formula in a path consisting of a subset of
 *                states of a given StateSpace.
 *
 * Arguments:     reachability_graph  --  A reference to a constant StateSpace
 *                                        from which the path states will be
 *                                        chosen.
 *                path_states         --  A vector of state identifiers
 *                                        representing the sequence of states
 *                                        which belong to the path.
 *                loop_state          --  Index to the `path_states' vector to
 *                                        indicate the beginning of the
 *                                        infinite state cycle in the path.
 *                result_set          --  Pointer to a Bitset for storing the
 *                                        evaluation result (i.e. truth value
 *                                        of the formula in the states of the
 *                                        path). If this argument is 0, the
 *                                        `global' result will be discarded.
 *
 * Returns:       Truth value of the formula in the first state of the path.
 *
 * ------------------------------------------------------------------------- */
{
  if (path_states.empty() || loop_state >= path_states.size())
    return false;

  map<const LtlFormula*, Bitset*, ptr_less, ALLOC(Bitset*) > result_cache;

  evalPath(reachability_graph, path_states, loop_state, result_cache);

  bool result = (*result_cache[this])[0];

  if (result_set != 0)
    *result_set = Bitset(*result_cache[this]);

  for (map<const LtlFormula*, Bitset*, ptr_less, ALLOC(Bitset*) >::iterator
         cache_iterator = result_cache.begin();
       cache_iterator != result_cache.end();
       ++cache_iterator)
    delete cache_iterator->second;

  return result;
}

/* ========================================================================= */
bool LtlFormula::displayPathProof
  (ostream& stream, int indent, const StateSpace& reachability_graph,
   const vector<StateSpace::size_type, ALLOC(StateSpace::size_type) >&
     path_states,
   StateSpace::size_type loop_state) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Displays a proof or a refutation for the formula in a path
 *                consisting of a subset of states of a given StateSpace.
 *
 * Arguments:     stream              --  A reference to an output stream.
 *                indent              --  Number of spaces to leave to the left
 *                                        of output.
 *                reachability_graph  --  A reference to a constant StateSpace
 *                                        from which the path states will be
 *                                        chosen.
 *                path_states         --  A vector of state identifiers
 *                                        representing the sequence of states
 *                                        which belong to the path.
 *                loop_state          --  Index to the `path_states' vector to
 *                                        indicate the beginning of the
 *                                        infinite state cycle in the path.
 *
 * Returns:       Truth value of the formula in the first state of the path.
 *
 * ------------------------------------------------------------------------- */
{
  if (path_states.empty() || loop_state >= path_states.size())
    return false;

  map<const LtlFormula*, Bitset*, ptr_less, ALLOC(Bitset*) > result_cache;

  evalPath(reachability_graph, path_states, loop_state, result_cache);

  bool result = (*result_cache[this])[0];
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);

  try
  {
    prettyPrintPathProof(estream, indent, result, 0, result_cache,
			 path_states, loop_state);
  }
  catch (...)
  {
    for (map<const LtlFormula*, Bitset*, ptr_less, ALLOC(Bitset*) >::iterator
	   cache_iterator = result_cache.begin();
	 cache_iterator != result_cache.end();
	 ++cache_iterator)
      delete cache_iterator->second;

    throw;
  }

  for (map<const LtlFormula*, Bitset*, ptr_less, ALLOC(Bitset*) >::iterator
	 cache_iterator = result_cache.begin();
       cache_iterator != result_cache.end();
       ++cache_iterator)
    delete cache_iterator->second;

  return result;
}

/* ========================================================================= */
bool LtlFormula::evaluatePath(const StateSpace& path, Bitset* result_set) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Evaluates the formula in a StateSpace, assumed to consist of
 *                a single infinite path.
 *
 * Arguments:     path        --  A reference to a constant StateSpace, assumed
 *                                to consist of a single infinite path.
 *                result_set  --  Pointer to a Bitset for storing the
 *                                evaluation result (i.e. truth value of the
 *                                formula in the states of the path). If this
 *                                argument is 0, the `global' result will be
 *                                discarded.
 *
 * Returns:       Truth value of the formula in the first state of the path.
 *
 * ------------------------------------------------------------------------- */
{
  if (path.empty())
    return false;

  vector<StateSpace::size_type, ALLOC(StateSpace::size_type) > path_states;
  map<StateSpace::size_type, StateSpace::size_type,
      less<StateSpace::size_type>, ALLOC(StateSpace::size_type) > ordering;

  StateSpace::size_type state = path.initialState();
  StateSpace::size_type state_count = 0;

  /*
   *  Construct a vector of state identifiers representing the path by
   *  traversing the state space until some state is encountered twice.
   *  Note: This algorithm will not work correctly if the assumption regarding
   *        the structure of the state space is violated!
   */

  while (1)
  {
    path_states.push_back(state);
    ordering[state] = state_count;
    state_count++;

    state = (*(path[state].edges().begin()))->targetNode();

    if (ordering.find(state) != ordering.end())
      break;
  }

  return evaluatePath(path, path_states, ordering[state], result_set);
}

/* ========================================================================= */
bool LtlFormula::displayPathProof
  (ostream& stream, int indent, const StateSpace& path) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Displays a proof or a refutation for the formula in a
 *                StateSpace (assumed to consist of a single infinite path).
 *
 * Arguments:     stream  --  A reference to an output stream.
 *                indent  --  Number of spaces to leave to the left of output.
 *                path    --  A reference to a constant StateSpace, assumed to
 *                            consist of a single infinite path.
 *
 * Returns:       Truth value of the formula in the first state of the path.
 *
 * ------------------------------------------------------------------------- */
{
  if (path.empty())
    return false;

  vector<StateSpace::size_type, ALLOC(StateSpace::size_type) > path_states;
  map<StateSpace::size_type, StateSpace::size_type,
      less<StateSpace::size_type>, ALLOC(StateSpace::size_type) > ordering;

  StateSpace::size_type state = path.initialState();
  StateSpace::size_type state_count = 0;

  /*
   *  Construct a vector of state identifiers representing the path by
   *  traversing the state space until some state is encountered twice.
   *  Note: This algorithm will not work correctly if the assumption regarding
   *        the structure of the state space is violated!
   */

  while (1)
  {
    path_states.push_back(state);
    ordering[state] = state_count;
    state_count++;

    state = (*(path[state].edges().begin()))->targetNode();

    if (ordering.find(state) != ordering.end())
      break;
  }

  return displayPathProof(stream, indent, path, path_states, ordering[state]);
}

/* ========================================================================= */
void LtlFormula::evalPath
  (const StateSpace& reachability_graph,
   const vector<StateSpace::size_type, ALLOC(StateSpace::size_type) >&
     path_states,
   StateSpace::size_type loop_state,
   map<const LtlFormula*, Bitset*, ptr_less, ALLOC(Bitset*) >& result_cache)
  const
/* ----------------------------------------------------------------------------
 *
 * Description:   Evaluates the LtlFormula in a given infinite state sequence
 *                (whose states are from a given StateSpace) consisting of a
 *                prefix and a loop. This is done by processing the formula
 *                structurally in a bottom-up order, i.e. evaluating first the
 *                `smallest' subformulae of the formula and using their truth
 *                values to compute the truth values of compound subformulae,
 *                until the truth value of the formula itself has been
 *                computed.
 *
 * Arguments:     reachability_graph  --  A reference to a constant StateSpace
 *                                        from which the states in the path
 *                                        are chosen.
 *                path_states         --  A vector of state identifiers forming
 *                                        the state sequence.
 *                loop_state          --  Index to the state identifier vector
 *                                        indicating the beginning of the
 *                                        infinite state cycle.
 *                result_cache        --  Storage location for partial results
 *                                        (truth values of the subformulae).
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  stack<const LtlFormula*, deque<const LtlFormula*,
                                 ALLOC(const LtlFormula*) > > subformula_stack;
  const LtlFormula* f;
  Bitset* val;

  /*
   *  Collect the subformulae of the formula onto a stack from which they can
   *  be accessed in an order which ensures that when processing a subformula,
   *  the truth values of its subformulae are available in the result cache.
   */

  recCollectSubformulae(subformula_stack);

  try
  {
    while (!subformula_stack.empty())
    {
      if (::user_break)
	throw UserBreakException();

      /*
       *  Pop a formula from the subformula stack.
       */

      f = subformula_stack.top();
      subformula_stack.pop();

      /*
       *  Discard the current formula if its truth value is already known.
       */

      if (result_cache.find(f) != result_cache.end())
	continue;

      /*
       *  Otherwise allocate space for the evaluation results of the current
       *  formula and then evaluate it in all states of the path, using the
       *  truth values of its subformulae which have already been stored in the
       *  result cache.
       */

      val = result_cache[f] = new Bitset(path_states.size());

      if (f->what() == LTL_UNTIL || f->what() == LTL_V
	  || f->what() == LTL_FINALLY || f->what() == LTL_GLOBALLY)
      {
	StateSpace::size_type marker_state;
	LtlFormula* g, *h;
	bool check_value;
	bool marker_valid = false;

	switch (f->what())
	{
	  case LTL_UNTIL :
	    g = static_cast<const Until*>(f)->subformula1;
	    h = static_cast<const Until*>(f)->subformula2;
	    check_value = true;
	    break;

	  case LTL_V :
	    val->set();
	    g = static_cast<const V*>(f)->subformula1;
	    h = static_cast<const V*>(f)->subformula2;
	    check_value = false;
	    break;

	  case LTL_FINALLY :
	    g = 0;
	    h = static_cast<const Finally*>(f)->subformula;
	    check_value = true;
	    break;

	  case LTL_GLOBALLY :
	    val->set();
	    g = 0;
	    h = static_cast<const Globally*>(f)->subformula;
	    check_value = false;
	    break;
	      
	  default:
	    break;
	}

	for (StateSpace::size_type state = 0; state < path_states.size();
	     state++)
	{
	  if (result_cache[h]->test(state) == check_value)
          {
	    val->flip(state);

	    if (marker_valid)
            {
	      while (marker_state < state)
              {
		val->flip(marker_state);
		marker_state++;
	      }
	      marker_valid = false;
	    } 
	  }
	  else if (g == 0 || result_cache[g]->test(state) == check_value)
	  {
	    if (!marker_valid)
            {
	      marker_valid = true;
	      marker_state = state;
	    }
	  }
	  else
	    marker_valid = false;
	}

	if (result_cache[f]->test(loop_state) == check_value)
        {
	  if (marker_valid)
	  {
	    while (marker_state < path_states.size())
            {
	      val->flip(marker_state);
	      marker_state++;
	    }
	  }
	}
      }
      else
      {
	switch (f->what())
	{
 	  case LTL_TRUE :
	    val->set();
	    break;

	  case LTL_FALSE :
	    break;

  	  case LTL_ATOM :
	    {
	      for (StateSpace::size_type s = 0; s < path_states.size(); s++)
	      {
		if (f->eval(reachability_graph[path_states[s]].
			      positiveAtoms()))
		    val->set(s);
	      }

	      break;
	    }

	  case LTL_NEGATION :
	    {
	      const LtlFormula* g = static_cast<const Not*>(f)->subformula;

	      for (StateSpace::size_type s = 0; s < path_states.size(); s++)
	      {
		if (!result_cache[g]->test(s))
		  val->set(s);
	      }

	      break;
	    }

	  case LTL_CONJUNCTION :
	    {
	      const LtlFormula* g = static_cast<const And*>(f)->subformula1;
	      const LtlFormula* h = static_cast<const And*>(f)->subformula2;

	      for (StateSpace::size_type s = 0; s < path_states.size(); s++)
	      {
		if (result_cache[g]->test(s) && result_cache[h]->test(s))
		  val->set(s);
	      }

	      break;
	    }

	  case LTL_DISJUNCTION :
	    {
	      const LtlFormula* g = static_cast<const Or*>(f)->subformula1;
	      const LtlFormula* h = static_cast<const Or*>(f)->subformula2;

	      for (StateSpace::size_type s = 0; s < path_states.size(); s++)
	      {
		if (result_cache[g]->test(s) || result_cache[h]->test(s))
		  val->set(s);
	      }

	      break;
	    }

	  case LTL_IMPLICATION :
	    {
	      const LtlFormula* g = static_cast<const Imply*>(f)->subformula1;
	      const LtlFormula* h = static_cast<const Imply*>(f)->subformula2;

	      for (StateSpace::size_type s = 0; s < path_states.size(); s++)
	      {
		if (!result_cache[g]->test(s) || result_cache[h]->test(s))
		  val->set(s);
	      }

	      break;
	    }

	  case LTL_EQUIVALENCE :
	    {
	      const LtlFormula* g = static_cast<const Equiv*>(f)->subformula1;
	      const LtlFormula* h = static_cast<const Equiv*>(f)->subformula2;
		  
	      for (StateSpace::size_type s = 0; s < path_states.size(); s++)
	      {
		if (result_cache[g]->test(s) == result_cache[h]->test(s))
		  val->set(s);
	      }

	      break;
	    }

	  case LTL_NEXT :
	    {
	      const LtlFormula* g = static_cast<const Next*>(f)->subformula;
	      StateSpace::size_type s;

	      for (s = 0; (s + 1) < path_states.size(); s++)
	      {
		if (result_cache[g]->test(s + 1))
		  val->set(s);
	      }

	      if (result_cache[g]->test(loop_state))
		val->set(s);

	      break;
	    }

	  default :
	    assert(0);
	    break;
	}
      }
    }
  }
  catch (...)
  {
    for (map<const LtlFormula*, Bitset*, ptr_less, ALLOC(Bitset*) >::iterator
	   cache_iterator = result_cache.begin();
	 cache_iterator != result_cache.end();
	 ++cache_iterator)
      delete cache_iterator->second;

    throw;
  }
}

/* ========================================================================= */
void LtlFormula::prettyPrintPathProof
  (Exceptional_ostream& estream, int indent, bool prove_true,
   StateSpace::size_type state,
   map<const LtlFormula*, Bitset*, ptr_less, ALLOC(Bitset*) >& result_cache,
   const vector<StateSpace::size_type, ALLOC(StateSpace::size_type) >& path,
   StateSpace::size_type loop_state)
  const
/* ----------------------------------------------------------------------------
 *
 * Description:   Displays a proof or a refutation for a LtlFormula in a path.
 *
 * Arguments:     estream       --  A reference to an Exceptional_ostream.
 *                indent        --  Number of spaces to leave to the left of
 *                                  output.
 *                prove_true    --  Determines whether to show a proof or a
 *                                  refutation for the formula.
 *                state         --  Index to the path vector representing the
 *                                  `current' state in the path.
 *                result_cache  --  A reference to the data structure
 *                                  containing the evaluation results for the
 *                                  formula and its subformula in the path.
 *                path          --  A constant reference to a vector of state
 *                                  identifiers representing the path.
 *                loop_state    --  Index to the path vector denoting the
 *                                  beginning of the infinite state cycle.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  static string line_prefix = "";

  estream << string(indent, ' ')
             + (line_prefix.size() == 0
		? ""
		: line_prefix.substr(0, line_prefix.size() - 4) + "+-> ")
             + "M,<s"
	  << path[state]
          << string(", ...> |") + (prove_true ? '=' : '/') + "= "
             + string(*this);

  switch (what())
  {
    case LTL_ATOM :
    case LTL_TRUE :
    case LTL_FALSE :
      estream << '\n';
      break;

    case LTL_NEGATION :
      /*
       *  To display a proof (refutation) for a negated formula, display a 
       *  refutation (proof) for the non-negated formula.
       */

      line_prefix += "    ";

      estream << " :\n";
      static_cast<const Not*>(this)->subformula->
        prettyPrintPathProof(estream, indent, !prove_true, state,
			     result_cache, path, loop_state);

      line_prefix.resize(line_prefix.size() - 4);
      break;

    case LTL_NEXT :
      {
	/*
	 *  To display a proof or refutation for a Next formula, display a
	 *  proof or refutation for the subformula of the Next operator in
	 *  the next state of the path.
	 */

	estream << " :\n" + string(indent, ' ') + line_prefix + "+-> s"
	        << path[state]
	        << " --> " << 's';

	state++;
	if (state == path.size())
	  state = loop_state;

	estream << path[state] << '\n';

	line_prefix += "    ";

	static_cast<const Next*>(this)->subformula->
	  prettyPrintPathProof(estream, indent, prove_true, state,
			       result_cache, path, loop_state);

	line_prefix.resize(line_prefix.size() - 4);
      }
      break;

    case LTL_CONJUNCTION :
      {
	estream << " :\n";

	line_prefix += "|   ";

	LtlFormula* f1 = static_cast<const And*>(this)->subformula1;
	LtlFormula* f2 = static_cast<const And*>(this)->subformula2;

	if (prove_true)
        {
	  /*
	   *  To display a proof for a conjunctive formula, display the proofs
	   *  for its subformulae.
	   */

	  f1->prettyPrintPathProof(estream, indent, true, state,
				   result_cache, path, loop_state);

	  line_prefix[line_prefix.size() - 4] = ' ';

	  f2->prettyPrintPathProof(estream, indent, true, state,
				   result_cache, path, loop_state);
	}
	else
        {
	  line_prefix[line_prefix.size() - 4] = ' ';

	  /*
	   *  To display a refutation for a conjunctive formula, display a
	   *  refutation for one of its false subformulae.
	   */

	  if (result_cache[f2]->test(state)
	      || (!result_cache[f1]->test(state)
		  && (f1->propositional() && !f2->propositional()
		      || (f1->constant() && !f2->constant())
		      || (string(*f1).length() <= string(*f2).length()))))
	    f1->prettyPrintPathProof(estream, indent, false, state,
				     result_cache, path, loop_state);
	  else
	    f2->prettyPrintPathProof(estream, indent, false, state,
				     result_cache, path, loop_state);
	}

	line_prefix.resize(line_prefix.size() - 4);
      }
      break;

    case LTL_DISJUNCTION :
      {
	estream << " :\n";

	line_prefix += "|   ";

	LtlFormula* f1 = static_cast<const Or*>(this)->subformula1;
	LtlFormula* f2 = static_cast<const Or*>(this)->subformula2;

	if (prove_true)
	{
	  line_prefix[line_prefix.size() - 4] = ' ';

	  /*
	   *  To display a proof for a disjunctive formula, display a proof
	   *  for one of its true subformulae.
	   */

	  if (!result_cache[f2]->test(state)
	      || (result_cache[f1]->test(state)
		  && ((f1->propositional() && !f2->propositional())
		      || (f1->constant() && !f2->constant())
		      || (string(*f1).length() <= string(*f2).length()))))
	    f1->prettyPrintPathProof(estream, indent, true, state,
				     result_cache, path, loop_state);
	  else
	    f2->prettyPrintPathProof(estream, indent, true, state,
				     result_cache, path, loop_state);
	}
	else
	{
	  /*
	   *  To display a refutation for a disjunctive formula, display
	   *  refutations for both of its subformulae.
	   */

	  f1->prettyPrintPathProof(estream, indent, false, state,
				   result_cache, path, loop_state);

	  line_prefix[line_prefix.size() - 4] = ' ';

	  f2->prettyPrintPathProof(estream, indent, false, state,
				   result_cache, path, loop_state);
	}

	line_prefix.resize(line_prefix.size() - 4);
      }
      break;

    case LTL_IMPLICATION :
      {
	estream << " :\n";

	line_prefix += "|   ";

	LtlFormula* f1 = static_cast<const Imply*>(this)->subformula1;
	LtlFormula* f2 = static_cast<const Imply*>(this)->subformula2;

	if (prove_true)
	{
	  line_prefix[line_prefix.size() - 4] = ' ';

	  /*
	   *  To display a proof for an implication, display either a
	   *  refutation for `f1' or a proof for `f2'.
	   */

	  if (!result_cache[f2]->test(state)
	      || (!result_cache[f1]->test(state)
		  && ((f1->propositional() && !f2->propositional())
		      || (f1->constant() && !f2->constant())
		      || (string(*f1).length() <= string(*f2).length()))))
	    f1->prettyPrintPathProof(estream, indent, false, state,
				     result_cache, path, loop_state);
	  else
	    f2->prettyPrintPathProof(estream, indent, true, state,
				     result_cache, path, loop_state);
	}
	else
	{
	  /*
	   *  To display a refutation for an implication, display a proof for
	   *  `f1' and a refutation for `f2'.
	   */

	  f1->prettyPrintPathProof(estream, indent, true, state,
				   result_cache, path, loop_state);

	  line_prefix[line_prefix.size() - 4] = ' ';

	  f2->prettyPrintPathProof(estream, indent, false, state,
				   result_cache, path, loop_state);
	}

	line_prefix.resize(line_prefix.size() - 4);
      }
      break;

    case LTL_EQUIVALENCE :
      {
	estream << " :\n";

	line_prefix += "|   ";

	/*
	 *  To display a proof or a refutation for an equivalence, display
	 *  either a proof or a refutation for both subformulae (a proof for
	 *  the equivalence) or a proof for the other subformula and a
	 *  refutation for the other (a refutation for the equivalence).
	 */

	LtlFormula* f1 = static_cast<const Equiv*>(this)->subformula1;
	LtlFormula* f2 = static_cast<const Equiv*>(this)->subformula2;

	bool f1_proof_style, f2_proof_style;

	if (prove_true)
	  f1_proof_style = f2_proof_style = result_cache[f1]->test(state);
	else
	{
	  f1_proof_style = result_cache[f1]->test(state);
	  f2_proof_style = !f1_proof_style;
	}
	
	f1->prettyPrintPathProof(estream, indent, f1_proof_style, state,
				 result_cache, path, loop_state);

	line_prefix[line_prefix.size() - 4] = ' ';

	f2->prettyPrintPathProof(estream, indent, f2_proof_style, state,
				 result_cache, path, loop_state);

	line_prefix.resize(line_prefix.size() - 4);
      }
      break;

    case LTL_UNTIL :
    case LTL_FINALLY :
      {
	estream << " :\n";

	line_prefix += "|   ";

	LtlFormula* f1;
	LtlFormula* f2;

	if (what() == LTL_UNTIL)
	{
	  f1 = static_cast<const Until*>(this)->subformula1;
	  f2 = static_cast<const Until*>(this)->subformula2;
	}
	else
	  f2 = static_cast<const Finally*>(this)->subformula;

	if (prove_true)
	{
	  /*
	   *  To display a proof for an Until or a Finally formula, scan
	   *  through the states of the path until a state where `f2' is true
	   *  is found and then display its proof. In case of an Until formula,
	   *  display also the proof of `f1' in every visited state where `f2'
	   *  does not hold.
	   */

	  while (!result_cache[f2]->test(state))
	  {
	    if (what() == LTL_UNTIL)
	      f1->prettyPrintPathProof(estream, indent, true, state,
				       result_cache, path, loop_state);

	    estream << string(indent, ' ')
	               + line_prefix.substr(0, line_prefix.size() - 4)
                       + "+-> s"
		    << path[state];

	    state++;
	    if (state == path.size())
	      state = loop_state;

	    estream << " --> " << 's' << path[state] << '\n';
	  }
	  
	  line_prefix[line_prefix.size() - 4] = ' ';

	  f2->prettyPrintPathProof(estream, indent, true, state,
				   result_cache, path, loop_state);
	}
	else
	{
	  /*
	   *  To display a refutation for an Until formula, scan through the
	   *  states of the path until a state where `f1' and `f2' are both
	   *  false is found or until an already visited state is found.
	   *  Unless in an already visited state at the end, display
	   *  refutation proofs for `f1' and `f2'. In each visited state where
	   *  `f1' holds but `f2' doesn't, display a refutation for `f2'.
	   *
	   *  Displaying a refutation for a Finally formula is equivalent to
	   *  the Until case where the search ends in an already visited state.
	   */

	  StateSpace::size_type start_state = state;

	  while (!result_cache[f2]->test(state)
		 && (what() == LTL_FINALLY || result_cache[f1]->test(state)))
	  {
	    f2->prettyPrintPathProof(estream, indent, false, state,
				     result_cache, path, loop_state);

	    estream << string(indent, ' ')
		       + line_prefix.substr(0, line_prefix.size() - 4)
                       + "+-> s"
		    << path[state];

	    state++;
	    if (state == path.size())
	    {
	      state = loop_state;
	      if (start_state < loop_state)
	      {
		estream << " --> " << 's' << path[state] << '\n';
		break;
	      }
	    }

	    estream << " --> " << 's' << path[state] << '\n';

	    if (state == start_state)
	      break;
	  }

	  if (what() == LTL_UNTIL && !result_cache[f1]->test(state))
	  {
	    f1->prettyPrintPathProof(estream, indent, false, state,
				     result_cache, path, loop_state);

	    line_prefix[line_prefix.size() - 4] = ' ';

	    f2->prettyPrintPathProof(estream, indent, false, state,
				     result_cache, path, loop_state);
	  }
	}

	line_prefix.resize(line_prefix.size() - 4);
      }
      break;

    case LTL_V :
    case LTL_GLOBALLY :
      {
	estream << " :\n";

	line_prefix += "|   ";

	LtlFormula* f1;
	LtlFormula* f2;

	if (what() == LTL_V)
	{
	  f1 = static_cast<const V*>(this)->subformula1;
	  f2 = static_cast<const V*>(this)->subformula2;
	}
	else
	  f2 = static_cast<const Globally*>(this)->subformula;

	if (prove_true)
	{
	  /*
	   *  To display a proof for a Release formula, scan through the states
	   *  of the path until a state where `f1' and `f2' are both true is
	   *  found or until an already visited state is found. Unless in an
	   *  already visited state at the end, display the proofs for `f1' and
	   *  `f2'. In each visited state where `f2' holds but `f1' doesn't,
	   *  display the proof for `f2'.
	   *
	   *  Showing a proof for a Globally formula is similar to the Release
	   *  case where the search ends in an already visited state.
	   */

	  StateSpace::size_type start_state = state;

	  while (result_cache[f2]->test(state)
		 && (what() == LTL_GLOBALLY || !result_cache[f1]->test(state)))
	  {
	    f2->prettyPrintPathProof(estream, indent, true,
				     state, result_cache, path, loop_state);

	    estream << string(indent, ' ')
                       + line_prefix.substr(0, line_prefix.size() - 4)
                       + "+-> s"
		    << path[state];

	    state++;
	    if (state == path.size())
	    {
	      state = loop_state;
	      if (start_state < loop_state)
	      {
		estream << " --> " << 's' << path[state] << '\n';
		break;
	      }
	    }

	    estream << " --> " << 's' << path[state] << '\n';

	    if (state == start_state)
	      break;
	  }

	  if (what() == LTL_V && result_cache[f1]->test(state))
	  {
	    f1->prettyPrintPathProof(estream, indent, true, state,
				     result_cache, path, loop_state);

	    line_prefix[line_prefix.size() - 4] = ' ';
	
	    f2->prettyPrintPathProof(estream, indent, true, state,
				     result_cache, path, loop_state);
	  }
	}
	else
	{
	  /*
	   *  To show a refutation for a Release or a Globally formula, scan
	   *  through the states of the path until a state where `f2' is false 
	   *  is found and then display a refutation for it. In case of a
	   *  Release formula, display also a refutation for `f1' in every
	   *  visited state where `f2' holds.
	   */

	  while (result_cache[f2]->test(state))
	  {
	    if (what() == LTL_V)
	      f1->prettyPrintPathProof(estream, indent, false, state,
				       result_cache, path, loop_state);

	    estream << string(indent, ' ')
		       + line_prefix.substr(0, line_prefix.size() - 4)
                       + "+-> s"
		    << path[state];

	    state++;
	    if (state == path.size())
	      state = loop_state;

	    estream << " --> " << 's' << path[state] << '\n';
	  }

	  line_prefix[line_prefix.size() - 4] = ' ';

	  f2->prettyPrintPathProof(estream, indent, false, state,
				   result_cache, path, loop_state);
	}

	line_prefix.resize(line_prefix.size() - 4);
      }
      break;
  }
}

/* ========================================================================= */
Bitset LtlFormula::findPropositionalModel() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Finds a model for a purely propositional formula (if the
 *                formula is satisfiable).
 *
 * Arguments:     None.
 *
 * Returns:       A Bitset defining a truth assignment for the propositional
 *                variables in the formula such that the formula is satisfied.
 *
 * ------------------------------------------------------------------------- */
{
  if (!info_flags.is_propositional)
    throw Exception("findPropositionalModel() called on a temporal formula");

  set<unsigned long int, less<unsigned long int>, ALLOC(unsigned long int) >
    positive_atoms;
  set<unsigned long int, less<unsigned long int>, ALLOC(unsigned long int) >
    negative_atoms;

  if (!sat_eval(true, positive_atoms, negative_atoms))
    throw Exception("formula is not satisfiable");

  unsigned long int max_id = 0;

  for (set<unsigned long int, less<unsigned long int>,
	   ALLOC(unsigned long int) >::const_iterator
	 id_iterator = positive_atoms.begin();
       id_iterator != positive_atoms.end();
       ++id_iterator)
  {
    if (*id_iterator > max_id)
      max_id = *id_iterator;
  }

  Bitset model(max_id + 1);

  for (set<unsigned long int, less<unsigned long int>,
	   ALLOC(unsigned long int) >::const_iterator
	 id_iterator = positive_atoms.begin();
       id_iterator != positive_atoms.end();
       ++id_iterator)
  {
    model.set(*id_iterator);
  }

  return model;
}



/******************************************************************************
 *
 * Initializations for the `symbol' attributes of classes LtlTrue, LtlFalse,
 * LtlNegation, LtlNext, LtlFinally, LtlGlobally, LtlDisjunction,
 * LtlConjunction, LtlImplication, LtlEquivalence, LtlUntil and LtlV.
 *
 *****************************************************************************/

const char* const LtlTrue::symbol        = "true";
const char* const LtlFalse::symbol       = "false";
const char* const LtlNegation::symbol    = "!";
const char* const LtlNext::symbol        = "X";
const char* const LtlFinally::symbol     = "<>";
const char* const LtlGlobally::symbol    = "[]";
const char* const LtlDisjunction::symbol = "\\/";
const char* const LtlConjunction::symbol = "/\\";
const char* const LtlImplication::symbol = "->";
const char* const LtlEquivalence::symbol = "<->";
const char* const LtlUntil::symbol       = "U";
const char* const LtlV::symbol           = "V";



/******************************************************************************
 *
 * Function definitions for class Atom.
 *
 *****************************************************************************/

/* ========================================================================= */
Atom::operator string() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Converts the atom to a string.
 *
 * Arguments:     None.
 *
 * Returns:       A string representation of the atom.
 *
 * ------------------------------------------------------------------------- */
{
  ostrstream atom_stream;
  atom_stream << 'p' << atom << ends;
  atom_stream.freeze();
  string result(atom_stream.str());
  atom_stream.freeze(0);
  return result;
}

/* ========================================================================= */
bool Atom::sat_eval
  (bool prove_true,
   set<unsigned long int, less<unsigned long int>, ALLOC(unsigned long int) >&
     positive_atoms,
   set<unsigned long int, less<unsigned long int>, ALLOC(unsigned long int) >& 
     negative_atoms) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Tests the satisfiability of the atomic proposition (or its
 *                negation). The proposition is satisfiable if and only if its
 *                truth value has not yet been fixed to be false. If the truth
 *                value of the proposition has not yet been fixed, set the
 *                proposition's truth value accordingly to make it (or its
 *                negation) satisfiable.
 *
 * Arguments:     prove_true      --  Determines whether the satisfiability
 *                                    check is to be made for the proposition
 *                                    or its negation.
 *                positive_atoms  --  A set of identifiers for propositions
 *                                    whose truth values have already been
 *                                    set to true.
 *                negative_atoms  --  A set of identifiers for propositions
 *                                    whose truth values have already been
 *                                    set to false.
 *
 * Returns:       A truth value according to the test whether the proposition
 *                or its negation is satisfiable in the ``current'' truth
 *                assignment for the propositions (given by the two sets of
 *                proposition identifiers in the parameters) or can be made
 *                satisfiable by extending the truth assignment.
 *
 * ------------------------------------------------------------------------- */
{
  if ((prove_true && negative_atoms.find(atom) != negative_atoms.end())
      || (!prove_true && positive_atoms.find(atom) != positive_atoms.end()))
    return false;

  if (prove_true)
    positive_atoms.insert(atom);
  else
    negative_atoms.insert(atom);

  return true;
}



/******************************************************************************
 *
 * Definition for the function for converting a FormulaString into an
 * LtlFormula.
 *
 *****************************************************************************/

/* ========================================================================= */
LtlFormula* formulaStringToLtlFormula(FormulaString& f)
/* ----------------------------------------------------------------------------
 *
 * Description:    Creates an LtlFormula from a FormulaString.
 *
 * Argument:       f  --  A reference to a FormulaString object.
 *
 * Returns:        A pointer to a new LtlFormula object constructed from the
 *                 FormulaString.
 *
 * --------------------------------------------------------------------------*/
{
  FormulaString::TraversalInfo<stack<LtlFormula*,
                                     deque<LtlFormula*,
                                           ALLOC(LtlFormula*) > > >
    info;

  /*
   * Construct the LTL formula by `traversing' the FormulaString, using a stack
   * of pointers to LTL formulae to maintain the state of the construction
   * process. If the construction succeeds, the stack will contain a pointer to
   * the desired LtlFormula object. In the case an exception is thrown during
   * the traversal, the stack is emptied and the exception is rethrown.
   */

  try
  {
    f.template traverse<AtomToLtl,
                        ConstantToLtl<True>,
                        ConstantToLtl<False>,
                        UnaryOperatorToLtl<Not>,
                        UnaryOperatorToLtl<Next>,
                        UnaryOperatorToLtl<Finally>,
                        UnaryOperatorToLtl<Globally>,
                        BinaryOperatorToLtl<And>,
                        BinaryOperatorToLtl<Or>,
                        BinaryOperatorToLtl<Imply>,
                        BinaryOperatorToLtl<Equiv>,
                        BinaryOperatorToLtl<Until>,
                        BinaryOperatorToLtl<V> >
      (FormulaString::POSTORDER, info);
  }
  catch (...)
  {
    while (!info.data.empty())
    {
      delete info.data.top();
      info.data.pop();
    }
    throw;
  }

  return info.data.top();
}

}
