/*
 *  Copyright (C) 1999, 2000  Heikki Tauriainen <Heikki.Tauriainen@hut.fi>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef _LtlFormula_h_
#define _LtlFormula_h_

#include <cassert>
#include <deque>
#include <iostream>
#include <map>
#include <set>
#include <stack>
#include <string>
#include <strstream>
#include <vector>
#include "Bitset.h"
#include "EdgeContainer.h"
#include "Exception.h"
#include "FormulaString.h"
#include "MacroDefs.h"
#include "StateSpace.h"

using namespace std;

extern bool user_break;

namespace LtlFormula
{

using FormulaString::FormulaString;
using Graph::StateSpace;

/******************************************************************************
 *
 *  Constants for identifying LTL formulae of various types.
 *
 *****************************************************************************/

enum FormulaType {LTL_TRUE, LTL_FALSE, LTL_ATOM, LTL_NEGATION, LTL_NEXT,
                  LTL_FINALLY, LTL_GLOBALLY, LTL_CONJUNCTION, LTL_DISJUNCTION,
                  LTL_IMPLICATION, LTL_EQUIVALENCE, LTL_UNTIL, LTL_V};



/******************************************************************************
 *
 * A base class for LTL formulae.
 *
 *****************************************************************************/

class LtlFormula
{
protected:
  class ptr_less;                                   /* `Less than' relation
						     * between two pointers to
						     * LtlFormulae.
						     */
public:
  LtlFormula();                                     /* Creates a new LTL
                                                     * formula.
                                                     */

  LtlFormula(const LtlFormula& f);                  /* Copy constructor. */

  virtual ~LtlFormula();                            /* Destructor. */

  LtlFormula& operator=(const LtlFormula& f);       /* Assignment operator. */

  virtual operator string() const = 0;              /* Converts the formula to
                                                     * a string.
						     */

  virtual LtlFormula* clone() const = 0;            /* Creates a copy of the
                                                     * formula object.
                                                     */

  bool evaluate                                     /* Evaluates the formula */
    (const Bitset& valuation) const;                /* in a given truth
                                                     * assignment for the
                                                     * propositional variables.
                                                     * (This function is
                                                     * intended to be used
                                                     * with pure propositional
                                                     * formulae. If the formula
						     * contains temporal
						     * operators, calling this
                                                     * function is interpreted 
                                                     * as a request to evaluate
                                                     * the formula in a state
                                                     * space consisting of a
                                                     * single state having a
						     * transition to itself,
						     * with the given truth
						     * assignment for the
						     * propositional
						     * variables.)
                                                     */

  bool evaluatePath                                 /* Computes the truth    */
    (const StateSpace& reachability_graph,          /* value of the formula  */
     const vector<StateSpace::size_type,            /* in the first state of */
                  ALLOC(StateSpace::size_type) >&   /* an infinite path (a   */
       path_states,                                 /* prefix and a loop).   */
     StateSpace::size_type loop_state,
     Bitset* result_set = 0) const;

  bool displayPathProof                             /* Same as above;       */
    (ostream& stream, int indent,                   /* however, this        */
     const StateSpace& reachability_graph,          /* function will also   */
     const vector<StateSpace::size_type,            /* output a proof or a  */
                  ALLOC(StateSpace::size_type) >&   /* refutation for the   */
       path_states,                                 /* formula to a stream. */
     StateSpace::size_type loop_state) const;       

  bool evaluatePath                                 /* Computes the truth    */
    (const StateSpace& path,                        /* value of the formula  */
     Bitset* result_set = 0) const;                 /* in the first state of
                                                     * a reachability graph,
						     * assumed to consist of
						     * a single path.
						     */

  bool displayPathProof                             /* Same as above;       */
    (ostream& stream, int indent,                   /* however, this        */
     const StateSpace& path) const;                 /* function will also
						     * output a proof or a
						     * refutation for the
						     * formula to a stream.
						     */

  bool propositional() const;                       /* Tells whether the
						     * formula is purely
						     * propositional, i.e. it
						     * contains no temporal
						     * operators.
						     */

  bool constant() const;                            /* Tells whether the
						     * formula is a constant,
						     * i.e. whether its all
						     * atoms are Boolean
						     * constants.
						     */

  bool satisfiable() const;                         /* Tells whether the
						     * formula is satisfiable.
						     * (This test is only
						     * allowed if the formula
						     * is fully propositional.)
						     */

  Bitset findPropositionalModel() const;            /* Finds a model for a
						     * purely propositional
						     * formula (if it is
						     * satisfiable).
						     */

  virtual FormulaType what() const = 0;             /* Tells the type of the
						     * formula.
						     */

  virtual void print(ostream& stream = cout) const; /* Prints the formula. */

  friend class LtlNegation;                         /* Friend declarations. */
  friend class LtlNext;
  friend class LtlFinally;
  friend class LtlGlobally;

  friend class LtlDisjunction;
  friend class LtlConjunction;
  friend class LtlImplication;
  friend class LtlEquivalence;
  friend class LtlUntil;
  friend class LtlV;

  friend class ResultCache;

  template<class Operator>
  friend class UnaryFormula;

  template<class Operator>
  friend class BinaryFormula;

protected:
  struct                                            /* These flags tell      */
  {                                                 /* whether the formula   */
    unsigned int is_propositional : 1;              /* contains any temporal */
    unsigned int is_constant      : 1;              /* operators and atomic  */
  } info_flags;                                     /* propositions.         */

  class ptr_less                                    /* Class for mapping    */
  {                                                 /* the `less than'      */
  public:                                           /* relation between two */
    bool operator()                                 /* pointers to          */
      (const LtlFormula* f1,                        /* LtlFormulae to the   */
       const LtlFormula* f2) const;                 /* (lexicographically)  */
  };                                                /* corresponding
						     * relation between the
						     * formulae itself.
						     */

private:
  virtual bool eval                                 /* Helper function for */
    (const Bitset& valuation) const = 0;            /* evaluating the
                                                     * formula in a given
                                                     * truth assignment.
                                                     */

  virtual void recCollectSubformulae                /* Collects all the      */
    (stack<const LtlFormula*,                       /* subformulae of the    */
           deque<const LtlFormula*,                 /* formula onto a stack. */
                 ALLOC(const LtlFormula*) > >&
       subformula_stack)
    const = 0;

  virtual bool sat_eval                             /* Helper function for */
    (bool prove_true,                               /* testing the         */
     set<unsigned long int,                         /* satisfiability of a */
         less<unsigned long int>,                   /* propositional       */
         ALLOC(unsigned long int) >&                /* formula.            */
       positive_atoms,
     set<unsigned long int,
         less<unsigned long int>,
         ALLOC(unsigned long int) >&
       negative_atoms) const = 0;

  void evalPath                                     /* Helper function for */
    (const StateSpace& reachability_graph,          /* computing the truth */
     const vector<StateSpace::size_type,            /* value of a path     */
                  ALLOC(StateSpace::size_type) >&   /* formula.            */
       path_states,
     StateSpace::size_type loop_state,
     map<const LtlFormula*, Bitset*, ptr_less,
         ALLOC(Bitset*) >&
       result_cache) const;

  void prettyPrintPathProof                         /* Helper function for  */
    (Exceptional_ostream& estream, int indent,      /* writing a proof or a */
     bool prove_true,                               /* refutation for a     */
     StateSpace::size_type state,                   /* path formula to a    */
     map<const LtlFormula*, Bitset*, ptr_less,      /* stream.              */
         ALLOC(Bitset*) >&
       result_cache,
     const vector<StateSpace::size_type,
                  ALLOC(StateSpace::size_type) >&
       path,
     StateSpace::size_type loop_state) const; 
};



/******************************************************************************
 *
 * A class for propositional atoms.
 *
 *****************************************************************************/

class Atom : public LtlFormula
{
public:
  explicit Atom(unsigned long int a);               /* Creates a new
                                                     * propositional atom.
                                                     */

  /* default copy constructor */

  ~Atom();                                          /* Destructor. */

  /* default assignment operator */

  operator string() const;                          /* Converts the atom to a
						     * string.
						     */

  Atom* clone() const;                              /* Creates a copy of the
                                                     * propositional atom.
                                                     */

  FormulaType what() const;                         /* Returns the type of
						     * the formula, i.e. the
						     * constant LTL_ATOM.
						     */

  friend class LtlNegation;                         /* Friend declarations. */
  friend class LtlNext;
  friend class LtlFinally;
  friend class LtlGlobally;

  friend class LtlDisjunction;
  friend class LtlConjunction;
  friend class LtlImplication;
  friend class LtlEquivalence;
  friend class LtlUntil;
  friend class LtlV;

  friend class ResultCache;

private:
  bool eval(const Bitset& valuation) const;         /* Returns the truth value
                                                     * of the atom in a given
                                                     * truth assignment.
                                                     */

  void recCollectSubformulae                        /* Pushes the atom onto */
    (stack<const LtlFormula*,                       /* a given stack used   */
           deque<const LtlFormula*,                 /* for recursively      */
                 ALLOC(const LtlFormula*) > >&      /* collecting the       */
     subformula_stack)                              /* subformulae of an    */
    const;                                          /* LtlFormula.          */

  bool sat_eval                                     /* Tests the             */
    (bool prove_true,                               /* satisfiability of the */
     set<unsigned long int,                         /* proposition.          */
         less<unsigned long int>,
         ALLOC(unsigned long int) >&
       positive_atoms,
     set<unsigned long int,
         less<unsigned long int>,
         ALLOC(unsigned long int) >&
       negative_atoms) const;

  unsigned long int atom;                           /* Identifier of a
                                                     * propositional variable.
                                                     */
};



/******************************************************************************
 *
 * A template class for Boolean constants. The class (`Symbol') used to
 * instantiate the template must satisfy the following requirements:
 *
 *    (1) The class contains a static public constant attribute `symbol'. The
 *        value of this attribute is used as the printable representation of
 *        the constant, so it must be of a type which supports the << operator
 *        for output.
 *
 *    (2) The class contains a static public constant attribute `type' of type
 *        FormulaType which can be used to identify the constant.
 *
 *    (3) The class contains a static public member function `eval' which can
 *        be called without any parameters. The function returns a truth value
 *        of the constant.
 *
 *****************************************************************************/

template<class Symbol>
class Constant : public LtlFormula
{
public:
  Constant();                                       /* Creates a Boolean
                                                     * constant. 
                                                     */

  /* default copy constructor */

  ~Constant();                                      /* Destructor. */

  /* default assignment operator */

  operator string() const;                          /* Converts the constant to
						     * a string.
						     */

  Constant<Symbol>* clone() const;                  /* Creates a copy of the
                                                     * Boolean constant
                                                     * object. 
                                                     */

  FormulaType what() const;                         /* Returns the type of
						     * the constant, i.e.
						     * Symbol::type.
						     */

  friend class LtlNegation;                         /* Friend declarations. */
  friend class LtlNext;
  friend class LtlFinally;
  friend class LtlGlobally;

  friend class LtlDisjunction;
  friend class LtlConjunction;
  friend class LtlImplication;
  friend class LtlEquivalence;
  friend class LtlUntil;
  friend class LtlV;

  friend class ResultCache;

private:
  bool eval(const Bitset&) const;                   /* These functions only  */
                                                    /* tell the truth value  */
                                                    /* of the constant. They */
                                                    /* are needed to satisfy */
                                                    /* the LtlFormula member */
                                                    /* function interface.   */

  void recCollectSubformulae                        /* Pushes the constant  */
    (stack<const LtlFormula*,                       /* onto a given stack   */
           deque<const LtlFormula*,                 /* used for recursively */
                 ALLOC(const LtlFormula*) > >&      /* collecting the       */
     subformula_stack)                              /* subformulae of some  */
    const;                                          /* LtlFormula.          */

  bool sat_eval                                     /* Tests the             */
    (bool prove_true,                               /* satisfiability of the */
     set<unsigned long int,                         /* constant.             */
         less<unsigned long int>,
         ALLOC(unsigned long int) >&,
     set<unsigned long int,
         less<unsigned long int>,
         ALLOC(unsigned long int) >&) const;
};



/******************************************************************************
 *
 * A class for the Boolean constant `true'.
 *
 *****************************************************************************/

class LtlTrue
{
public:
  static const char* const symbol;                  /* Used for printing the
                                                     * constant.
						     */

  static const FormulaType type = LTL_TRUE;         /* Type of the constant. */

  static bool eval();                               /* Returns the truth value
                                                     * of the constant.
						     */
};



/******************************************************************************
 *
 * A class for the Boolean constant `false'.
 *
 *****************************************************************************/

class LtlFalse
{
public:
  static const char* const symbol;                  /* Used for printing the
                                                     * constant.
						     */

  static const FormulaType type = LTL_FALSE;        /* Type of the constant. */

  static bool eval();                               /* Returns the truth value
                                                     * of the constant.
						     */
};



/******************************************************************************
 *
 * Type definitions for propositional constants to make them accessible
 * without using the template instantiation syntax.
 *
 *****************************************************************************/

typedef Constant<LtlTrue> True;
typedef Constant<LtlFalse> False;



/******************************************************************************
 *
 * A template class for LTL formulae consisting of a unary operator (e.g.
 * negation) and a subformula. The class `Operator' used to instantiate the
 * template must satisfy the following requirements:
 *
 *    (1) The class contains a static public constant attribute `symbol'. The
 *        value of this attribute is used as the printable representation of
 *        the unary operator, so it must be of a type which supports the <<
 *        operator for output.
 *
 *    (2) The class contains a static public constant attribute `type' of type
 *        FormulaType which can be used to identify the exact type of the
 *        formula.
 *
 *    (3) The class contains a static public member function `eval' which can
 *        be called with two parameters:
 *          - a Bitset object assigning truth values to atomic propositions
 *          - a pointer to an LtlFormula object (the subformula associated with
 *            the unary operator).
 *        The function must return a truth value (assumed to be the truth value
 *        of the formula in the truth assignment given by the Bitset object).
 *
 *****************************************************************************/

template<class Operator>
class UnaryFormula : public LtlFormula
{
public:
  UnaryFormula(LtlFormula* f);                      /* Creates a new formula
                                                     * with a unary operator.
                                                     */

  UnaryFormula(const LtlFormula& f);                /* Creates a formula from
                                                     * a reference to another.
                                                     */

  UnaryFormula                                      /* Copy constructor. */
    (const UnaryFormula<Operator>& f);    

  ~UnaryFormula();                                  /* Destructor. */

  UnaryFormula<Operator>& operator=                 /* Assignment operator. */
    (const UnaryFormula<Operator>& f);

  operator string() const;                          /* Converts the unary
						     * formula to a string.
						     */

  UnaryFormula<Operator>* clone() const;            /* Creates an identical
                                                     * copy of the formula.
                                                     */

  FormulaType what() const;                         /* Tells the type of the
						     * formula.
						     */

  friend class LtlFormula;                          /* Friend declarations. */

  friend class LtlNegation;
  friend class LtlNext;
  friend class LtlFinally;
  friend class LtlGlobally;

  friend class LtlDisjunction;
  friend class LtlConjunction;
  friend class LtlImplication;
  friend class LtlEquivalence;
  friend class LtlUntil;
  friend class LtlV;

  friend class ResultCache;

private:
  bool eval(const Bitset& valuation) const;         /* Returns the truth value
                                                     * of the formula in a
                                                     * given truth assignment.
                                                     */

  void recCollectSubformulae                        /* Recursively collects */
    (stack<const LtlFormula*,                       /* the subformulae of   */
           deque<const LtlFormula*,                 /* the formula onto a   */
                 ALLOC(const LtlFormula*) > >&      /* stack.               */
     subformula_stack)
    const;

  bool sat_eval                                     /* Tests the             */
    (bool prove_true,                               /* satisfiability of the */
     set<unsigned long int,                         /* unary (propositional) */
         less<unsigned long int>,                   /* formula.              */
         ALLOC(unsigned long int) >&
       positive_atoms,
     set<unsigned long int,
         less<unsigned long int>,
         ALLOC(unsigned long int) >&
       negative_atoms) const;

  LtlFormula* subformula;                           /* Pointer to the operand
                                                     * formula of the unary
                                                     * operator.
                                                     */
};



/******************************************************************************
 *
 * A class for negation operator.
 *
 *****************************************************************************/

class LtlNegation
{
public:
  static const char* const symbol;                  /* Used for printing the
                                                     * formula.
						     */

  static const FormulaType type = LTL_NEGATION;     /* Type of the formula. */

  static bool eval                                  /* Evaluates a negated  */
    (const Bitset& valuation, LtlFormula* f);       /* formula in a given 
                                                     * truth assignment (a
                                                     * single-state
                                                     * reachability graph).
                                                     */
};



/******************************************************************************
 *
 * A class for "Next" operator.
 *
 *****************************************************************************/

class LtlNext
{
public:
  static const char* const symbol;                  /* Used for printing the
                                                     * formula.
						     */

  static const FormulaType type = LTL_NEXT;         /* Type of the formula. */

  static bool eval                                  /* Evaluates a "Next"   */
    (const Bitset& valuation, LtlFormula* f);       /* formula in a given
                                                     * truth assignment (a
                                                     * single-state
                                                     * reachability graph).
                                                     */
};



/******************************************************************************
 *
 * A class for "Finally" operator.
 *
 *****************************************************************************/

class LtlFinally
{
public:
  static const char* const symbol;                  /* Used for printing the
                                                     * formula.
						     */

  static const FormulaType type = LTL_FINALLY;      /* Type of the formula. */

  static bool eval(const Bitset& valuation,         /* Evaluates a "Finally" */
                   LtlFormula* f);                  /* formula in a given
                                                     * truth assignment (a
                                                     * single-state
                                                     * reachability graph).
                                                     */
};



/******************************************************************************
 *
 * A class for "Globally" operator.
 *
 *****************************************************************************/

class LtlGlobally
{
public:
  static const char* const symbol;                  /* Used for printing the
                                                     * formula.
						     */

  static const FormulaType type = LTL_GLOBALLY;     /* Type of the formula. */

  static bool eval                                  /* Evaluates a           */
    (const Bitset& valuation, LtlFormula* f);       /* "Globally" formula in
                                                     * a given truth
                                                     * assignment (a single-
                                                     * state reachability
                                                     * graph).
                                                     */
};



/******************************************************************************
 *
 * Type definitions for unary LTL formulae to make them accessible without
 * using the template instantiation syntax.
 *
 *****************************************************************************/
 
typedef UnaryFormula<LtlNegation> Not;
typedef UnaryFormula<LtlNext> Next;
typedef UnaryFormula<LtlFinally> Finally;
typedef UnaryFormula<LtlGlobally> Globally;



/******************************************************************************
 *
 * A template class for LTL formulae consisting of a binary operator
 * (e.g. conjunction) with two subformulae. The class `Operator' used when
 * instantiating the template must satisfy the following requirements:
 *
 *    (1) The class contains a static public attribute `symbol'. The value
 *        of this attribute is used as the printable representation of the
 *        binary operator, so it must be of a type which supports the <<
 *        operator for output.
 *
 *    (2) The class contains a static public constant attribute `type' of type
 *        FormulaType which can be used to identify the exact type of the
 *        formula.
 *
 *    (3) The class contains a static public member function `eval' which can
 *        be called with three parameters:
 *          - a Bitset object assigning truth values to atomic propositions
 *          - two pointers to LtlFormula objects (the subformulae associated
 *            with the binary operator)
 *        The function must return a truth value (assumed to be the truth value
 *        of the formula in the truth assignment given by the Bitset object).
 *
 *****************************************************************************/

template<class Operator>
class BinaryFormula : public LtlFormula
{
public:
  BinaryFormula(LtlFormula* f1, LtlFormula* f2);    /* Constructors for a  */
                                                    /* binary LTL formula. */
  BinaryFormula
    (const LtlFormula& f1, LtlFormula* f2);
  BinaryFormula
    (LtlFormula* f1, const LtlFormula& f2);
  BinaryFormula
    (const LtlFormula& f1, const LtlFormula& f2);

  BinaryFormula                                     /* Copy constructor. */
    (const BinaryFormula<Operator>& f);  

  ~BinaryFormula();                                 /* Destructor. */

  BinaryFormula<Operator>& operator=                /* Assignment operator. */
    (const BinaryFormula<Operator>& f);

  operator string() const;                          /* Converts the binary
						     * formula to a string.
						     */

  BinaryFormula<Operator>* clone() const;           /* Creates a copy of the
                                                     * formula.
						     */

  FormulaType what() const;

  friend class LtlFormula;                          /* Friend declarations. */

  friend class LtlNegation;
  friend class LtlNext;
  friend class LtlFinally;
  friend class LtlGlobally;

  friend class LtlDisjunction;
  friend class LtlConjunction;
  friend class LtlImplication;
  friend class LtlEquivalence;
  friend class LtlUntil;
  friend class LtlV;

  friend class ResultCache;

private:
  bool eval(const Bitset& valuation) const;         /* Returns the truth value
						     * of the formula in a
						     * given truth assignment.
						     */

  void recCollectSubformulae                        /* Recursively collects */
    (stack<const LtlFormula*,                       /* the subformulae of   */
           deque<const LtlFormula*,                 /* the formula onto a   */
                 ALLOC(const LtlFormula*) > >&      /* stack.               */
     subformula_stack)
    const;

  bool sat_eval                                     /* Tests the             */
    (bool prove_true,                               /* satisfiability of the */
     set<unsigned long int,                         /* binary                */
         less<unsigned long int>,                   /* (propositional)       */
         ALLOC(unsigned long int) >&                /* formula.              */
       positive_atoms,
     set<unsigned long int,
         less<unsigned long int>,
         ALLOC(unsigned long int) >&
       negative_atoms) const;

  LtlFormula* subformula1;                          /* Pointer to the first
                                                     * operand of the binary
                                                     * operator.
						     */

  LtlFormula* subformula2;                          /* Pointer to the second
                                                     * operand.
						     */
};



/******************************************************************************
 *
 * A class for disjunction operator.
 *
 *****************************************************************************/

class LtlDisjunction
{
public:
  static const char* const symbol;                  /* Used for printing the
                                                     * formula.
						     */

  static const FormulaType type = LTL_DISJUNCTION;  /* Type of the formula. */

  static bool eval                                  /* Evaluates the       */
    (const Bitset& valuation,                       /* disjunctive formula */
     LtlFormula* f1,                                /* in a given truth    */
     LtlFormula* f2);                               /* assignment.         */
};



/******************************************************************************
 *
 * A class for conjunction operator.
 *
 *****************************************************************************/

class LtlConjunction
{
public:
  static const char* const symbol;                  /* Used for printing the
                                                     * formula.
						     */

  static const FormulaType type = LTL_CONJUNCTION;  /* Type of the formula. */

  static bool eval                                  /* Evaluates the       */
    (const Bitset& valuation,                       /* conjunctive formula */
     LtlFormula* f1,                                /* in a given truth    */
     LtlFormula* f2);                               /* assignment.         */
};



/******************************************************************************
 *
 * A class for implication operator.
 *
 *****************************************************************************/

class LtlImplication
{
public:
  static const char* const symbol;                  /* Used for printing the
                                                     * formula.
						     */

  static const FormulaType type = LTL_IMPLICATION;  /* Type of the formula. */

  static bool eval                                  /* Evaluates the       */
    (const Bitset& valuation,                       /* implication formula */
     LtlFormula* f1,                                /* in a given truth    */
     LtlFormula* f2);                               /* assignment.         */
};



/******************************************************************************
 *
 * A class for equivalence operator.
 *
 *****************************************************************************/

class LtlEquivalence
{
public:
  static const char* const symbol;                  /* Used for printing the
                                                     * formula.
						     */

  static const FormulaType type = LTL_EQUIVALENCE;  /* Type of the formula. */

  static bool eval                                  /* Evaluates the       */
    (const Bitset& valuation,                       /* equivalence formula */
     LtlFormula* f1,                                /* in a given truth    */
     LtlFormula* f2);                               /* assignment.         */
};



/******************************************************************************
 *
 * A class for `Until' operator.
 *
 *****************************************************************************/

class LtlUntil
{
public:
  static const char* const symbol;                  /* Used for printing the
                                                     * formula.
						     */

  static const FormulaType type = LTL_UNTIL;        /* Type of the formula. */

  static bool eval                                  /* Evaluates the        */
    (const Bitset& valuation,                       /* "Until" formula in a */
     LtlFormula*,                                   /* given truth          */
     LtlFormula* f2);                               /* assignment.          */
};



/******************************************************************************
 *
 * A class for `V' (the dual of `Until') operator.
 *
 *****************************************************************************/

class LtlV
{
public:
  static const char* const symbol;                  /* Used for printing the
                                                     * formula.
						     */

  static const FormulaType type = LTL_V;            /* Type of the formula. */

  static bool eval                                  /* Evaluates the "V"  */
    (const Bitset& valuation,                       /* formula in a given */
     LtlFormula*,                                   /* truth assignment.  */
     LtlFormula* f2);
};



/******************************************************************************
 *
 * Type definitions for the different binary LTL formulae to make them
 * accessible without using the template instantiation syntax.
 *
 *****************************************************************************/

typedef BinaryFormula<LtlDisjunction> Or;
typedef BinaryFormula<LtlConjunction> And;
typedef BinaryFormula<LtlImplication> Imply;
typedef BinaryFormula<LtlEquivalence> Equiv;
typedef BinaryFormula<LtlUntil> Until;
typedef BinaryFormula<LtlV> V;



/******************************************************************************
 *
 * Declarations for classes used by the function
 *      LtlFormula::formulaStringToLtlFormula(FormulaString& f)
 * for creating an LtlFormula from a FormulaString object. These classes are
 * used when `traversing' the FormulaString.
 *
 *****************************************************************************/

/* ----------------------------------------------------------------------------
 *  A class for translating a propositional atom into a LtlFormula.
 */

class AtomToLtl
{
public:
  static void eval
    (FormulaString::TraversalInfo<stack<LtlFormula*,
                                        deque<LtlFormula*,
                                              ALLOC(LtlFormula*) > > >& info,
     const string& atom);
};

/* ----------------------------------------------------------------------------
 *  A template class for translating a propositional constant into a
 *  LtlFormula.
 */

template<class T>
class ConstantToLtl
{
public:
  static void eval
    (FormulaString::TraversalInfo<stack<LtlFormula*,
                                        deque<LtlFormula*,
                                              ALLOC(LtlFormula*) > > >& info);
};

/* ----------------------------------------------------------------------------
 *  A template class for translating a unary formula into a LtlFormula.
 */

template<class T>
class UnaryOperatorToLtl
{
public:
  static void eval
    (FormulaString::TraversalInfo<stack<LtlFormula*,
                                        deque<LtlFormula*,
                                              ALLOC(LtlFormula*) > > >& info);
};

/* ----------------------------------------------------------------------------
 *  A template class for translating a binary formula into a LtlFormula.
 */

template<class T>
class BinaryOperatorToLtl
{
public:
  static void eval
    (FormulaString::TraversalInfo<stack<LtlFormula*,
                                        deque<LtlFormula*,
                                              ALLOC(LtlFormula*) > > >& info);
};



/******************************************************************************
 *
 * Inline function definitions for class LtlFormula.
 *
 *****************************************************************************/

/* ========================================================================= */
inline LtlFormula::LtlFormula()
/* ----------------------------------------------------------------------------
 *
 * Description:    Constructor for class LtlFormula. Initializes the attributes
 *                 of the base class LtlFormula.
 *
 * Arguments:      None.
 *
 * Returns:        Nothing.
 *
 * --------------------------------------------------------------------------*/
{
  info_flags.is_propositional = 0;
  info_flags.is_constant = 0;
}

/* ========================================================================= */
inline LtlFormula::LtlFormula(const LtlFormula& f) :
  info_flags(f.info_flags)
/* ----------------------------------------------------------------------------
 *
 * Description:   Copy constructor for class LtlFormula. Creates a new LTL
 *                formula object and initializes it with a copy of another one.
 *
 * Argument:      f  --  A reference to a constant LTL formula object of which
 *                       the new object is made a copy.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline LtlFormula::~LtlFormula()
/* ----------------------------------------------------------------------------
 *
 * Description:   Destructor for class LtlFormula. Deallocates the memory
 *                reserved for the formula.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline bool LtlFormula::evaluate(const Bitset& valuation) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Computes the truth value of the formula in a given truth
 *                assignment for the atomic propositions in the formula. This
 *                function can be used to evaluate purely propositional
 *                formulae without the requirement to evaluate the formula with
 *                respect to some reachability graph. However, if the formula
 *                contains temporal operators, calling this function for the
 *                formula is equivalent to evaluating the formula in a
 *                reachability graph consisting of a single state and a
 *                transition from the state to itself (with the given
 *                assignment for the propositions in the state).
 *
 * Arguments:     valuation  --  A reference to a constant Bitset representing
 *                               the truth assignment.
 *
 * Returns:       The truth value of the formula.
 *
 * ------------------------------------------------------------------------- */
{
  return eval(valuation);
}

/* ========================================================================= */
inline bool LtlFormula::propositional() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Tells whether the formula is purely propositional or not
 *                (i.e. whether it contains any temporal operators).
 *
 * Arguments:     None.
 *
 * Returns:       true or false according to the test whether the formula is
 *                purely propositional or not.
 *
 * ------------------------------------------------------------------------- */
{
  return info_flags.is_propositional;
}

/* ========================================================================= */
inline bool LtlFormula::constant() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Tells whether the formula contains any non-constant atoms.
 *
 * Arguments:     None.
 *
 * Returns:       A truth value according to the test for the property.
 *
 * ------------------------------------------------------------------------- */
{
  return info_flags.is_constant;
}

/* ========================================================================= */
inline bool LtlFormula::satisfiable() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Tells whether the formula is satisfiable. (This check is
 *                allowed only on fully propositional formulae.)
 *
 * Arguments:     None.
 *
 * Returns:       A truth value according to the result of the test.
 *
 * ------------------------------------------------------------------------- */
{
  if (!info_flags.is_propositional)
    throw Exception("satisfiable() called on a temporal formula");

  set<unsigned long int, less<unsigned long int>, ALLOC(unsigned long int) >
    positive_atoms;
  set<unsigned long int, less<unsigned long int>, ALLOC(unsigned long int) >
    negative_atoms;

  return sat_eval(true, positive_atoms, negative_atoms);
}

/* ========================================================================= */
inline void LtlFormula::print(ostream& stream) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Outputs the formula to a stream.
 *
 * Argument:      stream  --  A reference to the output stream.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);
  estream << string(*this);
}

/* ========================================================================= */
inline ostream& operator<<(ostream& stream, const LtlFormula& f)
/* ----------------------------------------------------------------------------
 *
 * Description:   Alternative method for printing an LTL formula by using the
 *                << operator.
 *
 * Arguments:     stream  --  A reference to the output stream to which the
 *                            formula is to be printed.
 *                f       --  A reference to the constant LTL formula.
 *
 * Returns:       A reference to the output stream (to support chaining of the
 *                << operators).
 *
 * --------------------------------------------------------------------------*/
{
  f.print(stream);
  return stream;
}



/******************************************************************************
 *
 * Inline function definitions for class LtlFormula::ptr_less.
 *
 *****************************************************************************/

/* ========================================================================= */
inline bool LtlFormula::ptr_less::operator()
  (const LtlFormula* f1, const LtlFormula* f2) const
/* ----------------------------------------------------------------------------
 *
 * Description:   This function maps the `less than' relation between two
 *                pointers to LtlFormula objects to the corresponding relation
 *                in the lexicographical ordering of the string representations
 *                of the formulae. Used when storing pointers to LtlFormulae
 *                into STL containers.
 *
 * Arguments:     f1, f2  --  Two constant pointers to LtlFormulae.
 *
 * Returns:       A truth value according to the relation between the string
 *                representations.
 *
 * ------------------------------------------------------------------------- */
{
  return (string(*f1) < string(*f2));
}



/******************************************************************************
 *
 * Inline function definitions for class Atom.
 *
 *****************************************************************************/

/* ========================================================================= */
inline Atom::Atom(unsigned long int a) :
  LtlFormula(), atom(a)
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructor for class Atom. Creates a new propositional atom.
 *
 * Argument:      a  --  Identifier of a propositional variable.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  info_flags.is_propositional = 1;
  info_flags.is_constant = 0;
}

/* ========================================================================= */
inline Atom::~Atom()
/* ----------------------------------------------------------------------------
 *
 * Description:   Destructor for class Atom.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline Atom* Atom::clone() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Creates a copy of the propositional atom.
 *
 * Arguments:     None.
 *
 * Returns:       A pointer to the new propositional atom.
 *
 * ------------------------------------------------------------------------- */
{
  return new Atom(atom);
}

/* ========================================================================= */
inline FormulaType Atom::what() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Returns the `formula type' of the atom, i.e. the constant
 *                LTL_ATOM.
 *
 * Arguments:     None.
 *
 * Returns:       The constant LTL_ATOM.
 *
 * ------------------------------------------------------------------------- */
{
  return LTL_ATOM;
}

/* ========================================================================= */
inline bool Atom::eval(const Bitset& valuation) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Returns the truth value of the atom in a given truth
 *                assignment.
 *
 * Argument:      valuation  --  A reference to a constant Bitset representing
 *                               the truth assignment.
 *
 * Returns:       The truth value of the atom in the assignment. (If the
 *                atom identifier exceeds the size of the Bitset representing
 *                the assignment, the truth value of the atom is assumed to be
 *                false.)
 *
 * ------------------------------------------------------------------------- */
{
  return (atom < valuation.size() && valuation.test(atom));
}

/* ========================================================================= */
inline void Atom::recCollectSubformulae
  (stack<const LtlFormula*, deque<const LtlFormula*,
                                  ALLOC(const LtlFormula*) > >&
     subformula_stack) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Pushes the atom onto a given stack used for recursively
 *                collecting the subformulae of an LtlFormula object.
 *
 * Arguments:     subformula_stack  --  A reference to a stack of pointers to
 *                                      constant LtlFormulae.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  subformula_stack.push(this);
}



/******************************************************************************
 *
 * Inline function definitions for template class Constant<Symbol>.
 *
 *****************************************************************************/

/* ========================================================================= */
template<class Symbol>
inline Constant<Symbol>::Constant()
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructor for a LTL constant.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  info_flags.is_propositional = 1;
  info_flags.is_constant = 1;
}

/* ========================================================================= */
template<class Symbol>
inline Constant<Symbol>::~Constant()
/* ----------------------------------------------------------------------------
 *
 * Description:   Destructor for a LTL constant.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
template<class Symbol>
inline Constant<Symbol>::operator string() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Converts the constant to a string.
 *
 * Arguments:     None.
 *
 * Returns:       A string representation of the constant.
 *
 * ------------------------------------------------------------------------- */
{
  return Symbol::symbol;
}

/* ========================================================================= */
template<class Symbol>
inline Constant<Symbol>* Constant<Symbol>::clone() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Creates a copy of the LTL constant object.
 *
 * Arguments:     None.
 *
 * Returns:       A pointer to the copy.
 *
 * ------------------------------------------------------------------------- */
{
  return new Constant<Symbol>;
}

/* ========================================================================= */
template<class Symbol>
inline FormulaType Constant<Symbol>::what() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Tells the type of the constant.
 *
 * Arguments:     None.
 *
 * Returns:       Type of the constant as a value of type FormulaType.
 *
 * ------------------------------------------------------------------------- */
{
  return Symbol::type;
}

/* ========================================================================= */
template<class Symbol>
inline bool Constant<Symbol>::eval(const Bitset&) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Evaluates the constant in a truth assignment, i.e. simply
 *                returns the constant's truth value.
 *
 * Argument:      A reference to a constant Bitset (required in order to
 *                satisfy the class interface restrictions).
 *
 * Returns:       Truth value of the constant.
 *
 * ------------------------------------------------------------------------- */
{
  return Symbol::eval();
}

/* ========================================================================= */
template<class Symbol>
inline void Constant<Symbol>::recCollectSubformulae
  (stack<const LtlFormula*, deque<const LtlFormula*,
                                  ALLOC(const LtlFormula*) > >&
     subformula_stack) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Pushes the constant onto a given stack used for recursively
 *                collecting the subformulae of a LtlFormula.
 *
 * Argument:      subformula_stack  --  A reference to a stack of pointers to
 *                                      constant LtlFormulae.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  subformula_stack.push(this);
}



/******************************************************************************
 *
 * Function definitions for template class Constant<Symbol>.
 *
 *****************************************************************************/

/* ========================================================================= */
template<class Symbol>
bool Constant<Symbol>::sat_eval
  (bool prove_true,
   set<unsigned long int, less<unsigned long int>, ALLOC(unsigned long int) >&,
   set<unsigned long int, less<unsigned long int>, ALLOC(unsigned long int) >&)
  const
/* ----------------------------------------------------------------------------
 *
 * Description:   Tells whether the propositional constant or its negation is
 *                satisfiable.
 *
 * Arguments:     prove_true  --  Determines whether the satisfiability check
 *                                is to be made on the proposition or its
 *                                negation.
 *                The other two arguments are needed to support the function
 *                interface.
 *
 * Returns:       A truth value according to the test whether the proposition
 *                or its negation is satisfiable.
 *
 * ------------------------------------------------------------------------- */
{
  return (prove_true == Symbol::eval());
}



/******************************************************************************
 *
 * Inline function definitions for class True.
 *
 *****************************************************************************/

/* ========================================================================= */
inline bool LtlTrue::eval()
/* ----------------------------------------------------------------------------
 *
 * Description:   Tells the truth value of a `true' constant.
 *
 * Arguments:     None.
 *
 * Returns:       true
 *
 * ------------------------------------------------------------------------- */
{
  return true;
}



/******************************************************************************
 *
 * Inline function definitions for class False.
 *
 *****************************************************************************/

/* ========================================================================= */
inline bool LtlFalse::eval()
/* ----------------------------------------------------------------------------
 *
 * Description:   Tells the truth value of a `false' constant.
 *
 * Arguments:     None.
 *
 * Returns:       false
 *
 * ------------------------------------------------------------------------- */
{
  return false;
}



/******************************************************************************
 *
 * Inline function definitions for template class UnaryFormula<Operator>.
 *
 *****************************************************************************/

/* ========================================================================= */
template<class Operator>
inline UnaryFormula<Operator>::UnaryFormula(LtlFormula* f) :
  LtlFormula(), subformula(f)
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructs an LTL formula with a unary operator from a
 *                pointer to another LTL formula. The argument formula will be
 *                directly linked to the new object, i.e. no copies are made.
 *
 * Argument:      f  --  Pointer to a LTL formula to be associated with the new
 *                       formula.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  info_flags.is_propositional = (f->propositional()
				 && Operator::symbol == LtlNegation::symbol);
  info_flags.is_constant = f->constant();
}

/* ========================================================================= */
template<class Operator>
inline UnaryFormula<Operator>::UnaryFormula(const LtlFormula& f) :
  LtlFormula(), subformula(f.clone())
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructs an LTL formula with a unary operator from another
 *                LTL formula. This constructor will make a copy of the
 *                argument formula.
 *
 * Argument:      f  --  A reference to a constant LTL formula.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  info_flags.is_propositional = (f.propositional()
				 && Operator::symbol == LtlNegation::symbol);
  info_flags.is_constant = f.constant();
}

/* ========================================================================= */
template<class Operator>
inline UnaryFormula<Operator>::UnaryFormula(const UnaryFormula<Operator>& f) :
  LtlFormula(f), subformula(f.subformula->clone())
/* ----------------------------------------------------------------------------
 *
 * Description:   Copy constructor for unary LTL formulae. Creates a new unary
 *                LTL formula and initializes it with a copy of another one of
 *                the same type.
 *
 * Argument:      f  --  A reference to a constant unary LTL formula of which
 *                       the new formula is made a copy.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
template<class Operator>
inline UnaryFormula<Operator>::~UnaryFormula()
/* ----------------------------------------------------------------------------
 *
 * Description:   Destructor for unary LTL formulae. Releases the memory
 *                allocated for the operand of the unary operator.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  delete subformula;
}

/* ========================================================================= */
template<class Operator>
inline UnaryFormula<Operator>::operator string() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Converts the unary formula to a string.
 *
 * Arguments:     None.
 *
 * Returns:       A string representation of the unary formula.
 *
 * ------------------------------------------------------------------------- */
{
  return string(Operator::symbol) + ' ' + string(*subformula);
}

/* ========================================================================= */
template<class Operator>
inline UnaryFormula<Operator>* UnaryFormula<Operator>::clone() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Creates an independent copy of a unary LTL formula.
 *
 * Arguments:     None.
 *
 * Returns:       A pointer to the copy.
 *
 * ------------------------------------------------------------------------- */
{
  return new UnaryFormula<Operator>(*this);
}

/* ========================================================================= */
template<class Operator>
inline FormulaType UnaryFormula<Operator>::what() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Tells the exact type of the unary formula.
 *
 * Arguments:     None.
 *
 * Returns:       Type of the formula (a FormulaType constant).
 *
 * ------------------------------------------------------------------------- */
{
  return Operator::type;
}

/* ========================================================================= */
template<class Operator>
inline bool UnaryFormula<Operator>::eval(const Bitset& valuation) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Evaluates the unary formula in a given truth assignment (in
 *                a state space consisting of a single state with a transition
 *                to itself).
 *
 * Argument:      A reference to a constant Bitset, defining the truth
 *                assignment.
 *
 * Returns:       Truth value of the formula in the assignment.
 *
 * ------------------------------------------------------------------------- */
{
  return Operator::eval(valuation, subformula);
}



/******************************************************************************
 *
 * Function definitions for template class UnaryFormula<Operator>.
 *
 *****************************************************************************/

/* ========================================================================= */
template<class Operator>
UnaryFormula<Operator>& UnaryFormula<Operator>::operator=
  (const UnaryFormula<Operator>& f)
/* ----------------------------------------------------------------------------
 *
 * Description:   Assignment operator for unary LTL formulae. Assigns a unary
 *                LTL formula of the same type to another one.
 *
 * Argument:      f  --  A reference to a constant unary LTL formula.
 *
 * Returns:       A reference to the formula which was assigned a new value.
 *
 * ------------------------------------------------------------------------- */
{
  if (&f != this)
  {
    LtlFormula::operator=(f);

    delete subformula;
    subformula = f.subformula->clone();
  }

  return *this;
}

/* ========================================================================= */
template<class Operator>
void UnaryFormula<Operator>::recCollectSubformulae
  (stack<const LtlFormula*, deque<const LtlFormula*,
                                  ALLOC(const LtlFormula*) > >&
     subformula_stack) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Pushes the formula and its subformulae onto a given stack
 *                used for recursively collecting the subformulae of a
 *                LtlFormula.
 *
 * Argument:      subformula_stack  --  A reference to a stack of pointers to
 *                                      constant LtlFormulae.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  subformula_stack.push(this);
  subformula->recCollectSubformulae(subformula_stack);
}

/* ========================================================================= */
template<class Operator>
bool UnaryFormula<Operator>::sat_eval
  (bool prove_true,
   set<unsigned long int, less<unsigned long int>, ALLOC(unsigned long int) >&
     positive_atoms,
   set<unsigned long int, less<unsigned long int>, ALLOC(unsigned long int) >&
     negative_atoms) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Tests the satisfiability of a propositional formula with a
 *                unary operator.
 *
 * Arguments:     prove_true      --  Determines whether the satisfiability
 *                                    test is to be made for the formula or its
 *                                    negation.
 *                positive_atoms  --  A set of identifiers for atomic
 *                                    propositions whose truth values have been
 *                                    fixed to true.
 *                negative_atoms  --  A set of identifiers for atomic
 *                                    propositions whose truth values have been
 *                                    fixed to false.
 *
 * Returns:       A truth value according to the test whether the formula or
 *                its negation is (or can be made) satisfiable.
 *
 * ------------------------------------------------------------------------- */
{
  switch (what())
  {
    case LTL_NEGATION :
      return subformula->sat_eval(!prove_true, positive_atoms, negative_atoms);

    default:
      throw Exception("sat_eval error");
  }
}



/******************************************************************************
 *
 * Inline function definitions for class LtlNegation.
 *
 *****************************************************************************/

/* ========================================================================= */
inline bool LtlNegation::eval(const Bitset& valuation, LtlFormula* f)
/* ----------------------------------------------------------------------------
 *
 * Description:   Evaluates the negated formula in a given assignment for the
 *                propositional variables.
 *
 * Arguments:     valuation  --  A reference to a constant Bitset representing
 *                               the truth assignment.
 *                f          --  A pointer to the subformula of the negation
 *                               operator.
 *
 * Returns:       The truth value of the formula in the truth assignment.
 *
 * ------------------------------------------------------------------------- */
{
  return (!f->eval(valuation));
}



/******************************************************************************
 *
 * Inline function definitions for class LtlNext.
 *
 *****************************************************************************/

/* ========================================================================= */
inline bool LtlNext::eval(const Bitset& valuation, LtlFormula* f)
/* ----------------------------------------------------------------------------
 *
 * Description:   Evaluates the `Next' formula in a given truth assignment for
 *                the propositional variables (in a state space consisting of a
 *                single state with a transition to itself).
 *
 * Arguments:     valuation  --  A reference to a constant Bitset representing
 *                               the truth assignment.
 *                f          --  A pointer to the subformula of the `Next'
 *                               operator.
 *
 * Returns:       The truth value of the formula in the assignment.
 *
 * ------------------------------------------------------------------------- */
{
  return f->eval(valuation);
}



/******************************************************************************
 *
 * Inline function definitions for class LtlFinally.
 *
 *****************************************************************************/

/* ========================================================================= */
inline bool LtlFinally::eval(const Bitset& valuation, LtlFormula* f)
/* ----------------------------------------------------------------------------
 *
 * Description:   Evaluates the `Finally' formula in a given truth assignment
 *                for the propositional variables (in a state space consisting
 *                of a single state with a transition to itself).
 *
 * Arguments:     valuation  --  A reference to a constant Bitset representing
 *                               the truth assignment.
 *                f          --  A pointer to the subformula of the `Finally'
 *                               operator.
 *
 * Returns:       The truth value of the formula in the assignment.
 *
 * ------------------------------------------------------------------------- */
{
  return f->eval(valuation);
}



/******************************************************************************
 *
 * Inline function definitions for class LtlGlobally.
 *
 *****************************************************************************/

/* ========================================================================= */
inline bool LtlGlobally::eval(const Bitset& valuation, LtlFormula* f)
/* ----------------------------------------------------------------------------
 *
 * Description:   Evaluates the `Globally' formula in a given truth assignment
 *                for the propositional variables (in a state space consisting
 *                of a single state with a transition to itself).
 *
 * Arguments:     valuation  --  A reference to a constant Bitset representing
 *                               the truth assignment.
 *                f          --  A pointer to the subformula of the `Globally'
 *                               operator.
 *
 * Returns:       The truth value of the formula in the assignment.
 *
 * ------------------------------------------------------------------------- */
{
  return f->eval(valuation);
}



/******************************************************************************
 *
 * Inline function definitions for template class BinaryFormula<Operator>.
 *
 *****************************************************************************/

/* ========================================================================= */
template<class Operator>
inline BinaryFormula<Operator>::BinaryFormula(LtlFormula* f1, LtlFormula* f2) :
  LtlFormula(), subformula1(f1), subformula2(f2)
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructs an LTL formula with a binary operator from two
 *                pointers to other LTL formulae. The argument formulae will 
 *                be directly linked to the new object, i.e. no copies are
 *                made.
 *
 * Arguments:     f1  --  Pointer to an LTL formula to be associated with the
 *                        new formula as its left operand.
 *                f2  --  Pointer to an LTL formula to be associated with the
 *                        new formula as its right operand.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  info_flags.is_propositional = (f1->propositional()
				 && f2->propositional()
				 && Operator::symbol != LtlUntil::symbol
				 && Operator::symbol != LtlV::symbol);
  info_flags.is_constant = (f1->constant() && f2->constant() ? 1 : 0);
}

/* ========================================================================= */
template<class Operator>
inline BinaryFormula<Operator>::BinaryFormula
  (const LtlFormula& f1, LtlFormula* f2) :
  LtlFormula(), subformula1(f1.clone()), subformula2(f2)
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructs an LTL formula with a binary operator from an
 *                LTL formula and a pointer to another formula. The pointer
 *                argument formula will be directly linked to the new object;
 *                the another formula will be copied.
 *
 * Arguments:     f1  --  A reference to a constant LTL formula whose copy is
 *                        to be made the left operand of the new formula.
 *                f2  --  Pointer to an LTL formula to be associated with the
 *                        new formula as its right operand.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  info_flags.is_propositional = (f1.propositional()
				 && f2->propositional()
				 && Operator::symbol != LtlUntil::symbol
				 && Operator::symbol != LtlV::symbol);
  info_flags.is_constant = (f1.constant() && f2->constant() ? 1 : 0);
}

/* ========================================================================= */
template<class Operator>
inline BinaryFormula<Operator>::BinaryFormula
  (LtlFormula* f1, const LtlFormula& f2) :
  LtlFormula(), subformula1(f1), subformula2(f2.clone())
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructs an LTL formula with a binary operator from a LTL
 *                formula and a pointer to another formula. The pointer
 *                argument formula will be directly linked to the new object;
 *                the another formula will be copied.
 *
 * Arguments:     f1  --  Pointer to an LTL formula to be associated with the
 *                        new formula as its left operand.
 *                f2  --  A reference to a constant LTL formula whose copy is
 *                        to be made the right operand of the new formula.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  info_flags.is_propositional = (f1->propositional()
				 && f2.propositional()
				 && Operator::symbol != LtlUntil::symbol
				 && Operator::symbol != LtlV::symbol);
  info_flags.is_constant = (f1->constant() && f2.constant() ? 1 : 0);
}

/* ========================================================================= */
template<class Operator>
inline BinaryFormula<Operator>::BinaryFormula
  (const LtlFormula& f1, const LtlFormula& f2) :
  LtlFormula(), subformula1(f1.clone()), subformula2(f2.clone())
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructs an LTL formula with a binary operator from two
 *                other LTL formulae. This constructor will create copies of
 *                the two subformulae.
 *
 * Arguments:     f1  --  A reference to a constant LTL formula whose copy is
 *                        to be made the left operand of the new formula.
 *                f2  --  A reference to a constant LTL formula whose copy is
 *                        to be made the right operand of the new formula.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  info_flags.is_propositional = (f1.propositional()
				 && f2.propositional()
				 && Operator::symbol != LtlUntil::symbol
				 && Operator::symbol != LtlV::symbol);
  info_flags.is_constant = (f1.constant() && f2.constant() ? 1 : 0);
}

/* ========================================================================= */
template<class Operator>
inline BinaryFormula<Operator>::BinaryFormula<Operator>
  (const BinaryFormula<Operator>& f) : 
  LtlFormula(f), subformula1(f.subformula1->clone()),
  subformula2(f.subformula2->clone())
/* ----------------------------------------------------------------------------
 *
 * Description:   Copy constructor for binary LTL formulae. Creates a new LTL
 *                formula with two operands and initializes it with a copy of
 *                another one of the same type.
 *
 * Argument:      f  --  A reference to a constant binary LTL formula of which
 *                       the new formula is made a copy.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
template<class Operator>
inline BinaryFormula<Operator>::~BinaryFormula()
/* ----------------------------------------------------------------------------
 *
 * Description:   Destructor for binary LTL formulae. Releases the memory
 *                allocated for the operands of the binary operator.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  delete subformula2;
  delete subformula1;
}

/* ========================================================================= */
template<class Operator>
inline BinaryFormula<Operator>::operator string() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Converts the binary formula to a string.
 *
 * Arguments:     None.
 *
 * Returns:       A string representation of the binary formula.
 *
 * ------------------------------------------------------------------------- */
{
  return '(' + string(*subformula1) + ' ' + Operator::symbol + ' '
         + string(*subformula2) + ')';
}

/* ========================================================================= */
template<class Operator>
inline BinaryFormula<Operator>* BinaryFormula<Operator>::clone() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Creates an independent copy of a binary LTL formula.
 *
 * Arguments:     None.
 *
 * Returns:       A pointer to the copy.
 *
 * ------------------------------------------------------------------------- */
{
  return new BinaryFormula<Operator>(*this);
}

/* ========================================================================= */
template<class Operator>
inline FormulaType BinaryFormula<Operator>::what() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Tells the exact type of the binary formula.
 *
 * Arguments:     None.
 *
 * Returns:       Type of the formula (a FormulaType constant).
 *
 * ------------------------------------------------------------------------- */
{
  return Operator::type;
}

/* ========================================================================= */
template<class Operator>
inline bool BinaryFormula<Operator>::eval(const Bitset& valuation) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Evaluates the binary formula in a given truth assignment (in
 *                a state space consisting of a single state with a transition
 *                to itself).
 *
 * Argument:      A reference to a constant Bitset, defining the truth
 *                assignment.
 *
 * Returns:       Truth value of the formula in the assignment.
 *
 * ------------------------------------------------------------------------- */
{
  return Operator::eval(valuation, subformula1, subformula2);
}



/******************************************************************************
 *
 * Function definitions for template class BinaryFormula<Operator>.
 *
 *****************************************************************************/

/* ========================================================================= */
template<class Operator>
BinaryFormula<Operator>& BinaryFormula<Operator>::operator=
  (const BinaryFormula<Operator>& f)
/* ----------------------------------------------------------------------------
 *
 * Description:   Assignment operator for binary LTL formulae. Assigns a binary
 *                LTL formula of the same type to another one.
 *
 * Argument:      f  --  A reference to a constant binary LTL formula.
 *
 * Returns:       A reference to the formula which was assigned a new value.
 *
 * ------------------------------------------------------------------------- */
{
  if (&f != this) 
  {
    LtlFormula::operator=(f);

    delete subformula2;
    delete subformula1;
    subformula1 = f.subformula1->clone();
    subformula2 = f.subformula2->clone();
  }

  return *this;
}

/* ========================================================================= */
template<class Operator>
void BinaryFormula<Operator>::recCollectSubformulae
  (stack<const LtlFormula*, deque<const LtlFormula*,
                                  ALLOC(const LtlFormula*) > >&
     subformula_stack) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Pushes the formula and its subformulae onto a given stack
 *                used for recursively collecting the subformulae of a
 *                LtlFormula.
 *
 * Argument:      subformula_stack  --  A reference to a stack of pointers to
 *                                      constant LtlFormulae.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  subformula_stack.push(this);
  subformula1->recCollectSubformulae(subformula_stack);
  subformula2->recCollectSubformulae(subformula_stack);
}

/* ========================================================================= */
template<class Operator>
bool BinaryFormula<Operator>::sat_eval
  (bool prove_true,
   set<unsigned long int, less<unsigned long int>, ALLOC(unsigned long int) >&
     positive_atoms,
   set<unsigned long int, less<unsigned long int>, ALLOC(unsigned long int) >&
     negative_atoms) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Tests the satisfiability of a propositional formula with a
 *                binary operator.
 *
 * Arguments:     prove_true      --  Determines whether the satisfiability
 *                                    test is to be made for the formula or its
 *                                    negation.
 *                positive_atoms  --  A set of identifiers for atomic
 *                                    propositions whose truth values have been
 *                                    fixed to true.
 *                negative_atoms  --  A set of identifiers for atomic
 *                                    propositions whose truth values have been
 *                                    fixed to false.
 *
 * Returns:       A truth value according to the test whether the formula or
 *                its negation is (or can be made) satisfiable.
 *
 * ------------------------------------------------------------------------- */
{
  bool result;

  switch (what())
  {
    case LTL_DISJUNCTION :
    case LTL_IMPLICATION :
    {
      bool disj = (what() == LTL_DISJUNCTION);

      if (prove_true)
      {
	set<unsigned long int, less<unsigned long int>,
	    ALLOC(unsigned long int) >
	  old_positive_atoms(positive_atoms);
	set<unsigned long int, less<unsigned long int>,
	    ALLOC(unsigned long int) >
	  old_negative_atoms(negative_atoms);

	if (subformula1->sat_eval(disj, positive_atoms, negative_atoms))
	  result = true;
	else
	{
	  positive_atoms = old_positive_atoms;
	  negative_atoms = old_negative_atoms;

	  result = subformula2->sat_eval(true, positive_atoms, negative_atoms);
	}
      }
      else
	result = (subformula1->sat_eval(!disj, positive_atoms, negative_atoms)
		  && subformula2->sat_eval(false, positive_atoms,
					   negative_atoms));

      break;
    }

    case LTL_CONJUNCTION :
      if (prove_true)
	result = (subformula1->sat_eval(true, positive_atoms, negative_atoms)
		  && subformula2->sat_eval(true, positive_atoms,
					   negative_atoms));
      else
      {
	set<unsigned long int, less<unsigned long int>,
	    ALLOC(unsigned long int) >
	  old_positive_atoms(positive_atoms);
	set<unsigned long int, less<unsigned long int>,
	    ALLOC(unsigned long int) >
	  old_negative_atoms(negative_atoms);

	if (subformula1->sat_eval(false, positive_atoms, negative_atoms))
	  result = true;
	else
	{
	  positive_atoms = old_positive_atoms;
	  negative_atoms = old_negative_atoms;

	  result = subformula2->sat_eval(false, positive_atoms,
					 negative_atoms);
	}
      }

      break;

    case LTL_EQUIVALENCE :
    {
      set<unsigned long int, less<unsigned long int>,
	  ALLOC(unsigned long int) >
	old_positive_atoms(positive_atoms);
      set<unsigned long int, less<unsigned long int>,
	  ALLOC(unsigned long int) >
	old_negative_atoms(negative_atoms);

      if (subformula1->sat_eval(true, positive_atoms, negative_atoms)
	  && subformula2->sat_eval(prove_true, positive_atoms, negative_atoms))
	result = true;
      else
      {
	positive_atoms = old_positive_atoms;
	negative_atoms = old_negative_atoms;

	result = (subformula1->sat_eval(false, positive_atoms, negative_atoms)
		   && subformula2->sat_eval(!prove_true, positive_atoms,
					    negative_atoms));
      }

      break;
    }

    default :
      throw Exception("sat_eval error");
  }

  return result;
}



/******************************************************************************
 *
 * Inline function definitions for class LtlDisjunction.
 *
 *****************************************************************************/

/* ========================================================================= */
inline bool LtlDisjunction::eval
  (const Bitset& valuation, LtlFormula* f1, LtlFormula* f2)
/* ----------------------------------------------------------------------------
 *
 * Description:   Evaluates the disjunctive formula in a given assignment for
 *                the propositional variables.
 *
 * Arguments:     valuation  --  A reference to a constant Bitset representing
 *                               the truth assignment.
 *                f1, f2     --  Pointers to the subformulae of the disjunction
 *                               operator.
 *
 * Returns:       The truth value of the formula in the truth assignment.
 *
 * ------------------------------------------------------------------------- */
{
  return (f1->eval(valuation) || f2->eval(valuation));
}



/******************************************************************************
 *
 * Inline function definitions for class LtlConjunction.
 *
 *****************************************************************************/

/* ========================================================================= */
inline bool LtlConjunction::eval
  (const Bitset& valuation, LtlFormula* f1, LtlFormula* f2)
/* ----------------------------------------------------------------------------
 *
 * Description:   Evaluates the conjunctive formula in a given assignment for
 *                the propositional variables.
 *
 * Arguments:     valuation  --  A reference to a constant Bitset representing
 *                               the truth assignment.
 *                f1, f2     --  Pointers to the subformulae of the conjunction
 *                               operator.
 *
 * Returns:       The truth value of the formula in the truth assignment.
 *
 * ------------------------------------------------------------------------- */
{
  return (f1->eval(valuation) && f2->eval(valuation));
}



/******************************************************************************
 *
 * Inline function definitions for class LtlImplication.
 *
 *****************************************************************************/

/* ========================================================================= */
inline bool LtlImplication::eval
  (const Bitset& valuation, LtlFormula* f1, LtlFormula* f2)
/* ----------------------------------------------------------------------------
 *
 * Description:   Evaluates the implication in a given assignment for the
 *                propositional variables.
 *
 * Arguments:     valuation  --  A reference to a constant Bitset representing
 *                               the truth assignment.
 *                f1, f2     --  Pointers to the subformulae of the implication
 *                               operator.
 *
 * Returns:       The truth value of the formula in the truth assignment.
 *
 * ------------------------------------------------------------------------- */
{
  return (!f1->eval(valuation) || f2->eval(valuation));
}



/******************************************************************************
 *
 * Inline function definitions for class LtlEquivalence.
 *
 *****************************************************************************/

/* ========================================================================= */
inline bool LtlEquivalence::eval
  (const Bitset& valuation, LtlFormula* f1, LtlFormula* f2)
/* ----------------------------------------------------------------------------
 *
 * Description:   Evaluates the equivalence in a given assignment for the
 *                propositional variables.
 *
 * Arguments:     valuation  --  A reference to a constant Bitset representing
 *                               the truth assignment.
 *                f1, f2     --  Pointers to the subformulae of the equivalence
 *                               operator.
 *
 * Returns:       The truth value of the formula in the truth assignment.
 *
 * ------------------------------------------------------------------------- */
{
  return (f1->eval(valuation) == f2->eval(valuation));
}



/******************************************************************************
 *
 * Inline function definitions for class LtlUntil.
 *
 *****************************************************************************/

/* ========================================================================= */
inline bool LtlUntil::eval
  (const Bitset& valuation, LtlFormula*, LtlFormula* f2)
/* ----------------------------------------------------------------------------
 *
 * Description:   Evaluates the `Until' formula in a given truth assignment
 *                for the propositional variables (in a state space consisting
 *                of a single state with a transition to itself).
 *
 * Arguments:     valuation  --  A reference to a constant Bitset representing
 *                               the truth assignment.
 *                f1, f2     --  Pointers to the subformulae of the `Until'
 *                               operator.
 *
 * Returns:       The truth value of the formula in the assignment.
 *
 * ------------------------------------------------------------------------- */
{
  return f2->eval(valuation);
}



/******************************************************************************
 *
 * Inline function definitions for class LtlV.
 *
 *****************************************************************************/

/* ========================================================================= */
inline bool LtlV::eval(const Bitset& valuation, LtlFormula*, LtlFormula* f2)
/* ----------------------------------------------------------------------------
 *
 * Description:   Evaluates the `Release' formula in a given truth assignment
 *                for the propositional variables (in a state space consisting
 *                of a single state with a transition to itself).
 *
 * Arguments:     valuation  --  A reference to a constant Bitset representing
 *                               the truth assignment.
 *                f1, f2     --  A pointer to the subformula of the `V'
 *                               operator.
 *
 * Returns:       The truth value of the formula in the assignment.
 *
 * ------------------------------------------------------------------------- */
{
  return f2->eval(valuation);
}



/******************************************************************************
 *
 * Function for constructing a LtlFormula from a FormulaString.
 *
 *****************************************************************************/

LtlFormula* formulaStringToLtlFormula(FormulaString& f);



/******************************************************************************
 *
 * Inline function definitions for class AtomToLtl.
 *
 *****************************************************************************/

/* ========================================================================= */
inline void AtomToLtl::eval
  (FormulaString::TraversalInfo<stack<LtlFormula*,
                                      deque<LtlFormula*,
                                            ALLOC(LtlFormula*) > > >&
     info,
   const string& atom)
/* ----------------------------------------------------------------------------
 *
 * Description:   Function for creating a new LtlAtom object from the
 *                printable representation of the atom. Used by
 *                LtlFormula::formulaStringToLtlFormula(FormulaString& f) when
 *                translating a FormulaString to a LtlFormula object (during
 *                the traversal of the FormulaString object).
 *
 * Argument:      info  --  A reference to a stack of pointers to LTL formulae.
 *                          The new propositional atom will be pushed on this
 *                          stack.
 *                atom  --  A propositional atom (a string of the form
 *                          "p<n>", where <n> is an unsigned long integer).
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  info.data.push(new Atom(strtoul(atom.c_str() + 1, 0, 10)));
}



/******************************************************************************
 *
 * Function definitions for template class ConstantToLtl.
 *
 *****************************************************************************/

/* ========================================================================= */
template<class T>
inline void ConstantToLtl<T>::eval
  (FormulaString::TraversalInfo<stack<LtlFormula*,
                                      deque<LtlFormula*,
                                            ALLOC(LtlFormula*) > > >&
     info)
/* ----------------------------------------------------------------------------
 *
 * Description:   Function for creating a new propositional constant. Used by
 *                LtlFormula::formulaStringToLtlFormula(FormulaString& f) when
 *                translating a FormulaString to an LtlFormula object (during
 *                the traversal of the FormulaString object).
 *
 * Argument:      info  --  A reference to a stack of pointers to LTL formulae.
 *                          The propositional constant object will be pushed on
 *                          this stack.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  info.data.push(new T);
}



/******************************************************************************
 *
 * Function definitions for template class UnaryOperatorToLtl.
 *
 *****************************************************************************/

/* ========================================================================= */
template<class T>
inline void UnaryOperatorToLtl<T>::eval
  (FormulaString::TraversalInfo<stack<LtlFormula*,
                                      deque<LtlFormula*,
                                            ALLOC(LtlFormula*) > > >&
     info)
/* ----------------------------------------------------------------------------
 *
 * Description:   Function for creating an LTL formula with a unary operator.
 *                Used by
 *                LtlFormula::formulaStringToLtlFormula(FormulaString& f) when
 *                translating a FormulaString to a LtlFormula object (during
 *                the traversal of the FormulaString object).
 *
 * Argument:      info  --  A reference to a stack of pointers to LTL formulae,
 *                          storing the partial results of the traversal.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  LtlFormula* f = info.data.top();
  info.data.top() = new T(f);
}



/******************************************************************************
 *
 * Function definitions for template class BinaryOperatorToLtl.
 *
 *****************************************************************************/

/* ========================================================================= */
template<class T>
inline void BinaryOperatorToLtl<T>::eval
  (FormulaString::TraversalInfo<stack<LtlFormula*,
                                      deque<LtlFormula*,
                                            ALLOC(LtlFormula*) > > >&
     info)
/* ----------------------------------------------------------------------------
 *
 * Description:   Function for creating an LTL formula with a binary operator.
 *                Used by
 *                LtlFormula::formulaStringToLtlFormula(FormulaString& f) when
 *                translating a FormulaString to a LtlFormula object (during
 *                the traversal of the FormulaString object).
 *
 * Argument:      info  --  A reference to a stack of pointers to propositional
 *                          formulae holding the partial results of the
 *                          traversal.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  LtlFormula* f2 = info.data.top();
  info.data.pop();
  LtlFormula* f1 = info.data.top();
  info.data.top() = new T(f1, f2);
}

}

#endif /* _LtlFormula_h_ */
