/*
 *  Copyright (C) 1999, 2000  Heikki Tauriainen <Heikki.Tauriainen@hut.fi>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef _RandomFormulaParameters_h_
#define _RandomFormulaParameters_h_

#include <utility>
#include <vector>
#include "MacroDefs.h"

using namespace std;

/******************************************************************************
 *
 * A class for storing random formula generation information.
 *
 *****************************************************************************/

class RandomFormulaParameters
{
public:

  RandomFormulaParameters();                        /* Constructor. */

  /* default copy constructor */

  ~RandomFormulaParameters();                       /* Destructor. */

  /* default assignment operator */

  enum Symbol {NO_SYMBOL = 0, ATOM = 1, TRUE = 2,   /* Enumeration constants */
               FALSE = 3, AND = 4, OR = 5,          /* for the different     */
               UNTIL = 6, V = 7, IMPLICATION = 8,   /* symbols (constants,   */
               EQUIVALENCE = 9, NOT = 10,           /* variables or          */
               NEXT = 11, FINALLY = 12,             /* operators) from which */
               GLOBALLY = 13};                      /* the formulae are
                                                     * constructed.
                                                     */

  void reset();                                     /* Clears the formula
                                                     * generation parameters.
                                                     */

  void usePropositionalSymbol                       /* Inserts an operand    */
    (const pair<int, int>& symbol_with_priority);   /* (a propositional
                                                     * variable or a Boolean
                                                     * constant) into the
                                                     * set of operands used
                                                     * for generating random
                                                     * formulae. An operand
                                                     * will be included in a
                                                     * formula with a given
                                                     * priority.            
                                                     */

  void useShortFormulaOperator                      /* Inserts an element */
    (const pair<int, int>& symbol_with_priority);   /* into the set of
                                                     * operators used for
                                                     * generating random
                                                     * formulae of length
                                                     * two.
                                                     */

  void useLongFormulaOperator                       /* Inserts an element */
    (const pair<int, int>& symbol_with_priority);   /* into the set of
                                                     * operators used for
                                                     * generating random
                                                     * formulae of length
                                                     * greater than two.
                                                     */

  unsigned long int number_of_available_variables;  /* Size of the set of
                                                     * propositional variables
                                                     * used in constructing
                                                     * random formulae.
                                                     */

  unsigned long int length;                         /* Target length of the
                                                     * generated formulae (in
                                                     * symbols).
                                                     */

  unsigned long int max_length;                     /* Maximum length of the
                                                     * generated formulae.
                                                     */

  typedef pair<int, int> IntegerPair;

  vector<IntegerPair, ALLOC(IntegerPair) >          /* Operand symbols and */
    propositional_symbol_priorities;                /* their priorities in
                                                     * random formulae.
                                                     */

  vector<IntegerPair, ALLOC(IntegerPair) >          /* Operators and their  */
    short_formula_operators;                        /* priorities in random
                                                     * formulae of length
                                                     * two.
                                                     */

  vector<IntegerPair, ALLOC(IntegerPair) >          /* Operators and their  */
    long_formula_operators;                         /* priorities in random
                                                     * formulae of length
                                                     * greater than two.
                                                     */
};



/******************************************************************************
 *
 * Inline function definitions for class RandomFormulaParameters.
 *
 *****************************************************************************/

/* ========================================================================= */
inline RandomFormulaParameters::RandomFormulaParameters()
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructor for class RandomFormulaParameter. Creates a new
 *                formula parameter information object, initializing the
 *                generation parameters with their default values.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  reset();
}

/* ========================================================================= */
inline RandomFormulaParameters::~RandomFormulaParameters()
/* ----------------------------------------------------------------------------
 *
 * Description:   Destructor for class RandomFormulaParameters.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline void RandomFormulaParameters::reset()
/* ----------------------------------------------------------------------------
 *
 * Description:   Clears the random formula generation parameters.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  number_of_available_variables = 5;
  length = 5;
  max_length = 5;
  propositional_symbol_priorities.clear();
  short_formula_operators.clear();
  long_formula_operators.clear();
  propositional_symbol_priorities.push_back(make_pair(NO_SYMBOL, 0));
  short_formula_operators.push_back(make_pair(NO_SYMBOL, 0));
  long_formula_operators.push_back(make_pair(NO_SYMBOL, 0));
}

/* ========================================================================= */
inline void RandomFormulaParameters::usePropositionalSymbol
  (const pair<int, int>& symbol_with_priority)
/* ----------------------------------------------------------------------------
 *
 * Description:   Inserts an operand (a propositional variable or a Boolean
 *                constant) into the set of operands used for generating random
 *                LTL formulae. An operand will be inserted into a formula with
 *                a given priority when generating the formula. A priority of 0
 *                for an operand type means that the type is not used.
 *
 * Argument:      symbol_with_priority  --  A (symbol, priority) pair where
 *                                          `symbol' is one of ATOM, TRUE or
 *                                          FALSE and `priority' is a
 *                                          nonnegative integer.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  propositional_symbol_priorities.push_back(symbol_with_priority);
  propositional_symbol_priorities[0].second += symbol_with_priority.second;
}

/* ========================================================================= */
inline void RandomFormulaParameters::useShortFormulaOperator
  (const pair<int, int>& symbol_with_priority)
/* ----------------------------------------------------------------------------
 *
 * Description:   Inserts an element into the set of operators used for
 *                generating random LTL formulae of length two. An operator
 *                will be inserted into a formula with a given priority when
 *                generating the formula. A priority of 0 for an operator
 *                means that the operator is not used.
 *
 * Argument:      symbol_with_priority  --  A (symbol, priority) pair where
 *                                          `symbol' is one of NOT, NEXT,
 *                                          FINALLY or GLOBALLY and `priority'
 *                                          is a nonnegative integer.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  short_formula_operators.push_back(symbol_with_priority);
  short_formula_operators[0].second += symbol_with_priority.second;
}

/* ========================================================================= */
inline void RandomFormulaParameters::useLongFormulaOperator
  (const pair<int, int>& symbol_with_priority)
/* ----------------------------------------------------------------------------
 *
 * Description:   Inserts an element into the set of operators used for
 *                generating random LTL formulae of length greater than two. An
 *                operator will be inserted into a formula with a given
 *                priority when generating the formula. A priority of 0 for an
 *                operator means that the operator is not used.
 *
 * Argument:      symbol_with_priority  --  A (symbol, priority) pair where
 *                                          `symbol' is one of NOT, NEXT,
 *                                          AND, OR, UNTIL, V, IMPLICATION,
 *                                          EQUIVALENCE, FINALLY or GLOBALLY
 *                                          and `priority' is a nonnegative
 *                                          integer.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  long_formula_operators.push_back(symbol_with_priority);
  long_formula_operators[0].second += symbol_with_priority.second;
}

#endif  /* _RandomFormulaParameters.h _ */
