/*
 *  Copyright (C) 1999, 2000  Heikki Tauriainen <Heikki.Tauriainen@hut.fi>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef _SinglePathIterator_h_
#define _SinglePathIterator_h_

#include "Bitset.h"
#include "StateSpace.h"

using namespace std;

namespace Graph
{

/******************************************************************************
 *
 * An iterator class for systematically enumerating all reachability graphs
 * consisting of a single infinite path (a prefix and a loop) with a given
 * number of states and a given number of atomic propositions in each state.
 *
 *****************************************************************************/

class SinglePathIterator
{
public:
  SinglePathIterator                                /* Constructor. */
    (unsigned long int propositions_per_state,
     StateSpace::size_type number_of_states);

  /* default copy constructor */

  ~SinglePathIterator();                            /* Destructor. */

  /* default assignment operator */

  bool operator==(const SinglePathIterator& it)     /* Equivalence operator. */
    const;

  bool operator!=(const SinglePathIterator& it)     /* Inequivalence operator.
						     */
    const;

  const StateSpace& operator*() const;              /* Dereferencing */
  const StateSpace* operator->() const;             /* operators.    */

  const StateSpace& operator++();                   /* Increment operators. */
  const StateSpace operator++(int);

  bool atEnd() const;                               /* Tells whether the
						     * iterator has enumerated
						     * all the reachability
						     * graphs in the range
						     * determined by the
						     * parameters with which
						     * the iterator was
						     * initialized.
						     */
						     
private:
  void computeNextPath();                           /* Updates the reachability
						     * graph currently pointed
						     * to by the iterator.
						     */

  StateSpace path;                                  /* The reachability graph
						     * (the path) currently
						     * pointed to by the
						     * iterator.
						     */

  StateSpace::size_type loop_target_state;          /* Identifier of the target
						     * state of the last state
						     * in the path.
						     */
};



/******************************************************************************
 *
 * Inline function definitions for class SinglePathIterator.
 *
 *****************************************************************************/

/* ========================================================================= */
inline SinglePathIterator::~SinglePathIterator()
/* ----------------------------------------------------------------------------
 *
 * Description:   Destructor for class SinglePathIterator.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline const StateSpace& SinglePathIterator::operator*() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Dereferencing operator for class SinglePathIterator. Gives
 *                access to the reachability graph currently pointed to by the
 *                iterator.
 *
 * Arguments:     None.
 *
 * Returns:       A constant reference to the reachability graph currently
 *                pointed to by the iterator.
 *
 * ------------------------------------------------------------------------- */
{
  return path;
}

/* ========================================================================= */
inline const StateSpace* SinglePathIterator::operator->() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Dereferencing operator for class SinglePathIterator. Gives
 *                access to the reachability graph currently pointed to by the
 *                iterator.
 *
 * Arguments:     None.
 *
 * Returns:       A pointer to a constant reachability graph currently pointed
 *                to by the iterator.
 *
 * ------------------------------------------------------------------------- */
{
  return &path;
}

/* ========================================================================= */
inline const StateSpace& SinglePathIterator::operator++()
/* ----------------------------------------------------------------------------
 *
 * Description:   Prefix increment operator for class SinglePathIterator.
 *                Computes the next path in the graph sequence and returns
 *                a constant reference to it.
 *
 * Arguments:     None.
 *
 * Returns:       A constant reference to the updated reachability graph
 *                pointed to by the iterator.
 *
 * ------------------------------------------------------------------------- */
{
  computeNextPath();
  return path;
}

/* ========================================================================= */
inline const StateSpace SinglePathIterator::operator++(int)
/* ----------------------------------------------------------------------------
 *
 * Description:   Postfix increment operator for class SinglePathIterator.
 *                Computes the next path in the graph sequence and returns
 *                the graph pointed to by the iterator after this operation.
 *
 * Arguments:     None.
 *
 * Returns:       A constant reference to the reachability graph pointed to by
 *                the iterator before computing the next path in the sequence.
 *
 * ------------------------------------------------------------------------- */
{
  StateSpace old_path(path);
  computeNextPath();
  return old_path;
}

/* ========================================================================= */
inline bool SinglePathIterator::atEnd() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Tells whether all possible paths have been enumerated by the
 *                iterator.
 *
 * Arguments:     None.
 *
 * Returns:       A truth value according to the result of the test.
 *
 * ------------------------------------------------------------------------- */
{
  return (loop_target_state == path.size());
}
 
}

#endif  /* _SinglePathIterator_h_ */
