/*
 *  Copyright (C) 1999, 2000  Heikki Tauriainen <Heikki.Tauriainen@hut.fi>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef _StateSpace_h_
#define _StateSpace_h_

#include <cstdlib>
#include <map>
#include <vector>
#include "Bitset.h"
#include "EdgeContainer.h"
#include "Exception.h"
#include "Graph.h"

using namespace std;

extern bool user_break;

namespace Graph
{

/******************************************************************************
 *
 * A class for representing reachability graphs.
 *
 *****************************************************************************/

class StateSpace : public Graph<GraphEdgeContainer>
{
private:
  unsigned long int atoms_per_state;                /* Number of propositional
                                                     * variables per state in
                                                     * the reachability graph.
                                                     */

  size_type initial_state;                          /* Index of the initial
                                                     * state of the
                                                     * reachability graph.
                                                     */

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

  class State :                                     /* A class for         */
    public Graph<GraphEdgeContainer>::Node          /* representing the    */
  {                                                 /* states of the
						     * reachability graph.
                                                     */
  public:
    explicit State                                  /* Constructs a state */
      (unsigned long int number_of_atoms = 0);      /* with a false truth
                                                     * assignment for the
                                                     * propositional variables.
                                                     */

    State(const Bitset& pos_atoms);                 /* Constructs a state
                                                     * using a given truth
                                                     * assignment.
                                                     */

    /* default copy constructor */

    ~State();                                       /* Destructor. */

    /* default assignment operator */

    State* clone() const;                           /* Creates a copy of the
                                                     * state.
                                                     */

    /* `edges' inherited from Graph<EdgeContainer>::Node */

    /* `connectTo' inherited from Graph<EdgeContainer>::Node */

    /* `disconnectFrom' inherited from Graph<EdgeContainer>::Node */

    /* `connected' inherited from Graph<EdgeContainer>::Node */
    
    bool holds(unsigned long int atom) const;       /* Test whether a given
                                                     * proposition holds in the
                                                     * state.
                                                     */
    
    const Bitset& positiveAtoms() const;            /* Get or set the truth  */
    Bitset& positiveAtoms();                        /* assignment for the
                                                     * propositional atoms in
                                                     * a state.
                                                     */
    
    void print                                      /* Writes information   */
      (ostream& stream = cout,                      /* about the state to a */
       int indent = 0,                              /* stream.              */
       GraphOutputFormat fmt = NORMAL)
      const;    
                                                     
  private:
    Bitset positive_atoms;                          /* The set of propositions
                                                     * holding in the state.
                                                     */
  };

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

public:
  explicit StateSpace                               /* Constructor. */
    (unsigned long int propositions_per_state = 0,
     size_type initial_number_of_states = 0);

  /* default copy constructor */

  ~StateSpace();                                    /* Destructor. */

  void randomize                                    /* Initializes a random */
    (size_type number_of_states,                    /* reachability graph.  */
     unsigned long int number_of_atoms,
     double edge_probability,
     double truth_probability);

  void randomizeConnected                           /* Initializes a random */
    (size_type number_of_states,                    /* connected            */
     unsigned long int number_of_atoms,             /* reachability graph   */
     double edge_probability,                       /* whose all states are */
     double truth_probability);                     /* reachable from its
						     * initial state.
						     */

  void randomizePath                                /* Initializes a random  */
    (size_type number_of_states,                    /* reachability graph to */
     unsigned long int number_of_atoms,             /* a single path.        */
     double truth_probability);    

  StateSpace& operator=                             /* Assignment operator. */
    (const StateSpace& statespace);

  State& operator[](size_type index) const;         /* Indexing operator. No
						     * range checks are
						     * performed on the
						     * argument.
						     */

  State& node(size_type index) const;               /* Synonym for the indexing
                                                     * operator. This function
						     * will also check the
						     * range of its argument.
                                                     */

  /* `size' inherited from Graph<EdgeContainer> */

  /* `empty' inherited from Graph<EdgeContainer> */

  /* `clear' inherited from Graph<EdgeContainer> */

  size_type expand(size_type node_count = 1);       /* Inserts states to the
                                                     * reachability graph.
                                                     */

  /* `connect' inherited from Graph<EdgeContainer> */

  /* `disconnect' inherited from Graph<EdgeContainer> */

  /* `connected' inherited from Graph<EdgeContainer> */

  /* `stats' inherited from Graph<EdgeContainer> */

  /* `subgraphStats' inherited from Graph<EdgeContainer> */

  size_type initialState() const;                   /* Get or set the       */
  size_type& initialState();                        /* initial state of the
                                                     * reachability graph.
                                                     */

  void print                                        /* Writes information    */
    (ostream& stream = cout,                        /* about the state space */
     int indent = 0,                                /* to a stream.          */
     GraphOutputFormat fmt = NORMAL) const;
};



/******************************************************************************
 *
 * Inline function definitions for class StateSpace.
 *
 *************************************************************************** */

/* ========================================================================= */
inline StateSpace::~StateSpace()
/* ----------------------------------------------------------------------------
 *
 * Description:   Destructor for class StateSpace.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline StateSpace::State& StateSpace::operator[](size_type index) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Indexing operator for class StateSpace. This function can
 *                be used to refer to the individual states of the reachability
 *                graph. This function does not perform any range checks on its
 *                argument.
 *
 * Argument:      index  --  Index of a state.
 *
 * Returns:       A reference to a state of the reachability graph.
 *
 * ------------------------------------------------------------------------- */
{
  return static_cast<State&>(*nodes[index]);
}

/* ========================================================================= */
inline StateSpace::State& StateSpace::node(size_type index) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Function for referring to a single state of a StateSpace.
 *                This function also performs a range check on its argument.
 *
 * Argument:      index  --  Index of a state.
 *
 * Returns:       A reference to a state of the reachability graph.
 *
 * ------------------------------------------------------------------------- */
{
  return static_cast<State&>(Graph<GraphEdgeContainer>::node(index));
}

/* ========================================================================= */
inline StateSpace::size_type StateSpace::initialState() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Returns the number of the initial state of the StateSpace by
 *                value.
 *
 * Arguments:     None.
 *
 * Returns:       Index of the initial state of the reachability graph.
 *
 * ------------------------------------------------------------------------- */
{
  return initial_state;
}

/* ========================================================================= */
inline StateSpace::size_type& StateSpace::initialState()
/* ----------------------------------------------------------------------------
 *
 * Description:   Returns the number of the initial state of the StateSpace by 
 *                reference. This function can therefore be used to change the
 *                initial state.
 *
 * Arguments:     None.
 *
 * Returns:       A reference to the index of the initial state.
 *
 * ------------------------------------------------------------------------- */
{
  return initial_state;
}



/******************************************************************************
 *
 * Inline function definitions for class StateSpace::State.
 *
 *****************************************************************************/

/* ========================================================================= */
inline StateSpace::State::State(unsigned long int number_of_atoms) :
  Graph<GraphEdgeContainer>::Node(), positive_atoms(number_of_atoms)
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructor for class StateSpace::State. Creates a new state
 *                with a given number of atomic propositions, all of which are
 *                initially false.
 *
 * Argument:      number_of_atoms  --  Number of atomic propositions in the
 *                                     state.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline StateSpace::State::State(const Bitset& positive_atoms) :
  Graph<GraphEdgeContainer>::Node(), positive_atoms(positive_atoms)
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructor for class StateSpace::State. Creates a new state,
 *                using a given truth assignment for propositional variables  
 *                for initialization.
 *
 * Argument:      positive_atoms  --  A truth assignment for propositional
 *                                    variables.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline StateSpace::State::~State()
/* ----------------------------------------------------------------------------
 *
 * Description:   Destructor for class StateSpace::State.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline StateSpace::State* StateSpace::State::clone() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Creates an identical copy of a reachability graph state.
 *
 * Arguments:     None.
 *
 * Returns:       A pointer to the copy.
 *
 * ------------------------------------------------------------------------- */
{
  return new State(*this);
}

/* ========================================================================= */
inline bool StateSpace::State::holds(unsigned long int atom) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Tests whether a given proposition holds in a state.
 *
 * Arguments:     atom  --  Identifier of the proposition.
 *
 * Returns:       Truth value of the proposition in the state.
 *
 * ------------------------------------------------------------------------- */
{
  if (atom >= positive_atoms.size())
    return false;

  return positive_atoms.test(atom);
}

/* ========================================================================= */
inline const Bitset& StateSpace::State::positiveAtoms() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Returns the truth assignment for the propositions in a state.
 *
 * Arguments:     None.
 *
 * Returns:       A reference to the truth assignment, represented as a
 *                constant Bitset.
 *
 * ------------------------------------------------------------------------- */
{
  return positive_atoms;
}

/* ========================================================================= */
inline Bitset& StateSpace::State::positiveAtoms()
/* ----------------------------------------------------------------------------
 *
 * Description:   Returns the truth assignment for the propositions in a state.
 *                This function can be also used to change the truth
 *                assignment.
 *
 * Arguments:     None.
 *
 * Returns:       A reference to the truth assignment, represented as a Bitset.
 *
 * ------------------------------------------------------------------------- */
{
  return positive_atoms;
}

}

#endif  /* _StateSpace_h_ */
