/*
 *  Copyright (C) 1999, 2000  Heikki Tauriainen <Heikki.Tauriainen@hut.fi>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "TestOperations.h"

/******************************************************************************
 *
 * Implementations for the operations used in the main test loop.
 *
 *****************************************************************************/

/* ========================================================================= */
void openFile
  (const char* filename, ifstream& stream, ios::openmode mode, int indent)
/* ----------------------------------------------------------------------------
 *
 * Description:   Function for opening a file for input.
 *
 * Arguments:     filename  --  A pointer to a constant C-style string
 *                              containing the name of the file to be opened.
 *                stream    --  A reference to the input stream which is to be
 *                              associated with the file.
 *                mode      --  An constant of type `ios::openmode' determining
 *                              the open mode.
 *                indent    --  Number of spaces to leave to the left of
 *                              messages given to the user.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  if (configuration.global_options.verbosity == 5)
  {
    cout << string(indent, ' ') + "<opening `" + filename + "'>";
    cout.flush();
  }

  stream.open(filename, mode);

  if (!stream.good())
  {
    if (configuration.global_options.verbosity == 5)
      cout << " error" << endl;

    throw FileOpenException(filename);
  }
  else if (configuration.global_options.verbosity == 5)
    cout << " ok" << endl;
}

/* ========================================================================= */
void openFile
  (const char* filename, ofstream& stream, ios::openmode mode, int indent)
/* ----------------------------------------------------------------------------
 *
 * Description:   Function for opening a file for output.
 *
 * Arguments:     filename  --  A pointer to a constant C-style string with
 *                              the name of the file to be opened.
 *                stream    --  A reference to the output stream which is to be
 *                              associated with the file.
 *                mode      --  An constant of type `ios::openmode' determining
 *                              the open mode.
 *                indent    --  Number of spaces to leave to the left of
 *                              messages given to the user.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  if (configuration.global_options.verbosity == 5)
  {
    cout << string(indent, ' ') + '<'
            + (mode & ios::trunc ? "creat" : "open")
            + "ing `" + filename + "'>";
    cout.flush();
  }

  stream.open(filename, mode);

  if (!stream.good())
  {
    if (configuration.global_options.verbosity == 5)
      cout << " error" << endl;

    if (mode & ios::trunc)
      throw FileCreationException(filename);
    else
      throw FileOpenException(filename);
  }
  else if (configuration.global_options.verbosity == 5)
    cout << " ok" << endl;
}

/* ========================================================================= */
void removeFile(const char* filename, int indent)
/* ----------------------------------------------------------------------------
 *
 * Description:   Removes a file.
 *
 * Arguments:     filename  --  A pointer to a constant C-style string with
 *                              the name of the file to be removed.
 *                indent    --  Number of spaces to leave to the left of
 *                              messages given to the user.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  if (configuration.global_options.verbosity == 5)
  {
    cout << string(indent, ' ') + "<removing `" + filename + "'>";
    cout.flush();
  }

  if (remove(filename) == 0)
  {
    if (configuration.global_options.verbosity == 5)
      cout << " ok" << endl;
  }
  else if (configuration.global_options.verbosity == 5)
    cout << " error" << endl;
}

/* ========================================================================= */
void printFileContents
  (ostream& stream, const char* message, const char* filename, int indent,
   const char* line_prefix)
/* ----------------------------------------------------------------------------
 *
 * Description:   Outputs the contents of a file into a stream.
 *
 * Arguments:     stream       --  A reference to the output stream into which
 *                                 the file contents are to be outputted.
 *                message      --  A pointer to a constant C-style string
 *                                 containing a message to be outputted to the
 *                                 stream before the contents of the file. (To
 *                                 display only the file contents, use a null
 *                                 pointer for the message; an empty string
 *                                 results in an empty line to be outputted
 *                                 before the file contents.)
 *                filename     --  A pointer to a constant C-style string
 *                                 containing the name of the input file.
 *                indent       --  Number of spaces to leave to the left of
 *                                 output.
 *                line_prefix  --  A pointer to a constant C-style string
 *                                 which will be prepended to each outputted
 *                                 line.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  ifstream file;
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);

  try
  {
    openFile(filename, file, ios::in, indent);

    bool first_line_printed = false;
    string message_line;

    if (configuration.global_options.verbosity == 5)
      cout << string(indent, ' ') + "<reading from `" + filename + "'>"
	   << endl;

    while (file.good())
    {
      message_line = "";

      getline(file, message_line, '\n');

      if (!file.eof())
      {
	if (!first_line_printed && message != 0)
	{
	  first_line_printed = true;
	  estream << string(indent, ' ') + message + '\n';
	}
                    
	estream << string(indent, ' ') + line_prefix + message_line + '\n';
      }
    }

    estream.flush();

    file.close();
  }
  catch (FileOpenException& e)
  {
    estream << string(indent, ' ') + e.what() << endl;
  }
  catch (IOException&)
  {
  }
}

/* ========================================================================= */
void writeToTranscript
  (ofstream& transcript_file, unsigned long int round, const string& message)
/* ----------------------------------------------------------------------------
 *
 * Description:   Writes a message into a transcript file.
 *
 * Arguments:     transcript_file  --  A reference to the transcript file 
 *                                     stream.
 *                round            --  Number for a test round.
 *                message          --  A message to be written to the file.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  if (transcript_file.good())
    transcript_file << "Round " << round << ": " << message << endl;
}

/* ========================================================================= */
StateSpace* generateStateSpace()
/* ----------------------------------------------------------------------------
 *
 * Description:   Generates a random state space. The global variable
 *                `configuration.statespace_options.generate_mode' determines
 *                whether a random graph or a random path is generated.
 *
 * Arguments:     None.
 *
 * Returns:       A pointer to a newly allocated StateSpace object.
 *
 * ------------------------------------------------------------------------- */
{
  if (configuration.global_options.verbosity >= 2)
  {
    if (configuration.global_options.verbosity == 2)
      cout << "    Generating random state space...\n\n";
    else
    {
      cout << "    Random state space:\n";
      if (configuration.global_options.verbosity >= 4)
        cout << string(6, ' ') + "<generating>";
    }
    cout.flush();
  }

  /*
   *  Determine the number of states in the state space to be generated. If
   *  the size can be chosen from an interval instead of a single value, choose
   *  the size randomly in the interval.
   */

  StateSpace::size_type size = configuration.statespace_options.size;

  if (configuration.statespace_options.max_size > size)
    size += LRAND(0, configuration.statespace_options.max_size - size + 1);

  /*
   *  Determine the type of the state space to be generated according to the
   *  configuration options, then generate it.
   */

  StateSpace* statespace = new StateSpace();

  try
  {
    switch (configuration.statespace_options.generate_mode)
    {
      case Configuration::RANDOMGRAPH :
	statespace->randomize
	  (size,
	   configuration.statespace_options.atoms_per_state,
	   configuration.statespace_options.edge_probability,
	   configuration.statespace_options.
	   truth_probability);
	break;

      case Configuration::RANDOMCONNECTEDGRAPH :
	statespace->randomizeConnected
	  (size,
	   configuration.statespace_options.atoms_per_state,
	   configuration.statespace_options.edge_probability,
	   configuration.statespace_options.
	   truth_probability);
	break;

      default :
        statespace->randomizePath
	  (size,
	   configuration.statespace_options.atoms_per_state,
	   configuration.statespace_options.truth_probability);
	break;
    }
  }
  catch (...)    
  {
    delete statespace;

    try
    {
      throw;
    }
    catch (bad_alloc&)
    {
      if (configuration.global_options.verbosity >= 3)
      {
	if (configuration.global_options.verbosity >= 4)
	  cout << " error (out of memory)" << endl;
	else
	  cout << string(6, ' ') + "[Out of memory]" << endl;
      }

      throw StateSpaceGenerationException();
    }
    catch (UserBreakException&)
    {
      if (configuration.global_options.verbosity >= 3)
      {
	if (configuration.global_options.verbosity >= 4)
	  cout << " user break" << endl;
	else
	  cout << string(6, ' ') + "[User break]\n" << endl;
      }

      throw;
    }
    catch (...)
    {
      throw;
    }
  }

  if (configuration.global_options.verbosity >= 4)
    cout << " ok" << endl;

  if (configuration.global_options.verbosity >= 3)
  {
    cout << string(6, ' ') + "number of states:  " << size
         << '\n' + string(6, ' ') + "variables/state:   "
	 << configuration.statespace_options.atoms_per_state
         << '\n' + string(6, ' ');

    if (configuration.statespace_options.generate_mode & Configuration::GRAPH)
      cout << "edge probability:  "
	   << configuration.statespace_options.edge_probability
	   << '\n' + string(6, ' ');

    cout << "truth probability: "
         << configuration.statespace_options.truth_probability
         << '\n' << endl;
  }

  return statespace;
}

/* ========================================================================= */
void generateFormulae
  (FormulaString::FormulaString formulae[4], istream* formula_input_stream)
/* ----------------------------------------------------------------------------
 *
 * Description:   Generates a random LTL formula according to the current
 *                configuration and stores the formula, its negation and their
 *                negated normal forms into an array.
 *
 * Arguments:     formulae              --  An array of four FormulaStrings for
 *                                          storing the results. The
 *                                          correspondence between the
 *                                          different formulae and the array
 *                                          indices is as follows:
 *                                            formulae[0] will contain the
 *                                                        formula expanded into
 *                                                        negation normal form.
 *                                            formulae[1] will contain the
 *                                                        negated formula
 *                                                        expanded into
 *                                                        negation normal form.
 *                                            formulae[2] will contain the
 *                                                        formula as generated.
 *                                            formulae[3] will contain the
 *                                                        negated formula as
 *                                                        generated.
 *
 *                formula_input_stream  --  A pointer to an input stream. If
 *                                          the pointer is nonzero, no random
 *                                          formula will be generated; instead,
 *                                          a formula will be read from the
 *                                          stream.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  using FormulaString::FormulaString;

  if (configuration.global_options.verbosity >= 2)
  {
    cout << "    ";
    if (configuration.global_options.verbosity == 2)
      cout << (formula_input_stream == 0 ? "Generating random" : "Processing")
	   << " LTL formula...\n\n";
    else
    {
      cout << (formula_input_stream == 0 ? "Random " : "")
	   << "LTL formula:\n";

      if (configuration.global_options.verbosity >= 4)
        cout << string(6, ' ') + '<'
                + (formula_input_stream == 0 ? "generat" : "process")
	        + "ing>";
    }
    cout.flush();
  }

  if (formula_input_stream != 0)
  {
    formulae[0].read(*formula_input_stream);
    if (configuration.global_options.verbosity >= 4)
      cout << " ok\n";
  }
  else
  {
    unsigned long int random_formula_length
      = formulae[0].randomize
          (configuration.formula_options.generation_parameters);

    if (configuration.global_options.verbosity >= 3)
    {
      if (configuration.global_options.verbosity >= 4)
	cout << " ok\n";
      
      cout << string(6, ' ') + "variables:" + string(17, ' ')
	   << configuration.formula_options.generation_parameters.
	        number_of_available_variables
	   << '\n' + string(6, ' ') + "length:" + string(20, ' ')
	   << random_formula_length
	   << '\n';
    }
  }

  formulae[1] = formulae[0];
  formulae[2] = formulae[0];

  if (configuration.global_options.verbosity >= 4)
  {
    cout << string(6, ' ') + "<converting to negation normal form>";
    cout.flush();
  }

  formulae[0].toNegationNormalForm();

  if (configuration.global_options.verbosity >= 4)
  {
    cout << " ok\n" + string(6, ' ') + "<negating formula>";
    cout.flush();
  }
  
  formulae[1].negate();
  formulae[3] = formulae[1];

  if (configuration.global_options.verbosity >= 4)
  {
    cout << " ok\n" + string(6, ' ')
            + "<converting negated formula to negation normal form>";
    cout.flush();
  }

  formulae[1].toNegationNormalForm();

  if (configuration.global_options.verbosity >= 3)
  {
    if (configuration.global_options.verbosity >= 4)
      cout << " ok\n";

    cout << string(6, ' ') + "formula:" + string(19, ' ');
    formulae[2].prettyPrint();

    if (configuration.formula_options.output_mode == Configuration::NNF)
    {
      cout << '\n' + string(8, ' ') + "in negation normal form: ";
      formulae[0].prettyPrint();
    }

    cout << '\n' + string(6, ' ') + "negated formula:" + string(11, ' ');    
    formulae[3].prettyPrint();

    if (configuration.formula_options.output_mode == Configuration::NNF)
    {
      cout << '\n' + string(8, ' ') + "in negation normal form: ";
      formulae[1].prettyPrint();
    }

    cout << endl << endl;
  }
}

/* ========================================================================= */
void writeFormulaToFile
  (const FormulaString::FormulaString& formula, const char* filename)
/* ----------------------------------------------------------------------------
 *
 * Description:   Writes a FormulaString into a file.
 *
 * Arguments:     formula   --  A reference to a constant FormulaString.
 *                filename  --  A pointer to a constant C-style string with
 *                              the name of the file where the formula is to be
 *                              stored.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  bool file_opened = false;

  try
  {
    ofstream formula_file;
    Exceptional_ostream eformula_file(&formula_file,
                                      ios::failbit | ios::badbit);

    openFile(filename, formula_file, ios::out | ios::trunc, 6);
    file_opened = true;

    if (configuration.global_options.verbosity == 5)
    {
      cout << string(6, ' ') + "<writing formula to `" + filename + "'>";
      cout.flush();
    }

    eformula_file << formula << endl;
    formula_file.close();
  }
  catch (IOException& e)
  {
    if (configuration.global_options.verbosity >= 2)
    {
      if (file_opened && configuration.global_options.verbosity == 5)
	cout << " error\n";

      if (configuration.global_options.verbosity >= 2)
	cout << string(6, ' ');

      cout << string("Error: ") + e.what() << endl;
    }
    
    throw;
  }

  if (configuration.global_options.verbosity == 5)
    cout << " ok\n";
}

/* ========================================================================= */
BuchiAutomaton* generateBuchiAutomaton
  (const char* formula_file_name,
   const Configuration::AlgorithmInformation& algorithm,
   AutomatonStats& automaton_stats, int indent)
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructs a newly allocated BuchiAutomaton by invoking an
 *                external program which will perform the conversion of a
 *                LTL formula (stored into a file) into a Bchi automaton.
 *
 * Arguments:     formula_file_name  --  Name of the file where the formula is
 *                                       stored.
 *                algorithm          --  Reference to a constant
 *                                       AlgorithmInformation object specifying
 *                                       the external program to use.
 *                automaton_stats    --  A reference to an AutomatonStats
 *                                       object for storing statistics on the
 *                                       generated automaton.
 *                indent             --  Number of spaces to leave on the left
 *                                       of all output messages.
 *
 * Returns:       A pointer to a newly allocated BuchiAutomaton object.
 *
 * ------------------------------------------------------------------------- */
{
  BuchiAutomaton* automaton = new BuchiAutomaton;

  struct tms timing_information_begin, timing_information_end;

  char automaton_file_name[L_tmpnam + 1];
  char cout_capture_file[L_tmpnam + 1];
  char cerr_capture_file[L_tmpnam + 1];
  automaton_file_name[0] = '\0';
  cout_capture_file[0] = '\0';
  cerr_capture_file[0] = '\0';

  try
  {
    if (tmpnam(automaton_file_name) == 0)
      throw FileCreationException("a temporary file");

    tmpnam(cout_capture_file);
    tmpnam(cerr_capture_file);

    /*
     *  Generate the command line to be used for executing the program that is
     *  supposed to generate the automaton from a formula. The name of the
     *  executable and any extra command line parameters are stored in the
     *  program configuration data structures. The command line will be of the
     *  form
     *    [path to executable] [additional parameters] [input file]
     *                         [output file] 1>[stdout capture file]
     *                         2>[stderr capture file],
     *  so the program used for generating the automaton is assumed to follow
     *  this format for passing the necessary input and output filenames.
     *  Note: This kind of output redirection requires that the call to
     *        `system' invokes a POSIX compatible shell!
     *
     *  The output of stdout and stderr will be redirected to two temporary
     *  files if possible.
     */

    string command_line = *(algorithm.path_to_program) + ' ';

    if (algorithm.extra_parameters != 0)
      command_line += *(algorithm.extra_parameters) + ' ';

    command_line += string(formula_file_name) + ' ' + automaton_file_name;

    if (cout_capture_file[0] != '\0')
      command_line += string(" 1>") + cout_capture_file;

    if (cerr_capture_file[0] != '\0')
      command_line += string(" 2>") + cerr_capture_file;

    if (configuration.global_options.verbosity >= 4)
    {
      cout << string(indent, ' ')
              + (configuration.global_options.verbosity == 4
                 ? "<computing Bchi automaton>"
                 : "<executing `" + command_line + "'>");
      cout.flush();
    }

    times(&timing_information_begin);

    int exitcode = system(command_line.c_str());

    if (exitcode != 0)
    {
      times(&timing_information_end);

      ExecFailedException e;

      ostrstream msg;

      msg << "Execution of `" + *(algorithm.path_to_program)
             + "' failed (exit code "
	  << exitcode;

      if (timing_information_begin.tms_utime != static_cast<clock_t>(-1)
	  && timing_information_begin.tms_cutime != static_cast<clock_t>(-1)
	  && timing_information_end.tms_utime != static_cast<clock_t>(-1)
	  && timing_information_end.tms_cutime != static_cast<clock_t>(-1))
      {
	int old_precision = msg.precision(2);
	ios::fmtflags old_flags = msg.flags(ios::fixed);
	msg << ", "
	    << static_cast<double>(timing_information_end.tms_utime
				   + timing_information_end.tms_cutime
				   - timing_information_begin.tms_utime
				   - timing_information_begin.tms_cutime)
	       / CLK_TCK
	    << " seconds elapsed";
	msg.precision(old_precision);
	msg.flags(old_flags);
      }

      msg << ")" << ends;

      msg.freeze();
      e.changeMessage(msg.str());
      msg.freeze(0);

      throw e;
    }

    times(&timing_information_end);

    if (configuration.global_options.verbosity >= 4)
      cout << " ok" << endl;

    ifstream automaton_file;
    openFile(automaton_file_name, automaton_file, ios::in, indent);

    if (configuration.global_options.verbosity == 5)
    {
      cout << string(indent, ' ') + "<reading automaton description>";
      cout.flush();
    }

    automaton_file >> *automaton;

    if (configuration.global_options.verbosity == 5)
      cout << " ok" << endl;

    if (timing_information_begin.tms_utime == static_cast<clock_t>(-1)
        || timing_information_begin.tms_cutime == static_cast<clock_t>(-1)
        || timing_information_end.tms_utime == static_cast<clock_t>(-1)
        || timing_information_end.tms_cutime == static_cast<clock_t>(-1))
      automaton_stats.buchi_generation_time = -1.0;
    else
      automaton_stats.buchi_generation_time =
        static_cast<double>(timing_information_end.tms_utime
                            + timing_information_end.tms_cutime
                            - timing_information_begin.tms_utime
                            - timing_information_begin.tms_cutime) / CLK_TCK;

    automaton_file.close();
  }
  catch (Exception& e)
  {
    delete automaton;

    if (automaton_file_name[0] != '\0' &&
	configuration.global_options.verbosity >= 4)
      cout << " error\n";

    if (configuration.global_options.verbosity >= 2)
    {
      cout << string(indent, ' ') + "Error";

      try
      {
	throw;
      }
      catch (ExecFailedException& e)
      {
	cout << ": " << e.what();
      }
      catch (BuchiAutomaton::AutomatonParseException& e)
      {
	cout << " parsing input: " << e.what();
      }
      catch (...)
      {
	cout << ": " << e.what();
      }

      cout << endl;
    }

    if (automaton_file_name[0] != '\0')
      removeFile(automaton_file_name, indent);

    if (cout_capture_file[0] != '\0')
    {
      if (configuration.global_options.verbosity >= 3)
	printFileContents(cout, "Contents of stdout:",
			  cout_capture_file, indent + 2, "> ");

      removeFile(cout_capture_file, indent);
    }

    if (cerr_capture_file[0] != '\0')
    {
      if (configuration.global_options.verbosity >= 3)
	printFileContents(cout, "Contents of stderr:", cerr_capture_file,
			  indent + 2, "> ");

      removeFile(cerr_capture_file, indent);
    }

    throw BuchiAutomatonGenerationException();
  }

  removeFile(automaton_file_name, indent);

  if (cout_capture_file[0] != '\0')
  {
    if (configuration.global_options.verbosity >= 3)
      printFileContents(cout, "Contents of stdout:", cout_capture_file,
			indent + 2, "> ");

    removeFile(cout_capture_file, indent);
  }

  if (cerr_capture_file[0] != '\0')
  {
    if (configuration.global_options.verbosity >= 3)
      printFileContents(cout, "Contents of stderr:", cerr_capture_file, 
			indent + 2, "> ");

    removeFile(cerr_capture_file, indent);
  }

  if (configuration.global_options.verbosity >= 4)
  {
    cout << string(indent, ' ') + "<computing statistics>";
    cout.flush();
  }

  pair<BuchiAutomaton::size_type, unsigned long int> buchi_stats =
    automaton->stats();

  automaton_stats.number_of_buchi_states = buchi_stats.first;
  automaton_stats.number_of_buchi_transitions = buchi_stats.second;
  automaton_stats.number_of_acceptance_sets
    = automaton->numberOfAcceptanceSets();

  if (configuration.global_options.verbosity >= 4)
    cout << " ok" << endl;

  return automaton;
}

/* ========================================================================= */
ProductAutomaton* generateProductAutomaton
  (const BuchiAutomaton& automaton, const StateSpace& statespace,
   AutomatonStats& automaton_stats, int indent)
/* ----------------------------------------------------------------------------
 *
 * Description:   Computes the product of a Bchi automaton with a reachability
 *                graph.
 *
 * Arguments:     automaton        --  A reference to a constant
 *                                     BuchiAutomaton.
 *                statespace       --  A reference to a constant StateSpace.
 *                automaton_stats  --  Data structure for storing information
 *                                     about the product automaton.
 *                indent           --  Number of spaces to leave on the left of
 *                                     the messages given to the user.
 *
 * Returns:       A pointer to a newly allocated ProductAutomaton.
 *
 * ------------------------------------------------------------------------- */
{
  ProductAutomaton* product_automaton = new ProductAutomaton();

  if (configuration.global_options.verbosity >= 4)
  {
    cout << string(indent, ' ') + "<computing product automaton>";
    cout.flush();
  }

  /*
   *  Initialize the product automaton. The variable
   *  `configuration.global_options.product_mode' is used to determine whether
   *  the product is to be computed with respect to all the states of the
   *  reachability graph or only the initial state of the graph.
   */

  try
  {
    product_automaton->computeProduct
      (automaton, statespace,
       configuration.global_options.product_mode == Configuration::GLOBAL);
  }
  catch (...)
  {
    delete product_automaton;

    try
    {
      throw;
    }
    catch (ProductAutomaton::ProductSizeException&)
    {
      if (configuration.global_options.verbosity >= 3)
      {
	if (configuration.global_options.verbosity >= 4)
	  cout << " error (product may be too big)" << endl;
	else
	  cout << string(10, ' ') + "[Error: Product may be too big]" << endl;
      }

      throw ProductAutomatonGenerationException();
    }
    catch (UserBreakException&)
    {
      if (configuration.global_options.verbosity >= 3)
      {
	if (configuration.global_options.verbosity >= 4)
	  cout << " user break" << endl;
	else
	  cout << string(10, ' ') + "[User break]\n" << endl;
      }

      throw;
    }
    catch (bad_alloc&)
    {
      if (configuration.global_options.verbosity >= 3)
      {
	if (configuration.global_options.verbosity >= 4)
	  cout << " error (out of memory)" << endl;
	else
	  cout << string(10, ' ') + "[Out of memory]" << endl;
      }

      throw ProductAutomatonGenerationException();
    }
    catch (...)
    {
      throw;
    }
  }

  /*
   *  Determine the number of states and transitions in the product.
   */

  if (configuration.global_options.verbosity >= 4)
  {
    cout << " ok\n" + string(indent, ' ') + "<computing statistics>";
    cout.flush();
  }

  pair<ProductAutomaton::size_type, unsigned long int> product_stats =
    product_automaton->stats();

  automaton_stats.number_of_product_states = product_stats.first;
  automaton_stats.number_of_product_transitions = product_stats.second;

  if (configuration.global_options.verbosity >= 4)
    cout << " ok" << endl;

  return product_automaton;
}

/* ========================================================================= */
void performEmptinessCheck
  (const ProductAutomaton& product_automaton, AutomatonStats& automaton_stats)
/* ----------------------------------------------------------------------------
 *
 * Description:   Performs the emptiness check on a ProductAutomaton, i.e.
 *                finds the states of the original reachability graph from
 *                which an accepting cycle of the Bchi automaton can be
 *                reached.
 *
 * Arguments:     product_automaton  --  A reference to a constant
 *                                       ProductAutomaton.
 *                automaton_stats    --  Data structure for storing the result
 *                                       of the emptiness check.
 *
 * Returns:       Nothing. The result of the emptiness check can be found
 *                in the Bitset object pointed to by
 *                `automaton_stats.emptiness_check_result'.
 *
 * ------------------------------------------------------------------------- */
{
  if (configuration.global_options.verbosity >= 4)
  {
    cout << string(10, ' ') + "<checking product automaton for emptiness>";
    cout.flush();
  }

  try
  {
    product_automaton.emptinessCheck(automaton_stats.emptiness_check_result);
    automaton_stats.emptiness_check_performed = true;
  }
  catch (...)
  {
    automaton_stats.emptiness_check_performed = false;

    try
    {
      throw;
    }
    catch (UserBreakException&)
    {
      if (configuration.global_options.verbosity >= 3)
      {
	if (configuration.global_options.verbosity >= 4)
	  cout << " user break" << endl;
	else
	  cout << string(10, ' ') + "[User break]\n" << endl;
      }

      throw;
    }
    catch (bad_alloc&)
    {
      if (configuration.global_options.verbosity >= 3)
      {
	if (configuration.global_options.verbosity >= 4)
	  cout << " error (out of memory)" << endl;
	else
	  cout << string(10, ' ') + "[Out of memory]" << endl;
      }

      throw EmptinessCheckFailedException();
    }
    catch (...)
    {
      throw;
    }
  }

  if (configuration.global_options.verbosity >= 4)
    cout << " ok" << endl;
}

/* ========================================================================= */
bool performConsistencyCheck
  (unsigned long int real_emptiness_check_size,
   AlgorithmTestResults& test_results)
/* ----------------------------------------------------------------------------
 *
 * Description:   Checks the model checking results for consistency for a
 *                particular LTL-to-Bchi conversion algorithm implementation,
 *                i.e. verifies that the model checking results for a formula
 *                and its negation are not contradictory.
 *
 * Arguments:     real_emptiness_check_size  --  Number of comparisons to be
 *                                               made.
 *                test_results               --  Data structure for storing
 *                                               information about the
 *                                               consistency check.
 *
 * Returns:       A truth value according to whether the check succeeded or
 *                not.
 *
 * ------------------------------------------------------------------------- */
{
  StateSpace::size_type state;
  const Bitset& acceptance_vector_for_formula
    = test_results.automaton_stats[0].emptiness_check_result;
  const Bitset& acceptance_vector_for_negation
    = test_results.automaton_stats[1].emptiness_check_result;

  if (configuration.global_options.verbosity >= 4)
  {
    cout << string(8, ' ') + "<comparing results>";
    cout.flush();
  }

  /*
   *  The consistency check will succeed if
   *  `test_results.consistency_check_result' has the value 1 at the end of the
   *  following loop.
   *
   *  The consistency check fails if there is a state in which both the
   *  formula and its negation are claimed to be false.
   */

  test_results.consistency_check_result = 1;

  for (state = 0; state < real_emptiness_check_size; ++state)
  {
    ++test_results.consistency_check_comparisons;
    if (!acceptance_vector_for_formula[state] &&
	!acceptance_vector_for_negation[state])
    {
      ++test_results.failed_consistency_check_comparisons;
      test_results.consistency_check_result = 0;
    }
  }

  if (configuration.global_options.verbosity >= 4)
    cout << (test_results.consistency_check_result == 1 ? " ok" : " failed")
         << endl;

  return (test_results.consistency_check_result == 1);
}      

/* ========================================================================= */
bool performPathCheck
  (const TestRoundInfo& round_info,
   vector<AlgorithmTestResults, ALLOC(AlgorithmTestResults) >& test_results,
   vector<TestStatistics, ALLOC(TestStatistics) >& stats)
/* ----------------------------------------------------------------------------
 *
 * Description:   Compares model checking results obtained using the different
 *                LTL-to-Bchi conversion algorithm implementation against
 *                those computed by the path checking algorithm.
 *
 * Arguments:     round_info    --  A reference to a constant TestRoundInfo
 *                                  object containing the result given by the
 *                                  path checking algorithm.
 *                test_results  --  A reference to a vector storing the model
 *                                  checking results obtained using different
 *                                  implementations.
 *                stats         --  Data structure storing the total number of
 *                                  path comparison checks performed for each
 *                                  pair of algorithms.
 *
 * Returns:       `true' if the check succeeded.
 *
 * ------------------------------------------------------------------------- */
{
  bool result = true;
  AutomatonStats* alg_stats;

  for (vector<AlgorithmTestResults, ALLOC(AlgorithmTestResults) >::size_type
	 alg = 0;
       alg < test_results.size();
       alg++)
  {
    for (int counter = 0; counter < 2; counter++)
    {
      alg_stats = &test_results[alg].automaton_stats[counter];
      if (alg_stats->emptiness_check_performed)
      {
	stats[alg].path_checks_performed[counter]++;

	alg_stats->path_check_result = 1;

	for (unsigned long int state = 0;
	     state < round_info.real_emptiness_check_size;
	     state++)
	{
	  if ((counter == 0
	       && alg_stats->emptiness_check_result[state]
	            != round_info.path_truth_values->test(state))
	      || (counter == 1
		  && alg_stats->emptiness_check_result[state]
		       == round_info.path_truth_values->test(state)))
	  {
	    stats[alg].path_check_failures[counter]++;
	    alg_stats->path_check_result = 0;
	    result = false;
	    break;
	  }
	}
      }
    }
  }

  return result;
}

/* ========================================================================= */
bool compareResults
  (vector<AlgorithmTestResults, ALLOC(AlgorithmTestResults) >& test_results,
   vector<TestStatistics, ALLOC(TestStatistics) >& stats)
/* ----------------------------------------------------------------------------
 *
 * Description:   Compares the model checking results obtained using different
 *                LTL->Bchi conversion algorithm implementations against each
 *                other.
 *
 * Arguments:     test_results  --  A reference to a vector storing the model
 *                                  checking results obtained using different
 *                                  implementations.
 *                stats         --  Data structure storing the total number of
 *                                  cross-comparisons performed for each pair
 *                                  of algorithms.
 *
 * Returns:       A truth value telling whether all the cross-comparison checks
 *                succeeded.
 *
 * ------------------------------------------------------------------------- */
{
  bool result = true;
  AutomatonStats* alg_1_stats;
  AutomatonStats* alg_2_stats;

  for (vector<AlgorithmTestResults, ALLOC(AlgorithmTestResults) >::size_type
	 alg_1 = 0;
       alg_1 < test_results.size();
       ++alg_1)
  {
    for (int counter = 0; counter < 2; counter++)
    {
      alg_1_stats = &test_results[alg_1].automaton_stats[counter];

      for (vector<AlgorithmTestResults, ALLOC(AlgorithmTestResults) >
	     ::size_type alg_2 = alg_1 + 1;
	   alg_2 < test_results.size();
	   ++alg_2)
      {
	alg_2_stats = &test_results[alg_2].automaton_stats[counter];

	if (configuration.algorithms[alg_1].enabled
	    && configuration.algorithms[alg_2].enabled
	    && alg_1_stats->emptiness_check_performed
	    && alg_2_stats->emptiness_check_performed)
	{
	  (stats[alg_1].cross_comparisons_performed[alg_2])++;
	  (stats[alg_2].cross_comparisons_performed[alg_1])++;

	  unsigned long int dist = alg_1_stats->emptiness_check_result
	                           - alg_2_stats->emptiness_check_result;

	  alg_1_stats->cross_comparison_stats[alg_2].first
	    = alg_2_stats->cross_comparison_stats[alg_1].first
	    = true;

	  alg_1_stats->cross_comparison_stats[alg_2].second
	    = alg_2_stats->cross_comparison_stats[alg_1].second
	    = dist;

	  if (dist > 0)
	  {
	    (stats[alg_1].cross_comparison_mismatches[alg_2])++;
	    (stats[alg_2].cross_comparison_mismatches[alg_1])++;

	    if (alg_1_stats->emptiness_check_result[0]
		!= alg_2_stats->emptiness_check_result[0])
	    {
	      (stats[alg_1].initial_cross_comparison_mismatches[alg_2])++;
	      (stats[alg_2].initial_cross_comparison_mismatches[alg_1])++;
	    }

	    result = false;
	  }
	}
      }
    }
  }

  return result;
}

/* ========================================================================= */
bool performBuchiIntersectionCheck
  (vector<AlgorithmTestResults, ALLOC(AlgorithmTestResults) >& test_results,
   vector<TestStatistics, ALLOC(TestStatistics) >& stats)
/* ----------------------------------------------------------------------------
 *
 * Description:   Tests the intersection of the Bchi automata constructed for
 *                the formula and its negation for emptiness.
 *
 * Arguments:     test_results       --  A reference to a vector storing the
 *                                       automata obtained using the different
 *                                       implementations.
 *                stats              --  Data structure storing the total
 *                                       number of intersection emptiness
 *                                       checks performed for each pair of
 *                                       algorithms.
 *
 * Returns:       A truth value according to whether the check was passed.
 *
 * ------------------------------------------------------------------------- */
{
  bool result = true;

  if (configuration.global_options.verbosity >= 4)
  {
    cout << string(8, ' ') + "<checking Bchi automata intersections for "
                             "emptiness>";
    cout.flush();
  }

  BuchiAutomaton* automaton_intersection = 0;

  for (vector<AlgorithmTestResults, ALLOC(AlgorithmTestResults) >::size_type
	 alg_1 = 0;
       alg_1 < test_results.size();
       ++alg_1)
  {
    for (vector<AlgorithmTestResults, ALLOC(AlgorithmTestResults) >::size_type
	   alg_2 = 0;
	 alg_2 < test_results.size();
	 ++alg_2)
    {
      try
      {
	/*
	 *  Compute the intersection of two Bchi automata constructed for
	 *  the positive and the negative formula, respectively.
	 */

	if (test_results[alg_1].automaton_stats[0].buchiAutomatonComputed()
	    && test_results[alg_2].automaton_stats[1].buchiAutomatonComputed())
	{
	  automaton_intersection
	    = BuchiAutomaton::intersect
	        (*(test_results[alg_1].automaton_stats[0].buchi_automaton),
		 *(test_results[alg_2].automaton_stats[1].buchi_automaton));

	  /*
	   *  Scan the nontrivial maximal strongly connected components of the
	   *  intersection automaton to check whether the automaton has any
	   *  accepting executions. If a MSCC with a state from every
	   *  acceptance set is found, the intersection emptiness check fails.
	   */

	  for (SccIterator<GraphEdgeContainer> scc(*automaton_intersection);
	       !scc.atEnd(); ++scc)
	  {
	    /*
	     *  MSCC nontriviality check (a MSCC is nontrivial iff it has at
	     *  least two states or if it consists of a single state connected
	     *  to itself).
	     */

	    if (scc->size() > 1
		|| (scc->size() == 1
		    && (*automaton_intersection)[*(scc->begin())].connected
		         (*(scc->begin()))))
	    {
	      Bitset acceptance_sets(automaton_intersection
				       ->numberOfAcceptanceSets());

	      unsigned long int accept_set;
	      unsigned long int acceptance_set_counter = 0;

	      for (set<GraphEdgeContainer::size_type,
	               less<GraphEdgeContainer::size_type>,
	               ALLOC(GraphEdgeContainer::size_type) >::const_iterator
		     state = scc->begin();
		   state != scc->end()
		     && acceptance_set_counter < acceptance_sets.size();
		   ++state)
	      {
		accept_set = acceptance_set_counter;
		while (accept_set < acceptance_sets.size())
		{
		  if ((*automaton_intersection)[*state].acceptanceSets().
		        test(accept_set))
		  {
		    acceptance_sets.set(accept_set);
		    if (accept_set == acceptance_set_counter)
		    {
		      do
			acceptance_set_counter++;
		      while (acceptance_set_counter < acceptance_sets.size()
			     && acceptance_sets.test(acceptance_set_counter));
		      accept_set = acceptance_set_counter;
		      continue;
		    }
		  }
		  accept_set++;
		}
	      }

	      if (acceptance_set_counter == acceptance_sets.size())
	      {
		test_results[alg_1].automaton_stats[0].
		  buchi_intersection_check_stats[alg_2] = 0;
		test_results[alg_2].automaton_stats[1].
		  buchi_intersection_check_stats[alg_1] = 0;

		stats[alg_1].buchi_intersection_check_failures[alg_2]++;
		if (alg_1 != alg_2)
		  stats[alg_2].buchi_intersection_check_failures[alg_1]++;

		result = false;

		break;
	      }
	    }
	  }

	  delete automaton_intersection;
	  automaton_intersection = 0;

	  if (test_results[alg_1].automaton_stats[0].
	        buchi_intersection_check_stats[alg_2] == -1)
	  {
	    test_results[alg_1].automaton_stats[0].
	      buchi_intersection_check_stats[alg_2] = 1;
	    test_results[alg_2].automaton_stats[1].
	      buchi_intersection_check_stats[alg_1] = 1;
	  }

	  stats[alg_1].buchi_intersection_checks_performed[alg_2]++;
	  if (alg_1 != alg_2)
	    stats[alg_2].buchi_intersection_checks_performed[alg_1]++;
	}
      }
      catch (UserBreakException&)
      {
	if (automaton_intersection != 0)
	  delete automaton_intersection;

	if (configuration.global_options.verbosity >= 3)
	{
	  if (configuration.global_options.verbosity >= 4)
	    cout << " (user break)" << endl;
	  else
	    cout << string(10, ' ') + "[User break]" << endl;
	}

	throw;
      }
      catch (bad_alloc&)
      {
	if (automaton_intersection != 0)
	  delete automaton_intersection;

	if (configuration.global_options.verbosity >= 3)
        {
	  if (configuration.global_options.verbosity >= 4)
	    cout << " error (out of memory)" << endl;
	  else
	    cout << string(10, ' ') + "[Out of memory]" << endl;
	}

	return false;
      }
    }
  }

  if (configuration.global_options.verbosity >= 4)
    cout << (result ? " ok" : " failed") << endl;

  return result;
}
