/*
 *  Copyright (C) 1999, 2000  Heikki Tauriainen <Heikki.Tauriainen@hut.fi>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "UserInterface.h"

/******************************************************************************
 *
 * Implementations for the user commands.
 *
 *****************************************************************************/

/* ========================================================================= */
void UserInterface::computeProductAutomaton
  (ProductAutomaton*& product_automaton,
   const BuchiAutomaton& buchi_automaton,
   pair<unsigned long int, bool>& last_automaton,
   const pair<unsigned long int, bool>& new_automaton) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Computes a product automaton.
 *
 * Arguments:     product_automaton  --  A reference to a pointer giving the
 *                                       storage location of the generated
 *                                       automaton.
 *                buchi_automaton    --  The Bchi automaton to be used for
 *                                       computing the product automaton.
 *                last_automaton     --  A pair telling the algorithm and the
 *                                       formula last used for computing an
 *                                       automaton (for testing whether the
 *                                       result is already available).
 *                new_automaton      --  A pair telling the algorithm and the
 *                                       formula which are to be used for
 *                                       computing the automaton.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  if (product_automaton == 0 || last_automaton != new_automaton)
  {
    AutomatonStats dummy_stats(0, 0);

    if (product_automaton != 0)
    {
      if (configuration.global_options.verbosity >= 4)
        cout << "  <cleaning up memory allocated for product automaton>"
             << endl;

      delete product_automaton;
      product_automaton = 0;
    }

    int old_verbosity = configuration.global_options.verbosity;
    configuration.global_options.verbosity = 4;

    try
    {
      product_automaton
	= generateProductAutomaton(buchi_automaton,
				   *(round_info.statespace),
				   dummy_stats,
				   2);
    }
    catch (...)
    {
      configuration.global_options.verbosity = old_verbosity;
      if (product_automaton != 0)
      {
	delete product_automaton;
	product_automaton = 0;
      }
      throw;
    }

    configuration.global_options.verbosity = old_verbosity;

    last_automaton = new_automaton;
  }
}

/* ========================================================================= */
void UserInterface::printAlgorithmList(ostream& stream, int indent) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Implements the user command `algorithms', i.e. writes a list
 *                of algorithms used in the tests to a stream.
 *
 * Arguments:     stream             --  A reference to an output stream.
 *                indent             --  Number of spaces to leave on the left
 *                                       of the output.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);

  vector<Configuration::AlgorithmInformation,
         ALLOC(Configuration::AlgorithmInformation) >::size_type algorithm_id;

  estream << string(indent, ' ') + "List of algorithms:\n";
    
  algorithm_id = 0;
  for (vector<Configuration::AlgorithmInformation,
              ALLOC(Configuration::AlgorithmInformation) >::const_iterator
         algorithm = configuration.algorithms.begin();
       algorithm != configuration.algorithms.end();
       ++algorithm)
  {
    estream << string(indent + 2, ' ') << algorithm_id
            << ": `" + *((*algorithm).name) + "' ("
               + (algorithm->enabled ? "en" : "dis") + "abled)\n";
    ++algorithm_id;
  }

  estream << endl;
}

/* ========================================================================= */
void UserInterface::synchronizePrefixAndCycle
  (deque<Graph::Graph<GraphEdgeContainer>::size_type,
         ALLOC(Graph::Graph<GraphEdgeContainer>::size_type) >& prefix,
   deque<Graph::Graph<GraphEdgeContainer>::size_type,
         ALLOC(Graph::Graph<GraphEdgeContainer>::size_type) >& cycle)
/* ----------------------------------------------------------------------------
 *
 * Description:   Helper function for `synchronizing' a sequence of states
 *                consisting of a prefix and a repeating cycle. This means
 *                removing from the end of the prefix the longest sequence of
 *                states which forms a postfix of the cycle. The states in the
 *                cycle will be `rotated' accordingly.
 *
 * Arguments:     prefix  --  A reference to a deque of state identifiers
 *                            forming the prefix of the state sequence.
 *                cycle   --  A reference to a deque of state identifiers
 *                            representing the states in the repeating cycle.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  while (!prefix.empty() && prefix.back() == cycle.back())
  {
    prefix.pop_back();
    cycle.push_front(cycle.back());
    cycle.pop_back();
  }
}

/* ========================================================================= */
void UserInterface::printCrossComparisonAnalysisResults
  (ostream& stream, int indent, bool formula_type,
   const string& option_1,
   const string& option_2,
   ProductAutomaton*& product_automaton,
   pair<unsigned long int, bool>& last_product_automaton)
/* ----------------------------------------------------------------------------
 *
 * Description:   Implements the user command `resultanalyze', i.e. analyzes a
 *                discrepancy between the results of two algorithms (or one
 *                algorithm against the path algorithm) by searching for a
 *                system execution producting contradicting results and then
 *                displaying the execution.
 *
 * Arguments:     stream                  --  A reference to an output stream.
 *                indent                  --  Number of spaces to leave on the
 *                                            left of the output.
 *                formula_type            --  Tells the LTL formula for which
 *                                            the analysis is to be performed.
 *                option_1,               --  References to constant strings   
 *                option_2                    specifying the algorithms for
 *                                            which the check is to be
 *                                            performed.
 *                product_automaton       --  A reference to a pointer telling
 *                                            the storage location of the
 *                                            generated product automaton.
 *                last_product_automaton  --  A pair telling the algorithm and
 *                                            the formula last used for
 *                                            computing a product automaton
 *                                            (for testing whether the result
 *                                            is already available).
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);

  unsigned long int algorithm1, algorithm2;
  bool path_compare = false;

  if (option_1 == "p")
  {
    path_compare = true;
    algorithm1 = parseNumber(option_2);
    algorithm2 = algorithm1 + 1;
  }
  else
  {
    algorithm1 = parseNumber(option_1);
    if (option_2 == "p")
    {
      path_compare = true;
      algorithm2 = algorithm1 + 1;
    }
    else
      algorithm2 = parseNumber(option_2);
  }

  if (path_compare
      && !(configuration.statespace_options.generate_mode
	   & Configuration::PATH))
    throw CommandErrorException("This feature is available only when using "
				"paths as state spaces.");

  verifyNumber(algorithm1, configuration.algorithms.size(),
	       "No such algorithm");

  int formula = (formula_type ? 0 : 1);
  const AutomatonStats* stats1;
  const AutomatonStats* stats2;

  stats1 = &test_results[algorithm1].automaton_stats[formula];

  if (!path_compare)
  {
    verifyNumber(algorithm2, configuration.algorithms.size(),
		 "No such algorithm");

    if (algorithm1 == algorithm2)
      throw CommandErrorException("Algorithm identifiers must be different.");

    stats2 = &test_results[algorithm2].automaton_stats[formula];

    if (!stats1->crossComparisonPerformed(algorithm2))
    {
      printTextBlock(stream, indent,
		     "Model checking result cross-comparison was not "
		     "performed between `"
		     + *(configuration.algorithms[algorithm1].name)
		     + "' and `" + *(configuration.algorithms[algorithm2].name)
		     + "'. \n",
		     78);
      return;
    }

    if (stats1->cross_comparison_stats[algorithm2].second == 0)
    {
      printTextBlock(stream, indent,
		     "No inconsistencies detected in the cross-comparison of "
		     "results given by `"
		     + *(configuration.algorithms[algorithm1].name) + "' and `"
		     + *(configuration.algorithms[algorithm2].name) + "'. \n",
		     78);
      return;
    }

    estream << string(indent, ' ') + "Model checking result cross-comparison ";
  }
  else
  {
    if (!stats1->emptiness_check_performed)
    {
      printTextBlock(stream, indent,
		     "Path check was not performed on the model checking "
		     "result given by `"
		     + *(configuration.algorithms[algorithm1].name) + "'. \n",
		     78);
      return;
    }

    if (stats1->path_check_result == 1)
    {
      printTextBlock(stream, indent,
		     '`' + *(configuration.algorithms[algorithm1].name)
                     + "' passed the path check on this input formula. \n",
		     78);
      return;
    }

    estream << string(indent, ' ') + "Path check inconsistency ";
  }

  estream << "analysis:\n" + string(indent + 2, ' ') + "Formula: ";

  round_info.formulae[configuration.formula_options.output_mode
	                == Configuration::NNF
		      ? formula
		      : 2 + formula].prettyPrint(stream);

  estream << "\n\n";

  if (!path_compare)
  {
    estream << string(indent + 2, ' ') + "The cross-comparison check failed "
                                         "in "
	    << stats1->cross_comparison_stats[algorithm2].second
	    << " (";

    changeStreamFormatting(stream, 0, 2, ios::fixed);
    estream << static_cast<double>
                 (stats1->cross_comparison_stats[algorithm2].second)
               / round_info.real_emptiness_check_size
               * 100.0;
    restoreStreamFormatting(stream);

    estream << "%) of "
	    << round_info.real_emptiness_check_size
	    << " test cases.\n\n";
  }

  /*
   *  Search for a system state in which the results of the two algorithms
   *  (or the result of an algorithm and the path checking algorithm) differ.
   */

  StateSpace::size_type state;

  if (!path_compare)
  {
    for (state = 0; 
	 stats1->emptiness_check_result[state] 
	   == stats2->emptiness_check_result[state];
	 ++state)
      ;
  }
  else
  {
    for (state = 0;
	 stats1->emptiness_check_result[state]
	   == (formula_type
	       ? (*round_info.path_truth_values)[state]
	       : !(*round_info.path_truth_values)[state]);
	 ++state)
      ;
  }

  /*
   *  Translate the formula to be analyzed into a Bchi automaton using the
   *  algorithm which claims the existence of an accepting cycle beginning in
   *  the state.
   */

  unsigned long int accepting_algorithm
    = (stats1->emptiness_check_result[state] ? algorithm1 : algorithm2);

  unsigned long int rejecting_algorithm = (accepting_algorithm == algorithm1
					   ? algorithm2
					   : algorithm1);

  vector<StateSpace::size_type, ALLOC(StateSpace::size_type) > path;

  deque<StateSpace::size_type, ALLOC(StateSpace::size_type) > system_prefix;
  deque<StateSpace::size_type, ALLOC(StateSpace::size_type) > system_cycle;
  deque<BuchiAutomaton::size_type, ALLOC(BuchiAutomaton::size_type) >
    automaton_prefix;
  deque<BuchiAutomaton::size_type, ALLOC(BuchiAutomaton::size_type) >
    automaton_cycle;

  if (!path_compare || accepting_algorithm == algorithm1)
  {
    /*
     *  Compute the synchronous product of the automaton and the reachability
     *  graph.
     */

    computeProductAutomaton(product_automaton,
			    *(test_results[accepting_algorithm].
			       automaton_stats[formula].buchi_automaton),
			    last_product_automaton,
			    make_pair(accepting_algorithm, formula_type));

    /*
     *  Search the product automaton for an accepting cycle.
     */

    pair<deque<ProductAutomaton::StateIdPair,
               ALLOC(ProductAutomaton::StateIdPair) >,
         deque<ProductAutomaton::StateIdPair,
               ALLOC(ProductAutomaton::StateIdPair) > >
      execution;

    try
    {
      cout << "  <searching for a system execution producing contradicting "
	      "results>";
      cout.flush();

      product_automaton->findAcceptingExecution(state, execution);
    }
    catch (...)
    {
      cout << " error" << endl;
      throw;
    }

    cout << " ok\n" << endl;

    /*
     *  Separate the parallel executions of the system and the automaton into
     *  prefixes and cycles.
     */

    while (!execution.first.empty())
    {
      automaton_prefix.push_back(execution.first.front().first);
      system_prefix.push_back(execution.first.front().second);
      execution.first.pop_front();
    }

    while (!execution.second.empty())
    {
      automaton_cycle.push_back(execution.second.front().first);
      system_cycle.push_back(execution.second.front().second);
      execution.second.pop_front();
    }

    synchronizePrefixAndCycle(automaton_prefix, automaton_cycle);
    synchronizePrefixAndCycle(system_prefix, system_cycle);

    state = execution.first.size();
  }
  else
  {
    StateSpace::size_type loop_state
      = (*((*round_info.statespace)[round_info.statespace->size() - 1].
	    edges().begin()))->targetNode();
				
    for (StateSpace::size_type path_state = state; path_state < loop_state;
	 path_state++)
      system_prefix.push_back(path_state);

    for (StateSpace::size_type path_state = loop_state;
	 path_state < round_info.statespace->size();
	 path_state++)
      system_cycle.push_back(path_state);

    state = loop_state;
  }

  /*
   *  Construct a path in the reachability graph from the system prefix and
   *  the cycle.
   */

  path.insert(path.end(), system_prefix.begin(), system_prefix.end());
  path.insert(path.end(), system_cycle.begin(), system_cycle.end());

  /*
   *  Write information about the execution to the output stream. That is,
   *  display the states in the infinite execution (a prefix and a cycle)
   *  and tell which one of the algorithms would accept and which one
   *  would reject the execution (or show whether the algorithm compared
   *  against the path checking algorithm accepted or rejected the
   *  execution).
   */

  estream << string(indent + 2, ' ') + "Execution M:\n";

  printPath(stream, indent + 4, system_prefix, system_cycle,
	    *(round_info.statespace));

  estream << string(indent + 4, ' ');

  if (!path_compare)
    estream << "accepted by: "
	    << accepting_algorithm
	    << ": `" + *(configuration.algorithms[accepting_algorithm].name)
               + "'\n" + string(indent + 4, ' ') + "rejected by: "
	    << rejecting_algorithm
	    << ": `" + *(configuration.algorithms[rejecting_algorithm].name)
               + "'\n";
  else
    estream << string(algorithm1 == accepting_algorithm ? "accep" : "rejec")
               + "ted by "
	    << algorithm1
	    << ": `" + *(configuration.algorithms[algorithm1].name) + "'\n";
               
  estream << endl;

  /*
   *  Verify the formula separately in the obtained execution to find out
   *  which one of the algorithms was in error.
   */

  LtlFormula::LtlFormula* f = 0;

  try
  {
    f = LtlFormula::formulaStringToLtlFormula
	  (round_info.formulae[configuration.formula_options.output_mode
	                         == Configuration::NNF
                               ? formula
	 	               : 2 + formula]);

    estream << string(indent + 2, ' ')
	    << "Analysis of the formula in the execution:\n";

    bool result = f->displayPathProof(stream, indent + 4,
				      *(round_info.statespace),
				      path, state);

    printTextBlock(stream, indent + 2,
		   string(" \n The formula is ") + (result ? "" : "not ")
                   + "satisfied in the execution, which should therefore be "
                   + (result ? "accep" : "rejec")
                   + "ted by all Bchi automata obtained from the formula."
		   + (path_compare
		      ? ""
		      : " This suggests an error in `"
                        + *(configuration.algorithms[result
                                                     ? rejecting_algorithm
				   	             : accepting_algorithm].
			      name)
		        + "' implementation.")
                   + " \n",
		   78);

    if (!result)
      printAcceptingCycle(stream, indent + 2, accepting_algorithm,
			  automaton_prefix, automaton_cycle,
			  *(test_results[accepting_algorithm].
			      automaton_stats[formula].buchi_automaton));

    delete f;
  }
  catch (...)
  {
    if (f != 0)
      delete f;
    throw;
  }
}

/* ========================================================================= */
void UserInterface::printAutomatonAnalysisResults
  (ostream& stream, int indent, unsigned long int algorithm1,
   unsigned long int algorithm2) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Implements the user command `buchianalyze', i.e. analyzes an
 *                inconsistency detected in the intersection emptiness check
 *                for two Bchi automata constructed for the formula and its
 *                negation.
 *
 * Arguments:     stream       --  A reference to an output stream.
 *                indent       --  Number of spaces to leave on the left of the
 *                                 output.
 *                algorithm1,  --  Identifiers of the algorithms for which the
 *                algorithm2       should be performed.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);

  verifyNumber(algorithm1, configuration.algorithms.size(),
	       "No such algorithm");
  verifyNumber(algorithm2, configuration.algorithms.size(),
	       "No such algorithm");

  if (test_results[algorithm1].automaton_stats[0].
        buchi_intersection_check_stats[algorithm2] == -1)
  {
    printTextBlock(stream, indent,
		   "Automata intersection emptiness check was not performed "
                   + string(algorithm1 == algorithm2 ? "o" : "betwee") + "n `"
                   + *(configuration.algorithms[algorithm1].name) + '\''
	           + (algorithm1 == algorithm2
		      ? ""
		      : " (positive formula) and `"
		        + *(configuration.algorithms[algorithm2].name)
                        + "' (negative formula)")
                   + " \n",
		   78);
    return;
  }

  if (test_results[algorithm1].automaton_stats[0].
        buchi_intersection_check_stats[algorithm2] == 1)
  {
    printTextBlock(stream, indent,
		   "The automata intersection emptiness check was successful "
		   + string(algorithm1 == algorithm2 ? "for" : "between")
	           + " `" + *(configuration.algorithms[algorithm1].name) + '\''
                   + (algorithm1 == algorithm2
		      ? ""
		      : " (positive formula) and `"
		        + *(configuration.algorithms[algorithm2].name)
		        + "' (negative formula)")
		   + " \n",
		   78);
    return;
  }

  estream << string(indent, ' ') + "Automata intersection emptiness check "
                                   "analysis:\n"
             + string(indent + 2, ' ') + "Positive formula: ";

  round_info.formulae[configuration.formula_options.output_mode
	                == Configuration::NNF
		      ? 0
		      : 2].prettyPrint(stream);

  estream << '\n' + string(indent + 2, ' ') + "Negative formula: ";
  round_info.formulae[configuration.formula_options.output_mode
		        == Configuration::NNF
		      ? 1
		      : 3].prettyPrint(stream);

  estream << '\n' + string(indent + 2, ' ') + "Algorithm"
             + (algorithm1 != algorithm2 ? "s:" : ": ") + string(7, ' ')
	  << algorithm1
	  <<  ": `" + *(configuration.algorithms[algorithm1].name) + '\'';

  if (algorithm1 != algorithm2)
    estream << " (positive formula)\n" + string(indent + 20, ' ')
	    << algorithm2
	    << ": `" + *(configuration.algorithms[algorithm2].name)
	       + "' (negative formula)";

  estream << '\n' << endl;

  /*
   *  Translate the formula and its negation into Bchi automata using the
   *  selected algorithm, then compute the intersection of the two
   *  automata.
   */
  
  BuchiAutomaton* a = 0;

  try
  {
    map<BuchiAutomaton::size_type, BuchiAutomaton::StateIdPair,
        less<BuchiAutomaton::size_type>, ALLOC(BuchiAutomaton::StateIdPair) >
      intersection_state_mapping;

    a = BuchiAutomaton::intersect
          (*(test_results[algorithm1].automaton_stats[0].buchi_automaton),
	   *(test_results[algorithm2].automaton_stats[1].buchi_automaton),
	   &intersection_state_mapping);

    /*
     *  Search the intersection automaton for an accepting execution. This is
     *  done `indirectly' as follows:
     *
     *      1.  Convert the intersection automaton into a StateSpace, i.e.
     *          construct a StateSpace which shares a similar transition
     *          relation with the automaton.
     *      2.  Construct a `dummy' single-state Bchi automaton which
     *          accepts all inputs.
     *      3.  Compute the synchronous product of the state space and the
     *          automaton. The obtained product automaton again has the same
     *          transition relation as the original intersection automaton
     *          (state identifiers may have been permuted; however, the
     *          product automaton contains information about the
     *          correspondence between the state identifiers of the product 
     *          automaton and the original intersection automaton).
     *      4.  Switch the roles of the state space and the Bchi automaton in
     *          the product to effectively transfer state acceptance
     *          information to the product automaton from the automaton
     *          intersection.
     *      5.  Search an accepting cycle in the product automaton. This then
     *          corresponds to an accepting execution of the intersection
     *          automaton.
     */
    
    StateSpace automaton_as_statespace(0, a->size());

    for (BuchiAutomaton::size_type state = 0; state < a->size(); state++)
    {
      for (GraphEdgeContainer::const_iterator transition
	     = (*a)[state].edges().begin();
	   transition != (*a)[state].edges().end();
	   ++transition)
	automaton_as_statespace[state].
	  connectTo((*transition)->targetNode());
    }

    BuchiAutomaton dummy_automaton(1, 0, 0);
    dummy_automaton[0].connectTo(0, new LtlFormula::True());

    ProductAutomaton p;
    p.computeProduct(dummy_automaton, automaton_as_statespace, false);

    p.buchi_automaton = a;
    p.statespace_size = 1;

    /*
     *  Search the product automaton for an accepting cycle.
     */

    pair<deque<BuchiAutomaton::StateIdPair,
               ALLOC(BuchiAutomaton::StateIdPair) >,
         deque<BuchiAutomaton::StateIdPair,
               ALLOC(BuchiAutomaton::StateIdPair) > >
      execution;

    try
    {
      cout << "  <searching for accepting execution in automata intersection>";
      cout.flush();

      p.findAcceptingExecution(0, execution);
    }
    catch (...)
    {
      cout << " error" << endl;
      throw;
    }

    cout << " ok\n" << endl;

    deque<BuchiAutomaton::size_type, ALLOC(BuchiAutomaton::size_type) > prefix;
    deque<BuchiAutomaton::size_type, ALLOC(BuchiAutomaton::size_type) > cycle;

    while (!execution.first.empty())
    {
      prefix.push_back(execution.first.front().first);
      execution.first.pop_front();
    }

    while (!execution.second.empty())
    {
      cycle.push_back(execution.second.front().first);
      execution.second.pop_front();
    }

    vector<StateSpace::size_type, ALLOC(StateSpace::size_type) >
      path(prefix.begin(), prefix.end());

    path.insert(path.end(), cycle.begin(), cycle.end());
    path.push_back(cycle.front());

    /*
     *  Construct an execution accepted by both of the automata. This is done
     *  by giving suitable truth assignments for the atomic proposition in
     *  selected states of the state space to which the intersection automaton
     *  was converted.
     */

    prefix.push_back(cycle.front());
    cycle.push_back(cycle.front());

    const deque<StateSpace::size_type, ALLOC(StateSpace::size_type) >*
      execution_states;
    deque<StateSpace::size_type, ALLOC(StateSpace::size_type) >::size_type
      index_shift;

    GraphEdgeContainer::const_iterator transition;

    for (int counter = 0; counter < 2; counter++)
    {
      if (counter == 0)
      {
	execution_states = &prefix;
	index_shift = 0;
      }
      else
      {
	execution_states = &cycle;
	index_shift = prefix.size() - 1;
      }
      
      for (vector<StateSpace::size_type, ALLOC(StateSpace::size_type) >
	     ::size_type state = 0;
	   state + 1 < execution_states->size();
	   state++)
      {
	for (transition = (*a)[path[state + index_shift]].edges().begin();
	     (*transition)->targetNode() != path[state + index_shift + 1];
	     ++transition)
	  ;

	automaton_as_statespace[path[state + index_shift]].positiveAtoms()
	  = static_cast<BuchiAutomaton::BuchiTransition*>(*transition)
	      ->guard().findPropositionalModel();
      }
    }

    delete a;
    a = 0;

    path.pop_back();
    prefix.pop_back();
    cycle.pop_back();

    /*
     *  Display the input sequence accepted by both automata.
     */

    estream << string(indent + 2, ' ')
               + "Execution M accepted by both automata:\n";

    printPath(stream, indent + 4, prefix, cycle, automaton_as_statespace);

    estream << '\n';

    /*
     *  For each of the original automata, display the accepting runs that
     *  these automata have on the input sequence.
     */

    deque<BuchiAutomaton::size_type, ALLOC(BuchiAutomaton::size_type) >
      aut_prefix;

    deque<BuchiAutomaton::size_type, ALLOC(BuchiAutomaton::size_type) >
      aut_cycle;

    const deque<BuchiAutomaton::size_type, ALLOC(BuchiAutomaton::size_type) >*
      original_execution_states;
    deque<BuchiAutomaton::size_type, ALLOC(BuchiAutomaton::size_type) >*
      new_execution_states;

    for (int i = 0; i < 2; i++)
    {
      aut_prefix.clear();
      aut_cycle.clear();

      for (int j = 0; j < 2; j++)
      {
	if (j == 0)
	{
	  original_execution_states = &prefix;
	  new_execution_states = &aut_prefix;
	}
	else
	{
	  original_execution_states = &cycle;
	  new_execution_states = &aut_cycle;
	}

	for (deque<BuchiAutomaton::size_type,
                   ALLOC(BuchiAutomaton::size_type) >::const_iterator
	       execution_state = original_execution_states->begin();
	   execution_state != original_execution_states->end();
	   ++execution_state)
	{
	  new_execution_states->push_back
	    (i == 0
	     ? intersection_state_mapping[*execution_state].first
	     : intersection_state_mapping[*execution_state].second);
	}
      }

      synchronizePrefixAndCycle(aut_prefix, aut_cycle);

      printAcceptingCycle(stream, indent + 2,
			  (i == 0 ? algorithm1 : algorithm2),
			  aut_prefix, aut_cycle,
			  *(test_results[(i == 0 ? algorithm1 : algorithm2)].
			      automaton_stats[i].buchi_automaton));
    }

    LtlFormula::LtlFormula* f = 0;

    try
    {
      f = LtlFormula::formulaStringToLtlFormula
	    (round_info.formulae[configuration.formula_options.output_mode
		  	           == Configuration::NNF
                                 ? 0
	 	                 : 2]);

      estream << string(indent + 2, ' ')
	      << "Analysis of the positive formula in the execution M:\n";

      bool result = f->displayPathProof(stream, indent + 4,
					automaton_as_statespace, path,
					prefix.size());

      printTextBlock(stream, indent + 2,
		     " \n The positive formula is "
		     + string(result ? "" : "not ")
		     + "satisfied in the execution. This suggests that the "
                       "Bchi automaton constructed for the "
		     + (result ? "nega" : "posi") + "tive formula "
                     + (algorithm1 == algorithm2
			? ""
			: "(the automaton constructed by `"
			  + *(configuration.algorithms[result
						       ? algorithm2
						       : algorithm1].name)
                          + "') ")
		     + "is incorrect. \n",
		     78);

      delete f;
    }
    catch (...)
    {
      if (f != 0)
	delete f;
      throw;
    }
  }
  catch (...)
  {
    if (a != 0)
      delete a;
  }
}

/* ========================================================================= */
void UserInterface::printPath
  (ostream& stream, int indent,
   const deque<StateSpace::size_type, ALLOC(StateSpace::size_type) >& prefix,
   const deque<StateSpace::size_type, ALLOC(StateSpace::size_type) >& cycle,
   const StateSpace& path)
/* ----------------------------------------------------------------------------
 *
 * Description:   Writes information about a single execution path to a stream.
 *
 * Arguments:     stream       --  A reference to an output stream.
 *                indent       --  Number of spaces to leave to the left of
 *                                 output.
 *                prefix       --  A reference to a constant deque of state
 *                                 identifiers forming the prefix of the
 *                                 execution.
 *                cycle        --  A reference to a constant deque of state
 *                                 identifiers forming the infinitely repeating
 *                                 cycle in the execution.
 *                path         --  A reference to the constant state space from
 *                                 which the state identifiers in the deques
 *                                 are taken.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::badbit | ios::failbit);

  const deque<StateSpace::size_type, ALLOC(StateSpace::size_type) >*
    execution_states;

  for (int counter = 0; counter < 2; counter++)
  {
    estream << string(indent, ' ');

    if (counter == 0)
    {
      execution_states = &prefix;
      estream << "prefix:      < ";
    }
    else
    {
      execution_states = &cycle;
      estream << "cycle:       < ";
    }

    if (!execution_states->empty())
    {
      deque<StateSpace::size_type, ALLOC(StateSpace::size_type) >
	::const_iterator execution_state = execution_states->begin();

      bool first_printed;

      do
      {
	if (execution_state != execution_states->begin())
	  estream << ",\n" + string(indent + 15, ' ');

	estream << 's' << *execution_state << "  {";

	first_printed = false;

	for (unsigned long int proposition = 0;
	     proposition < path[*execution_state].positiveAtoms().size();
	     proposition++)
	{
	  if (path[*execution_state].positiveAtoms().test(proposition))
	  {
	    if (first_printed)
	      estream << ", ";
	    else
	      first_printed = true;

	    estream << 'p' << proposition;
	  }
	}

	estream << '}';

	++execution_state;
      }
      while (execution_state != execution_states->end());
    }
    else
      estream << "empty";
    
    estream << " >\n";
  }
}

/* ========================================================================= */
void UserInterface::printAcceptingCycle
  (ostream& stream, int indent,
   vector<Configuration::AlgorithmInformation,
          ALLOC(Configuration::AlgorithmInformation) >::size_type
     algorithm_id,		
   const deque<BuchiAutomaton::size_type, ALLOC(BuchiAutomaton::size_type) >&
     prefix,
   const deque<BuchiAutomaton::size_type, ALLOC(BuchiAutomaton::size_type) >&
     cycle,
   const BuchiAutomaton& automaton)
/* ----------------------------------------------------------------------------
 *
 * Description:   Writes information about a single automaton execution to a
 *                stream.
 *
 * Arguments:     stream        --  A reference to an output stream.
 *                indent        --  Number of spaces to leave to the left of
 *                                  output.
 *                algorithm_id  --  Identifier for an algorithm.
 *                prefix        --  A reference to a constant deque of state
 *                                  identifiers forming the prefix of the
 *                                  execution.
 *                cycle         --  A reference to a constant deque of state
 *                                  identifiers forming the infinitely
 *                                  repeating cycle in the execution.
 *                automaton     --  A reference to a constant BuchiAutomaton
 *                                  from which the state identifiers in the
 *                                  deques are taken.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);

  const deque<BuchiAutomaton::size_type, ALLOC(BuchiAutomaton::size_type) >*
    execution_states;
  bool first_printed;

  ostrstream msg;
  msg << "On input M, the automaton constructed by `"
         + *(configuration.algorithms[algorithm_id].name)
         + "' (with "
      << automaton.numberOfAcceptanceSets()
      << " acceptance set"
      << (automaton.numberOfAcceptanceSets() == 1 ? "" : "s")
      << ") has the following accepting execution:"
      << ends;

  msg.freeze();
  printTextBlock(stream, indent, msg.str(), 78);
  msg.freeze(0);
		     
  for (int counter = 0; counter < 2; counter++)
  {
    estream << string(indent + 2, ' ');

    if (counter == 0)
    {
      estream << "prefix: < ";
      execution_states = &prefix;
    }
    else
    {
      estream << "cycle:  < ";
      execution_states = &cycle;
    }

    if (!execution_states->empty())
    {
      first_printed = false;
	  
      for (deque<BuchiAutomaton::size_type,
	         ALLOC(BuchiAutomaton::size_type) >::const_iterator
	     execution_state = execution_states->begin();
	   execution_state != execution_states->end();
	   ++execution_state)
      {
	if (first_printed)
	  estream << ", ";
	else
	  first_printed = true;
	estream << *execution_state;

	if (counter == 1 && automaton.numberOfAcceptanceSets() > 0)
	{
	  bool first_acceptance_set_printed = false;

	  for (unsigned long int accept_set = 0;
	       accept_set < automaton.numberOfAcceptanceSets();
	       accept_set++)
	  {
	    if (automaton[*execution_state].acceptanceSets().test(accept_set))
	    {
	      if (first_acceptance_set_printed)
		estream << ", ";
	      else
	      {
		first_acceptance_set_printed = true;
		estream << " (acc: ";
	      }
	      estream << accept_set;
	    }
	  }

	  if (first_acceptance_set_printed)
	    estream << ')';
	}
      }
    }
    else
      estream << "empty";

    estream << " >\n";
  }

  estream << '\n';
}
 
/* ========================================================================= */
void UserInterface::printBuchiAutomaton
  (ostream& stream, int indent, bool formula_type,
   vector<string, ALLOC(string) >& input_tokens,
   Graph::GraphOutputFormat fmt) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Implements the user command `buchi', i.e. writes information
 *                about a set of states of a Bchi automaton computed using a
 *                given algorithm to a stream.
 *
 * Arguments:     stream           --  A reference to an output stream.
 *                indent           --  Number of spaces to leave on the left of
 *                                     the output.
 *                formula_type     --  Determines the formula to be translated
 *                                     into a Bchi automaton.
 *                input_tokens     --  A reference to a vector containing the
 *                                     arguments of the user command (the
 *                                     algorithm and automaton state numbers).
 *                fmt              --  Determines the format of output.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);
  unsigned long int algorithm = parseNumber(input_tokens[1]);

  if (algorithm >= configuration.algorithms.size())
    throw CommandErrorException("No such algorithm ("
                                + input_tokens[1] + ").");
  else
  {
    if (!test_results[algorithm].automaton_stats[formula_type ? 0 : 1].
	  buchiAutomatonComputed())
    {
      printTextBlock(stream, indent,
		     "No automaton was generated using algorithm `"
                     + *(configuration.algorithms[algorithm].name)
                     + "' for the formula",
		     78);

      estream << string(indent, ' ');
      round_info.formulae[(configuration.formula_options.output_mode
			     == Configuration::NNF
			   ? 0
			   : 2)
			  + (formula_type ? 0 : 1)].prettyPrint(stream);
      estream << " .\n" << endl;

      return;
    }

    const BuchiAutomaton* automaton
      = test_results[algorithm].automaton_stats[formula_type ? 0 : 1].
	  buchi_automaton;

    if (fmt == Graph::NORMAL)
    {
      /*
       *  Display information about the states of the automaton. If no state
       *  list was given, display the whole automaton.
       */

      estream << string(indent, ' ') + "Bchi automaton information:\n"
                 + string(indent + 2, ' ') + "Formula: ";

      round_info.formulae[(configuration.formula_options.output_mode
			     == Configuration::NNF
			   ? 0
			   : 2)
			  + (formula_type ? 0 : 1)].prettyPrint(stream);

      estream << '\n' + string(indent + 2, ' ') + "Algorithm: "
                 + *(configuration.algorithms[algorithm].name) + '\n'
	         + string(indent + 2, ' ') + "Number of acceptance sets: "
	      << automaton->numberOfAcceptanceSets()
	      << '\n';

      if (automaton->empty())
        estream << string(indent + 2, ' ') + "The automaton is empty.\n";
      else
      {
        set<unsigned long int, less<unsigned long int>,
            ALLOC(unsigned long int) > states;

        if (input_tokens.size() == 2)
          input_tokens.push_back("*");

        parseInterval(input_tokens[2], states, 0, automaton->size() - 1);

        for (set<unsigned long int, less<unsigned long int>,
                 ALLOC(unsigned long int) >::const_iterator
	       state = states.begin();
             state != states.end(); ++state)
        {
	  verifyNumber(*state, automaton->size(),
		       "State identifier out of range");

          estream << string(indent + 2, ' ') + "State " << *state
                  << string(*state == automaton->initialState()
                            ? " (initial state)" : "") + ":\n";
          automaton->node(*state).print(stream, indent + 4);
        }
      }
    }
    else if (fmt == Graph::DOT)
      automaton->print(stream, indent, Graph::DOT);

    estream << endl;
  }
}

/* ========================================================================= */
void UserInterface::evaluateFormula
  (ostream& stream, int indent, bool formula_type, 
   vector<string, ALLOC(string) >& input_tokens) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Implements the user command `evaluate', i.e. tells whether
 *                there exists an accepting execution beginning at a given
 *                system state (or several system states) according to the
 *                product automaton constructed from a Bchi automaton computed
 *                using a given algorithm (or algorithms).
 *
 * Arguments:     stream        --  A reference to an output stream.
 *                indent        --  Number of spaces to leave on the left of
 *                                  the output.
 *                formula_type  --  Determines the LTL formula to be evaluated.
 *                input_tokens  --  A reference to a vector of strings
 *                                  containing the arguments of the user
 *                                  command.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);
  set<unsigned long int, less<unsigned long int>, ALLOC(unsigned long int) >
    algorithms;
  set<unsigned long int, less<unsigned long int>, ALLOC(unsigned long int) >
    system_states;
  string algorithm_name;
  int formula = (formula_type ? 0 : 1);

  /*
   *  If no list of algorithms was given as an argument, show the results of
   *  all algorithms.
   */

  if (input_tokens.size() < 2)
    input_tokens.push_back("*");

  /*
   *  If no list of states was given, show information about all states.
   *  (If only a local product was computed, show information only about the
   *  initial state.)
   */

  if (input_tokens.size() < 3)
    input_tokens.push_back("*");

  parseInterval(input_tokens[1], algorithms, 0,
                configuration.algorithms.size() - 1);
  parseInterval(input_tokens[2], system_states, 0,
                configuration.global_options.product_mode
                  == Configuration::GLOBAL ? round_info.statespace->size() - 1
                                           : 0);

  estream << string(indent, ' ') + "Acceptance information:\n"
             + string(indent + 2, ' ') + "CTL* formula: E ";

  round_info.formulae[configuration.formula_options.output_mode
		        == Configuration::NNF
		      ? formula
		      : 2 + formula].prettyPrint(stream);

  estream << '\n';

  for (set<unsigned long int, less<unsigned long int>,
           ALLOC(unsigned long int) >::const_iterator
	 state = system_states.begin();
       state != system_states.end();
       ++state)
  {
    verifyNumber(*state, (configuration.global_options.product_mode
			    == Configuration::GLOBAL
			  ? round_info.statespace->size()
			  : 1),
		 "State identifier out of range");

    estream << string(indent + 2, ' ') + "State "
            << *state << ":\n";

    for (set<unsigned long int, less<unsigned long int>,
             ALLOC(unsigned long int) >::const_iterator
	   algorithm = algorithms.begin();
         algorithm != algorithms.end();
         ++algorithm)
    {
      estream << string(indent + 4, ' ');
      verifyNumber(*algorithm, configuration.algorithms.size(),
		   "No such algorithm");

      algorithm_name =
        configuration.algorithms[*algorithm].name->substr(0, 20);
      estream << algorithm_name + ':'
                 + string(21 - algorithm_name.length(), ' ');

      if (!test_results[*algorithm].automaton_stats[formula].
             emptiness_check_performed)
        estream << "acceptance cycles not computed\n";
      else
        estream << (test_results[*algorithm].automaton_stats[formula].
                      emptiness_check_result[*state]
                    ? "true"
                    : "false")
                << '\n';
    }
  }

  estream << endl;
}    

/* ========================================================================= */
void UserInterface::printFormula
  (ostream& stream, int indent, bool formula_type) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Implements the user command `formula', i.e. displays the
 *                random LTL formula (or its negation) used for generating the
 *                Bchi automata. The formula is shown in its original form and
 *                possibly also in negation normal form if the user has
 *                requested the conversion.
 *
 * Arguments:     stream        --  A reference to an output stream.
 *                indent        --  Number of spaces to leave on the left of
 *                                  the output.
 *                formula_type  --  Identifies the formula to be displayed.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);

  estream << string(indent, ' ') + "Formula:" + string(17, ' ');
  round_info.formulae[formula_type ? 2 : 3].prettyPrint(stream);

  if (configuration.formula_options.output_mode == Configuration::NNF)
  {
    estream << '\n' + string(indent, ' ') + "In negation normal form: ";
    round_info.formulae[formula_type ? 0 : 1].prettyPrint(stream);
  }

  estream << "\n\n";
  estream.flush();
}
  
/* ========================================================================= */
void UserInterface::printCommandHelp
  (ostream& stream, int indent,
   const vector<string, ALLOC(string) >& input_tokens)
/* ----------------------------------------------------------------------------
 *
 * Description:   Implements the user command `help', i.e. gives instructions
 *                on the use of different commands.
 *
 * Arguments:     stream        --  A reference to an output stream.
 *                indent        --  Number of spaces to leave on the left of 
 *                                  the output.
 *                input_tokens  --  A reference to a constant vector of strings
 *                                  containing the arguments of the user
 *                                  command.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);

  TokenType command = _NO_INPUT;

  if (input_tokens.size() > 1)
    command = parseCommand(input_tokens[1]);

  if (command == UNKNOWN)
    estream << string(indent, ' ') + "Unknown command (`" + input_tokens[1]
                                 + "').\n\n";

  switch (command)
  {
    case ALGORITHMS :
      estream << string(indent, ' ') + "algorithms\n";

      printTextBlock(stream, indent + 4,
                     "Display a list of algorithm identifiers which are used "
                     "as parameters for other commands. Tell also whether the"
		     " algorithms are currently enabled for testing.",
                     78);
      break;

    case BUCHI :
      estream << string(indent, ' ') + "buchi [\"+\"|\"-\"] <algorithm> "
                                       "[states|\"dot\"]\n";

      printTextBlock(stream, indent + 4,
                     "Display information about a (list of) state(s) of a "
                     "Bchi automaton constructed either from the positive or "
                     "the negated formula using a given algorithm (+ denotes "
                     "the positive formula, - the negated formula). If no "
                     "formula is specified, use the positive formula. If no "
                     "state list is given, display all the states of the "
                     "automaton. The description of the entire automaton can "
                     "alternatively be obtained in a format which can be "
                     "input to the `dot' tool to obtain a graphical "
                     "representation of the automaton. To do this, "
                     "replace the state list with the string \"dot\".",
                     78);
      break;

    case BUCHIANALYZE :
      estream << string(indent, ' ') + "buchianalysis <algorithm 1> "
                                       "<algorithm 2>\n";

      printTextBlock(stream, indent + 4,
		     "Analyze a contradiction in the Bchi automata "
                     "intersection emptiness check between the automaton "
                     "constructed from the positive formula by `algorithm 1' "
                     "and the automaton constructed from the negated formula "
                     "by `algorithm 2'.",
		     78);
      break;

    case CONTINUE :
      estream << string(indent, ' ') + "continue [number of rounds]\n";

      printTextBlock(stream, indent + 4,
                     "Continue testing, pausing next time after a given "
                     "number of rounds or when another error occurs (if "
                     "using the `OnError' value for the `Interactive' "
                     "option in the program configuration file).  The "
                     "number of rounds defaults to 1 if no value is "
                     "specified. \n \n Note: The output of this command is "
                     "not redirectable to a file.",
                     78);
      break;

    case DISABLE :
      estream << string(indent, ' ') + "disable [algorithms]\n";

      printTextBlock(stream, indent + 4,
		     "Disables a given set of algorithms to use in the tests. "
		     "If no algorithms are specified, disables all algorithms."
		     " \n \n Note: The output of this command is not "
		     "redirectable to a file.",
		     78);
      break;

    case ENABLE :
      estream << string(indent, ' ') + "enable [algorithms]\n";

      printTextBlock(stream, indent + 4,
		     "Enables a given set of algorithms to use in the tests. "
		     "If no algorithms are specified, enables all algorithms."
		     " \n \n Note: The output of this command is not "
		     "redirectable to a file.",
		     78);
      break;

    case EVALUATE :
      estream << string(indent, ' ') + "evaluate [\"+\"|\"-\"] "
                                       "[algorithms] [states]\n";

      printTextBlock(stream, indent + 4,
                     "Determine whether the Bchi automaton constructed from "
                     "a formula (+ denotes the positive formula, - the "
                     "negated formula) using a given algorithm accpets some "
                     "system execution starting from a given system state. "
                     "If no formula is specified, use the positive formula. "
                     "Leaving the list of algorithms or the list of states "
                     "unspecified will display the results for all algorithms "
                     "or all system states, respectively.",
                     78);
      break;

    case FORMULA :
      estream << string(indent, ' ') + "formula [\"+\"|\"-\"]\n";

      printTextBlock(stream, indent + 4,
                     "Display the LTL formula used in the last round for "
                     "generating Bchi automata (+ denotes the positive "
                     "formula, - the negated formula). If no formula is "
                     "specified, use the positive formula.",
                     78);
      break;

    case HELP :
      estream << string(indent, ' ') + "help [command]\n";

      printTextBlock(stream, indent + 4,
                     "Display help about a specific command. If no command "
                     "name is given, give a list of all available commands."
                     " \n \n In command-specific help, arguments in angle "
                     "brackets < > are obligatory, while square bracketed [ ] "
                     "arguments are optional. A vertical bar | denotes "
                     "selection between several possibilities. Arguments "
                     "in quotes must be entered literally (without the "
                     "quotes itself).",
                     78);
      break;

    case INCONSISTENCIES :
      estream << string(indent, ' ') + "inconsistencies [algorithms]\n";

      printTextBlock(stream, indent + 4,
                     "List the system states where the consistency check "
                     "failed for a given (set of) algorithm(s).",
                     78);
      break;

    case PATHINCONSISTENCIES :
      estream << string(indent, ' ') + "pathinconsistencies [\"+\"|\"-\"] "
	                               "[algorithms]\n";

      printTextBlock(stream, indent + 4,
		     "List the system states where the path check failed on "
		     "the positive or the negated input formula for "
		     "a given (set of) algorithm(s). \n \n Note: This command "
		     "is available only when using paths as reachability "
		     "graphs.",
		     78);
      break;

    case QUIT :
      estream << string(indent, ' ') + "quit\n";

      printTextBlock(stream, indent + 4, "Abort testing.", 78);
      break;

    case RESULTANALYZE :
      estream << string(indent, ' ') + "resultanalysis [\"+\"|\"-\"] "
                                       "<algorithm> <algorithm>\n";

      printTextBlock(stream, indent + 4,
                     "Analyze a contradiction in the model checking results "
                     "of two algorithms on a formula (+ denotes the positive "
                     "formula, - the negated formula). If no formula "
                     "is specified, use the positive formula. When using "
                     "paths as state spaces, one of algorithm identifiers "
                     "can be replaced by the letter \"p\" to analyze the "
                     "result given by some algorithm against that given by "
                     "the path checking algorithm.",
                     78);
      break;

    case RESULTS :
      estream << string(indent, ' ') + "results [algorithms]\n";

      printTextBlock(stream, indent + 4,
                     "Display this round's test results for a given (set of) "
                     "algorithm(s). If no algorithms are specified, show "
                     "the results for all algorithms.",
                     78);
      break;

    case SKIP :
      estream << string(indent, ' ') + "skip [number of rounds]\n";

      printTextBlock(stream, indent + 4,
                     "Skip a given number of rounds (defaults to 1 if no "
                     "number is specified). \n \n Note: The output of this "
                     "command is not redirectable to a file.",
                     78);
      break;

    case STATESPACE :
      estream << string(indent, ' ') + "statespace [states|\"dot\"]\n";

      printTextBlock(stream, indent + 4,
                     "Display information about a given (list of) system "
                     "state(s). Display the whole reachability graph if no "
                     "states are specified. Alternatively, the state "
                     "space description can be obtained in a format which "
                     "can be input to the `dot' tool to obtain a graphical "
                     "representation of the reachability graph. To do this, "
                     "replace the state list with the string \"dot\".",
                     78);
      break;

    case STATISTICS :
      estream << string(indent, ' ') + "statistics\n";

      printTextBlock(stream, indent + 4,
                     "Display test statistics over all tests performed.",
                     78);
      break;
    
    case VERBOSITY :
      estream << string(indent, ' ') + "verbosity [0-5]\n";

      printTextBlock(stream, indent + 4,
                     "Change the verbosity of the output produced by the "
                     "program. If no value is given, display the current "
                     "setting. \n \n Note: The output of this command is "
                     "not redirectable to a file.",
                     78);
      break;
    
    default :
      printTextBlock(stream, indent,
                     "List of available commands (use `help "
                     "[command]' for command specific help): \n",
                     78);

      printTextBlock(stream, indent + 2,
                     "algorithms \n "
                     "buchi \n "
                     "buchianalysis \n "
                     "continue \n "
		     "disable \n "
		     "enable \n "
                     "evaluate \n "
                     "help \n "
                     "inconsistencies \n "
		     "pathinconsistencies \n "
                     "quit \n "
                     "resultanalysis \n "
                     "results \n "
                     "skip \n "
                     "statespace \n "
                     "statistics \n "
                     "verbosity \n",
                     78);

      printTextBlock(stream, indent,
                     "Command names can be abbreviated to the shortest prefix "
		     "sufficient for identifying the command unambiguously. "
                     "For example, the prefix `c' can be used for the "
		     "`continue' command. \n \n "
                     "Lists of algorithm or state identifiers may be "
                     "specified as comma-separated intervals (with no spaces "
		     "in between), e.g., the command `statespace "
		     "-5,8,14-18,22-' would display the list of all states "
		     "with an identifier less than or equal to 5, state 8 and "
		     "states 14--18, together with the states whose "
		     "identifier is greater than or equal to 22. \n \n "
                     "The output of most commands can be redirected or "
                     "appended to a file by ending the command line with "
                     "`>filename' or `>>filename', respectively.",
                     78);
      break;
  }

  estream << endl;
  estream.flush();
}

/* ========================================================================= */
void UserInterface::printInconsistencies
  (ostream& stream, int indent, unsigned long int round,
   vector<string, ALLOC(string) >& input_tokens) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Implements the user command `inconsistencies', i.e. lists
 *                the states where the consistency check failed for a set of
 *                algorithms.
 *
 * Arguments:     stream        --  A reference to an output stream.
 *                indent        --  Number of spaces to leave on the left of
 *                                  the output.
 *                round         --  Number of test round.
 *                input_tokens  --  A reference to a vector of strings
 *                                  containing the parameters of the user
 *                                  command.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);

  set<unsigned long int, less<unsigned long int>, ALLOC(unsigned long int) >
    number_set;

  estream << string(indent, ' ') + "Model checking result consistency check "
                                   "results for round "
          << round << ":\n";

  indent += 2;

  if (input_tokens.size() == 1)
    input_tokens.push_back("*");

  parseInterval(input_tokens[1], number_set, 0,
                configuration.algorithms.size() - 1);

  for (set<unsigned long int, less<unsigned long int>,
           ALLOC(unsigned long int) >::const_iterator
	 algorithm = number_set.begin();
       algorithm != number_set.end(); ++algorithm)
  {
    estream << '\n';

    if (*algorithm >= configuration.algorithms.size())
      throw CommandErrorException("No such algorithm (" + input_tokens[1]
                                  + ").");

    estream << string(indent, ' ') + '`'
               + *(configuration.algorithms[*algorithm].name) + "'\n"
               + string(indent + 2, ' ');

    if (test_results[*algorithm].consistency_check_result == -1)
      estream << "Model checking result consistency check was not performed on"
	         " this\n" + string(indent + 2, ' ') + "algorithm.\n";
    else
    {
      if (test_results[*algorithm].consistency_check_result > 0)
        estream << "The algorithm passed the model checking result consistency"
	           " check.\n";
      else
      {
	bool first_printed = false;

	estream << "System states failing consistency check: {";

        for (unsigned long int state = 0;
             state < round_info.real_emptiness_check_size;
             state++)
        {
          if (!test_results[*algorithm].automaton_stats[0].
                 emptiness_check_result[state]
              && !test_results[*algorithm].automaton_stats[1].
                    emptiness_check_result[state])
	  {
	    if (first_printed)
	      estream << ", ";
	    else
	      first_printed = true;

	    estream << state;
	  }
        }

        estream << "}\n";
      }
    }
    estream.flush();
  }

  estream << endl;
}

/* ========================================================================= */
void UserInterface::printPathInconsistencies
  (ostream& stream, int indent, bool formula_type, TestRoundInfo& round_info,
   vector<string, ALLOC(string) >& input_tokens) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Implements the user command `pathinconsistencies', i.e. lists
 *                the states where the path check failed for a set of
 *                algorithms on either the positive or the negated input
 *                formula.
 *
 * Arguments:     stream        --  A reference to an output stream.
 *                indent        --  Number of spaces to leave on the left of
 *                                  the output.
 *                formula_type  --  Selects between the positive and the
 *                                  negated input formula.
 *                round_info    --  Information about the current test round
 *                                  (provides access to the number of the round
 *                                  and the path check result).
 *                input_tokens  --  A reference to a vector of strings
 *                                  containing the parameters of the user
 *                                  command.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);

  set<unsigned long int, less<unsigned long int>, ALLOC(unsigned long int) >
    number_set;

  estream << string(indent, ' ') + "Path check results for round "
          << round_info.current_round << ":\n";

  indent += 2;

  if (input_tokens.size() == 1)
    input_tokens.push_back("*");

  parseInterval(input_tokens[1], number_set, 0,
                configuration.algorithms.size() - 1);

  estream << string(indent, ' ') + "Formula: ";
  round_info.formulae[(configuration.formula_options.output_mode
		         == Configuration::NNF
		       ? 0
		       : 2) + (formula_type ? 0 : 1)].prettyPrint(stream);
  estream << '\n';

  for (set<unsigned long int, less<unsigned long int>,
           ALLOC(unsigned long int) >::const_iterator
	 algorithm = number_set.begin();
       algorithm != number_set.end(); ++algorithm)
  {
    estream << '\n';

    if (*algorithm >= configuration.algorithms.size())
      throw CommandErrorException("No such algorithm (" + input_tokens[1]
                                  + ").");

    estream << string(indent, ' ') + '`'
               + *(configuration.algorithms[*algorithm].name) + "'\n"
               + string(indent + 2, ' ');

    if (test_results[*algorithm].automaton_stats[formula_type ? 0 : 1].
          path_check_result == -1)
      estream << "Path check was not performed on this input formula.\n";
    else
    {
      if (test_results[*algorithm].automaton_stats[formula_type ? 0 : 1].
	    path_check_result > 0)
        estream << "The algorithm passed the path check on this input "
	           "formula.\n";
      else
      {
	bool first_printed = false;

	estream << "System states failing the path check: {";

        for (unsigned long int state = 0;
             state < round_info.real_emptiness_check_size;
             state++)
        {
          if (test_results[*algorithm].automaton_stats[formula_type ? 0 : 1].
	        emptiness_check_result[state]
	        != (formula_type ? round_info.path_truth_values->test(state)
		                 : !round_info.path_truth_values->test(state)))
	  {
	    if (first_printed)
	      estream << ", ";
	    else
	      first_printed = true;

	    estream << state;
	  }
        }

        estream << "}\n";
      }
    }
    estream.flush();
  }

  estream << endl;
}

/* ========================================================================= */
void UserInterface::printTestResults
  (ostream& stream, int indent, unsigned long int round,
   vector<string, ALLOC(string) >& input_tokens)
/* ----------------------------------------------------------------------------
 *
 * Description:   Implements the user command `results', i.e. displays the
 *                current round's test results for a given algorithm.
 *
 * Arguments:     stream        --  A reference to an output stream.
 *                indent        --  Number of spaces to leave on the left of
 *                                  the output.
 *                round         --  Number of test round.
 *                input_tokens  --  A reference to a vector of strings
 *                                  containing the arguments of the user
 *                                  command.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);
  set<unsigned long int, less<unsigned long int>, ALLOC(unsigned long int) >
    number_set;

  estream << string(indent, ' ') + "Test results for round " << round << ":\n";

  if (input_tokens.size() == 1)
    input_tokens.push_back("*");

  parseInterval(input_tokens[1], number_set, 0,
                configuration.algorithms.size() - 1);

  for (set<unsigned long int, less<unsigned long int>,
           ALLOC(unsigned long int) >::const_iterator
	 algorithm = number_set.begin();
       algorithm != number_set.end(); ++algorithm)
  {
    if (*algorithm < configuration.algorithms.size())
    {
      printAllStats(stream, indent, *algorithm);

      if (configuration.statespace_options.generate_mode & Configuration::PATH)
      {
	printHeading(stream, indent,
		     "Model checking result comparison with the path checking "
		     "algorithm:");
	printPathCheckStats(stream, indent + 2, *algorithm);
      }

      printHeading(stream, indent,
		   "Model checking result cross-comparison:");
      printCrossComparisonStats(stream, indent + 2, *algorithm);
      printHeading(stream, indent,
		   "Bchi automata intersection emptiness check:");
      printBuchiIntersectionCheckStats(stream, indent + 2, *algorithm);
    }
    else
      throw CommandErrorException("No such algorithm (" + input_tokens[1]
                                  + ").");
  }
}

/* ========================================================================= */
void UserInterface::printStateSpace
  (ostream& stream, int indent, vector<string, ALLOC(string) >& input_tokens,
   Graph::GraphOutputFormat fmt) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Implements the user command `statespace', i.e. displays
 *                information about a given set of states of the reachability
 *                graph.
 *
 * Arguments:     stream        --  A reference to an output stream.
 *                indent        --  Number of spaces to leave on the left of
 *                                  the output.
 *                input_tokens  --  A reference to a vector of strings
 *                                  containing the arguments of the user
 *                                  command.
 *                fmt           --  Determines the format of output.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);
  set<unsigned long int, less<unsigned long int>, ALLOC(unsigned long int) >
    number_set;

  if (fmt == Graph::NORMAL)
  {
    estream << string(indent, ' ') + "Reachability graph information:\n";

    if (input_tokens.size() == 1)
      input_tokens.push_back("*");

    parseInterval(input_tokens[1], number_set, 0,
		  round_info.statespace->size() - 1);

    for (set<unsigned long int, less<unsigned long int>,
             ALLOC(unsigned long int) >::const_iterator
	   state = number_set.begin();
         state != number_set.end(); ++state)
    {
      verifyNumber(*state, round_info.statespace->size(), "No such state");

      estream << string(indent, ' ') + "State "
	      << *state
	      << string(*state == round_info.statespace->initialState()
			? " (initial state)"
			: "")
	         + ":\n";

      (*(round_info.statespace))[*state].print(stream, indent + 2);
    }
  }
  else if (fmt == Graph::DOT)
    round_info.statespace->print(stream, indent, Graph::DOT);

  estream << endl;
}

/* ========================================================================= */
void UserInterface::changeVerbosity
  (const vector<string, ALLOC(string) >& input_tokens) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Implements the user command `verbosity', i.e. displays or
 *                changes the output verbosity level.
 *
 * Argument:      input_tokens  --  A reference to a constant vector of strings
 *                                  containing the argument of the command.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  if (input_tokens.size() == 1)
    cout << "  Output verbosity is currently set to "
         << configuration.global_options.verbosity
         << ".\n" << endl;
  else
  {
    unsigned long int new_verbosity = parseNumber(input_tokens[1]);
    if (new_verbosity > 5)
      throw CommandErrorException("Verbosity level must be between 0 and 5 "
                                  "inclusive.");

    configuration.global_options.verbosity = new_verbosity;

    cout << "  Output verbosity level set to " << new_verbosity << ".\n"
	 << endl;
  }
}

/* ========================================================================= */
void UserInterface::changeAlgorithmState
  (vector<string, ALLOC(string) >& input_tokens, bool enable) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Changes the enabledness of a set of algorithms used in the
 *                tests.
 *
 * Argument:      input_tokens  --  A reference to a constant vector of strings
 *                                  containing the argument of the command.
 *                enable        --  Determines whether the algorithms are to be
 *                                  enabled or disabled.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  set<unsigned long int, less<unsigned long int>, ALLOC(unsigned long int) >
    algorithms;

  if (input_tokens.size() < 2)
    input_tokens.push_back("*");

  parseInterval(input_tokens[1], algorithms, 0,
                configuration.algorithms.size() - 1);

  for (set<unsigned long int, less<unsigned long int>,
           ALLOC(unsigned long int) >::const_iterator
	 alg = algorithms.begin();
       alg != algorithms.end(); alg++)
  {
    verifyNumber(*alg, configuration.algorithms.size(), "No such algorithm");

    cout << string("  ") + (enable ? "En" : "Dis") + "abling algorithm "
	 << *alg 
	 << " (`" + *(configuration.algorithms[*alg].name) + "').\n";
    
    configuration.algorithms[*alg].enabled = enable;
  }

  cout << endl;
}
