/*
 *  Copyright (C) 1999, 2000  Heikki Tauriainen <Heikki.Tauriainen@hut.fi>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef _UserInterface_h_
#define _UserInterface_h_

#include <deque>
#include <fstream>
#include <iostream>
#include <set>
#include <stack>
#include <string>
#include <strstream>
#include <utility>
#include <vector>
#include "Bitset.h"
#include "BuchiAutomaton.h"
#include "Configuration.h"
#include "Exception.h"
#include "FormulaString.h"
#include "LtlFormula.h"
#include "MacroDefs.h"
#include "ProductAutomaton.h"
#include "StateSpace.h"
#include "TestOperations.h"
#include "TestRoundInfo.h"
#include "TestStatistics.h"

using namespace std;

extern Configuration configuration;

/******************************************************************************
 *
 * A wrapper class for various functions for interacting with the user.
 *
 *****************************************************************************/

class UserInterface
{                                                    
public:                                              
  UserInterface                                     /* Constructor. */
    (TestRoundInfo& test_round_info,
     const vector<AlgorithmTestResults,
                  ALLOC(AlgorithmTestResults) >&
       results,
     const vector<TestStatistics,
                  ALLOC(TestStatistics) >&
       final_stats);

  ~UserInterface();                                 /* Destructor. */

  void changeStreamFormatting                       /* Changes the state of */
    (ostream& stream, int width, int precision,     /* an output stream.    */
     ios::fmtflags flags);                          /* saving its previous
                                                     * state.
						     */

  void restoreStreamFormatting(ostream& stream);    /* Restores a previously
                                                     * saved state of an
                                                     * output stream.
						     */

  void printHeading                                 /* Writes a single line */
    (ostream& stream, int indent,                   /* of text to a stream. */ 
     const string& heading_text) const;   


  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   * 
   * The following functions all extract the information they display from a
   * previously computed TestStatistics structure. The implementations of these
   * functions are in file StatDisplay.cc.
   */

  void printBuchiAutomatonStats                     /* Displays information */
    (ostream& stream,                               /* about a Bchi        */
     int indent,                                    /* automaton.           */
     vector<AlgorithmTestResults,
            ALLOC(AlgorithmTestResults) >::size_type
       algorithm,
     int result_id);

  void printProductAutomatonStats                   /* Displays information */
    (ostream& stream,                               /* about a product      */
     int indent,                                    /* automaton.           */
     vector<AlgorithmTestResults,
            ALLOC(AlgorithmTestResults) >::size_type
       algorithm,
     int result_id);

  void printAcceptanceCycleStats                    /* Displays information */
    (ostream& stream,                               /* about the acceptance */
     int indent,                                    /* cycles of a product  */
     vector<AlgorithmTestResults,                   /* automaton.           */
            ALLOC(AlgorithmTestResults) >::size_type
       algorithm,
     int result_id);

  void printConsistencyCheckStats                   /* Displays the result */
    (ostream& stream,                               /* of the consistency  */
     int indent,                                    /* check for a given   */
     vector<AlgorithmTestResults,                   /* algorithm.          */
            ALLOC(AlgorithmTestResults) >::size_type
       algorithm);

  void printPathCheckStats                          /* Displays the result */
    (ostream& stream,                               /* of the path check   */
     int indent,                                    /* for a given         */
     vector<AlgorithmTestResults,                   /* algorithm.          */
            ALLOC(AlgorithmTestResults) >
       ::size_type
       algorithm) const;

  void printCrossComparisonStats                    /* Displays information */
    (ostream& stream,                               /* about the model      */
     int indent,                                    /* checking result      */
     vector<AlgorithmTestResults,                   /* cross-comparison     */
            ALLOC(AlgorithmTestResults) >           /* check.               */
       ::size_type
       algorithm);

  void printBuchiIntersectionCheckStats             /* Displays the results  */
    (ostream& stream, int indent,                   /* of the Bchi automata */
     vector<AlgorithmTestResults,                   /* intersection          */
            ALLOC(AlgorithmTestResults) >           /* emptiness checks.     */
       ::size_type
       algorithm);

  void printAllStats                                /* A shorthand for       */
    (ostream& stream,                               /* showing all the       */
     int indent,                                    /* information displayed */
     vector<TestStatistics,                         /* by the previous five  */
            ALLOC(TestStatistics)>::size_type       /* functions.            */
       algorithm);  

  void printCollectiveCrossComparisonStats          /* Displays a single     */
    (ostream& stream,                               /* `cell' of the final   */
     vector<TestStatistics,                         /* result cross-         */
            ALLOC(TestStatistics) >::size_type      /* comparison table.     */
       algorithm_y,                                 
     vector<TestStatistics,
            ALLOC(TestStatistics) >::size_type
       algorithm_x,
     int data_type);

  void printCollectiveStats                         /* Displays average test */
    (ostream& stream, int indent);                  /* data over all the
                                                     * test rounds
						     * performed so far.
						     */

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

  void executeUserCommands();                       /* Function for reading and
                                                     * executing user commands
                                                     * at the end of a test
                                                     * round.
						     */

private:
  UserInterface(const UserInterface& ui);           /* Copy constructor (hidden
                                                     * from external usage).
						     */

  UserInterface& operator=                          /* Assignment operator   */
    (const UserInterface& ui);                      /* (hidden from external
						     * usage).
						     */

  TestRoundInfo& round_info;                        /* A reference to a test
						     * round information
						     * object.
						     */

  const vector<AlgorithmTestResults,                /* Results of a single */
               ALLOC(AlgorithmTestResults) >&       /* test round.         */
    test_results;

  const vector<TestStatistics,                      /* Average test data. */
               ALLOC(TestStatistics) >&
    final_statistics;

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

  struct StreamFormatting;                          /* Data type for storing
                                                     * the state of an output
                                                     * stream.
						     */

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
                                                   
  stack<StreamFormatting,                           /* Stack for storing the */
        deque<StreamFormatting,                     /* previous states of an */
              ALLOC(StreamFormatting) > >           /* output stream.        */
    stream_formatting_stack;

  enum TokenType                                    /* User commands. */
    {ALGORITHMS, BUCHI, BUCHIANALYZE, CONTINUE,
     DISABLE, ENABLE, EVALUATE, FORMULA, HELP,
     INCONSISTENCIES, PATHINCONSISTENCIES, QUIT,
     RESULTANALYZE, RESULTS, SKIP, STATESPACE,
     STATISTICS, VERBOSITY, UNKNOWN, _NO_INPUT};

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   * 
   * Functions implementing the user commands. The definitions for these
   * functions are in file UserCommands.cc.
   */

  void computeProductAutomaton                      /* Computes a product */
    (ProductAutomaton*& product_automaton,          /* automaton.         */
     const BuchiAutomaton& buchi_automaton,
     pair<unsigned long int, bool>& last_automaton,
     const pair<unsigned long int, bool>&
       new_automaton) const;

  void printAlgorithmList                           /* Displays a list of */
    (ostream& stream, int indent) const;            /* algorithms used in
                                                     * the tests.    
						     */

  static void synchronizePrefixAndCycle             /* Synchronizes a prefix */
    (deque<Graph::Graph<GraphEdgeContainer>         /* with a cycle in a     */
             ::size_type,                           /* sequence of graph     */
           ALLOC(Graph::Graph<GraphEdgeContainer>   /* state identifiers.    */
		   ::size_type) >&
       prefix,
     deque<Graph::Graph<GraphEdgeContainer>
             ::size_type,
           ALLOC(Graph::Graph<GraphEdgeContainer>
	           ::size_type) >&
       cycle);

  void printCrossComparisonAnalysisResults          /* Analyzes a            */
    (ostream& stream, int indent,                   /* contradiction between */
     bool formula_type,                             /* test results of two   */
     const string& option_1,                        /* algorithms.           */
     const string& option_2,
     ProductAutomaton*& product_automaton,
     pair<unsigned long int, bool>&
       last_product_automaton);

  void printAutomatonAnalysisResults                /* Analyzes a           */
    (ostream& stream, int indent,                   /* contradiction in the */
     unsigned long int algorithm1,                  /* Bchi automata       */
     unsigned long int algorithm2) const;           /* intersection
                                                     * emptiness check.
						     */

  static void printPath                             /* Displays information */
    (ostream& stream, int indent,                   /* about a single       */
     const deque<StateSpace::size_type,             /* system execution.    */
                 ALLOC(StateSpace::size_type) >&
       prefix,
     const deque<StateSpace::size_type,
                 ALLOC(StateSpace::size_type) >&
       cycle,
     const StateSpace& path);

  static void printAcceptingCycle                   /* Displays information */
    (ostream& stream, int indent,                   /* a single automaton   */
     vector<Configuration::AlgorithmInformation,    /* execution.           */
            ALLOC(Configuration::AlgorithmInformation) >
       ::size_type
       algorithm_id,
     const deque<StateSpace::size_type,
                 ALLOC(StateSpace::size_type) >&
       prefix,
     const deque<StateSpace::size_type,
                 ALLOC(StateSpace::size_type) >&
       cycle,
     const BuchiAutomaton& automaton);

  void printBuchiAutomaton                          /* Displays information */
    (ostream& stream, int indent,                   /* about a Bchi        */
     bool formula_type,                             /* automaton.           */
     vector<string, ALLOC(string) >& input_tokens,
     Graph::GraphOutputFormat fmt) const;

  void evaluateFormula                              /* Displays information */
    (ostream& stream, int indent,                   /* about existence of   */
     bool formula_type,                             /* accepting system     */
     vector<string, ALLOC(string) >& input_tokens)  /* executions.          */
    const;

  void printFormula                                 /* Displays a formula */
    (ostream& stream, int indent,                   /* used for testing.  */
     bool formula_type) const;             

  static void printCommandHelp                      /* Displays help about */
    (ostream& stream, int indent,                   /* user commands.      */
     const vector<string, ALLOC(string) >&
       input_tokens);           

  void printInconsistencies                         /* Lists the system   */
    (ostream& stream, int indent,                   /* states failing the */
     unsigned long int round,                       /* consistency check  */
     vector<string, ALLOC(string) >& input_tokens)  /* for an algorihm.   */
    const;

  void printPathInconsistencies                     /* Lists the system    */
    (ostream& stream, int indent,                   /* states failing the  */
     bool formula_type,                             /* path check for an   */
     TestRoundInfo& round_info,                     /* algorithm on the    */
     vector<string, ALLOC(string) >& input_tokens)  /* positive or negated */
    const;                                          /* input formula.      */

  void printTestResults                             /* Displays the test   */
    (ostream& stream, int indent,                   /* results of the last */  
     unsigned long int round,                       /* round performed.    */
     vector<string, ALLOC(string) >& input_tokens);

  void printStateSpace                              /* Displays information */
    (ostream& stream, int indent,                   /* about a reachability */
     vector<string, ALLOC(string) >& input_tokens,  /* graph.               */
     Graph::GraphOutputFormat fmt) const;

  void changeVerbosity                              /* Displays or changes */
    (const vector<string, ALLOC(string) >&          /* the verbosity of    */
       input_tokens) const;                         /* output.             */

  void changeAlgorithmState                         /* Enables or disables a */
    (vector<string, ALLOC(string) >& input_tokens,  /* set of algorithms     */
     bool enable) const;                            /* used in the tests.    */

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   * 
   * Functions for reading and parsing user commands. The definitions for
   * these functions are in file UserCommandReader.cc.
   */

  static void printTextBlock                        /* Writes an indented    */
    (ostream& stream, int indent,                   /* multi-line text block */
     const string& text, int max_line_len);         /* to a stream.
						     */

  static void sliceString                           /* Breaks a string into */
    (const string& s, const char* slice_chars,      /* `slices', using a    */
     vector<string, ALLOC(string) >& slices);       /* given set of
						     * characters as
                                                     * separators.
						     */

  static TokenType parseCommand                     /* Translates a command  */
    (const string& token);                          /* name into its
                                                     * corresponding
                                                     * TokenType identifier.
						     */

  static void verifyArgumentCount                   /* Checks that the      */
    (const vector<string, ALLOC(string) >&          /* number of arguments  */
       arguments,                                   /* for a command is     */
     vector<string, ALLOC(string) >::size_type      /* within given limits. */
       min_arg_count,       
     vector<string, ALLOC(string) >::size_type
       max_arg_count);

  static pair<string, bool> parseRedirection        /* Checks whether the  */
    (vector<string, ALLOC(string) >& input_tokens); /* command given will
                                                     * require redirecting
                                                     * its output to a
                                                     * file.
						     */

  static bool parseFormulaType                      /* Checks whether an     */
    (vector<string, ALLOC(string) >& input_tokens); /* user command
						     * specified a positive
						     * or a negative
						     * formula.
						     */

  static unsigned long int parseNumber              /* Converts a string to  */
    (const string& number_string);                  /* an unsigned long
                                                     * integer.
						     */

  static void parseInterval                         /* Converts a string of  */
    (const string& token,                           /* number intervals to   */
     set<unsigned long int, less<unsigned long int>,/* the corresponding set */
         ALLOC(unsigned long int) >& number_set,    /* of numbers.           */
     unsigned long int min,                         
     unsigned long int max);

  static void verifyNumber                          /* Checks that a given   */
    (unsigned long int number,                      /* number is less than a */
     unsigned long int max,                         /* given maximum value.  */
     const char* error_message);

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

  class CommandErrorException;                      /* A class for reporting
                                                     * errors in user
                                                     * commands.
						     */
};



/******************************************************************************
 *
 * A data stucture for storing information about the state of an output stream.
 *
 *****************************************************************************/

struct UserInterface::StreamFormatting
{
  int width;                                        /* Field width. */

  int precision;                                    /* Floating-point
                                                     * precision.
						     */

  ios::fmtflags flags;                              /* Various flags affecting
                                                     * e.g. the justification
                                                     * of output.
						     */
};



/******************************************************************************
 *
 * An exception class for reporting errors in user commands.
 *
 *****************************************************************************/

class UserInterface::CommandErrorException : public Exception
{
public:
  CommandErrorException                             /* Constructor. */
    (const string& message     
       = "Syntax error in command.");          

  /* default copy constructor */

  ~CommandErrorException();                         /* Destructor. */

  CommandErrorException& operator=                  /* Assignment operator. */
    (const CommandErrorException& e);
};



/******************************************************************************
 *
 * Inline function definitions for class UserInterface.
 *
 *****************************************************************************/

/* ========================================================================= */
inline UserInterface::UserInterface
  (TestRoundInfo& test_round_info,
   const vector<AlgorithmTestResults, ALLOC(AlgorithmTestResults) >& results,
   const vector<TestStatistics, ALLOC(TestStatistics) >& final_stats) :
  round_info(test_round_info), test_results(results),
  final_statistics(final_stats)
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructor for class UserInterface. Initializes a
 *                UserInterface object with pointers and references to the
 *                locations from where data can be extracted for displaying it
 *                to the user.
 *
 * Arguments:     test_round_info       --  An object storing test information,
 *                                          providing access e.g. to the
 *                                          reachability graph and LTL formulae
 *                                          used in the tests.
 *                results               --  A reference to a constant vector of
 *                                          AlgorithmTestResults objects
 *                                          storing the results of a single
 *                                          test round.
 *                final_stats           --  A reference to a constant vector of
 *                                          TestStatistics objects storing the
 *                                          average data over all test rounds
 *                                          performed.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline UserInterface::~UserInterface()
/* ----------------------------------------------------------------------------
 *
 * Description:   Destructor for class UserInterface.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}



/******************************************************************************
 *
 * Inline function definitions for class UserInterface::CommandErrorException.
 *
 *****************************************************************************/

/* ========================================================================= */
inline UserInterface::CommandErrorException::CommandErrorException
  (const string& message) :
  Exception(message)
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructor for class UserInterface::CommandErrorException.
 *                Creates an exception object and initializes it with an error
 *                message.
 *
 * Arguments:     message  --  A reference to a constant string containing the
 *                             error message.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline UserInterface::CommandErrorException::~CommandErrorException()
/* ----------------------------------------------------------------------------
 *
 * Description:   Destructor for class UserInterface::CommandErrorException.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline UserInterface::CommandErrorException&
UserInterface::CommandErrorException::operator=(const CommandErrorException& e)
/* ----------------------------------------------------------------------------
 *
 * Description:   Assignment operator for class
 *                UserInterface::CommandErrorException. Copies the contents of
 *                an exception object to another.
 *
 * Arguments:     e  --  A reference to a constant CommandErrorException.
 *
 * Returns:       A reference to the object assigned to.
 *
 * ------------------------------------------------------------------------- */
{
  Exception::operator=(e);
  return *this;
}

#endif  /* _UserInterface_h_ */
