// This file is based on code originally written by Mauno Rnkk,
// Copyright (C) 1998 Mauno Rnkk <mronkko@abo.fi>.
// Minor modifications by Heikki Tauriainen <Heikki.Tauriainen@hut.fi>.
//
// The original version of this file can be obtained through
//   <URL: http://www.abo.fi/~mauno.ronkko/ASA/ltlalg.html>.
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


#ifndef __Set_h__

#define __Set_h__

#include <iostream.h>


/**
 * A template class for sets.
 *
 *
 * This class provides:
 * creation and destruction, addition and substraction of elements
 * tests (emptyness, inclusion, and equalities), 
 * function for returning some value from set, search
 * of matching element with a given comparison function
 * and criteria, and output to cout.
 *
 *
 * The elements are treated as dummy elements;
 * if an element contains any pointers to allocated memory, 
 * those must be deallocated before the destructor is called.
 **/
template <class T> class Set {

 private:
  /**
   * A template class for nodes in Set.
   *
   *
   * A Node is a container for an element with knowledge
   * of at most another Setnode. Thus, a set is a linked list.
   *
   *
   * The class for a SetNode is a dummy template class 
   * providing the services only for creation and destruction.
   * The destruction of a SetNode does not destroy
   * any other (linked) nodes. Also, if the element contains
   * any pointers to allocated memory, those must be
   * deallocated before the destructor is called.
   **/
  class Node {

  public:
    /// Stored information in a node.
    T    m_value; 
    /// Information of a 'next' existing node.
    Node *m_next;

    /// Creates a new node with given initial values.
    Node (const T &value, Node *next)
      { m_value=value; m_next=next; }
    /// Remember to deallocate pointed objects before this is called.
    ~Node ()
      { }
  };
  
  Node *m_nodes;

  /**
   * Destroys this set.
   * Remember to deallocate pointed objects in T before this is called.
   **/
  void destruct();
  
 public:
  /// Creates an empty set.
  Set()
    { m_nodes=NULL; }
  
  /// Creates a set with one element.
  Set(const T &value)
    { m_nodes=new Node(value,NULL); }
  
  /// Creates a set from another Set.
  Set(const Set &s);
  
  /**
   * Destroys this set.
   * Remember to deallocate pointed objects in T before this is called.
   **/
  ~Set()
    { destruct(); }

  /// Copy the content from a Set.
  const Set& operator = (const Set &s);

  /// Copy the content from a Set of one element.
  const Set& operator = (const T &value)
    { *this = Set<T>(value); return *this; }

  /// Returns the union of two Sets.
  friend const Set operator + <> (const Set &as, const Set &bs);
  
  /// Returns the union of a Set and an element.
  friend const Set operator + <> (const Set &as, const T &value);
  
  /// Returns the subtraction of a Set by another Set.
  friend const Set operator - <> (const Set &as, const Set &bs);
  
  /// Returns the subtraction of a Set by an element.
  friend const Set operator - <> (const Set &as, const T &value);
  
  /// Does this set have a certain value?
  const bool operator >= (const T &value) const;
  
  /// Is this set a superset of another?
  const bool operator >= (const Set &s) const;
  
  /// Are the sets equal?
  const bool operator == (const Set& s) const
    { return ((*this >= s) && (s >= *this)); } 
  
  /// Are the sets inequal?
  const bool operator != (const Set& s) const
    { return *this==s ? false : true; } 
  
  /// Is this set empty?
  const bool isEmpty () const
    { return (m_nodes==NULL) ? true : false; }
  
  /**
   * Returns a pointer to some value from this set, OR NULL, 
   * if the set is empty.
   **/
  T* some() const
    { return m_nodes==NULL ? (T*)NULL : &(m_nodes->m_value); }

  /**
   * Returns a pointer to the contents of an element,
   * which by using "cmpFn (E,criteria)" fulfills the
   * given "criteria",
   * OR NULL, if there is no such element.
   **/
  T  *matching (const bool (*cmpFn)(const T &a, const T &c), 
		const T &criteria) const;

  /// Output a Set to ostream.
  friend ostream& operator << <> (ostream& str, const Set &s);
};


template <class T> void Set<T>::destruct()
{
  Set<T>::Node *next;
  
  while (m_nodes!=NULL) {
    next=m_nodes->m_next;
    delete m_nodes;
    m_nodes=next;
  }
}


template <class T> Set<T>::Set (const Set<T> &s)
{
  Set<T>::Node *node=s.m_nodes;
  
  m_nodes=NULL;
  while (node!=NULL) {
    m_nodes=new Set<T>::Node(node->m_value, m_nodes);
    node=node->m_next;
  }
}


template <class T> const Set<T> &Set<T>::operator = (const Set<T> &s)
{
  Set<T> temp(s);

  destruct();
  m_nodes=temp.m_nodes;
  temp.m_nodes=NULL;

  return *this;
}


template <class T> const Set<T> operator + (const Set<T> &as, 
					    const Set<T> &bs)
{
  Set<T>       result(as);
  Set<T>::Node *node=bs.m_nodes;

  while (node!=NULL) {
    if (!(as >= node->m_value))
      result.m_nodes=new Set<T>::Node(node->m_value, result.m_nodes);
    node=node->m_next;
  }
  return result;
}


template <class T> const Set<T> operator + (const Set<T> &as, const T &value)
{
  return as + Set<T>(value);
}


template <class T> const Set<T> operator - (const Set<T> &as, 
					    const Set<T> &bs)
{
  Set<T>       result;
  Set<T>::Node *node=as.m_nodes;

  while (node!=NULL) {
    if (!(bs >= node->m_value))
      result.m_nodes=new Set<T>::Node(node->m_value, result.m_nodes);
    node=node->m_next;
  }
  return result;
}


template <class T> const Set<T> operator - (const Set<T> &as, const T &value)
{
  return as - Set<T>(value);
}


template <class T> const bool Set<T>::operator >= (const T &value) const
{
  Set<T>::Node *node=m_nodes;
  
  while (node!=NULL)
    if (node->m_value==value)
      return true;
    else
      node=node->m_next;
  return false;
}


template <class T> const bool Set<T>::operator >= (const Set<T>& s) const
{
  Set<T>::Node *node=s.m_nodes;
  
  while (node!=NULL)
    if (*this >= node->m_value)
      node=node->m_next;
    else
      return false;
  return true;
}


template <class T> T *Set<T>::matching (const bool (*cmpFn)(const T &a, 
							    const T &c),
					const T &criteria) const
{
  Set<T>::Node *node=m_nodes;
  T            *match=NULL;
  
  while ((node!=NULL) && (match==NULL))
    if (cmpFn (node->m_value, criteria))
      match=&(node->m_value);
    else
      node=node->m_next;
  return match;
}


template <class T> ostream& operator << (ostream& str, const Set<T>& s)
{ 
  Set<T>::Node *next;
  bool         first=true;
      
  next=s.m_nodes;
  str << "{";
  while (next!=NULL) {
    if (!first) str << ", ";
    first=false;
    str << next->m_value;
    next=next->m_next;
  }
  return str << "}";
}


#endif


