/*
 *  Copyright (C) 1999, 2000  Heikki Tauriainen <Heikki.Tauriainen@hut.fi>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "main.h"

/******************************************************************************
 *
 * Handler for the SIGINT signal.
 *
 *****************************************************************************/

void breakHandler(int)
{
  user_break = true;
}



/******************************************************************************
 *
 * Test loop.
 *
 *****************************************************************************/

void testLoop(const char* command_line_formula)
{
  const Configuration::GlobalConfiguration& global_options
    = configuration.global_options;

  const Configuration::StatespaceConfiguration& statespace_options
    = configuration.statespace_options;

  /*
   *  Initialize the data structure with information about the current test
   *  round.
   */

  TestRoundInfo round_info;

  round_info.abort = false;
  round_info.next_round_to_run = 1;

  round_info.next_round_to_stop
    = (global_options.interactive == Configuration::ALWAYS
       ? 1
       : global_options.number_of_rounds + 1);

  round_info.num_generated_statespaces = 0;
  round_info.total_statespace_states = 0;
  round_info.total_statespace_transitions = 0;
  round_info.num_processed_formulae = 0;
  round_info.statespace = 0;
  round_info.number_of_rounds_to_change_statespace = 0;
  round_info.number_of_rounds_to_change_formula = 0;

  if (tmpnam(round_info.formula_file_name) == 0
      || tmpnam(round_info.negated_formula_file_name) == 0)
    throw Exception("Unable to allocate names for temporary files.");

  /*
   *  If a formula file name was given in the configuration, open the file
   *  for reading.
   */

  ifstream formula_input_file;

  if (!global_options.formula_input_filename.empty()
      && command_line_formula == 0)
    openFile(global_options.formula_input_filename.c_str(), formula_input_file,
	     ios::in, 0);

  /*
   *  If a name for the transcript file was given in the configuration,
   *  create the file.
   */

  ofstream transcript_file;

  if (!global_options.transcript_filename.empty())
  {
    time_t current_time;

    time(&current_time);

    openFile(global_options.transcript_filename.c_str(), transcript_file,
	     ios::out | ios::trunc, 0);

    Exceptional_ostream etranscript_file(&transcript_file,
					 ios::badbit | ios::failbit);

    etranscript_file << "lbtt error log file, created on "
		     << ctime(&current_time) << endl;

    configuration.print(transcript_file);
  }

  /*
   *  Pointer to a Graph::SinglePathIterator to be used if the state spaces
   *  are enumerated paths.
   */

  Graph::SinglePathIterator* path_iterator = 0;

  /*
   *  If the state spaces are paths, allocate space for the results of the
   *  direct model checking algorithm when evaluating the LTL formulae in
   *  the generated paths.
   */

  if (statespace_options.generate_mode & Configuration::PATH)
    round_info.path_truth_values = new Bitset(statespace_options.max_size);
  else
    round_info.path_truth_values = 0;

  /*
   *  Intialize a vector for storing the test results for each algorithm.
   */

  vector<AlgorithmTestResults, ALLOC(AlgorithmTestResults) > test_results
    (configuration.algorithms.size(),
     AlgorithmTestResults(configuration.algorithms.size(),
			  (global_options.product_mode == Configuration::GLOBAL
			   ? statespace_options.max_size
			   : 1)));

  /*
   *  Initialize a vector for collecting the test statistics for each
   *  algorithm.
   */

  vector<TestStatistics, ALLOC(TestStatistics) >
    final_statistics(configuration.algorithms.size(),
                     TestStatistics(configuration.algorithms.size()));

  UserInterface user_interface
                  (round_info, test_results, final_statistics);

  for (round_info.current_round = 1;
       !round_info.abort
         && round_info.current_round <= global_options.number_of_rounds;
       ++round_info.current_round)
  {
    user_break = false;
    round_info.error = false;
    round_info.skip = false;

    if (global_options.verbosity >= 1)
      cout << "Round "
	   << round_info.current_round
	   << '/'
	   << global_options.number_of_rounds
	   << endl
	   << endl;

    for (vector<AlgorithmTestResults, ALLOC(AlgorithmTestResults) >::iterator
	   result = test_results.begin();
	 result != test_results.end(); ++result)
      result->reset();

    try
    {
      /*
       *  Test whether it is necessary to generate a new state space.
       */

      if (round_info.number_of_rounds_to_change_statespace == 0)
      {
	round_info.number_of_rounds_to_change_statespace
	  = global_options.statespace_change_interval;

	try
	{
	  if (statespace_options.generate_mode
	        != Configuration::ENUMERATEDPATH)
	  {
	    if (round_info.statespace != 0)
	      delete round_info.statespace;

	    round_info.statespace = generateStateSpace();
	  }
	  else
	  {
	    if (global_options.verbosity >= 2)
	      user_interface.printHeading(cout, 4,
					  "Generating next state space...");

	    if (path_iterator == 0)
	    {
	      path_iterator = new Graph::SinglePathIterator
		                (statespace_options.atoms_per_state,
				 statespace_options.size);
	      round_info.statespace = &(**path_iterator);
	    }
	    else
	      ++(*path_iterator);

	    if (path_iterator->atEnd())
	    {
	      if (global_options.verbosity >= 2)
		user_interface.printHeading(cout, 6,
					    "[All state spaces have been "
					    "enumerated. Starting over.]");

	      delete path_iterator;
	      path_iterator = new Graph::SinglePathIterator
		                    (statespace_options.atoms_per_state,
				     statespace_options.size);
	      round_info.statespace = &(**path_iterator);
	    }

	    if (global_options.verbosity >= 2)
	      cout << endl;
	  }

	  pair<StateSpace::size_type, unsigned long int>
	    statespace_stats(round_info.statespace->stats());

	  round_info.total_statespace_states += statespace_stats.first;
	  round_info.total_statespace_transitions += statespace_stats.second;

	  ++round_info.num_generated_statespaces;

	  round_info.real_emptiness_check_size
	    = (global_options.product_mode == Configuration::GLOBAL
	       ? round_info.statespace->size()
	       : 1);
	}
	catch (...)
	{
	  round_info.statespace = 0;

	  if (global_options.statespace_change_interval == 0)
	    round_info.next_round_to_run = global_options.number_of_rounds + 1;
	  else if (round_info.next_round_to_run <
		     round_info.current_round
		     + round_info.number_of_rounds_to_change_statespace)
	    round_info.next_round_to_run
	      = round_info.current_round
	        + round_info.number_of_rounds_to_change_statespace;

	  try
	  {
	    throw;
	  }
	  catch (StateSpaceGenerationException&)
	  {
	    if (!global_options.transcript_filename.empty())
	      writeToTranscript(transcript_file,
				round_info.current_round,
				"State space generation failed, no tests "
				"performed");
	  }
	}
      }

      /*
       *  Test whether it is necessary to generate a new LTL formula. If a
       *  formula was given as a command line parameter, use that formula.
       */

      if (command_line_formula != 0)
      {
	round_info.num_processed_formulae = 1;
	istrstream formula_stream(command_line_formula);
	generateFormulae(round_info.formulae, &formula_stream);
      }
      else if (round_info.number_of_rounds_to_change_formula == 0)
      {
	try
	{
	  generateFormulae(round_info.formulae,
			   (formula_input_file.is_open()
			    ? &formula_input_file
			    : 0));
	  ++round_info.num_processed_formulae;
	}
	catch (FormulaString::FormulaString::ParseErrorException&)
	{
	  if (global_options.verbosity >= 2)
	    user_interface.printHeading(cout, 6,
					string("[Error parsing the ")
					+ (formula_input_file.is_open()
					   ? "formula file"
					   : "input formula")
					+ ". Aborting.]");

	  if (!global_options.transcript_filename.empty())
	    writeToTranscript(transcript_file, round_info.current_round,
			      "Error parsing input formula. Testing aborted.");

	  break;
	}
	catch (IOException&)
	{
	  /*
	   *  Abort testing if an IOException is caught. (Such an exception
	   *  will also be thrown if there are no more input formulas, so this
	   *  is not necessarily an error situation.)
	   */

	  if (global_options.verbosity >= 2)
	    user_interface.printHeading(cout, 6,
					string("[")
					+ (!global_options.
					     formula_input_filename.empty()
					   && formula_input_file.eof()
					   ? "No more input formulae"
					   : "Error processing formula")
					+ ". Aborting.]");

	  if (!global_options.transcript_filename.empty()
	      && (global_options.formula_input_filename.empty()
		  || !formula_input_file.eof()))
	    writeToTranscript(transcript_file, round_info.current_round,
			      "Error processing input formula. "
			      "Testing aborted.");
	  
	  break;
	}
      
	round_info.number_of_rounds_to_change_formula
	  = global_options.formula_change_interval;
      }

      if (round_info.current_round < round_info.next_round_to_run)
      {
	if (user_break)
	{
	  if (global_options.verbosity >= 3)
	    user_interface.printHeading(cout, 4, "[User break]\n");

	  throw UserBreakException();
	}

	if (global_options.verbosity >= 1)
	  cout << "    Skipped round " << round_info.current_round << ".\n\n";
	round_info.skip = true;
      }
      else
      {
	/*
	 *  If the generated reachability graphs are paths, model check the
	 *  formula separately in the path.
	 */

	if (statespace_options.generate_mode & Configuration::PATH)
	{
	  if (global_options.verbosity >= 2)
	  {
	    user_interface.printHeading
	      (cout, 4, "Model checking formula in the path...");
	    cout << endl;
	  }

	  LtlFormula::LtlFormula* formula;
	  round_info.path_truth_values->clear();

	  formula
	    = LtlFormula::formulaStringToLtlFormula(round_info.formulae[2]);

	  try
	  {
	    formula->evaluatePath(*(round_info.statespace),
				  round_info.path_truth_values);
	    delete formula;
	  }
	  catch (...)
	  {
	    delete formula;

	    if (!global_options.transcript_filename.empty())
	      writeToTranscript(transcript_file,
				round_info.current_round,
				"Formula evaluation failure, no tests "
				"performed");

	    throw;
	  }
	}

	/*
	 *  Write the formula and its negation into two files (in negation
	 *  normal form if necessary).
	 */

	try
        {
	  writeFormulaToFile
	    (round_info.formulae[configuration.formula_options.output_mode
		                   == Configuration::NNF ? 0 : 2],
	     round_info.formula_file_name);

	  writeFormulaToFile
	    (round_info.formulae[configuration.formula_options.output_mode
			           == Configuration::NNF ? 1 : 3],
	     round_info.negated_formula_file_name);
	}
	catch (IOException&)
	{
	  round_info.error = true;
	}

	if (!round_info.error)
	{
	  for (vector<Configuration::AlgorithmInformation,
		      ALLOC(Configuration::AlgorithmInformation) >::size_type
		 algorithm_id = 0;
	       algorithm_id < configuration.algorithms.size();
	       ++algorithm_id)
	  {
	    if (global_options.verbosity >= 2)
	      cout << "    "
		   << algorithm_id
		   << ": `" + *(configuration.algorithms[algorithm_id].name)
                      + '\''
		   << endl;

	    if (!configuration.algorithms[algorithm_id].enabled)
	    {
	      if (global_options.verbosity >= 2)
		user_interface.printHeading(cout, 6, "[Disabled]\n");
	      continue;
	    }

	    for (int counter = 0; counter < 2; counter++)
	    {
	      if (user_break)
	      {
		if (global_options.verbosity >= 3)
		  user_interface.printHeading(cout, 6, "[User break]\n");
		throw UserBreakException();
	      }

	      if (global_options.verbosity >= 2)
		user_interface.printHeading
		  (cout, 6, string(counter == 1 ? "Negated" : "Positive")
		            + " formula:");

	      AutomatonStats& automaton_stats = 
		test_results[algorithm_id].automaton_stats[counter];

	      round_info.product_automaton = 0;

	      try
	      {
		/*
		 *  Generate a Bchi automaton using the current algorithm.
		 *  `counter' determines the formula which is to be
		 *  translated into an automaton; 0 denotes the positive and
		 *  1 the negated formula.
		 */

		final_statistics[algorithm_id].
		  buchi_automaton_count[counter]++;

		if (global_options.verbosity >= 2)
		  user_interface.printHeading
		    (cout, 8, (global_options.verbosity == 2
			       ? "Computing Bchi automaton..."
			       : "Bchi automaton:"));
            
		automaton_stats.buchi_automaton
		  = generateBuchiAutomaton
		      ((counter == 1 ? round_info.negated_formula_file_name
			             : round_info.formula_file_name),
		       configuration.algorithms[algorithm_id],
		       automaton_stats,
		       (global_options.verbosity >= 2 ? 10 : 0));
            
		if (global_options.verbosity >= 3)
		  user_interface.printBuchiAutomatonStats
		    (cout, 10, algorithm_id, counter);
		
		/*
		 *  Update Bchi automaton statistics for the current
		 *  algorithm.
		 */

		final_statistics[algorithm_id].
		  total_number_of_buchi_states[counter] +=
		  automaton_stats.number_of_buchi_states;
		final_statistics[algorithm_id].	
		  total_number_of_buchi_transitions[counter] +=
		  automaton_stats.number_of_buchi_transitions;
		final_statistics[algorithm_id].
		  total_number_of_acceptance_sets[counter] +=
		  automaton_stats.number_of_acceptance_sets;
		  
		if (final_statistics[algorithm_id].
		      total_buchi_generation_time[counter] < 0.0
		    || automaton_stats.buchi_generation_time < 0.0)
		  final_statistics[algorithm_id].
		    total_buchi_generation_time[counter] = -1.0;
		else
		  final_statistics[algorithm_id].
		    total_buchi_generation_time[counter] +=
		    automaton_stats.buchi_generation_time;

		/*
		 *  Compute the product of the Bchi automaton with the
		 *  reachability graph.
		 */

		if (global_options.verbosity >= 2)
		  user_interface.printHeading
		    (cout, 8, (global_options.verbosity == 2
			       ? "Computing product automaton..."
			       : "Product automaton:"));

		final_statistics[algorithm_id].
		  product_automaton_count[counter]++;

		round_info.product_automaton = generateProductAutomaton
		  (*(automaton_stats.buchi_automaton),
		   *(round_info.statespace), automaton_stats, 10);
        
		if (global_options.verbosity >= 3)
		  user_interface.printProductAutomatonStats
		    (cout, 10, algorithm_id, counter);

		/*
		 *  Update product automaton statistics for the current
		 *  algorithm.
		 */

		final_statistics[algorithm_id].
		  total_number_of_product_states[counter] +=
		  automaton_stats.number_of_product_states;
		final_statistics[algorithm_id].
		  total_number_of_product_transitions[counter] +=
		  automaton_stats.number_of_product_transitions;

		/*
		 *  Find the system states from which an accepting execution
		 *  cycle can be reached by checking the product automaton
		 *  for emptiness.
		 */

		if (global_options.verbosity >= 2)
		  user_interface.printHeading
		    (cout, 8, (global_options.verbosity == 2
			       ? "Performing emptiness check..."
			       : "Accepting cycles:"));

		performEmptinessCheck
		  (*(round_info.product_automaton), automaton_stats);

		if (global_options.verbosity >= 3)
		  user_interface.printAcceptanceCycleStats
		    (cout, 10, algorithm_id, counter);
		
		/*
		 *  Release the memory allocated for the product automaton.
		 */

		if (global_options.verbosity >= 4)
		{
		  cout << string(8, ' ')
		          + "<cleaning up memory allocated for product "
		            "automaton>";
		  cout.flush();
		}

		delete round_info.product_automaton;
		round_info.product_automaton = 0;
	      
		if (global_options.verbosity >= 4)
		  cout << " ok" << endl;
	      }
	      catch (...)
	      {
		if (round_info.product_automaton != 0)
		{
		  delete round_info.product_automaton;
		  round_info.product_automaton = 0;
		}
		
		try
		{
		  throw;
		}
		catch (BuchiAutomatonGenerationException&)
		{
		  if (automaton_stats.buchi_automaton != 0)
		  {
		    delete automaton_stats.buchi_automaton;
		    automaton_stats.buchi_automaton = 0;
		  }

		  if (!global_options.transcript_filename.empty())
		    writeToTranscript(transcript_file,
				      round_info.current_round,
				      string("`") + *configuration.algorithms
                                                      [algorithm_id].name
                                      + "' failed to generate a Bchi "
				        "automaton from "
                                      + (counter == 1 ? "nega" : "posi")
                                      + "tive formula");

		  round_info.error = true;
		  final_statistics[algorithm_id].
		    failures_to_compute_buchi_automaton[counter]++;
		}
		catch (ProductAutomatonGenerationException&)
		{
		  round_info.error = true;

		  if (!global_options.transcript_filename.empty())
		    writeToTranscript(transcript_file,
				      round_info.current_round,
				      string("Product automaton generation "
                                             "failed (`")
                                      + *configuration.algorithms
				          [algorithm_id].name
				      + "', "
                                      + (counter == 1 ? "nega" : "posi")
                                      + "tive formula)");

		  final_statistics[algorithm_id].
		    failures_to_compute_product_automaton[counter]++;
		}
		catch (EmptinessCheckFailedException&)
		{
		  round_info.error = true;

		  if (!global_options.transcript_filename.empty())
		    writeToTranscript(transcript_file,
				      round_info.current_round,
				      string("Emptiness check not performed "
                                             "(`")
                                      + *configuration.algorithms
				          [algorithm_id].name
				      + "', "
                                      + (counter == 1 ? "nega" : "posi")
                                      + "tive formula)");
		}
		catch (...)
		{
		  throw;
		}
	      }
	    }

	    /*
	     *  If the emptiness check was performed successfully for the
	     *  product automata constructed from both the positive and
	     *  negated formulae, test whether the emptiness check results
	     *  are consistent with each other. (It should not be possible
	     *  for both the formula and its negation to be true in any
	     *  state.)
	     */

	    if (test_results[algorithm_id].automaton_stats[0].
		  emptiness_check_performed
		&& test_results[algorithm_id].automaton_stats[1].
		     emptiness_check_performed)
	    {
	      final_statistics[algorithm_id].consistency_checks_performed++;

	      if (global_options.verbosity >= 2)
		user_interface.printHeading
		  (cout, 6, (global_options.verbosity == 2
			     ? "Performing model checking result consistency "
			       "check..."
			     : "Result consistency check:"));

	      if (!performConsistencyCheck(round_info.
					     real_emptiness_check_size,
					   test_results[algorithm_id]))
	      {
		round_info.error = true;

		if (!global_options.transcript_filename.empty())
		  writeToTranscript(transcript_file,
				    round_info.current_round,
				    string("`") + *configuration.algorithms
                                                    [algorithm_id].name
				    + "' fails the model checking result "
				      "consistency check");

		final_statistics[algorithm_id].consistency_check_failures++;
	      }

	      if (global_options.verbosity >= 3)
		user_interface.printConsistencyCheckStats
		  (cout, 8, algorithm_id);
	    }

	    if (global_options.verbosity >= 2)
	      cout << endl;
	  }

	  /*
	   *  If the generated state spaces are paths, perform the model
	   *  checking result comparison against the results obtained by
	   *  evaluating the LTL formula directly.
	   */

	  if (statespace_options.generate_mode & Configuration::PATH)
	  {
	    if (global_options.verbosity >= 2)
	      user_interface.printHeading
		(cout, 4, (global_options.verbosity == 2
			   ? "Performing result comparison with the path "
			     "checking algorithm..."
			   : "Model checking result comparison with the "
			     "path checking algorithm:"));

	    if (!performPathCheck(round_info, test_results, final_statistics))
	    {
	      round_info.error = true;
	      
	      if (!global_options.transcript_filename.empty())
	      {
		writeToTranscript(transcript_file,
				  round_info.current_round,
				  "Model checking result comparison with the "
				  "path checking algorithm failed");
		user_interface.printPathCheckStats
		  (transcript_file, 4, configuration.algorithms.size());
	      }
	    }

	    if (configuration.global_options.verbosity >= 3)
	      user_interface.printPathCheckStats
		(cout, 6, configuration.algorithms.size());
	  }
			   
	  /*
	   *  Perform the pairwise comparisons of the emptiness check results
	   *  obtained using the different algorithms.
	   */

	  if (configuration.algorithms.size() > 1)
	  {
	    if (global_options.verbosity >= 2)
	      user_interface.printHeading
		(cout, 4, (global_options.verbosity == 2
			   ? "Performing model checking result "
			     "cross-comparison...\n"
			   : "Model checking result cross-comparison:"));

	    if (!compareResults(test_results, final_statistics))
	    {
	      round_info.error = true;

	      if (!global_options.transcript_filename.empty())
	      {
		writeToTranscript(transcript_file,
				  round_info.current_round,
				  "Model checking result cross-comparison "
				  "check failed");
		user_interface.printCrossComparisonStats
		  (transcript_file, 4, configuration.algorithms.size());
	      }
	    }
	      
	    if (global_options.verbosity >= 3)
	      user_interface.printCrossComparisonStats
		(cout, 6, configuration.algorithms.size());
	  }

	  /*
	   *  Perform the pairwise intersection emptiness checks on the
	   *  Bchi automata computed during this test round using the
	   *  different algorithms.
	   */

	  if (global_options.verbosity >= 2)
	    user_interface.printHeading
	      (cout, 4, (global_options.verbosity == 2
			 ? "Performing Bchi automata intersection "
			   "emptiness check...\n"
			 : "Bchi automata intersection emptiness check:"));

	  if (!performBuchiIntersectionCheck(test_results, final_statistics))
	  {
	    round_info.error = true;

	    if (!global_options.transcript_filename.empty())
	    {
	      writeToTranscript(transcript_file,
				round_info.current_round,
				"Bchi automata intersection emptiness check "
				"failed");
	      user_interface.printBuchiIntersectionCheckStats
		(transcript_file, 4, configuration.algorithms.size());
	    }
	  }

	  if (global_options.verbosity >= 3)
	    user_interface.printBuchiIntersectionCheckStats
	      (cout, 6, configuration.algorithms.size());
	}
      }
    }
    catch (UserBreakException&)
    {
      user_break = false;
      round_info.next_round_to_stop = round_info.current_round;
    }

    /*
     *  Determine from the program configuration and the error status whether
     *  the testing is to be paused to wait for user commands.
     */

    if (round_info.error &&
        global_options.interactive == Configuration::ONERROR)
      round_info.next_round_to_stop = round_info.current_round;

    if (round_info.next_round_to_stop == round_info.current_round)
      user_interface.executeUserCommands();

    if (round_info.next_round_to_run == round_info.current_round)
      round_info.next_round_to_run++;

    /*
     *  If the current round was not skipped, deallocate the memory reserved
     *  for the results of the current round. Delete also the temporary formula
     *  files in this case.
     */

    if (!round_info.skip)
    {
      for (vector<AlgorithmTestResults, ALLOC(AlgorithmTestResults) >
	     ::reverse_iterator
	     result = test_results.rbegin();
           result != test_results.rend();
           ++result)
      {
	if (global_options.verbosity >= 4)
	{
	  cout << string(8, ' ') + "<releasing allocated memory>";
	  cout.flush();
	}

	for (int counter = 1; counter >= 0; counter--)
	{
	  if (result->automaton_stats[counter].buchi_automaton != 0)
	    delete result->automaton_stats[counter].buchi_automaton;
	}

	if (global_options.verbosity >= 4)
	  cout << " ok" << endl;
      }

      removeFile(round_info.formula_file_name);
      removeFile(round_info.negated_formula_file_name);
    }

    /*
     *  Update the counters determining when to generate a new reachability
     *  graph or a new LTL formula.
     */

    if (global_options.statespace_change_interval == 0)
      round_info.number_of_rounds_to_change_statespace = 1;
    else
      round_info.number_of_rounds_to_change_statespace--;

    if (global_options.formula_change_interval == 0)
      round_info.number_of_rounds_to_change_formula = 1;
    else
      round_info.number_of_rounds_to_change_formula--;
  }

  if (round_info.path_truth_values != 0)
    delete round_info.path_truth_values;
  if (path_iterator != 0)
    delete path_iterator;
  else if (round_info.statespace != 0)
    delete round_info.statespace;

  if (round_info.current_round > global_options.number_of_rounds)
    round_info.current_round = global_options.number_of_rounds;

  if (global_options.verbosity >= 1)
    user_interface.printCollectiveStats(cout, 0);

  if (transcript_file.is_open())
  {
    Exceptional_ostream etranscript_file(&transcript_file,
					 ios::badbit | ios::failbit);

    etranscript_file << endl;
    user_interface.printCollectiveStats(transcript_file, 0);
    etranscript_file << endl;

    time_t current_time;

    time(&current_time);

    etranscript_file << "lbtt error log finished on "
		     << ctime(&current_time)
		     << endl;

    transcript_file.close();
  }

  if (formula_input_file.is_open())
    formula_input_file.close();
}



/******************************************************************************
 *
 * Main function.
 *
 *****************************************************************************/

int main(int argc, char* argv[])
{
  try 
  {
    configuration.read(argc, argv);
    configuration.readCommandLineParameters(argc, argv);
    configuration.validateAndInitialize();
  }
  catch (Configuration::ConfigurationException& e)
  {
    cerr << argv[0] << ": Configuration error: ";
    if (e.line_number != -1)
      cerr << "Line " << e.line_number << ": ";
    cerr << e.what() << endl;
    exit(-1);
  }

  if (configuration.global_options.verbosity >= 3)
    configuration.print(cout);

  user_break = false;
  signal(SIGINT, breakHandler);

  try 
  {
    testLoop(0);
  }
  catch (Exception& e)
  {
    cerr << endl << argv[0] << ": " << e.what() << endl;
    exit(-1);
  }
  catch (bad_alloc&)
  {
    cerr << endl << argv[0] << ": out of memory" << endl;
    exit(-1);
  }

  return 0;
}
