/*
 *  Copyright (C) 1999, 2000  Heikki Tauriainen <Heikki.Tauriainen@hut.fi>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "spin_lbt.h"

static char* program_name;

/******************************************************************************
 *
 * Signal handler for debugging purposes.
 *
 *****************************************************************************/

void signalHandler(int signal_number)
{
  cerr << program_name << ": received signal " << signal_number << endl;
  signal(signal_number, SIG_DFL);
  raise(signal_number);
}



/******************************************************************************
 *
 * Main function.
 *
 *****************************************************************************/

int main(int argc, char* argv[])
{
  using namespace spin_buchi;
  using spin_buchi::FormulaString;

  program_name = argv[0];

  if (argc != 4)
  {
    cerr << program_name << ": Invalid number of command line arguments."
	 << endl;
    exit(-1);
  }

  char spin_input_filename[L_tmpnam + 1];
  char spin_output_filename[L_tmpnam + 1];
  bool spin_input_file_created = false, spin_output_file_created = false;

  int exitstatus = 0;

  signal(SIGHUP, signalHandler);
  signal(SIGINT, signalHandler);
  signal(SIGQUIT, signalHandler);
  signal(SIGILL, signalHandler);
  signal(SIGABRT, signalHandler);
  signal(SIGFPE, signalHandler);
  signal(SIGSEGV, signalHandler);
  signal(SIGPIPE, signalHandler);
  signal(SIGALRM, signalHandler);
  signal(SIGTERM, signalHandler);

  try
  {
    if (tmpnam(spin_input_filename) == 0 || tmpnam(spin_output_filename) == 0)
      throw FileCreationException("a temporary file");

    /*  
     *  Read the formula string from input file.
     */

    FormulaString formula;

    formula.read(argv[2]);

    /*
     *  Translate the formula into Spin's input file format. This is done by
     *  `traversing' the FormulaString object, i.e. scanning through the
     *  formula and applying suitable translations to the symbols in the
     *  formula. The result of the translation will be found in the string
     *  `spin_formula'.
     */

    string spin_formula;
    FormulaString::TraversalInfo<string*> info;
    info.data = &spin_formula;

    formula.template traverse<TranslateAtom,
                              TranslateConstant<SPIN_TRUE>,
                              TranslateConstant<SPIN_FALSE>,
                              TranslateUnary<SPIN_NOT>,
                              TranslateUnary<SPIN_NEXT>,
                              TranslateUnary<SPIN_FINALLY>,
                              TranslateUnary<SPIN_GLOBALLY>,
                              TranslateBinary<SPIN_AND>,
                              TranslateBinary<SPIN_OR>,
                              TranslateBinary<SPIN_IMPLICATION>,
                              TranslateBinary<SPIN_EQUIVALENCE>,
                              TranslateBinary<SPIN_UNTIL>,
                              TranslateBinary<SPIN_V> >
      (FormulaString::PREORDER | FormulaString::INORDER |
       FormulaString::POSTORDER, info);

    /*
     *  Create a file containing the translated formula.
     */

    ofstream spin_input_file;
    spin_input_file.open(spin_input_filename, ios::out | ios::trunc);
    if (!spin_input_file.good())
      throw FileCreationException(spin_input_filename);

    spin_input_file_created = true;

    Exceptional_ostream espin_input_file(&spin_input_file,
                                         ios::failbit | ios::badbit);

    espin_input_file << spin_formula << '\n';
    espin_input_file.flush();
    spin_input_file.close();

    /*
     *  Execute Spin with the formula as input.
     */

    string spin_command_line;

    spin_command_line = string(argv[1]) + " -F " + spin_input_filename + " >"
                        + spin_output_filename;

    int exitcode = system(spin_command_line.c_str());

    spin_output_file_created = true;

    if (exitcode != 0)
      throw ExecFailedException(argv[1]);

    /*
     *  Parse Spin's output file and save the constructed Bchi automaton
     *  to the final output file.
     */

    NeverClaimParser::NeverClaimAutomaton automaton;
    automaton.read(spin_output_filename);
    automaton.write(argv[3]);
  }
  catch (...)
  {
    cerr << program_name << ": ";
    exitstatus = -1;

    try
    {
      throw;
    }
    catch (IOException& e)
    {
      cerr << e.what();
    }
    catch (NeverClaimParser::ParseErrorException& e)
    {
      cerr << e.what();
    }
    catch (FormulaString::FormulaString::ParseErrorException& e)
    {
      cerr << e.what();
    }
    catch (ExecFailedException& e)
    {
      cerr << e.what();
    }
    catch (...)
    {
      cerr << "Fatal error, aborting.";
    }

    cerr << endl;
  }

  if (spin_input_file_created)
    remove(spin_input_filename);

  if (spin_output_file_created)
    remove(spin_output_filename);

  return exitstatus;
}
